@@ -10,12 +10,14 @@
 #include <limits.h>
 #include <pthread.h>
 
+#include "src/base/fpu.h"
 #include "src/base/logging.h"
 #if defined(__DragonFly__) || defined(__FreeBSD__) || defined(__OpenBSD__)
 #include <pthread_np.h>  // for pthread_set_name_np
 #endif
 #include <fcntl.h>
 #include <sched.h>  // for sched_yield
+#include <signal.h>
 #include <stdio.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
@@ -35,11 +37,12 @@
 
 #include <cmath>
 #include <cstdlib>
-
-#include "src/base/platform/platform-posix.h"
+#include <optional>
 
 #include "src/base/lazy-instance.h"
 #include "src/base/macros.h"
+#include "src/base/platform/mutex.h"
+#include "src/base/platform/platform-posix.h"
 #include "src/base/platform/platform.h"
 #include "src/base/platform/time.h"
 #include "src/base/utils/random-number-generator.h"
@@ -55,7 +58,9 @@
 #if V8_OS_DARWIN
 #include <mach/mach.h>
 #include <malloc/malloc.h>
-#else
+#elif V8_OS_OPENBSD
+#include <sys/malloc.h>
+#elif !V8_OS_ZOS
 #include <malloc.h>
 #endif
 
@@ -69,7 +74,7 @@
 #include <sys/resource.h>
 #endif
 
-#if !defined(_AIX) && !defined(V8_OS_FUCHSIA)
+#if !defined(_AIX) && !defined(V8_OS_FUCHSIA) && !V8_OS_ZOS
 #include <sys/syscall.h>
 #endif
 
@@ -99,9 +104,11 @@ namespace base {
 namespace {
 
 // 0 is never a valid thread id.
+#if V8_OS_ZOS
+const pthread_t kNoThread = {0};  // pthread_t is a struct on z/OS
+#else
 const pthread_t kNoThread = static_cast<pthread_t>(0);
-
-bool g_hard_abort = false;
+#endif
 
 const char* g_gc_fake_mmap = nullptr;
 
@@ -109,7 +116,7 @@ DEFINE_LAZY_LEAKY_OBJECT_GETTER(RandomNumberGenerator,
                                 GetPlatformRandomNumberGenerator)
 static LazyMutex rng_mutex = LAZY_MUTEX_INITIALIZER;
 
-#if !V8_OS_FUCHSIA
+#if !V8_OS_FUCHSIA && !V8_OS_ZOS
 #if V8_OS_DARWIN
 // kMmapFd is used to pass vm_alloc flags to tag the region with the user
 // defined tag 255 This helps identify V8-allocated regions in memory analysis
@@ -134,7 +141,8 @@ int GetFlagsForMemoryPermission(OS::MemoryPermission access,
                                 PageType page_type) {
   int flags = MAP_ANONYMOUS;
   flags |= (page_type == PageType::kShared) ? MAP_SHARED : MAP_PRIVATE;
-  if (access == OS::MemoryPermission::kNoAccess) {
+  if (access == OS::MemoryPermission::kNoAccess ||
+      access == OS::MemoryPermission::kNoAccessWillJitLater) {
 #if !V8_OS_AIX && !V8_OS_FREEBSD && !V8_OS_QNX
     flags |= MAP_NORESERVE;
 #endif  // !V8_OS_AIX && !V8_OS_FREEBSD && !V8_OS_QNX
@@ -147,7 +155,8 @@ int GetFlagsForMemoryPermission(OS::MemoryPermission access,
   // hardened runtime/memory protection is enabled, which is optional (via code
   // signing) on Intel-based Macs but mandatory on Apple silicon ones. See also
   // https://developer.apple.com/documentation/apple-silicon/porting-just-in-time-compilers-to-apple-silicon.
-  if (access == OS::MemoryPermission::kNoAccessWillJitLater) {
+  if (access == OS::MemoryPermission::kNoAccessWillJitLater ||
+      access == OS::MemoryPermission::kReadWriteExecute) {
     flags |= MAP_JIT;
   }
 #endif  // V8_OS_DARWIN
@@ -184,7 +193,7 @@ void* Allocate(void* hint, size_t size, OS::MemoryPermission access,
   return result;
 }
 
-#endif  // !V8_OS_FUCHSIA
+#endif  // !V8_OS_FUCHSIA && !V8_OS_ZOS
 
 }  // namespace
 
@@ -251,25 +260,64 @@ bool OS::ArmUsingHardFloat() {
 #endif  // def __arm__
 #endif
 
-void PosixInitializeCommon(bool hard_abort, const char* const gc_fake_mmap) {
-  g_hard_abort = hard_abort;
+void PosixInitializeCommon(AbortMode abort_mode,
+                           const char* const gc_fake_mmap) {
+  g_abort_mode = abort_mode;
   g_gc_fake_mmap = gc_fake_mmap;
 }
 
 #if !V8_OS_FUCHSIA
-void OS::Initialize(bool hard_abort, const char* const gc_fake_mmap) {
-  PosixInitializeCommon(hard_abort, gc_fake_mmap);
+void OS::Initialize(AbortMode abort_mode, const char* const gc_fake_mmap) {
+  PosixInitializeCommon(abort_mode, gc_fake_mmap);
 }
 #endif  // !V8_OS_FUCHSIA
 
+bool OS::IsHardwareEnforcedShadowStacksEnabled() { return false; }
+
+void OS::EnsureAlternativeSignalStackIsAvailableForCurrentThread() {
+// sigaltstack() is forbidden on tvOS and its usage causes build errors.
+#if !V8_OS_TVOS
+  stack_t ss, old_ss;
+  memset(&ss, 0, sizeof(stack_t));
+  memset(&old_ss, 0, sizeof(stack_t));
+
+  if (sigaltstack(nullptr, &old_ss) == -1) {
+    FATAL("sigaltstack failed with error %d: %s", errno, strerror(errno));
+  }
+
+  // If an alternative stack is already registered, there's nothing left to do.
+  if (!(old_ss.ss_flags & SS_DISABLE)) return;
+
+  // The default alternative stack size of SIGSTKSZ can be too small (e.g.
+  // 8192). For example the profiler signal handler requires a larger stack.
+  const size_t kStackSize = 1024 * 1024;
+  CHECK_GE(kStackSize, SIGSTKSZ);
+  // Allocate the alternative stack with guard regions on both sides to
+  // ensure that we catch stack overflows (and similar issues) on that stack.
+  const size_t kPageSize = AllocatePageSize();
+  const size_t kAllocationSize = kStackSize + 2 * kPageSize;
+  void* ptr = Allocate(nullptr, kAllocationSize, kPageSize,
+                       MemoryPermission::kNoAccess);
+  CHECK_NE(ptr, nullptr);
+  void* stack_ptr = reinterpret_cast<char*>(ptr) + kPageSize;
+  CHECK(SetPermissions(stack_ptr, kStackSize, MemoryPermission::kReadWrite));
+
+  ss.ss_sp = stack_ptr;
+  ss.ss_size = kStackSize;
+  ss.ss_flags = 0;
+
+  if (sigaltstack(&ss, nullptr) == -1) {
+    FATAL("sigaltstack failed with error %d: %s", errno, strerror(errno));
+  }
+#endif
+}
+
 int OS::ActivationFrameAlignment() {
 #if V8_TARGET_ARCH_ARM
   // On EABI ARM targets this is required for fp correctness in the
   // runtime system.
   return 8;
-#elif V8_TARGET_ARCH_MIPS
-  return 8;
-#elif V8_TARGET_ARCH_S390
+#elif V8_TARGET_ARCH_S390X
   return 8;
 #else
   // Otherwise we just assume 16 byte alignment, i.e.:
@@ -329,21 +377,27 @@ void* OS::GetRandomMmapAddr() {
   raw_addr &= 0x007fffff0000ULL;
   raw_addr += 0x7e8000000000ULL;
 #else
-#if V8_TARGET_ARCH_X64 || V8_TARGET_ARCH_ARM64
+#if V8_TARGET_ARCH_X64
   // Currently available CPUs have 48 bits of virtual addressing.  Truncate
   // the hint address to 46 bits to give the kernel a fighting chance of
   // fulfilling our placement request.
   raw_addr &= uint64_t{0x3FFFFFFFF000};
+#elif V8_TARGET_ARCH_ARM64
+#if defined(V8_TARGET_OS_LINUX) || defined(V8_TARGET_OS_ANDROID)
+  // On Linux, the default virtual address space is limited to 39 bits when
+  // using 4KB pages, see arch/arm64/Kconfig. We truncate to 38 bits.
+  raw_addr &= uint64_t{0x3FFFFFF000};
+#else
+  // On macOS and elsewhere, we use 46 bits, same as on x64.
+  raw_addr &= uint64_t{0x3FFFFFFFF000};
+#endif
 #elif V8_TARGET_ARCH_PPC64
 #if V8_OS_AIX
-  // AIX: 64 bits of virtual addressing, but we limit address range to:
-  //   a) minimize Segment Lookaside Buffer (SLB) misses and
+  // AIX: 64 bits of virtual addressing, but we limit address range to minimize
+  // Segment Lookaside Buffer (SLB) misses.
   raw_addr &= uint64_t{0x3FFFF000};
   // Use extra address space to isolate the mmap regions.
   raw_addr += uint64_t{0x400000000000};
-#elif V8_TARGET_BIG_ENDIAN
-  // Big-endian Linux: 42 bits of virtual addressing.
-  raw_addr &= uint64_t{0x03FFFFFFF000};
 #else
   // Little-endian Linux: 46 bits of virtual addressing.
   raw_addr &= uint64_t{0x3FFFFFFF0000};
@@ -353,10 +407,6 @@ void* OS::GetRandomMmapAddr() {
   // of virtual addressing.  Truncate to 40 bits to allow kernel chance to
   // fulfill request.
   raw_addr &= uint64_t{0xFFFFFFF000};
-#elif V8_TARGET_ARCH_S390
-  // 31 bits of virtual addressing.  Truncate to 29 bits to allow kernel chance
-  // to fulfill request.
-  raw_addr &= 0x1FFFF000;
 #elif V8_TARGET_ARCH_MIPS64
   // 42 bits of virtual addressing. Truncate to 40 bits to allow kernel chance
   // to fulfill request.
@@ -404,6 +454,7 @@ void* OS::GetRandomMmapAddr() {
 
 // TODO(bbudge) Move Cygwin and Fuchsia stuff into platform-specific files.
 #if !V8_OS_CYGWIN && !V8_OS_FUCHSIA
+#if !V8_OS_ZOS
 // static
 void* OS::Allocate(void* hint, size_t size, size_t alignment,
                    MemoryPermission access) {
@@ -541,10 +592,8 @@ bool OS::RecommitPages(void* address, size_t size, MemoryPermission access) {
 #if defined(V8_OS_DARWIN)
   while (madvise(address, size, MADV_FREE_REUSE) == -1 && errno == EAGAIN) {
   }
-  return true;
-#else
-  return SetPermissions(address, size, access);
 #endif  // defined(V8_OS_DARWIN)
+  return true;
 }
 
 // static
@@ -605,12 +654,27 @@ bool OS::DecommitPages(void* address, size_t size) {
   return true;
 }
 #endif  // !defined(_AIX)
+#endif  // !V8_OS_ZOS
+
+// static
+bool OS::SealPages(void* address, size_t size) {
+#ifdef V8_ENABLE_MEMORY_SEALING
+#if V8_OS_LINUX && defined(__NR_mseal)
+  long ret = syscall(__NR_mseal, address, size, 0);
+  return ret == 0;
+#else
+  return false;
+#endif
+#else  // V8_ENABLE_MEMORY_SEALING
+  return false;
+#endif
+}
 
 // static
 bool OS::CanReserveAddressSpace() { return true; }
 
 // static
-Optional<AddressSpaceReservation> OS::CreateAddressSpaceReservation(
+std::optional<AddressSpaceReservation> OS::CreateAddressSpaceReservation(
     void* hint, size_t size, size_t alignment,
     MemoryPermission max_permission) {
   // On POSIX, address space reservations are backed by private memory mappings.
@@ -672,6 +736,7 @@ void OS::DestroySharedMemoryHandle(PlatformSharedMemoryHandle handle) {
 }
 #endif  // !defined(V8_OS_DARWIN)
 
+#if !V8_OS_ZOS
 // static
 bool OS::HasLazyCommits() {
 #if V8_OS_AIX || V8_OS_LINUX || V8_OS_DARWIN
@@ -681,6 +746,7 @@ bool OS::HasLazyCommits() {
   return false;
 #endif
 }
+#endif  // !V8_OS_ZOS
 #endif  // !V8_OS_CYGWIN && !V8_OS_FUCHSIA
 
 const char* OS::GetGCFakeMMapFile() {
@@ -694,8 +760,15 @@ void OS::Sleep(TimeDelta interval) {
 
 
 void OS::Abort() {
-  if (g_hard_abort) {
-    IMMEDIATE_CRASH();
+  switch (g_abort_mode) {
+    case AbortMode::kExitWithSuccessAndIgnoreDcheckFailures:
+      _exit(0);
+    case AbortMode::kExitWithFailureAndIgnoreDcheckFailures:
+      _exit(-1);
+    case AbortMode::kImmediateCrash:
+      IMMEDIATE_CRASH();
+    case AbortMode::kDefault:
+      break;
   }
   // Redirect to std abort to signal abnormal program termination.
   abort();
@@ -713,13 +786,15 @@ void OS::DebugBreak() {
   asm("break");
 #elif V8_HOST_ARCH_LOONG64
   asm("break 0");
-#elif V8_HOST_ARCH_PPC || V8_HOST_ARCH_PPC64
+#elif V8_HOST_ARCH_PPC64
   asm("twge 2,2");
 #elif V8_HOST_ARCH_IA32
   asm("int $3");
 #elif V8_HOST_ARCH_X64
   asm("int $3");
-#elif V8_HOST_ARCH_S390
+#elif V8_OS_ZOS
+  asm(" dc x'0001'");
+#elif V8_HOST_ARCH_S390X
   // Software breakpoint instruction is 0x0001
   asm volatile(".word 0x0001");
 #elif V8_HOST_ARCH_RISCV64
@@ -731,7 +806,7 @@ void OS::DebugBreak() {
 #endif
 }
 
-
+#if !V8_OS_ZOS
 class PosixMemoryMappedFile final : public OS::MemoryMappedFile {
  public:
   PosixMemoryMappedFile(FILE* file, void* memory, size_t size)
@@ -801,14 +876,13 @@ PosixMemoryMappedFile::~PosixMemoryMappedFile() {
   if (memory_) OS::Free(memory_, RoundUp(size_, OS::AllocatePageSize()));
   fclose(file_);
 }
-
+#endif  // !V8_OS_ZOS
 
 int OS::GetCurrentProcessId() {
   return static_cast<int>(getpid());
 }
 
-
-int OS::GetCurrentThreadId() {
+int OS::GetCurrentThreadIdInternal() {
 #if V8_OS_DARWIN || (V8_OS_ANDROID && defined(__APPLE__))
   return static_cast<int>(pthread_mach_thread_np(pthread_self()));
 #elif V8_OS_LINUX
@@ -821,6 +895,8 @@ int OS::GetCurrentThreadId() {
   return static_cast<int>(zx_thread_self());
 #elif V8_OS_SOLARIS
   return static_cast<int>(pthread_self());
+#elif V8_OS_ZOS
+  return gettid();
 #else
   return static_cast<int>(reinterpret_cast<intptr_t>(pthread_self()));
 #endif
@@ -853,6 +929,9 @@ int OS::GetPeakMemoryUsageKb() {
 #if defined(V8_OS_FUCHSIA)
   // Fuchsia does not implement getrusage()
   return -1;
+#elif defined(V8_OS_ZOS)
+  // TODO(v8:342445981): zos - rusage struct doesn't yet include ru_maxrss
+  return -1;
 #else
   struct rusage usage;
   if (getrusage(RUSAGE_SELF, &usage) < 0) return -1;
@@ -964,6 +1043,7 @@ void OS::PrintError(const char* format, ...) {
   va_start(args, format);
   VPrintError(format, args);
   va_end(args);
+  fflush(stderr);
 }
 
 
@@ -1015,7 +1095,8 @@ void OS::StrNCpy(char* dest, int length, const char* src, size_t n) {
 
 #if !V8_OS_CYGWIN && !V8_OS_FUCHSIA
 
-Optional<AddressSpaceReservation> AddressSpaceReservation::CreateSubReservation(
+std::optional<AddressSpaceReservation>
+AddressSpaceReservation::CreateSubReservation(
     void* address, size_t size, OS::MemoryPermission max_permission) {
   DCHECK(Contains(address, size));
   DCHECK_EQ(0, size % OS::AllocatePageSize());
@@ -1049,6 +1130,8 @@ bool AddressSpaceReservation::Free(void* address, size_t size) {
   return OS::DecommitPages(address, size);
 }
 
+// z/OS specific implementation in platform-zos.cc.
+#if !defined(V8_OS_ZOS)
 // Darwin specific implementation in platform-darwin.cc.
 #if !defined(V8_OS_DARWIN)
 bool AddressSpaceReservation::AllocateShared(void* address, size_t size,
@@ -1068,6 +1151,7 @@ bool AddressSpaceReservation::FreeShared(void* address, size_t size) {
   return mmap(address, size, PROT_NONE, MAP_FIXED | MAP_ANONYMOUS | MAP_PRIVATE,
               -1, 0) == address;
 }
+#endif  // !V8_OS_ZOS
 
 bool AddressSpaceReservation::SetPermissions(void* address, size_t size,
                                              OS::MemoryPermission access) {
@@ -1146,6 +1230,10 @@ static void SetThreadName(const char* name) {
 }
 
 static void* ThreadEntry(void* arg) {
+  // Reset denormals state to default, in case we picked up a non-default one
+  // with the posix clone() call.
+  base::FPU::SetFlushDenormals(false);
+
   Thread* thread = reinterpret_cast<Thread*>(arg);
   // We take the lock here to make sure that pthread_create finished first since
   // we don't know which thread will run first (the original thread or the new
@@ -1166,7 +1254,7 @@ static void* ThreadEntry(void* arg) {
     case Thread::Priority::kDefault:
       break;
   }
-#elif V8_OS_LINUX
+#elif V8_OS_LINUX || V8_OS_ZOS
   switch (thread->priority()) {
     case Thread::Priority::kBestEffort:
       setpriority(PRIO_PROCESS, 0, 10);
@@ -1319,6 +1407,18 @@ bool MainThreadIsCurrentThread() {
 
 // static
 Stack::StackSlot Stack::ObtainCurrentThreadStackStart() {
+#if V8_OS_ZOS
+  return __get_stack_start();
+#elif V8_OS_OPENBSD
+  stack_t stack;
+  int error = pthread_stackseg_np(pthread_self(), &stack);
+  if(error) {
+    DCHECK(MainThreadIsCurrentThread());
+    return nullptr;
+  }
+  void* stack_start = reinterpret_cast<uint8_t*>(stack.ss_sp) + stack.ss_size;
+  return stack_start;
+#else
   pthread_attr_t attr;
   int error = pthread_getattr_np(pthread_self(), &attr);
   if (error) {
@@ -1348,6 +1448,7 @@ Stack::StackSlot Stack::ObtainCurrentThreadStackStart() {
   }
 #endif  // !defined(V8_LIBC_GLIBC)
   return stack_start;
+#endif  // V8_OS_ZOS
 }
 
 #endif  // !defined(V8_OS_FREEBSD) && !defined(V8_OS_DARWIN) &&
