@@ -7,6 +7,7 @@
 #include <atomic>
 
 #include "include/v8-platform.h"
+#include "src/base/fpu.h"
 #include "src/base/platform/mutex.h"
 #include "src/base/platform/time.h"
 #include "src/codegen/compiler.h"
@@ -92,18 +93,19 @@ namespace {
 // into that. Since we have two optional slots (preparse data and job), this
 // gets a little messy.
 void SetUncompiledDataJobPointer(LocalIsolate* isolate,
-                                 Handle<SharedFunctionInfo> shared_info,
+                                 DirectHandle<SharedFunctionInfo> shared_info,
                                  Address job_address) {
-  Tagged<UncompiledData> uncompiled_data = shared_info->uncompiled_data();
-  switch (uncompiled_data->map(isolate)->instance_type()) {
+  Tagged<UncompiledData> uncompiled_data =
+      shared_info->uncompiled_data(isolate);
+  switch (uncompiled_data->map()->instance_type()) {
     // The easy cases -- we already have a job slot, so can write into it and
     // return.
     case UNCOMPILED_DATA_WITH_PREPARSE_DATA_AND_JOB_TYPE:
-      UncompiledDataWithPreparseDataAndJob::cast(uncompiled_data)
+      Cast<UncompiledDataWithPreparseDataAndJob>(uncompiled_data)
           ->set_job(job_address);
       break;
     case UNCOMPILED_DATA_WITHOUT_PREPARSE_DATA_WITH_JOB_TYPE:
-      UncompiledDataWithoutPreparseDataWithJob::cast(uncompiled_data)
+      Cast<UncompiledDataWithoutPreparseDataWithJob>(uncompiled_data)
           ->set_job(job_address);
       break;
 
@@ -113,10 +115,10 @@ void SetUncompiledDataJobPointer(LocalIsolate* isolate,
     case UNCOMPILED_DATA_WITH_PREPARSE_DATA_TYPE: {
       Handle<String> inferred_name(uncompiled_data->inferred_name(), isolate);
       Handle<PreparseData> preparse_data(
-          UncompiledDataWithPreparseData::cast(uncompiled_data)
+          Cast<UncompiledDataWithPreparseData>(uncompiled_data)
               ->preparse_data(),
           isolate);
-      Handle<UncompiledDataWithPreparseDataAndJob> new_uncompiled_data =
+      DirectHandle<UncompiledDataWithPreparseDataAndJob> new_uncompiled_data =
           isolate->factory()->NewUncompiledDataWithPreparseDataAndJob(
               inferred_name, uncompiled_data->start_position(),
               uncompiled_data->end_position(), preparse_data);
@@ -128,10 +130,11 @@ void SetUncompiledDataJobPointer(LocalIsolate* isolate,
     case UNCOMPILED_DATA_WITHOUT_PREPARSE_DATA_TYPE: {
       DCHECK(IsUncompiledDataWithoutPreparseData(uncompiled_data));
       Handle<String> inferred_name(uncompiled_data->inferred_name(), isolate);
-      Handle<UncompiledDataWithoutPreparseDataWithJob> new_uncompiled_data =
-          isolate->factory()->NewUncompiledDataWithoutPreparseDataWithJob(
-              inferred_name, uncompiled_data->start_position(),
-              uncompiled_data->end_position());
+      DirectHandle<UncompiledDataWithoutPreparseDataWithJob>
+          new_uncompiled_data =
+              isolate->factory()->NewUncompiledDataWithoutPreparseDataWithJob(
+                  inferred_name, uncompiled_data->start_position(),
+                  uncompiled_data->end_position());
 
       new_uncompiled_data->set_job(job_address);
       shared_info->set_uncompiled_data(*new_uncompiled_data);
@@ -160,7 +163,7 @@ void LazyCompileDispatcher::Enqueue(
   SetUncompiledDataJobPointer(isolate, shared_info,
                               reinterpret_cast<Address>(job));
 
-  // Post a a background worker task to perform the compilation on the worker
+  // Post a background worker task to perform the compilation on the worker
   // thread.
   {
     base::MutexGuard lock(&mutex_);
@@ -181,15 +184,16 @@ void LazyCompileDispatcher::Enqueue(
 }
 
 bool LazyCompileDispatcher::IsEnqueued(
-    Handle<SharedFunctionInfo> function) const {
+    DirectHandle<SharedFunctionInfo> shared) const {
+  if (!shared->HasUncompiledData(isolate_)) return false;
   Job* job = nullptr;
-  Tagged<Object> function_data = function->function_data(kAcquireLoad);
-  if (IsUncompiledDataWithPreparseDataAndJob(function_data)) {
+  Tagged<UncompiledData> data = shared->uncompiled_data(isolate_);
+  if (IsUncompiledDataWithPreparseDataAndJob(data)) {
     job = reinterpret_cast<Job*>(
-        UncompiledDataWithPreparseDataAndJob::cast(function_data)->job());
-  } else if (IsUncompiledDataWithoutPreparseDataWithJob(function_data)) {
+        Cast<UncompiledDataWithPreparseDataAndJob>(data)->job());
+  } else if (IsUncompiledDataWithoutPreparseDataWithJob(data)) {
     job = reinterpret_cast<Job*>(
-        UncompiledDataWithoutPreparseDataWithJob::cast(function_data)->job());
+        Cast<UncompiledDataWithoutPreparseDataWithJob>(data)->job());
   }
   return job != nullptr;
 }
@@ -210,7 +214,8 @@ void LazyCompileDispatcher::WaitForJobIfRunningOnBackground(
       // vector.
       pending_background_jobs_.erase(
           std::remove(pending_background_jobs_.begin(),
-                      pending_background_jobs_.end(), job));
+                      pending_background_jobs_.end(), job),
+          pending_background_jobs_.end());
       job->state = Job::State::kPendingToRunOnForeground;
       NotifyRemovedBackgroundJob(lock);
     } else {
@@ -222,7 +227,8 @@ void LazyCompileDispatcher::WaitForJobIfRunningOnBackground(
       // TODO(leszeks): Remove from finalizable jobs without walking the whole
       // vector.
       finalizable_jobs_.erase(
-          std::remove(finalizable_jobs_.begin(), finalizable_jobs_.end(), job));
+          std::remove(finalizable_jobs_.begin(), finalizable_jobs_.end(), job),
+          finalizable_jobs_.end());
       job->state = Job::State::kFinalizingNow;
     }
     return;
@@ -240,11 +246,13 @@ void LazyCompileDispatcher::WaitForJobIfRunningOnBackground(
   // TODO(leszeks): Remove from finalizable jobs without walking the whole
   // vector.
   finalizable_jobs_.erase(
-      std::remove(finalizable_jobs_.begin(), finalizable_jobs_.end(), job));
+      std::remove(finalizable_jobs_.begin(), finalizable_jobs_.end(), job),
+      finalizable_jobs_.end());
   job->state = Job::State::kFinalizingNow;
 }
 
-bool LazyCompileDispatcher::FinishNow(Handle<SharedFunctionInfo> function) {
+bool LazyCompileDispatcher::FinishNow(
+    DirectHandle<SharedFunctionInfo> function) {
   TRACE_EVENT0(TRACE_DISABLED_BY_DEFAULT("v8.compile"),
                "V8.LazyCompilerDispatcherFinishNow");
   RCS_SCOPE(isolate_, RuntimeCallCounterId::kCompileFinishNowOnDispatcher);
@@ -281,7 +289,7 @@ bool LazyCompileDispatcher::FinishNow(Handle<SharedFunctionInfo> function) {
       job->task.get(), isolate_, Compiler::KEEP_EXCEPTION);
   job->state = Job::State::kFinalized;
 
-  DCHECK_NE(success, isolate_->has_pending_exception());
+  DCHECK_NE(success, isolate_->has_exception());
   DeleteJob(job);
 
   // Opportunistically finalize all other jobs for a maximum time of
@@ -295,7 +303,8 @@ bool LazyCompileDispatcher::FinishNow(Handle<SharedFunctionInfo> function) {
   return success;
 }
 
-void LazyCompileDispatcher::AbortJob(Handle<SharedFunctionInfo> shared_info) {
+void LazyCompileDispatcher::AbortJob(
+    DirectHandle<SharedFunctionInfo> shared_info) {
   if (trace_compiler_dispatcher_) {
     PrintF("LazyCompileDispatcher: aborting job for ");
     ShortPrint(*shared_info);
@@ -316,7 +325,8 @@ void LazyCompileDispatcher::AbortJob(Handle<SharedFunctionInfo> shared_info) {
 
       pending_background_jobs_.erase(
           std::remove(pending_background_jobs_.begin(),
-                      pending_background_jobs_.end(), job));
+                      pending_background_jobs_.end(), job),
+          pending_background_jobs_.end());
       job->state = Job::State::kAbortingNow;
       NotifyRemovedBackgroundJob(lock);
     } else if (job->state == Job::State::kReadyToFinalize) {
@@ -325,7 +335,8 @@ void LazyCompileDispatcher::AbortJob(Handle<SharedFunctionInfo> shared_info) {
           1);
 
       finalizable_jobs_.erase(
-          std::remove(finalizable_jobs_.begin(), finalizable_jobs_.end(), job));
+          std::remove(finalizable_jobs_.begin(), finalizable_jobs_.end(), job),
+          finalizable_jobs_.end());
       job->state = Job::State::kAbortingNow;
     } else {
       UNREACHABLE();
@@ -368,14 +379,15 @@ void LazyCompileDispatcher::AbortAll() {
 }
 
 LazyCompileDispatcher::Job* LazyCompileDispatcher::GetJobFor(
-    Handle<SharedFunctionInfo> shared, const base::MutexGuard&) const {
-  Tagged<Object> function_data = shared->function_data(kAcquireLoad);
-  if (IsUncompiledDataWithPreparseDataAndJob(function_data)) {
+    DirectHandle<SharedFunctionInfo> shared, const base::MutexGuard&) const {
+  if (!shared->HasUncompiledData(isolate_)) return nullptr;
+  Tagged<UncompiledData> data = shared->uncompiled_data(isolate_);
+  if (IsUncompiledDataWithPreparseDataAndJob(data)) {
     return reinterpret_cast<Job*>(
-        UncompiledDataWithPreparseDataAndJob::cast(function_data)->job());
-  } else if (IsUncompiledDataWithoutPreparseDataWithJob(function_data)) {
+        Cast<UncompiledDataWithPreparseDataAndJob>(data)->job());
+  } else if (IsUncompiledDataWithoutPreparseDataWithJob(data)) {
     return reinterpret_cast<Job*>(
-        UncompiledDataWithoutPreparseDataWithJob::cast(function_data)->job());
+        Cast<UncompiledDataWithoutPreparseDataWithJob>(data)->job());
   }
   return nullptr;
 }
@@ -394,6 +406,8 @@ void LazyCompileDispatcher::ScheduleIdleTaskFromAnyThread(
 }
 
 void LazyCompileDispatcher::DoBackgroundWork(JobDelegate* delegate) {
+  base::FlushDenormalsScope flush_denormals_scope(isolate_->flush_denormals());
+
   TRACE_EVENT0(TRACE_DISABLED_BY_DEFAULT("v8.compile"),
                "V8.LazyCompileDispatcherDoBackgroundWork");
 
