@@ -84,7 +84,8 @@ void SimplifiedLoweringVerifier::ReportInvalidTypeCombination(
 }
 
 bool IsModuloTruncation(const Truncation& truncation) {
-  return truncation.IsUsedAsWord32() || truncation.IsUsedAsWord64() ||
+  return truncation.IsUsedAsWord32() ||
+         (Is64() && truncation.IsUsedAsWord64()) ||
          Truncation::Any().IsLessGeneralThan(truncation);
 }
 
@@ -464,6 +465,7 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(DeoptimizeUnless)
       CASE(TrapIf)
       CASE(TrapUnless)
+      CASE(Assert)
       CASE(TailCall)
       CASE(Terminate)
       CASE(Throw)
@@ -475,6 +477,7 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(NumberConstant)
       CASE(PointerConstant)
       CASE(CompressedHeapConstant)
+      CASE(TrustedHeapConstant)
       CASE(RelocatableInt32Constant)
       CASE(RelocatableInt64Constant)
       // Inner operators
@@ -512,12 +515,14 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(ChangeUint32ToTagged)
       CASE(ChangeFloat64ToTagged)
       CASE(ChangeFloat64ToTaggedPointer)
+      CASE(ChangeFloat64OrUndefinedToTagged)
       CASE(ChangeTaggedToBit)
       CASE(ChangeBitToTagged)
       CASE(ChangeInt64ToBigInt)
       CASE(ChangeUint64ToBigInt)
       CASE(TruncateTaggedToWord32)
       CASE(TruncateTaggedToFloat64)
+      CASE(TruncateTaggedToFloat64PreserveUndefined)
       CASE(TruncateTaggedPointerToBit)
       // Simplified checked operators
       CASE(CheckedInt32Add)
@@ -527,6 +532,8 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(CheckedUint32Div)
       CASE(CheckedUint32Mod)
       CASE(CheckedInt32Mul)
+      CASE(CheckedAdditiveSafeIntegerAdd)
+      CASE(CheckedAdditiveSafeIntegerSub)
       CASE(CheckedInt64Add)
       CASE(CheckedInt64Sub)
       CASE(CheckedInt64Mul)
@@ -534,6 +541,7 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(CheckedInt64Mod)
       CASE(CheckedInt32ToTaggedSigned)
       CASE(CheckedInt64ToInt32)
+      CASE(CheckedInt64ToAdditiveSafeInteger)
       CASE(CheckedInt64ToTaggedSigned)
       CASE(CheckedUint32Bounds)
       CASE(CheckedUint32ToInt32)
@@ -542,12 +550,14 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(CheckedUint64ToInt32)
       CASE(CheckedUint64ToInt64)
       CASE(CheckedUint64ToTaggedSigned)
+      CASE(CheckedFloat64ToAdditiveSafeInteger)
       CASE(CheckedFloat64ToInt64)
       CASE(CheckedTaggedSignedToInt32)
       CASE(CheckedTaggedToInt32)
       CASE(CheckedTaggedToArrayIndex)
       CASE(CheckedTruncateTaggedToWord32)
       CASE(CheckedTaggedToFloat64)
+      CASE(CheckedTaggedToAdditiveSafeInteger)
       CASE(CheckedTaggedToInt64)
       SIMPLIFIED_COMPARE_BINOP_LIST(CASE)
       SIMPLIFIED_NUMBER_BINOP_LIST(CASE)
@@ -649,7 +659,9 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(BitcastWord32ToWord64)
       CASE(ChangeInt64ToFloat64)
       CASE(ChangeUint32ToFloat64)
+      CASE(ChangeFloat16RawBitsToFloat64)
       CASE(TruncateFloat64ToFloat32)
+      CASE(TruncateFloat64ToFloat16RawBits)
       CASE(TruncateInt64ToInt32)
       CASE(RoundFloat64ToInt32)
       CASE(RoundInt32ToFloat32)
@@ -672,8 +684,8 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(Float64Select)
       CASE(LoadStackCheckOffset)
       CASE(LoadFramePointer)
-      CASE(LoadStackPointer)
-      CASE(SetStackPointer)
+      IF_WASM(CASE, LoadStackPointer)
+      IF_WASM(CASE, SetStackPointer)
       CASE(LoadParentFramePointer)
       CASE(LoadRootRegister)
       CASE(UnalignedLoad)
@@ -705,6 +717,7 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(JSAsyncFunctionReject)
       CASE(JSAsyncFunctionResolve)
       CASE(JSCallRuntime)
+      CASE(JSDetachContextCell)
       CASE(JSForInEnumerate)
       CASE(JSForInNext)
       CASE(JSForInPrepare)
@@ -716,7 +729,7 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
       CASE(JSGetImportMeta)
       CASE(JSGeneratorStore)
       CASE(JSGeneratorRestoreContinuation)
-      CASE(JSGeneratorRestoreContext)
+      CASE(JSGeneratorRestoreContextNoCell)
       CASE(JSGeneratorRestoreRegister)
       CASE(JSGeneratorRestoreInputOrDebugPos)
       CASE(JSFulfillPromise)
@@ -731,7 +744,7 @@ void SimplifiedLoweringVerifier::VisitNode(Node* node,
         break;
       }
       MACHINE_SIMD128_OP_LIST(CASE)
-      MACHINE_SIMD256_OP_LIST(CASE)
+      IF_WASM(MACHINE_SIMD256_OP_LIST, CASE)
       IF_WASM(SIMPLIFIED_WASM_OP_LIST, CASE) {
         // SIMD operators should not be in the graph, yet.
         UNREACHABLE();
