@@ -6,6 +6,7 @@
 #define V8_COMPILER_TURBOSHAFT_TYPE_INFERENCE_ANALYSIS_H_
 
 #include <limits>
+#include <optional>
 
 #include "src/base/logging.h"
 #include "src/base/vector.h"
@@ -41,15 +42,15 @@ class TypeInferenceAnalysis {
       : graph_(graph),
         // TODO(nicohartmann@): Might put types back into phase_zone once we
         // don't store them in the graph anymore.
-        types_(graph.op_id_count(), Type{}, graph.graph_zone()),
+        types_(graph.op_id_count(), Type{}, graph.graph_zone(), &graph),
         table_(phase_zone),
-        op_to_key_mapping_(phase_zone),
-        block_to_snapshot_mapping_(graph.block_count(), base::nullopt,
+        op_to_key_mapping_(phase_zone, &graph),
+        block_to_snapshot_mapping_(graph.block_count(), std::nullopt,
                                    phase_zone),
         predecessors_(phase_zone),
         graph_zone_(graph.graph_zone()) {}
 
-  GrowingSidetable<Type> Run(
+  GrowingOpIndexSidetable<Type> Run(
       GrowingBlockSidetable<std::vector<std::pair<OpIndex, Type>>>*
           block_refinements = nullptr) {
 #ifdef DEBUG
@@ -97,7 +98,7 @@ class TypeInferenceAnalysis {
     {
       predecessors_.clear();
       for (const Block* pred : block.PredecessorsIterable()) {
-        base::Optional<table_t::Snapshot> pred_snapshot =
+        std::optional<table_t::Snapshot> pred_snapshot =
             block_to_snapshot_mapping_[pred->index()];
         if (pred_snapshot.has_value()) {
           predecessors_.push_back(pred_snapshot.value());
@@ -130,8 +131,8 @@ class TypeInferenceAnalysis {
 
     // Check if the predecessor is a branch that allows us to refine a few
     // types.
-    DCHECK_IMPLIES(revisit_loop_header, block.HasExactlyNPredecessors(2));
-    if (block.HasExactlyNPredecessors(1)) {
+    DCHECK_IMPLIES(revisit_loop_header, block.PredecessorCount() == 2);
+    if (block.PredecessorCount() == 1) {
       Block* predecessor = block.LastPredecessor();
       const Operation& terminator = predecessor->LastOperation(graph_);
       if (const BranchOp* branch = terminator.TryCast<BranchOp>()) {
@@ -163,6 +164,7 @@ class TypeInferenceAnalysis {
         case Opcode::kDebugPrint:
 #if V8_ENABLE_WEBASSEMBLY
         case Opcode::kGlobalSet:
+        case Opcode::kWasmIncCoverageCounter:
         case Opcode::kTrapIf:
 #endif
         case Opcode::kCheckException:
@@ -189,7 +191,7 @@ class TypeInferenceAnalysis {
           ProcessProjection(index, op.Cast<ProjectionOp>());
           break;
         case Opcode::kWordBinop:
-          ProcessWordBinop(index, op.Cast<WordBinopOp>());
+          ProcessWordBinop(V<Word>::Cast(index), op.Cast<WordBinopOp>());
           break;
         case Opcode::kWord32PairBinop:
         case Opcode::kAtomicWord32Pair:
@@ -222,115 +224,7 @@ class TypeInferenceAnalysis {
           break;
         }
 
-        case Opcode::kWordUnary:
-        case Opcode::kFloatUnary:
-        case Opcode::kShift:
-        case Opcode::kEqual:
-        case Opcode::kChange:
-        case Opcode::kChangeOrDeopt:
-        case Opcode::kTryChange:
-        case Opcode::kBitcastWord32PairToFloat64:
-        case Opcode::kTaggedBitcast:
-        case Opcode::kSelect:
-        case Opcode::kLoad:
-        case Opcode::kAtomicRMW:
-        case Opcode::kMemoryBarrier:
-        case Opcode::kAllocate:
-        case Opcode::kDecodeExternalPointer:
-        case Opcode::kStackCheck:
-        case Opcode::kParameter:
-        case Opcode::kOsrValue:
-        case Opcode::kStackPointerGreaterThan:
-        case Opcode::kStackSlot:
-        case Opcode::kFrameConstant:
-        case Opcode::kCall:
-        case Opcode::kCatchBlockBegin:
-        case Opcode::kTailCall:
-        case Opcode::kDidntThrow:
-        case Opcode::kObjectIs:
-        case Opcode::kFloatIs:
-        case Opcode::kObjectIsNumericValue:
-        case Opcode::kConvert:
-        case Opcode::kConvertUntaggedToJSPrimitive:
-        case Opcode::kConvertUntaggedToJSPrimitiveOrDeopt:
-        case Opcode::kConvertJSPrimitiveToUntagged:
-        case Opcode::kConvertJSPrimitiveToUntaggedOrDeopt:
-        case Opcode::kTruncateJSPrimitiveToUntagged:
-        case Opcode::kTruncateJSPrimitiveToUntaggedOrDeopt:
-        case Opcode::kConvertJSPrimitiveToObject:
-        case Opcode::kNewConsString:
-        case Opcode::kNewArray:
-        case Opcode::kDoubleArrayMinMax:
-        case Opcode::kLoadFieldByIndex:
-        case Opcode::kBigIntBinop:
-        case Opcode::kBigIntEqual:
-        case Opcode::kBigIntComparison:
-        case Opcode::kBigIntUnary:
-        case Opcode::kStringAt:
-#ifdef V8_INTL_SUPPORT
-        case Opcode::kStringToCaseIntl:
-#endif  // V8_INTL_SUPPORT
-        case Opcode::kStringLength:
-        case Opcode::kStringIndexOf:
-        case Opcode::kStringFromCodePointAt:
-        case Opcode::kStringSubstring:
-        case Opcode::kStringConcat:
-        case Opcode::kStringEqual:
-        case Opcode::kStringComparison:
-        case Opcode::kArgumentsLength:
-        case Opcode::kNewArgumentsElements:
-        case Opcode::kLoadTypedElement:
-        case Opcode::kLoadDataViewElement:
-        case Opcode::kLoadStackArgument:
-        case Opcode::kStoreTypedElement:
-        case Opcode::kStoreDataViewElement:
-        case Opcode::kTransitionAndStoreArrayElement:
-        case Opcode::kCompareMaps:
-        case Opcode::kCheckMaps:
-        case Opcode::kAssumeMap:
-        case Opcode::kCheckedClosure:
-        case Opcode::kCheckEqualsInternalizedString:
-        case Opcode::kLoadMessage:
-        case Opcode::kStoreMessage:
-        case Opcode::kSameValue:
-        case Opcode::kFloat64SameValue:
-        case Opcode::kFastApiCall:
-        case Opcode::kRuntimeAbort:
-        case Opcode::kEnsureWritableFastElements:
-        case Opcode::kMaybeGrowFastElements:
-        case Opcode::kTransitionElementsKind:
-        case Opcode::kFindOrderedHashEntry:
-#if V8_ENABLE_WEBASSEMBLY
-        // TODO(14108): Implement.
-        case Opcode::kGlobalGet:
-        case Opcode::kIsNull:
-        case Opcode::kNull:
-        case Opcode::kAssertNotNull:
-        case Opcode::kRttCanon:
-        case Opcode::kWasmTypeCheck:
-        case Opcode::kWasmTypeCast:
-        case Opcode::kExternInternalize:
-        case Opcode::kExternExternalize:
-        case Opcode::kStructGet:
-        case Opcode::kStructSet:
-        case Opcode::kArrayGet:
-        case Opcode::kArraySet:
-        case Opcode::kArrayLength:
-        case Opcode::kSimd128Constant:
-        case Opcode::kSimd128Binop:
-        case Opcode::kSimd128Unary:
-        case Opcode::kSimd128Shift:
-        case Opcode::kSimd128Test:
-        case Opcode::kSimd128Splat:
-        case Opcode::kSimd128Ternary:
-        case Opcode::kSimd128ExtractLane:
-        case Opcode::kSimd128ReplaceLane:
-        case Opcode::kSimd128LaneMemory:
-        case Opcode::kSimd128LoadTransform:
-        case Opcode::kSimd128Shuffle:
-        case Opcode::kStringAsWtf16:
-        case Opcode::kStringPrepareForGetCodeUnit:
-#endif
+        default:
           // TODO(nicohartmann@): Support remaining operations. For now we
           // compute fallback types.
           if (op.outputs_rep().size() > 0) {
@@ -395,6 +289,22 @@ class TypeInferenceAnalysis {
   }
 
   void ProcessConstant(OpIndex index, const ConstantOp& constant) {
+    if (constant.kind == ConstantOp::Kind::kFloat64 &&
+        constant.float64().is_hole_nan()) {
+      // TODO(nicohartmann): figure out how to type Float64 NaN holes. Typing
+      // them simply as NaN is not always correct and can lead to replacing NaN
+      // holes with regular NaNs.
+      SetType(index, Type::Any());
+      return;
+    }
+#ifdef V8_ENABLE_EXPERIMENTAL_UNDEFINED_DOUBLE
+    if (constant.kind == ConstantOp::Kind::kFloat64 &&
+        constant.float64().is_undefined_nan()) {
+      // TODO(nicohartmann): Ignore this case for now.
+      SetType(index, Type::Any());
+      return;
+    }
+#endif  // V8_ENABLE_EXPERIMENTAL_UNDEFINED_DOUBLE
     Type type = Typer::TypeConstant(constant.kind, constant.storage);
     SetType(index, type);
   }
@@ -476,7 +386,7 @@ class TypeInferenceAnalysis {
     SetType(index, result_type);
   }
 
-  void ProcessWordBinop(OpIndex index, const WordBinopOp& binop) {
+  void ProcessWordBinop(V<Word> index, const WordBinopOp& binop) {
     Type left_type = GetType(binop.left());
     Type right_type = GetType(binop.right());
 
@@ -636,12 +546,12 @@ class TypeInferenceAnalysis {
 
  private:
   const Graph& graph_;
-  GrowingSidetable<Type> types_;
+  GrowingOpIndexSidetable<Type> types_;
   using table_t = SnapshotTable<Type>;
   table_t table_;
   const Block* current_block_ = nullptr;
-  GrowingSidetable<base::Optional<table_t::Key>> op_to_key_mapping_;
-  GrowingBlockSidetable<base::Optional<table_t::Snapshot>>
+  GrowingOpIndexSidetable<std::optional<table_t::Key>> op_to_key_mapping_;
+  GrowingBlockSidetable<std::optional<table_t::Snapshot>>
       block_to_snapshot_mapping_;
   // {predecessors_} is used during merging, but we use an instance variable for
   // it, in order to save memory and not reallocate it for each merge.
