@@ -5,6 +5,8 @@
 #ifndef V8_HEAP_FACTORY_BASE_H_
 #define V8_HEAP_FACTORY_BASE_H_
 
+#include <string_view>
+
 #include "src/base/export-template.h"
 #include "src/base/strings.h"
 #include "src/common/globals.h"
@@ -25,8 +27,8 @@ class CoverageInfo;
 class DeoptimizationLiteralArray;
 class DeoptimizationFrameTranslation;
 class FixedArray;
-template <typename T>
-class FixedIntegerArray;
+template <typename T, typename Base>
+class FixedIntegerArrayBase;
 class FreshlyAllocatedBigInt;
 class FunctionLiteral;
 class HeapObject;
@@ -56,8 +58,8 @@ class FactoryBase;
 
 enum class NumberCacheMode { kIgnore, kSetOnly, kBoth };
 
-using FixedInt32Array = FixedIntegerArray<int32_t>;
-using FixedUInt32Array = FixedIntegerArray<uint32_t>;
+using FixedInt32Array = FixedIntegerArrayBase<int32_t, ByteArray>;
+using FixedUInt32Array = FixedIntegerArrayBase<uint32_t, ByteArray>;
 
 // Putting Torque-generated definitions in a superclass allows to shadow them
 // easily when they shouldn't be used and to reference them when they happen to
@@ -74,8 +76,9 @@ class EXPORT_TEMPLATE_DECLARE(V8_EXPORT_PRIVATE) TorqueGeneratedFactory {
 struct NewCodeOptions {
   CodeKind kind;
   Builtin builtin;
+  bool is_context_specialized;
   bool is_turbofanned;
-  int stack_slots;
+  uint16_t parameter_count;
   int instruction_size;
   int metadata_size;
   unsigned int inlined_bytecode_size;
@@ -83,9 +86,12 @@ struct NewCodeOptions {
   int handler_table_offset;
   int constant_pool_offset;
   int code_comments_offset;
+  int32_t jump_table_info_offset;
   int32_t unwinding_info_offset;
-  Handle<HeapObject> bytecode_or_deoptimization_data;
-  Handle<ByteArray> bytecode_offsets_or_source_position_table;
+  MaybeHandle<TrustedObject> bytecode_or_interpreter_data;
+  MaybeHandle<DeoptimizationData> deoptimization_data;
+  MaybeHandle<TrustedByteArray> bytecode_offset_table;
+  MaybeHandle<TrustedByteArray> source_position_table;
   // Either instruction_stream is set and instruction_start is kNullAddress, or
   // instruction_stream is empty and instruction_start a valid target.
   MaybeHandle<InstructionStream> instruction_stream;
@@ -97,25 +103,28 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
  public:
   Handle<Code> NewCode(const NewCodeOptions& options);
 
+  DirectHandle<CodeWrapper> NewCodeWrapper();
+
   // Converts the given boolean condition to JavaScript boolean value.
   inline Handle<Boolean> ToBoolean(bool value);
 
 #define ROOT_ACCESSOR(Type, name, CamelName) inline Handle<Type> name();
   READ_ONLY_ROOT_LIST(ROOT_ACCESSOR)
+  MUTABLE_ROOT_LIST(ROOT_ACCESSOR)
 #undef ROOT_ACCESSOR
 
   // Numbers (e.g. literals) are pretenured by the parser.
   // The return value may be a smi or a heap number.
   template <AllocationType allocation = AllocationType::kYoung>
-  inline Handle<Object> NewNumber(double value);
+  inline Handle<Number> NewNumber(double value);
   template <AllocationType allocation = AllocationType::kYoung>
-  inline Handle<Object> NewNumberFromInt(int32_t value);
+  inline Handle<Number> NewNumberFromInt(int32_t value);
   template <AllocationType allocation = AllocationType::kYoung>
-  inline Handle<Object> NewNumberFromUint(uint32_t value);
+  inline Handle<Number> NewNumberFromUint(uint32_t value);
   template <AllocationType allocation = AllocationType::kYoung>
-  inline Handle<Object> NewNumberFromSize(size_t value);
+  inline DirectHandle<Number> NewNumberFromSize(size_t value);
   template <AllocationType allocation = AllocationType::kYoung>
-  inline Handle<Object> NewNumberFromInt64(int64_t value);
+  inline DirectHandle<Number> NewNumberFromInt64(int64_t value);
   template <AllocationType allocation = AllocationType::kYoung>
   inline Handle<HeapNumber> NewHeapNumber(double value);
   template <AllocationType allocation = AllocationType::kYoung>
@@ -123,6 +132,9 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
   template <AllocationType allocation = AllocationType::kYoung>
   inline Handle<HeapNumber> NewHeapNumberWithHoleNaN();
 
+  template <AllocationType allocation = AllocationType::kYoung>
+  inline Handle<HeapNumber> NewHeapInt32(int32_t value);
+
   template <AllocationType allocation>
   Handle<HeapNumber> NewHeapNumber();
 
@@ -136,10 +148,18 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
   Handle<FixedArray> NewFixedArray(
       int length, AllocationType allocation = AllocationType::kYoung);
 
+  // Allocates a trusted fixed array in trusted space, initialized with zeros.
+  Handle<TrustedFixedArray> NewTrustedFixedArray(
+      int length, AllocationType allocation = AllocationType::kTrusted);
+
+  // Allocates a protected fixed array in trusted space, initialized with zeros.
+  Handle<ProtectedFixedArray> NewProtectedFixedArray(int length,
+                                                     bool shared = false);
+
   // Allocates a fixed array-like object with given map and initialized with
   // undefined values.
   Handle<FixedArray> NewFixedArrayWithMap(
-      Handle<Map> map, int length,
+      DirectHandle<Map> map, int length,
       AllocationType allocation = AllocationType::kYoung);
 
   // Allocate a new fixed array with non-existing entries (the hole).
@@ -147,7 +167,7 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
       int length, AllocationType allocation = AllocationType::kYoung);
 
   // Allocate a new fixed array with Tagged<Smi>(0) entries.
-  Handle<FixedArray> NewFixedArrayWithZeroes(
+  DirectHandle<FixedArray> NewFixedArrayWithZeroes(
       int length, AllocationType allocation = AllocationType::kYoung);
 
   // Allocate a new uninitialized fixed double array.
@@ -168,21 +188,36 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
   Handle<WeakFixedArray> NewWeakFixedArray(
       int length, AllocationType allocation = AllocationType::kYoung);
 
+  // Allocates a trusted weak fixed array in trusted space, initialized with
+  // zeros.
+  Handle<TrustedWeakFixedArray> NewTrustedWeakFixedArray(int length);
+
+  // Allocates a protected weak fixed array in trusted space, initialized with
+  // zeros.
+  Handle<ProtectedWeakFixedArray> NewProtectedWeakFixedArray(int length);
+
   // The function returns a pre-allocated empty byte array for length = 0.
   Handle<ByteArray> NewByteArray(
       int length, AllocationType allocation = AllocationType::kYoung);
 
-  Handle<ExternalPointerArray> NewExternalPointerArray(
-      int length, AllocationType allocation = AllocationType::kYoung);
+  // Allocates a trusted byte array in trusted space, initialized with zeros.
+  Handle<TrustedByteArray> NewTrustedByteArray(
+      int length, AllocationType allocation_type = AllocationType::kTrusted);
 
-  Handle<DeoptimizationLiteralArray> NewDeoptimizationLiteralArray(int length);
-  Handle<DeoptimizationFrameTranslation> NewDeoptimizationFrameTranslation(
+  DirectHandle<DeoptimizationLiteralArray> NewDeoptimizationLiteralArray(
       int length);
+  DirectHandle<DeoptimizationFrameTranslation>
+  NewDeoptimizationFrameTranslation(int length);
 
-  Handle<BytecodeArray> NewBytecodeArray(int length,
-                                         const uint8_t* raw_bytecodes,
-                                         int frame_size, int parameter_count,
-                                         Handle<FixedArray> constant_pool);
+  Handle<BytecodeArray> NewBytecodeArray(
+      int length, const uint8_t* raw_bytecodes, int frame_size,
+      uint16_t parameter_count, uint16_t max_arguments,
+      DirectHandle<TrustedFixedArray> constant_pool,
+      DirectHandle<TrustedByteArray> handler_table,
+      AllocationType allocation = AllocationType::kTrusted);
+
+  DirectHandle<BytecodeWrapper> NewBytecodeWrapper(
+      AllocationType allocation = AllocationType::kOld);
 
   // Allocates a fixed array for name-value pairs of boilerplate properties and
   // calculates the number of properties we need to store in the backing store.
@@ -191,57 +226,67 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
 
   // Create a new ArrayBoilerplateDescription struct.
   Handle<ArrayBoilerplateDescription> NewArrayBoilerplateDescription(
-      ElementsKind elements_kind, Handle<FixedArrayBase> constant_values);
+      ElementsKind elements_kind, DirectHandle<FixedArrayBase> constant_values);
 
-  Handle<RegExpBoilerplateDescription> NewRegExpBoilerplateDescription(
-      Handle<FixedArray> data, Handle<String> source, Tagged<Smi> flags);
+  DirectHandle<RegExpDataWrapper> NewRegExpDataWrapper();
+
+  DirectHandle<RegExpBoilerplateDescription> NewRegExpBoilerplateDescription(
+      DirectHandle<RegExpData> data, DirectHandle<String> source,
+      Tagged<Smi> flags);
 
   // Create a new TemplateObjectDescription struct.
   Handle<TemplateObjectDescription> NewTemplateObjectDescription(
-      Handle<FixedArray> raw_strings, Handle<FixedArray> cooked_strings);
+      DirectHandle<FixedArray> raw_strings,
+      DirectHandle<FixedArray> cooked_strings);
 
   Handle<Script> NewScript(
-      Handle<PrimitiveHeapObject> source,
+      DirectHandle<UnionOf<String, Undefined>> source,
       ScriptEventType event_type = ScriptEventType::kCreate);
   Handle<Script> NewScriptWithId(
-      Handle<PrimitiveHeapObject> source, int script_id,
+      DirectHandle<UnionOf<String, Undefined>> source, int script_id,
       ScriptEventType event_type = ScriptEventType::kCreate);
 
-  Handle<SloppyArgumentsElements> NewSloppyArgumentsElements(
-      int length, Handle<Context> context, Handle<FixedArray> arguments,
+  DirectHandle<SloppyArgumentsElements> NewSloppyArgumentsElements(
+      int length, DirectHandle<Context> context,
+      DirectHandle<FixedArray> arguments,
       AllocationType allocation = AllocationType::kYoung);
-  Handle<ArrayList> NewArrayList(
+  DirectHandle<ArrayList> NewArrayList(
       int size, AllocationType allocation = AllocationType::kYoung);
 
   Handle<SharedFunctionInfo> NewSharedFunctionInfoForLiteral(
-      FunctionLiteral* literal, Handle<Script> script, bool is_toplevel);
+      FunctionLiteral* literal, DirectHandle<Script> script, bool is_toplevel);
 
   // Create a copy of a given SharedFunctionInfo for use as a placeholder in
   // off-thread compilation
   Handle<SharedFunctionInfo> CloneSharedFunctionInfo(
-      Handle<SharedFunctionInfo> other);
+      DirectHandle<SharedFunctionInfo> other);
+
+  DirectHandle<SharedFunctionInfoWrapper> NewSharedFunctionInfoWrapper(
+      DirectHandle<SharedFunctionInfo> sfi);
 
   Handle<PreparseData> NewPreparseData(int data_length, int children_length);
 
-  Handle<UncompiledDataWithoutPreparseData>
-  NewUncompiledDataWithoutPreparseData(Handle<String> inferred_name,
+  DirectHandle<UncompiledDataWithoutPreparseData>
+  NewUncompiledDataWithoutPreparseData(DirectHandle<String> inferred_name,
                                        int32_t start_position,
                                        int32_t end_position);
 
-  Handle<UncompiledDataWithPreparseData> NewUncompiledDataWithPreparseData(
-      Handle<String> inferred_name, int32_t start_position,
-      int32_t end_position, Handle<PreparseData>);
+  DirectHandle<UncompiledDataWithPreparseData>
+  NewUncompiledDataWithPreparseData(DirectHandle<String> inferred_name,
+                                    int32_t start_position,
+                                    int32_t end_position,
+                                    DirectHandle<PreparseData>);
 
-  Handle<UncompiledDataWithoutPreparseDataWithJob>
-  NewUncompiledDataWithoutPreparseDataWithJob(Handle<String> inferred_name,
-                                              int32_t start_position,
-                                              int32_t end_position);
+  DirectHandle<UncompiledDataWithoutPreparseDataWithJob>
+  NewUncompiledDataWithoutPreparseDataWithJob(
+      DirectHandle<String> inferred_name, int32_t start_position,
+      int32_t end_position);
 
-  Handle<UncompiledDataWithPreparseDataAndJob>
-  NewUncompiledDataWithPreparseDataAndJob(Handle<String> inferred_name,
+  DirectHandle<UncompiledDataWithPreparseDataAndJob>
+  NewUncompiledDataWithPreparseDataAndJob(DirectHandle<String> inferred_name,
                                           int32_t start_position,
                                           int32_t end_position,
-                                          Handle<PreparseData>);
+                                          DirectHandle<PreparseData>);
 
   // Allocates a FeedbackMetadata object and zeroes the data section.
   Handle<FeedbackMetadata> NewFeedbackMetadata(
@@ -262,7 +307,7 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
       base::Vector<const uint8_t> str, uint32_t raw_hash_field);
   Handle<SeqTwoByteString> NewTwoByteInternalizedString(
       base::Vector<const base::uc16> str, uint32_t raw_hash_field);
-  Handle<SeqOneByteString> NewOneByteInternalizedStringFromTwoByte(
+  DirectHandle<SeqOneByteString> NewOneByteInternalizedStringFromTwoByte(
       base::Vector<const base::uc16> str, uint32_t raw_hash_field);
 
   Handle<SeqOneByteString> AllocateRawOneByteInternalizedString(
@@ -284,26 +329,39 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
         .ToHandleChecked();
   }
 
+  inline Handle<String> NewStringFromAsciiChecked(
+      std::string_view str,
+      AllocationType allocation = AllocationType::kYoung) {
+    return NewStringFromOneByte(base::OneByteVector(str.data(), str.length()),
+                                allocation)
+        .ToHandleChecked();
+  }
+
   // Allocates and partially initializes an one-byte or two-byte String. The
   // characters of the string are uninitialized. Currently used in regexp code
   // only, where they are pretenured.
   V8_WARN_UNUSED_RESULT MaybeHandle<SeqOneByteString> NewRawOneByteString(
-      int length, AllocationType allocation = AllocationType::kYoung);
+      int length, AllocationType allocation = AllocationType::kYoung,
+      AllocationHint hint = AllocationHint());
   V8_WARN_UNUSED_RESULT MaybeHandle<SeqTwoByteString> NewRawTwoByteString(
-      int length, AllocationType allocation = AllocationType::kYoung);
+      int length, AllocationType allocation = AllocationType::kYoung,
+      AllocationHint hint = AllocationHint());
   // Create a new cons string object which consists of a pair of strings.
-  V8_WARN_UNUSED_RESULT MaybeHandle<String> NewConsString(
-      Handle<String> left, Handle<String> right,
+  template <template <typename> typename HandleType>
+    requires(std::is_convertible_v<HandleType<String>, DirectHandle<String>>)
+  V8_WARN_UNUSED_RESULT HandleType<String>::MaybeType NewConsString(
+      HandleType<String> left, HandleType<String> right,
       AllocationType allocation = AllocationType::kYoung);
 
   V8_WARN_UNUSED_RESULT Handle<String> NewConsString(
-      Handle<String> left, Handle<String> right, int length, bool one_byte,
-      AllocationType allocation = AllocationType::kYoung);
+      DirectHandle<String> left, DirectHandle<String> right, int length,
+      bool one_byte, AllocationType allocation = AllocationType::kYoung);
 
   V8_WARN_UNUSED_RESULT Handle<String> NumberToString(
-      Handle<Object> number, NumberCacheMode mode = NumberCacheMode::kBoth);
-  V8_WARN_UNUSED_RESULT Handle<String> HeapNumberToString(
-      Handle<HeapNumber> number, double value,
+      DirectHandle<Object> number,
+      NumberCacheMode mode = NumberCacheMode::kBoth);
+  V8_WARN_UNUSED_RESULT Handle<String> DoubleToString(
+      double value, bool canonicalize = true,
       NumberCacheMode mode = NumberCacheMode::kBoth);
   V8_WARN_UNUSED_RESULT Handle<String> SmiToString(
       Tagged<Smi> number, NumberCacheMode mode = NumberCacheMode::kBoth);
@@ -316,13 +374,13 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
   // Allocates a new BigInt with {length} digits. Only to be used by
   // MutableBigInt::New*.
   Handle<FreshlyAllocatedBigInt> NewBigInt(
-      int length, AllocationType allocation = AllocationType::kYoung);
+      uint32_t length, AllocationType allocation = AllocationType::kYoung);
 
   // Create a serialized scope info.
   Handle<ScopeInfo> NewScopeInfo(int length,
                                  AllocationType type = AllocationType::kOld);
 
-  Handle<SourceTextModuleInfo> NewSourceTextModuleInfo();
+  DirectHandle<SourceTextModuleInfo> NewSourceTextModuleInfo();
 
   Handle<DescriptorArray> NewDescriptorArray(
       int number_of_descriptors, int slack = 0,
@@ -337,13 +395,20 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
   Handle<SwissNameDictionary> NewSwissNameDictionaryWithCapacity(
       int capacity, AllocationType allocation);
 
-  Handle<FunctionTemplateRareData> NewFunctionTemplateRareData();
+  DirectHandle<FunctionTemplateRareData> NewFunctionTemplateRareData();
 
-  MaybeHandle<Map> GetInPlaceInternalizedStringMap(Tagged<Map> from_string_map);
+  MaybeDirectHandle<Map> GetInPlaceInternalizedStringMap(
+      Tagged<Map> from_string_map);
 
   AllocationType RefineAllocationTypeForInPlaceInternalizableString(
       AllocationType allocation, Tagged<Map> string_map);
 
+#ifdef V8_ENABLE_LEAPTIERING
+  JSDispatchHandle NewJSDispatchHandle(uint16_t parameter_count,
+                                       DirectHandle<Code> code,
+                                       JSDispatchTable::Space* space);
+#endif
+
  protected:
   // Must be large enough to fit any double, int, or size_t.
   static constexpr int kNumberToStringBufferSize = 32;
@@ -363,36 +428,44 @@ class FactoryBase : public TorqueGeneratedFactory<Impl> {
 
   Tagged<HeapObject> AllocateRawWithImmortalMap(
       int size, AllocationType allocation, Tagged<Map> map,
-      AllocationAlignment alignment = kTaggedAligned);
+      AllocationAlignment alignment = kTaggedAligned,
+      AllocationHint hint = AllocationHint());
   Tagged<HeapObject> NewWithImmortalMap(Tagged<Map> map,
                                         AllocationType allocation);
 
-  Handle<FixedArray> NewFixedArrayWithFiller(Handle<Map> map, int length,
-                                             Handle<HeapObject> filler,
+  Handle<FixedArray> NewFixedArrayWithFiller(DirectHandle<Map> map, int length,
+                                             DirectHandle<HeapObject> filler,
                                              AllocationType allocation);
 
   Handle<SharedFunctionInfo> NewSharedFunctionInfo(AllocationType allocation);
   Handle<SharedFunctionInfo> NewSharedFunctionInfo(
-      MaybeHandle<String> maybe_name,
-      MaybeHandle<HeapObject> maybe_function_data, Builtin builtin,
+      MaybeDirectHandle<String> maybe_name,
+      MaybeDirectHandle<HeapObject> maybe_function_data, Builtin builtin,
+      int len, AdaptArguments adapt,
       FunctionKind kind = FunctionKind::kNormalFunction);
 
   Handle<String> MakeOrFindTwoCharacterString(uint16_t c1, uint16_t c2);
 
   template <typename SeqStringT>
-  MaybeHandle<SeqStringT> NewRawStringWithMap(int length, Tagged<Map> map,
-                                              AllocationType allocation);
+  MaybeHandle<SeqStringT> NewRawStringWithMap(
+      int length, Tagged<Map> map, AllocationType allocation,
+      AllocationHint hint = AllocationHint());
 
  private:
   Impl* impl() { return static_cast<Impl*>(this); }
   auto isolate() { return impl()->isolate(); }
   ReadOnlyRoots read_only_roots() { return impl()->read_only_roots(); }
 
-  Tagged<HeapObject> AllocateRaw(
-      int size, AllocationType allocation,
-      AllocationAlignment alignment = kTaggedAligned);
+  Tagged<HeapObject> AllocateRaw(int size, AllocationType allocation,
+                                 AllocationAlignment alignment = kTaggedAligned,
+                                 AllocationHint hint = AllocationHint());
 
   friend TorqueGeneratedFactory<Impl>;
+  template <class Derived, class Shape, class Super>
+  friend class TaggedArrayBase;
+  template <class Derived, class Shape, class Super>
+  friend class PrimitiveArrayBase;
+  friend class DoubleStringCache;
 };
 
 extern template class EXPORT_TEMPLATE_DECLARE(V8_EXPORT_PRIVATE)
