@@ -8,8 +8,8 @@
 #include "src/common/globals.h"
 #include "src/heap/free-list-inl.h"
 #include "src/heap/heap.h"
-#include "src/heap/memory-chunk-inl.h"
-#include "src/heap/page-inl.h"
+#include "src/heap/mutable-page-metadata-inl.h"
+#include "src/heap/page-metadata-inl.h"
 #include "src/objects/free-space-inl.h"
 
 namespace v8 {
@@ -18,13 +18,17 @@ namespace internal {
 // -----------------------------------------------------------------------------
 // Free lists for old object spaces implementation
 
-void FreeListCategory::Reset(FreeList* owner) {
+void FreeListCategory::Unlink(FreeList* owner) {
   if (is_linked(owner) && !top().is_null()) {
     owner->DecreaseAvailableBytes(available_);
   }
-  set_top(FreeSpace());
   set_prev(nullptr);
   set_next(nullptr);
+}
+
+void FreeListCategory::Reset(FreeList* owner) {
+  Unlink(owner);
+  set_top(Tagged<FreeSpace>());
   available_ = 0;
 }
 
@@ -32,10 +36,10 @@ Tagged<FreeSpace> FreeListCategory::PickNodeFromList(size_t minimum_size,
                                                      size_t* node_size) {
   Tagged<FreeSpace> node = top();
   DCHECK(!node.is_null());
-  DCHECK(Page::FromHeapObject(node)->CanAllocate());
+  DCHECK(MemoryChunk::FromHeapObject(node)->CanAllocate());
   if (static_cast<size_t>(node->Size()) < minimum_size) {
     *node_size = 0;
-    return FreeSpace();
+    return Tagged<FreeSpace>();
   }
   set_top(node->next());
   *node_size = node->Size();
@@ -48,7 +52,7 @@ Tagged<FreeSpace> FreeListCategory::SearchForNodeInList(size_t minimum_size,
   Tagged<FreeSpace> prev_non_evac_node;
   for (Tagged<FreeSpace> cur_node = top(); !cur_node.is_null();
        cur_node = cur_node->next()) {
-    DCHECK(Page::FromHeapObject(cur_node)->CanAllocate());
+    DCHECK(MemoryChunk::FromHeapObject(cur_node)->CanAllocate());
     size_t size = cur_node->size(kRelaxedLoad);
     if (size >= minimum_size) {
       DCHECK_GE(available_, size);
@@ -57,9 +61,18 @@ Tagged<FreeSpace> FreeListCategory::SearchForNodeInList(size_t minimum_size,
         set_top(cur_node->next());
       }
       if (!prev_non_evac_node.is_null()) {
-        CodePageMemoryModificationScope code_modification_scope(
-            BasicMemoryChunk::FromHeapObject(prev_non_evac_node));
-        prev_non_evac_node->set_next(cur_node->next());
+        if (MemoryChunk::FromHeapObject(prev_non_evac_node)->executable()) {
+          WritableJitPage jit_page(prev_non_evac_node->address(),
+                                   prev_non_evac_node->Size());
+          WritableFreeSpace free_space = jit_page.FreeRange(
+              prev_non_evac_node->address(), prev_non_evac_node->Size());
+          prev_non_evac_node->SetNext(free_space, cur_node->next());
+        } else {
+          prev_non_evac_node->SetNext(
+              WritableFreeSpace::ForNonExecutableMemory(
+                  prev_non_evac_node->address(), prev_non_evac_node->Size()),
+              cur_node->next());
+        }
       }
       *node_size = size;
       return cur_node;
@@ -67,20 +80,17 @@ Tagged<FreeSpace> FreeListCategory::SearchForNodeInList(size_t minimum_size,
 
     prev_non_evac_node = cur_node;
   }
-  return FreeSpace();
+  return Tagged<FreeSpace>();
 }
 
-void FreeListCategory::Free(Address start, size_t size_in_bytes, FreeMode mode,
-                            FreeList* owner) {
+void FreeListCategory::Free(const WritableFreeSpace& writable_free_space,
+                            FreeMode mode, FreeList* owner) {
   Tagged<FreeSpace> free_space =
-      FreeSpace::cast(HeapObject::FromAddress(start));
-  DCHECK_EQ(free_space->Size(), size_in_bytes);
-  {
-    CodePageMemoryModificationScope memory_modification_scope(
-        BasicMemoryChunk::FromAddress(start));
-    free_space->set_next(top());
-  }
+      Cast<FreeSpace>(HeapObject::FromAddress(writable_free_space.Address()));
+  DCHECK_EQ(free_space->Size(), writable_free_space.Size());
+  free_space->SetNext(writable_free_space, top());
   set_top(free_space);
+  size_t size_in_bytes = writable_free_space.Size();
   available_ += size_in_bytes;
   if (mode == kLinkCategory) {
     if (is_linked(owner)) {
@@ -113,6 +123,11 @@ void FreeListCategory::Relink(FreeList* owner) {
 // ------------------------------------------------
 // Generic FreeList methods (alloc/free related)
 
+FreeList::FreeList(int number_of_categories, size_t min_block_size)
+    : number_of_categories_(number_of_categories),
+      last_category_(number_of_categories - 1),
+      min_block_size_(min_block_size) {}
+
 std::unique_ptr<FreeList> FreeList::CreateFreeList() {
   return std::make_unique<FreeListManyCachedOrigin>();
 }
@@ -125,7 +140,7 @@ Tagged<FreeSpace> FreeList::TryFindNodeIn(FreeListCategoryType type,
                                           size_t minimum_size,
                                           size_t* node_size) {
   FreeListCategory* category = categories_[type];
-  if (category == nullptr) return FreeSpace();
+  if (category == nullptr) return Tagged<FreeSpace>();
   Tagged<FreeSpace> node = category->PickNodeFromList(minimum_size, node_size);
   if (!node.is_null()) {
     DecreaseAvailableBytes(*node_size);
@@ -157,8 +172,10 @@ Tagged<FreeSpace> FreeList::SearchForNodeInList(FreeListCategoryType type,
   return node;
 }
 
-size_t FreeList::Free(Address start, size_t size_in_bytes, FreeMode mode) {
-  Page* page = Page::FromAddress(start);
+size_t FreeList::Free(const WritableFreeSpace& free_space, FreeMode mode) {
+  Address start = free_space.Address();
+  size_t size_in_bytes = free_space.Size();
+  PageMetadata* page = PageMetadata::FromAddress(start);
   page->DecreaseAllocatedBytes(size_in_bytes);
 
   // Blocks have to be a minimum size to hold free list items.
@@ -170,7 +187,7 @@ size_t FreeList::Free(Address start, size_t size_in_bytes, FreeMode mode) {
   // Insert other blocks at the head of a free list of the appropriate
   // magnitude.
   FreeListCategoryType type = SelectFreeListCategoryType(size_in_bytes);
-  page->free_list_category(type)->Free(start, size_in_bytes, mode, this);
+  page->free_list_category(type)->Free(free_space, mode, this);
   DCHECK_EQ(page->AvailableInFreeList(),
             page->AvailableInFreeListFromAllocatedBytes());
   return 0;
@@ -181,34 +198,18 @@ size_t FreeList::Free(Address start, size_t size_in_bytes, FreeMode mode) {
 
 constexpr unsigned int FreeListMany::categories_min[kNumberOfCategories];
 
-FreeListMany::FreeListMany() {
+FreeListMany::FreeListMany() : FreeList(kNumberOfCategories, kMinBlockSize) {
   // Initializing base (FreeList) fields
-  number_of_categories_ = kNumberOfCategories;
-  last_category_ = number_of_categories_ - 1;
-  min_block_size_ = kMinBlockSize;
   categories_ = new FreeListCategory*[number_of_categories_]();
-
   Reset();
 }
 
 FreeListMany::~FreeListMany() { delete[] categories_; }
 
-size_t FreeListMany::GuaranteedAllocatable(size_t maximum_freed) {
-  if (maximum_freed < categories_min[0]) {
-    return 0;
-  }
-  for (int cat = kFirstCategory + 1; cat <= last_category_; cat++) {
-    if (maximum_freed < categories_min[cat]) {
-      return categories_min[cat - 1];
-    }
-  }
-  return maximum_freed;
-}
-
-Page* FreeListMany::GetPageForSize(size_t size_in_bytes) {
+PageMetadata* FreeListMany::GetPageForSize(size_t size_in_bytes) {
   FreeListCategoryType minimum_category =
       SelectFreeListCategoryType(size_in_bytes);
-  Page* page = nullptr;
+  PageMetadata* page = nullptr;
   for (int cat = minimum_category + 1; !page && cat <= last_category_; cat++) {
     page = GetPageForCategoryType(cat);
   }
@@ -236,7 +237,7 @@ Tagged<FreeSpace> FreeListMany::Allocate(size_t size_in_bytes,
   }
 
   if (!node.is_null()) {
-    Page::FromHeapObject(node)->IncreaseAllocatedBytes(*node_size);
+    PageMetadata::FromHeapObject(node)->IncreaseAllocatedBytes(*node_size);
   }
 
   VerifyAvailable();
@@ -253,6 +254,11 @@ void FreeListManyCached::Reset() {
   FreeListMany::Reset();
 }
 
+void FreeListManyCached::ResetForNonBlackAllocatedPages() {
+  ResetCache();
+  FreeListMany::ResetForNonBlackAllocatedPages();
+}
+
 bool FreeListManyCached::AddCategory(FreeListCategory* category) {
   bool was_added = FreeList::AddCategory(category);
 
@@ -282,9 +288,11 @@ void FreeListManyCached::RemoveCategory(FreeListCategory* category) {
 #endif
 }
 
-size_t FreeListManyCached::Free(Address start, size_t size_in_bytes,
+size_t FreeListManyCached::Free(const WritableFreeSpace& free_space,
                                 FreeMode mode) {
-  Page* page = Page::FromAddress(start);
+  Address start = free_space.Address();
+  size_t size_in_bytes = free_space.Size();
+  PageMetadata* page = PageMetadata::FromAddress(start);
   page->DecreaseAllocatedBytes(size_in_bytes);
 
   // Blocks have to be a minimum size to hold free list items.
@@ -296,7 +304,7 @@ size_t FreeListManyCached::Free(Address start, size_t size_in_bytes,
   // Insert other blocks at the head of a free list of the appropriate
   // magnitude.
   FreeListCategoryType type = SelectFreeListCategoryType(size_in_bytes);
-  page->free_list_category(type)->Free(start, size_in_bytes, mode, this);
+  page->free_list_category(type)->Free(free_space, mode, this);
 
   // Updating cache
   if (mode == kLinkCategory) {
@@ -342,7 +350,7 @@ Tagged<FreeSpace> FreeListManyCached::Allocate(size_t size_in_bytes,
 #endif
 
   if (!node.is_null()) {
-    Page::FromHeapObject(node)->IncreaseAllocatedBytes(*node_size);
+    PageMetadata::FromHeapObject(node)->IncreaseAllocatedBytes(*node_size);
   }
 
   VerifyAvailable();
@@ -403,7 +411,7 @@ Tagged<FreeSpace> FreeListManyCachedFastPathBase::Allocate(
 
   if (!node.is_null()) {
     if (categories_[type] == nullptr) UpdateCacheAfterRemoval(type);
-    Page::FromHeapObject(node)->IncreaseAllocatedBytes(*node_size);
+    PageMetadata::FromHeapObject(node)->IncreaseAllocatedBytes(*node_size);
   }
 
 #ifdef DEBUG
@@ -441,14 +449,33 @@ void FreeList::Reset() {
   available_ = 0;
 }
 
-size_t FreeList::EvictFreeListItems(Page* page) {
+void FreeList::ResetForNonBlackAllocatedPages() {
+  DCHECK(v8_flags.black_allocated_pages);
+  ForAllFreeListCategories([this](FreeListCategory* category) {
+    if (!category->is_empty()) {
+      auto* chunk = MemoryChunk::FromHeapObject(category->top());
+      if (chunk->IsFlagSet(MemoryChunk::BLACK_ALLOCATED)) {
+        category->Unlink(this);
+        return;
+      }
+    }
+    category->Reset(this);
+  });
+  for (int i = kFirstCategory; i < number_of_categories_; i++) {
+    categories_[i] = nullptr;
+  }
+  wasted_bytes_ = 0;
+  available_ = 0;
+}
+
+void FreeList::EvictFreeListItems(PageMetadata* page) {
   size_t sum = 0;
   page->ForAllFreeListCategories([this, &sum](FreeListCategory* category) {
     sum += category->available();
     RemoveCategory(category);
     category->Reset(this);
   });
-  return sum;
+  page->add_wasted_memory(sum);
 }
 
 void FreeList::RepairLists(Heap* heap) {
@@ -516,7 +543,7 @@ size_t FreeListCategory::SumFreeList() {
     // We can't use "cur->map()" here because both cur's map and the
     // root can be null during bootstrapping.
     DCHECK(
-        cur->map_slot().contains_map_value(Page::FromHeapObject(cur)
+        cur->map_slot().contains_map_value(PageMetadata::FromHeapObject(cur)
                                                ->heap()
                                                ->isolate()
                                                ->root(RootIndex::kFreeSpaceMap)
