@@ -7,9 +7,11 @@
 #include <inttypes.h>
 
 #include <cmath>
+#include <optional>
 
 #include "src/base/logging.h"
 #include "src/base/platform/time.h"
+#include "src/common/globals.h"
 #include "src/execution/vm-state-inl.h"
 #include "src/flags/flags.h"
 #include "src/handles/global-handles.h"
@@ -18,6 +20,9 @@
 #include "src/heap/gc-tracer-inl.h"
 #include "src/heap/gc-tracer.h"
 #include "src/heap/heap-inl.h"
+#include "src/heap/heap-layout-inl.h"
+#include "src/heap/heap-visitor-inl.h"
+#include "src/heap/heap-visitor.h"
 #include "src/heap/heap.h"
 #include "src/heap/incremental-marking-job.h"
 #include "src/heap/mark-compact.h"
@@ -25,10 +30,8 @@
 #include "src/heap/marking-visitor-inl.h"
 #include "src/heap/marking-visitor.h"
 #include "src/heap/memory-chunk-layout.h"
-#include "src/heap/memory-chunk.h"
 #include "src/heap/minor-mark-sweep.h"
-#include "src/heap/objects-visiting-inl.h"
-#include "src/heap/objects-visiting.h"
+#include "src/heap/mutable-page-metadata.h"
 #include "src/heap/safepoint.h"
 #include "src/init/v8.h"
 #include "src/logging/runtime-call-stats-scope.h"
@@ -54,10 +57,10 @@ static constexpr v8::base::TimeDelta kMaxStepSizeOnAllocation =
 
 #ifndef DEBUG
 static constexpr size_t kV8ActivationThreshold = 8 * MB;
-static constexpr size_t kEmbedderActivationThreshold = 8 * MB;
+static constexpr size_t kGlobalActivationThreshold = 8 * MB;
 #else
 static constexpr size_t kV8ActivationThreshold = 0;
-static constexpr size_t kEmbedderActivationThreshold = 0;
+static constexpr size_t kGlobalActivationThreshold = 0;
 #endif  // DEBUG
 
 base::TimeDelta GetMaxDuration(StepOrigin step_origin) {
@@ -106,10 +109,14 @@ void IncrementalMarking::MarkBlackBackground(Tagged<HeapObject> obj,
                                              int object_size) {
   CHECK(marking_state()->TryMark(obj));
   base::MutexGuard guard(&background_live_bytes_mutex_);
-  background_live_bytes_[MemoryChunk::FromHeapObject(obj)] +=
+  background_live_bytes_[MutablePageMetadata::FromHeapObject(obj)] +=
       static_cast<intptr_t>(object_size);
 }
 
+bool IncrementalMarking::CanAndShouldBeStarted() const {
+  return CanBeStarted() && heap_->ShouldUseIncrementalMarking();
+}
+
 bool IncrementalMarking::CanBeStarted() const {
   // Only start incremental marking in a safe state:
   //   1) when incremental marking is turned on
@@ -121,33 +128,49 @@ bool IncrementalMarking::CanBeStarted() const {
 }
 
 bool IncrementalMarking::IsBelowActivationThresholds() const {
-  return heap_->OldGenerationSizeOfObjects() <= kV8ActivationThreshold &&
-         heap_->EmbedderSizeOfObjects() <= kEmbedderActivationThreshold;
+  return heap_->OldGenerationConsumedBytes() <= kV8ActivationThreshold &&
+         heap_->GlobalConsumedBytes() <= kGlobalActivationThreshold;
 }
 
 void IncrementalMarking::Start(GarbageCollector garbage_collector,
                                GarbageCollectionReason gc_reason) {
-  DCHECK(CanBeStarted());
-  DCHECK(!heap_->sweeping_in_progress());
-  DCHECK(IsStopped());
+  CHECK(IsStopped());
+  CHECK_IMPLIES(garbage_collector == GarbageCollector::MARK_COMPACTOR,
+                !heap_->sweeping_in_progress());
+  CHECK_IMPLIES(garbage_collector == GarbageCollector::MINOR_MARK_SWEEPER,
+                !heap_->minor_sweeping_in_progress());
+  // Do not invoke CanAndShouldBeStarted() here again because its return value
+  // might change across multiple invocations (its internal state could be
+  // updated concurrently from another thread between invocations).
+  CHECK(CanBeStarted());
+  // The "current isolate" must be set correctly so we can access pointer
+  // tables.
+  DCHECK_EQ(isolate(), Isolate::TryGetCurrent());
 
   if (V8_UNLIKELY(v8_flags.trace_incremental_marking)) {
     const size_t old_generation_size_mb =
         heap()->OldGenerationSizeOfObjects() / MB;
+    const size_t old_generation_waste_mb =
+        heap()->OldGenerationWastedBytes() / MB;
     const size_t old_generation_limit_mb =
         heap()->old_generation_allocation_limit() / MB;
     const size_t global_size_mb = heap()->GlobalSizeOfObjects() / MB;
+    const size_t global_waste_mb = heap()->GlobalWastedBytes() / MB;
     const size_t global_limit_mb = heap()->global_allocation_limit() / MB;
     isolate()->PrintWithTimestamp(
-        "[IncrementalMarking] Start (%s): (size/limit/slack) v8: %zuMB / %zuMB "
-        "/ %zuMB global: %zuMB / %zuMB / %zuMB\n",
-        ToString(gc_reason), old_generation_size_mb, old_generation_limit_mb,
-        old_generation_size_mb > old_generation_limit_mb
+        "[IncrementalMarking] Start (%s): (size/waste/limit/slack) v8: %zuMB / "
+        "%zuMB / %zuMB "
+        "/ %zuMB global: %zuMB / %zuMB / %zuMB / %zuMB\n",
+        ToString(gc_reason), old_generation_size_mb, old_generation_waste_mb,
+        old_generation_limit_mb,
+        old_generation_size_mb + old_generation_waste_mb >
+                old_generation_limit_mb
             ? 0
             : old_generation_limit_mb - old_generation_size_mb,
-        global_size_mb, global_limit_mb,
-        global_size_mb > global_limit_mb ? 0
-                                         : global_limit_mb - global_size_mb);
+        global_size_mb, global_waste_mb, global_limit_mb,
+        global_size_mb + global_waste_mb > global_limit_mb
+            ? 0
+            : global_limit_mb - global_size_mb);
   }
 
   Counters* counters = isolate()->counters();
@@ -183,93 +206,23 @@ void IncrementalMarking::Start(GarbageCollector garbage_collector,
 
   if (is_major) {
     StartMarkingMajor();
-    heap_->allocator()->AddAllocationObserver(&old_generation_observer_,
-                                              &new_generation_observer_);
-    if (incremental_marking_job()) {
-      incremental_marking_job()->ScheduleTask();
-    }
-    DCHECK_NULL(schedule_);
-    schedule_ =
-        v8_flags.incremental_marking_bailout_when_ahead_of_schedule
-            ? ::heap::base::IncrementalMarkingSchedule::
-                  CreateWithZeroMinimumMarkedBytesPerStep(v8_flags.predictable)
-            : ::heap::base::IncrementalMarkingSchedule::
-                  CreateWithDefaultMinimumMarkedBytesPerStep(
-                      v8_flags.predictable);
-    schedule_->NotifyIncrementalMarkingStart();
   } else {
-    // Allocation observers are not currently used by MinorMS because we don't
-    // do incremental marking.
     StartMarkingMinor();
   }
 }
 
-bool IncrementalMarking::WhiteToGreyAndPush(Tagged<HeapObject> obj) {
-  if (marking_state()->TryMark(obj)) {
-    local_marking_worklists()->Push(obj);
-    return true;
-  }
-  return false;
-}
-
-class IncrementalMarking::IncrementalMarkingRootMarkingVisitor final
-    : public RootVisitor {
- public:
-  explicit IncrementalMarkingRootMarkingVisitor(Heap* heap)
-      : heap_(heap), incremental_marking_(heap->incremental_marking()) {}
-
-  void VisitRootPointer(Root root, const char* description,
-                        FullObjectSlot p) override {
-    DCHECK(!MapWord::IsPacked((*p).ptr()));
-    MarkObjectByPointer(root, p);
-  }
-
-  void VisitRootPointers(Root root, const char* description,
-                         FullObjectSlot start, FullObjectSlot end) override {
-    for (FullObjectSlot p = start; p < end; ++p) {
-      DCHECK(!MapWord::IsPacked((*p).ptr()));
-      MarkObjectByPointer(root, p);
-    }
-  }
-
- private:
-  void MarkObjectByPointer(Root root, FullObjectSlot p) {
-    Tagged<Object> object = *p;
-    if (!IsHeapObject(object)) return;
-    DCHECK(!MapWord::IsPacked(object.ptr()));
-    Tagged<HeapObject> heap_object = HeapObject::cast(object);
-
-    if (heap_object.InAnySharedSpace() || heap_object.InReadOnlySpace()) return;
-
-    if (incremental_marking_->IsMajorMarking()) {
-      if (incremental_marking_->WhiteToGreyAndPush(heap_object)) {
-        if (V8_UNLIKELY(v8_flags.track_retaining_path)) {
-          heap_->AddRetainingRoot(root, heap_object);
-        }
-      }
-    } else if (Heap::InYoungGeneration(heap_object)) {
-      incremental_marking_->WhiteToGreyAndPush(heap_object);
-    }
-  }
-
-  Heap* const heap_;
-  IncrementalMarking* const incremental_marking_;
-};
-
 void IncrementalMarking::MarkRoots() {
-  CodePageHeaderModificationScope rwx_write_scope(
-      "Marking of builtins table entries require write access to Code page "
-      "header");
   if (IsMajorMarking()) {
-    IncrementalMarkingRootMarkingVisitor visitor(heap_);
+    RootMarkingVisitor root_visitor(heap_->mark_compact_collector());
     heap_->IterateRoots(
-        &visitor,
+        &root_visitor,
         base::EnumSet<SkipRoot>{SkipRoot::kStack, SkipRoot::kMainThreadHandles,
                                 SkipRoot::kTracedHandles, SkipRoot::kWeak,
                                 SkipRoot::kReadOnlyBuiltins});
   } else {
+    DCHECK(IsMinorMarking());
     YoungGenerationRootMarkingVisitor root_visitor(
-        heap_->minor_mark_sweep_collector()->main_marking_visitor());
+        heap_->minor_mark_sweep_collector());
     heap_->IterateRoots(
         &root_visitor,
         base::EnumSet<SkipRoot>{
@@ -277,7 +230,6 @@ void IncrementalMarking::MarkRoots() {
             SkipRoot::kExternalStringTable, SkipRoot::kGlobalHandles,
             SkipRoot::kTracedHandles, SkipRoot::kOldGeneration,
             SkipRoot::kReadOnlyBuiltins});
-
     isolate()->global_handles()->IterateYoungStrongAndDependentRoots(
         &root_visitor);
   }
@@ -311,20 +263,16 @@ void IncrementalMarking::StartMarkingMajor() {
   is_compacting_ = major_collector_->StartCompaction(
       MarkCompactCollector::StartCompactionMode::kIncremental);
 
-#ifdef V8_COMPRESS_POINTERS
-  heap_->external_pointer_space()->StartCompactingIfNeeded();
-#endif  // V8_COMPRESS_POINTERS
+  // The schedule is acquired for CppHeap as well. Initialize it early.
+  schedule_ =
+      ::heap::base::IncrementalMarkingSchedule::Create(v8_flags.predictable);
+  schedule_->NotifyIncrementalMarkingStart();
 
-  if (heap_->cpp_heap()) {
-    TRACE_GC(heap()->tracer(),
-             GCTracer::Scope::MC_INCREMENTAL_EMBEDDER_PROLOGUE);
-    // PrepareForTrace should be called before visitor initialization in
-    // StartMarking.
-    CppHeap::From(heap_->cpp_heap())
-        ->InitializeTracing(CppHeap::CollectionType::kMajor);
+  if (v8_flags.incremental_marking_unified_schedule) {
+    major_collector_->StartMarking(schedule_);
+  } else {
+    major_collector_->StartMarking();
   }
-
-  major_collector_->StartMarking();
   current_local_marking_worklists_ =
       major_collector_->local_marking_worklists();
 
@@ -352,17 +300,18 @@ void IncrementalMarking::StartMarkingMajor() {
   }
 
   if (heap()->cpp_heap()) {
-    // StartTracing may call back into V8 in corner cases, requiring that
+    // `StartMarking()` may call back into V8 in corner cases, requiring that
     // marking (including write barriers) is fully set up.
-    TRACE_GC(heap()->tracer(),
-             GCTracer::Scope::MC_INCREMENTAL_EMBEDDER_PROLOGUE);
-    CppHeap::From(heap()->cpp_heap())->StartTracing();
+    TRACE_GC(heap()->tracer(), GCTracer::Scope::MC_MARK_EMBEDDER_PROLOGUE);
+    CppHeap::From(heap()->cpp_heap())->StartMarking();
   }
 
   heap_->InvokeIncrementalMarkingEpilogueCallbacks();
 
-  if (v8_flags.minor_ms && heap_->new_space()) {
-    heap_->paged_new_space()->ForceAllocationSuccessUntilNextGC();
+  heap_->allocator()->AddAllocationObserver(&old_generation_observer_,
+                                            &new_generation_observer_);
+  if (incremental_marking_job()) {
+    incremental_marking_job()->ScheduleTask();
   }
 }
 
@@ -374,7 +323,9 @@ void IncrementalMarking::StartMarkingMinor() {
         "[IncrementalMarking] (MinorMS) Start marking\n");
   }
 
-  minor_collector_->StartMarking();
+  // We only reach this code if Heap::ShouldUseBackgroundThreads() returned
+  // true. So we can force the use of background threads here.
+  minor_collector_->StartMarking(true);
   current_local_marking_worklists_ =
       minor_collector_->local_marking_worklists();
 
@@ -403,23 +354,42 @@ void IncrementalMarking::StartMarkingMinor() {
   }
 
   DCHECK(!is_compacting_);
+
+  // Allocation observers are not currently used by MinorMS because we don't
+  // do incremental marking.
 }
 
 void IncrementalMarking::StartBlackAllocation() {
   DCHECK(!black_allocation_);
   DCHECK(IsMajorMarking());
   black_allocation_ = true;
-  heap()->allocator()->MarkLinearAllocationAreaBlack();
+  if (v8_flags.black_allocated_pages) {
+    heap()->allocator()->FreeLinearAllocationAreasAndResetFreeLists();
+  } else {
+    heap()->allocator()->MarkLinearAllocationAreasBlack();
+  }
   if (isolate()->is_shared_space_isolate()) {
-    DCHECK(!heap()->shared_space()->main_allocator()->IsLabValid());
     isolate()->global_safepoint()->IterateSharedSpaceAndClientIsolates(
         [](Isolate* client) {
-          client->heap()->MarkSharedLinearAllocationAreasBlack();
+          if (v8_flags.black_allocated_pages) {
+            client->heap()->FreeSharedLinearAllocationAreasAndResetFreeLists();
+          } else {
+            client->heap()->MarkSharedLinearAllocationAreasBlack();
+          }
         });
   }
   heap()->safepoint()->IterateLocalHeaps([](LocalHeap* local_heap) {
-    local_heap->MarkLinearAllocationAreaBlack();
+    if (v8_flags.black_allocated_pages) {
+      // The freelists of the underlying spaces must anyway be empty after the
+      // first call to FreeLinearAllocationAreasAndResetFreeLists(). However,
+      // don't call FreeLinearAllocationAreas(), since it also frees the
+      // shared-space areas.
+      local_heap->FreeLinearAllocationAreasAndResetFreeLists();
+    } else {
+      local_heap->MarkLinearAllocationAreasBlack();
+    }
   });
+  StartPointerTableBlackAllocation();
   if (v8_flags.trace_incremental_marking) {
     isolate()->PrintWithTimestamp(
         "[IncrementalMarking] Black allocation started\n");
@@ -428,16 +398,21 @@ void IncrementalMarking::StartBlackAllocation() {
 
 void IncrementalMarking::PauseBlackAllocation() {
   DCHECK(IsMajorMarking());
-  heap()->allocator()->UnmarkLinearAllocationArea();
-  if (isolate()->is_shared_space_isolate()) {
-    DCHECK(!heap()->shared_space()->main_allocator()->IsLabValid());
-    isolate()->global_safepoint()->IterateSharedSpaceAndClientIsolates(
-        [](Isolate* client) {
-          client->heap()->UnmarkSharedLinearAllocationAreas();
-        });
+  if (!v8_flags.black_allocated_pages) {
+    heap()->allocator()->UnmarkLinearAllocationsArea();
+
+    if (isolate()->is_shared_space_isolate()) {
+      isolate()->global_safepoint()->IterateSharedSpaceAndClientIsolates(
+          [](Isolate* client) {
+            client->heap()->UnmarkSharedLinearAllocationAreas();
+          });
+    }
+
+    heap()->safepoint()->IterateLocalHeaps([](LocalHeap* local_heap) {
+      local_heap->UnmarkLinearAllocationsArea();
+    });
   }
-  heap()->safepoint()->IterateLocalHeaps(
-      [](LocalHeap* local_heap) { local_heap->UnmarkLinearAllocationArea(); });
+  StopPointerTableBlackAllocation();
   if (v8_flags.trace_incremental_marking) {
     isolate()->PrintWithTimestamp(
         "[IncrementalMarking] Black allocation paused\n");
@@ -446,111 +421,71 @@ void IncrementalMarking::PauseBlackAllocation() {
 }
 
 void IncrementalMarking::FinishBlackAllocation() {
-  if (black_allocation_) {
-    black_allocation_ = false;
-    if (v8_flags.trace_incremental_marking) {
-      isolate()->PrintWithTimestamp(
-          "[IncrementalMarking] Black allocation finished\n");
-    }
+  if (!black_allocation_) {
+    return;
+  }
+  // Don't fixup the marking bitmaps of the black allocated pages, since the
+  // concurrent marker may still be running and will access the page flags.
+  black_allocation_ = false;
+  StopPointerTableBlackAllocation();
+  if (v8_flags.trace_incremental_marking) {
+    isolate()->PrintWithTimestamp(
+        "[IncrementalMarking] Black allocation finished\n");
   }
 }
 
-void IncrementalMarking::UpdateMarkingWorklistAfterScavenge() {
-  if (!IsMajorMarking()) return;
-  DCHECK(!v8_flags.separate_gc_phases);
-  DCHECK(IsMajorMarking());
-  // Minor MS never runs during incremental marking.
-  DCHECK(!v8_flags.minor_ms);
-
-  Tagged<Map> filler_map = ReadOnlyRoots(heap_).one_pointer_filler_map();
-
-  MarkingState* marking_state = heap()->marking_state();
-
-  major_collector_->local_marking_worklists()->Publish();
-  MarkingBarrier::PublishAll(heap());
-  PtrComprCageBase cage_base(isolate());
-  major_collector_->marking_worklists()->Update([this, marking_state, cage_base,
-                                                 filler_map](
-                                                    Tagged<HeapObject> obj,
-                                                    Tagged<HeapObject>* out)
-                                                    -> bool {
-    DCHECK(IsHeapObject(obj));
-    USE(marking_state);
-
-    // Only pointers to from space have to be updated.
-    if (Heap::InFromPage(obj)) {
-      MapWord map_word = obj->map_word(cage_base, kRelaxedLoad);
-      if (!map_word.IsForwardingAddress()) {
-        // There may be objects on the marking deque that do not exist
-        // anymore, e.g. left trimmed objects or objects from the root set
-        // (frames). If these object are dead at scavenging time, their
-        // marking deque entries will not point to forwarding addresses.
-        // Hence, we can discard them.
-        return false;
-      }
-      // Live young large objects are not relocated and directly promoted into
-      // the old generation before invoking this method. So they looke like any
-      // other pointer into the old space and we won't encounter them here in
-      // this code path.
-      DCHECK(!Heap::IsLargeObject(obj));
-      Tagged<HeapObject> dest = map_word.ToForwardingAddress(obj);
-      DCHECK_IMPLIES(marking_state->IsUnmarked(obj), IsFreeSpaceOrFiller(obj));
-      if (dest.InWritableSharedSpace() &&
-          !isolate()->is_shared_space_isolate()) {
-        // Object got promoted into the shared heap. Drop it from the client
-        // heap marking worklist.
-        return false;
-      }
-      // For any object not a DescriptorArray, transferring the object always
-      // increments live bytes as the marked state cannot distinguish fully
-      // processed from to-be-processed. Decrement the counter for such objects
-      // here.
-      if (!IsDescriptorArray(dest)) {
-        MemoryChunk::FromHeapObject(dest)->IncrementLiveBytesAtomically(
-            -ALIGN_TO_ALLOCATION_ALIGNMENT(dest->Size()));
-      }
-      *out = dest;
-      return true;
-    } else {
-      DCHECK(!Heap::InToPage(obj));
-      DCHECK_IMPLIES(marking_state->IsUnmarked(obj),
-                     IsFreeSpaceOrFiller(obj, cage_base));
-      // Skip one word filler objects that appear on the
-      // stack when we perform in place array shift.
-      if (obj->map(cage_base) != filler_map) {
-        *out = obj;
-        return true;
-      }
-      return false;
-    }
-  });
-
-  major_collector_->local_weak_objects()->Publish();
-  weak_objects_->UpdateAfterScavenge();
+void IncrementalMarking::StartPointerTableBlackAllocation() {
+#ifdef V8_COMPRESS_POINTERS
+  heap()->old_external_pointer_space()->set_allocate_black(true);
+  heap()->cpp_heap_pointer_space()->set_allocate_black(true);
+#endif  // V8_COMPRESS_POINTERS
+#ifdef V8_ENABLE_SANDBOX
+  heap()->code_pointer_space()->set_allocate_black(true);
+  heap()->trusted_pointer_space()->set_allocate_black(true);
+  if (isolate()->is_shared_space_isolate()) {
+    isolate()->shared_trusted_pointer_space()->set_allocate_black(true);
+  }
+#endif  // V8_ENABLE_SANDBOX
+#ifdef V8_ENABLE_LEAPTIERING
+  heap()->js_dispatch_table_space()->set_allocate_black(true);
+#endif  // V8_ENABLE_LEAPTIERING
 }
 
-void IncrementalMarking::UpdateMarkedBytesAfterScavenge(
-    size_t dead_bytes_in_new_space) {
-  if (!IsMajorMarking()) return;
-  // When removing the call, adjust the marking schedule to only support
-  // monotonically increasing mutator marked bytes.
-  main_thread_marked_bytes_ -=
-      std::min(main_thread_marked_bytes_, dead_bytes_in_new_space);
+void IncrementalMarking::StopPointerTableBlackAllocation() {
+#ifdef V8_COMPRESS_POINTERS
+  heap()->old_external_pointer_space()->set_allocate_black(false);
+  heap()->cpp_heap_pointer_space()->set_allocate_black(false);
+#endif  // V8_COMPRESS_POINTERS
+#ifdef V8_ENABLE_SANDBOX
+  heap()->code_pointer_space()->set_allocate_black(false);
+  heap()->trusted_pointer_space()->set_allocate_black(false);
+  if (isolate()->is_shared_space_isolate()) {
+    heap()->isolate()->shared_trusted_pointer_space()->set_allocate_black(
+        false);
+  }
+#endif  // V8_ENABLE_SANDBOX
+#ifdef V8_ENABLE_LEAPTIERING
+  heap()->js_dispatch_table_space()->set_allocate_black(false);
+#endif  // V8_ENABLE_LEAPTIERING
 }
 
-v8::base::TimeDelta IncrementalMarking::EmbedderStep(
-    v8::base::TimeDelta expected_duration) {
+std::pair<v8::base::TimeDelta, size_t> IncrementalMarking::CppHeapStep(
+    v8::base::TimeDelta max_duration, std::optional<size_t> marked_bytes_limit,
+    StepOrigin step_origin) {
   DCHECK(IsMarking());
   auto* cpp_heap = CppHeap::From(heap_->cpp_heap());
-  DCHECK_NOT_NULL(cpp_heap);
-  if (!cpp_heap->incremental_marking_supported()) {
+  if (!cpp_heap || !cpp_heap->incremental_marking_supported()) {
     return {};
   }
 
   TRACE_GC(heap()->tracer(), GCTracer::Scope::MC_INCREMENTAL_EMBEDDER_TRACING);
   const auto start = v8::base::TimeTicks::Now();
-  cpp_heap->AdvanceTracing(expected_duration);
-  return v8::base::TimeTicks::Now() - start;
+  cpp_heap->AdvanceMarking(
+      max_duration, marked_bytes_limit,
+      step_origin == StepOrigin::kTask
+          ? cppgc::internal::StackState::kNoHeapPointers
+          : cppgc::internal::StackState::kMayContainHeapPointers);
+  return {v8::base::TimeTicks::Now() - start, cpp_heap->last_bytes_marked()};
 }
 
 bool IncrementalMarking::Stop() {
@@ -559,13 +494,18 @@ bool IncrementalMarking::Stop() {
   if (v8_flags.trace_incremental_marking) {
     int old_generation_size_mb =
         static_cast<int>(heap()->OldGenerationSizeOfObjects() / MB);
+    int old_generation_waste_mb =
+        static_cast<int>(heap()->OldGenerationWastedBytes() / MB);
     int old_generation_limit_mb =
         static_cast<int>(heap()->old_generation_allocation_limit() / MB);
     isolate()->PrintWithTimestamp(
-        "[IncrementalMarking] Stopping: old generation %dMB, limit %dMB, "
+        "[IncrementalMarking] Stopping: old generation size %dMB, waste %dMB, "
+        "limit %dMB, "
         "overshoot %dMB\n",
-        old_generation_size_mb, old_generation_limit_mb,
-        std::max(0, old_generation_size_mb - old_generation_limit_mb));
+        old_generation_size_mb, old_generation_waste_mb,
+        old_generation_limit_mb,
+        std::max(0, old_generation_size_mb + old_generation_waste_mb -
+                        old_generation_limit_mb));
   }
 
   if (IsMajorMarking()) {
@@ -599,7 +539,7 @@ bool IncrementalMarking::Stop() {
 
   // Merge live bytes counters of background threads
   for (const auto& pair : background_live_bytes_) {
-    MemoryChunk* memory_chunk = pair.first;
+    MutablePageMetadata* memory_chunk = pair.first;
     intptr_t live_bytes = pair.second;
     if (live_bytes) {
       memory_chunk->IncrementLiveBytesAtomically(live_bytes);
@@ -690,6 +630,7 @@ bool IncrementalMarking::TryInitializeTaskTimeout() {
             : allowed_overshoot - optional_time_to_current_task.value();
     completion_task_timeout_ = now + delta;
   }
+  DCHECK_IMPLIES(!delaying, completion_task_timeout_ <= now);
   if (V8_UNLIKELY(v8_flags.trace_incremental_marking)) {
     isolate()->PrintWithTimestamp(
         "[IncrementalMarking] Completion: %s GC via stack guard, "
@@ -711,8 +652,14 @@ size_t IncrementalMarking::GetScheduledBytes(StepOrigin step_origin) {
   FetchBytesMarkedConcurrently();
   // TODO(v8:14140): Consider the size including young generation here as well
   // as the full marker marks both the young and old generations.
-  const size_t max_bytes_to_process =
-      schedule_->GetNextIncrementalStepDuration(OldGenerationSizeOfObjects());
+  size_t estimated_live_bytes = OldGenerationSizeOfObjects();
+  if (v8_flags.incremental_marking_unified_schedule) {
+    if (auto* cpp_heap = CppHeap::From(heap_->cpp_heap())) {
+      estimated_live_bytes += cpp_heap->used_size();
+    }
+  }
+  const size_t marked_bytes_limit =
+      schedule_->GetNextIncrementalStepDuration(estimated_live_bytes);
   if (V8_UNLIKELY(v8_flags.trace_incremental_marking)) {
     const auto step_info = schedule_->GetCurrentStepInfo();
     isolate()->PrintWithTimestamp(
@@ -720,7 +667,7 @@ size_t IncrementalMarking::GetScheduledBytes(StepOrigin step_origin) {
         "%.1f, marked: %zuKB (mutator: %zuKB, concurrent %zuKB), expected "
         "marked: %zuKB, estimated live: %zuKB, schedule delta: %+" PRIi64
         "KB\n",
-        max_bytes_to_process / KB, ToString(step_origin),
+        marked_bytes_limit / KB, ToString(step_origin),
         step_info.elapsed_time.InMillisecondsF(), step_info.marked_bytes() / KB,
         step_info.mutator_marked_bytes / KB,
         step_info.concurrent_marked_bytes / KB,
@@ -728,7 +675,7 @@ size_t IncrementalMarking::GetScheduledBytes(StepOrigin step_origin) {
         step_info.estimated_live_bytes / KB,
         step_info.scheduled_delta_bytes() / KB);
   }
-  return max_bytes_to_process;
+  return marked_bytes_limit;
 }
 
 void IncrementalMarking::AdvanceAndFinalizeIfComplete() {
@@ -756,21 +703,6 @@ void IncrementalMarking::AdvanceForTesting(v8::base::TimeDelta max_duration,
   Step(max_duration, max_bytes_to_mark, StepOrigin::kV8);
 }
 
-bool IncrementalMarking::IsAheadOfSchedule() const {
-  DCHECK(IsMajorMarking());
-
-  const ::heap::base::IncrementalMarkingSchedule* v8_schedule = schedule_.get();
-  if (v8_schedule->GetCurrentStepInfo().is_behind_expectation()) {
-    return false;
-  }
-  if (auto* cpp_heap = CppHeap::From(heap()->cpp_heap())) {
-    if (!cpp_heap->marker()->IsAheadOfSchedule()) {
-      return false;
-    }
-  }
-  return true;
-}
-
 void IncrementalMarking::AdvanceOnAllocation() {
   DCHECK_EQ(heap_->gc_state(), Heap::NOT_IN_GC);
   DCHECK(v8_flags.incremental_marking);
@@ -818,7 +750,7 @@ void IncrementalMarking::FetchBytesMarkedConcurrently() {
 }
 
 void IncrementalMarking::Step(v8::base::TimeDelta max_duration,
-                              size_t max_bytes_to_process,
+                              size_t marked_bytes_limit,
                               StepOrigin step_origin) {
   NestedTimedHistogramScope incremental_marking_scope(
       isolate()->counters()->gc_incremental_marking());
@@ -831,7 +763,7 @@ void IncrementalMarking::Step(v8::base::TimeDelta max_duration,
   DCHECK(IsMajorMarking());
   const auto start = v8::base::TimeTicks::Now();
 
-  base::Optional<SafepointScope> safepoint_scope;
+  std::optional<SafepointScope> safepoint_scope;
   // Conceptually an incremental marking step (even though it always runs on the
   // main thread) may introduce a form of concurrent marking when background
   // threads access the heap concurrently (e.g. concurrent compilation). On
@@ -841,26 +773,31 @@ void IncrementalMarking::Step(v8::base::TimeDelta max_duration,
   // marker doesn't rely on correct synchronization but e.g. on black allocation
   // and the on_hold worklist.
 #ifndef V8_ATOMIC_OBJECT_FIELD_WRITES
-  DCHECK(!v8_flags.concurrent_marking);
-  safepoint_scope.emplace(isolate(), SafepointKind::kIsolate);
+  {
+    DCHECK(!v8_flags.concurrent_marking);
+    // Ensure that the isolate has no shared heap. Otherwise a shared GC might
+    // happen when trying to enter the safepoint.
+    DCHECK(!isolate()->has_shared_space());
+    AllowGarbageCollection allow_gc;
+    safepoint_scope.emplace(isolate(), SafepointKind::kIsolate);
+  }
 #endif
 
-  size_t v8_bytes_processed = 0;
-  v8::base::TimeDelta embedder_duration;
-  v8::base::TimeDelta max_embedder_duration;
-
-  if (v8_flags.concurrent_marking) {
+  if (V8_LIKELY(v8_flags.concurrent_marking)) {
     // It is safe to merge back all objects that were on hold to the shared
     // work list at Step because we are at a safepoint where all objects
     // are properly initialized. The exception is the last allocated object
     // before invoking an AllocationObserver. This allocation had no way to
     // escape and get marked though.
     local_marking_worklists()->MergeOnHold();
+
+    heap()->mark_compact_collector()->MaybeEnableBackgroundThreadsInCycle(
+        MarkCompactCollector::CallOrigin::kIncrementalMarkingStep);
   }
   if (step_origin == StepOrigin::kTask) {
     // We cannot publish the pending allocations for V8 step origin because the
     // last object was allocated before invoking the step.
-    heap()->PublishPendingAllocations();
+    heap()->PublishMainThreadPendingAllocations();
   }
 
   // Perform a single V8 and a single embedder step. In case both have been
@@ -869,62 +806,84 @@ void IncrementalMarking::Step(v8::base::TimeDelta max_duration,
   // This ignores that case where the embedder finds new V8-side objects. The
   // assumption is that large graphs are well connected and can mostly be
   // processed on their own. For small graphs, helping is not necessary.
-  std::tie(v8_bytes_processed, std::ignore) =
-      major_collector_->ProcessMarkingWorklist(
-          max_duration, max_bytes_to_process,
-          MarkCompactCollector::MarkingWorklistProcessingMode::kDefault);
-  main_thread_marked_bytes_ += v8_bytes_processed;
-  schedule_->UpdateMutatorThreadMarkedBytes(main_thread_marked_bytes_);
-  const auto v8_time = v8::base::TimeTicks::Now() - start;
-  if (heap_->cpp_heap() && (v8_time < max_duration)) {
-    // The CppHeap only gets the remaining slice and not the exact same time.
-    // This is fine because CppHeap will schedule its own incremental steps. We
-    // want to help out here to be able to fully finalize when all worklists
-    // have been drained.
-    max_embedder_duration = max_duration - v8_time;
-    embedder_duration = EmbedderStep(max_embedder_duration);
+  //
+  // The idea of a unified incremental marking step is the following:
+  // - We use a single schedule for both V8 and CppHeap.
+  // - Process CppHeap first in here as there's some objects in there that can
+  //   only be processed on the main thread.
+  // - Use the left over time and bytes for a V8 step.
+  // - We ignore the case where both individual steps discover new references to
+  //   each other and assume that graphs are generally well connected.
+  // - Always flush objects to enable concurrent marking to make progress.
+
+  // Start with a CppHeap step as there's objects on CppHeap that must be marked
+  // on the main thread.
+  v8::base::TimeDelta cpp_heap_duration;
+  size_t cpp_heap_marked_bytes;
+  std::optional<size_t> cpp_heap_marked_bytes_limit;
+  if (v8_flags.incremental_marking_unified_schedule) {
+    cpp_heap_marked_bytes_limit.emplace(marked_bytes_limit);
+  }
+  std::tie(cpp_heap_duration, cpp_heap_marked_bytes) =
+      CppHeapStep(max_duration, cpp_heap_marked_bytes_limit, step_origin);
+
+  // Add an optional V8 step if we are not exceeding our limits.
+  size_t v8_marked_bytes = 0;
+  v8::base::TimeDelta v8_time;
+  if (cpp_heap_duration < max_duration &&
+      (!v8_flags.incremental_marking_unified_schedule ||
+       (cpp_heap_marked_bytes < marked_bytes_limit))) {
+    const auto v8_start = v8::base::TimeTicks::Now();
+    const size_t v8_marked_bytes_limit =
+        v8_flags.incremental_marking_unified_schedule
+            ? marked_bytes_limit - cpp_heap_marked_bytes
+            : marked_bytes_limit;
+    std::tie(v8_marked_bytes, std::ignore) =
+        major_collector_->ProcessMarkingWorklist(
+            max_duration - cpp_heap_duration, v8_marked_bytes_limit);
+    v8_time = v8::base::TimeTicks::Now() - v8_start;
+    heap_->tracer()->AddIncrementalMarkingStep(v8_time.InMillisecondsF(),
+                                               v8_marked_bytes);
   }
 
-  if (v8_flags.concurrent_marking) {
+  if (V8_LIKELY(v8_flags.concurrent_marking)) {
     local_marking_worklists()->ShareWork();
     heap_->concurrent_marking()->RescheduleJobIfNeeded(
         GarbageCollector::MARK_COMPACTOR);
   }
 
-  heap_->tracer()->AddIncrementalMarkingStep(v8_time.InMillisecondsF(),
-                                             v8_bytes_processed);
-
   if (V8_UNLIKELY(v8_flags.trace_incremental_marking)) {
+    const auto v8_max_duration = max_duration - cpp_heap_duration;
+    const auto v8_marked_bytes_limit =
+        marked_bytes_limit > cpp_heap_marked_bytes
+            ? marked_bytes_limit - cpp_heap_marked_bytes
+            : 0;
     isolate()->PrintWithTimestamp(
-        "[IncrementalMarking] Step: origin: %s, V8: %zuKB (%zuKB) in %.1f, "
-        "embedder: %fms (%fms) in %.1f (%.1f), V8 marking speed: %.fMB/s\n",
-        ToString(step_origin), v8_bytes_processed / KB,
-        max_bytes_to_process / KB, v8_time.InMillisecondsF(),
-        embedder_duration.InMillisecondsF(),
-        max_embedder_duration.InMillisecondsF(),
-        (v8::base::TimeTicks::Now() - start).InMillisecondsF(),
-        max_duration.InMillisecondsF(),
+        "[IncrementalMarking] Step: origin: %s overall: %.1fms "
+        "V8: %zuKB (%zuKB), %.1fms (%.1fms), %.1fMB/s "
+        "CppHeap: %zuKB (%zuKB), %.1fms (%.1fms)\n",
+        ToString(step_origin),
+        (v8::base::TimeTicks::Now() - start).InMillisecondsF(), v8_marked_bytes,
+        v8_marked_bytes_limit, v8_time.InMillisecondsF(),
+        v8_max_duration.InMillisecondsF(),
         heap()->tracer()->IncrementalMarkingSpeedInBytesPerMillisecond() *
-            1000 / MB);
+            1000 / MB,
+        cpp_heap_marked_bytes, marked_bytes_limit,
+        cpp_heap_duration.InMillisecondsF(), max_duration.InMillisecondsF());
   }
 }
 
 Isolate* IncrementalMarking::isolate() const { return heap_->isolate(); }
 
-IncrementalMarking::PauseBlackAllocationScope::PauseBlackAllocationScope(
-    IncrementalMarking* marking)
-    : marking_(marking) {
-  if (marking_->black_allocation()) {
-    paused_ = true;
-    marking_->PauseBlackAllocation();
-  }
-}
-
-IncrementalMarking::PauseBlackAllocationScope::~PauseBlackAllocationScope() {
-  if (paused_) {
-    marking_->StartBlackAllocation();
-  }
-}
+// The allocation observer step size determines the LAB size when marking is on.
+// Objects in the LAB are not marked until the LAB bounds are reset in marking
+// steps. As a result, the concurrent marker may pick up this many bytes. If
+// kStepSizeWhenNotMakingProgress is too small, we would consider marking
+// objects in the LAB as making progress which means we would not finalize as
+// long as we allocate objects in the LAB between steps.
+static_assert(
+    ::heap::base::IncrementalMarkingSchedule::kStepSizeWhenNotMakingProgress >=
+    kMajorGCYoungGenerationAllocationObserverStep);
 
 }  // namespace internal
 }  // namespace v8
