@@ -50,7 +50,7 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   Handle<BytecodeArray> FinalizeBytecode(IsolateT* isolate,
                                          Handle<Script> script);
   template <typename IsolateT>
-  Handle<ByteArray> FinalizeSourcePositionTable(IsolateT* isolate);
+  DirectHandle<TrustedByteArray> FinalizeSourcePositionTable(IsolateT* isolate);
 
   // Check if hint2 is same or the subtype of hint1.
   static bool IsSameOrSubTypeHint(TypeHint hint1, TypeHint hint2) {
@@ -73,13 +73,14 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   void VisitModuleDeclarations(Declaration::List* declarations);
   void VisitGlobalDeclarations(Declaration::List* declarations);
   void VisitDeclarations(Declaration::List* declarations);
-  void VisitStatements(const ZonePtrList<Statement>* statments);
+  void VisitStatements(const ZonePtrList<Statement>* statments, int start = 0);
 
  private:
   class AccumulatorPreservingScope;
   class ContextScope;
   class ControlScope;
   class ControlScopeForBreakable;
+  class ControlScopeForDerivedConstructor;
   class ControlScopeForIteration;
   class ControlScopeForTopLevel;
   class ControlScopeForTryCatch;
@@ -93,11 +94,13 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   class IteratorRecord;
   class MultipleEntryBlockContextScope;
   class LoopScope;
+  class ForInScope;
   class NaryCodeCoverageSlots;
   class OptionalChainNullLabelScope;
   class RegisterAllocationScope;
   class TestResultScope;
   class TopLevelDeclarationsBuilder;
+  class DisposablesStackScope;
   class ValueResultScope;
 
   using ToBooleanMode = BytecodeArrayBuilder::ToBooleanMode;
@@ -107,7 +110,7 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
 
   // An assignment has to evaluate its LHS before its RHS, but has to assign to
   // the LHS after both evaluations are done. This class stores the data
-  // computed in the LHS evaulation that has to live across the RHS evaluation,
+  // computed in the LHS evaluation that has to live across the RHS evaluation,
   // and is used in the actual LHS assignment.
   class AssignmentLhsData {
    public:
@@ -195,6 +198,16 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   };
 
   void GenerateBytecodeBody();
+
+  void GenerateBaseConstructorBody();
+  void GenerateDerivedConstructorBody();
+  void GenerateAsyncFunctionBody();
+  void GenerateAsyncGeneratorFunctionBody();
+
+  void GenerateBodyPrologue();
+  void GenerateBodyStatements(int start = 0);
+  void GenerateBodyStatementsWithoutImplicitFinalReturn(int start = 0);
+
   template <typename IsolateT>
   void AllocateDeferredConstants(IsolateT* isolate, Handle<Script> script);
 
@@ -249,6 +262,8 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   void VisitPropertyLoadForRegister(Register obj, Property* expr,
                                     Register destination);
 
+  bool BuildInitializationBlockForParametersIfExist();
+
   AssignmentLhsData PrepareAssignmentLhs(
       Expression* lhs, AccumulatorPreservingMode accumulator_preserving_mode =
                            AccumulatorPreservingMode::kNone);
@@ -273,10 +288,15 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
                              const AstRawString* name);
   void BuildStoreGlobal(Variable* variable);
 
+  void BuildLoadKeyedProperty(Register object, FeedbackSlot slot);
+
   bool IsVariableInRegister(Variable* var, Register reg);
 
   void SetVariableInRegister(Variable* var, Register reg);
 
+  Variable* GetPotentialVariableInAccumulator();
+
+  void AddDisposableValue(VariableMode mode);
   void BuildVariableLoad(Variable* variable, HoleCheckMode hole_check_mode,
                          TypeofMode typeof_mode = TypeofMode::kNotInside);
   void BuildVariableLoadForAccumulatorValue(
@@ -350,6 +370,8 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   void VisitArgumentsObject(Variable* variable);
   void VisitRestArgumentsArray(Variable* rest);
   void VisitCallSuper(Call* call);
+  void BuildInstanceInitializationAfterSuperCall(Register this_function,
+                                                 Register instance);
   void BuildInvalidPropertyAccess(MessageTemplate tmpl, Property* property);
   void BuildPrivateBrandCheck(Property* property, Register object);
   void BuildPrivateMethodIn(Variable* private_name,
@@ -372,6 +394,7 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
                                          Register instance);
   void BuildGeneratorObjectVariableInitialization();
   void VisitBlockDeclarationsAndStatements(Block* stmt);
+  void VisitBlockMaybeDispose(Block* stmt);
   void VisitLiteralAccessor(LiteralProperty* property, Register value_out);
   void VisitForInAssignment(Expression* expr);
   void VisitModuleNamespaceImports();
@@ -413,6 +436,9 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
 
   int AllocateBlockCoverageSlotIfEnabled(AstNode* node, SourceRangeKind kind);
   int AllocateNaryBlockCoverageSlotIfEnabled(NaryOperation* node, size_t index);
+  int AllocateConditionalChainBlockCoverageSlotIfEnabled(ConditionalChain* node,
+                                                         SourceRangeKind kind,
+                                                         size_t index);
 
   void BuildIncrementBlockCoverageCounterIfEnabled(AstNode* node,
                                                    SourceRangeKind kind);
@@ -430,10 +456,16 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
                        FinallyBodyFunc finally_body_func,
                        HandlerTable::CatchPrediction catch_prediction,
                        TryFinallyStatement* stmt_for_coverage = nullptr);
+  template <typename WrappedFunc>
+  void BuildDisposeScope(WrappedFunc wrapped_func, bool has_await_using);
 
   template <typename ExpressionFunc>
   void BuildOptionalChain(ExpressionFunc expression_func);
 
+  void BuildGetAndCheckSuperConstructor(Register this_function,
+                                        Register new_target,
+                                        Register constructor,
+                                        BytecodeLabel* super_ctor_call_done);
   void BuildSuperCallOptimization(Register this_function, Register new_target,
                                   Register constructor_then_instance,
                                   BytecodeLabel* super_ctor_call_done);
@@ -442,6 +474,9 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   // register, or just getting the effect. Some visitors return a TypeHint which
   // specifies the type of the result of the visited expression.
   TypeHint VisitForAccumulatorValue(Expression* expr);
+  TypeHint VisitForAccumulatorValueAsPropertyKey(Expression* expr);
+  TypeHint VisitForAccumulatorValueImpl(Expression* expr,
+                                        ValueResultScope* accumulator_scope);
   void VisitForAccumulatorValueOrTheHole(Expression* expr);
   V8_WARN_UNUSED_RESULT Register VisitForRegisterValue(Expression* expr);
   V8_INLINE void VisitForRegisterValue(Expression* expr, Register destination);
@@ -495,6 +530,7 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
                                            : ToBooleanMode::kConvertToBoolean;
   }
 
+  inline Register incoming_new_target() const;
   inline Register generator_object() const;
 
   inline BytecodeArrayBuilder* builder() { return &builder_; }
@@ -549,6 +585,21 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
     current_loop_scope_ = loop_scope;
   }
 
+  inline ForInScope* current_for_in_scope() const {
+    return current_for_in_scope_;
+  }
+  inline void set_current_for_in_scope(ForInScope* for_in_scope) {
+    current_for_in_scope_ = for_in_scope;
+  }
+
+  Register current_disposables_stack() const {
+    SBXCHECK(current_disposables_stack_.is_valid());
+    return current_disposables_stack_;
+  }
+  void set_current_disposables_stack(Register disposables_stack) {
+    current_disposables_stack_ = disposables_stack;
+  }
+
   LocalIsolate* local_isolate_;
   Zone* zone_;
   BytecodeArrayBuilder builder_;
@@ -575,12 +626,14 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   ZoneVector<std::pair<ClassLiteral*, size_t>> class_literals_;
   ZoneVector<std::pair<GetTemplateObject*, size_t>> template_objects_;
   ZoneVector<Variable*> vars_in_hole_check_bitmap_;
+  ZoneVector<std::pair<Call*, Scope*>> eval_calls_;
 
   ControlScope* execution_control_;
   ContextScope* execution_context_;
   ExpressionResultScope* execution_result_;
 
   Register incoming_new_target_or_generator_;
+  Register current_disposables_stack_;
 
   BytecodeLabels* optional_chaining_null_labels_;
 
@@ -598,6 +651,7 @@ class BytecodeGenerator final : public AstVisitor<BytecodeGenerator> {
   Variable::HoleCheckBitmap hole_check_bitmap_;
 
   LoopScope* current_loop_scope_;
+  ForInScope* current_for_in_scope_;
 
   HandlerTable::CatchPrediction catch_prediction_;
 };
