@@ -5,14 +5,18 @@
 #ifndef V8_MAGLEV_MAGLEV_INTERPRETER_FRAME_STATE_H_
 #define V8_MAGLEV_MAGLEV_INTERPRETER_FRAME_STATE_H_
 
-#include "src/base/logging.h"
+#include <optional>
+
 #include "src/base/threaded-list.h"
 #include "src/compiler/bytecode-analysis.h"
 #include "src/compiler/bytecode-liveness-map.h"
+#include "src/compiler/js-heap-broker.h"
 #include "src/interpreter/bytecode-register.h"
 #include "src/maglev/maglev-compilation-unit.h"
 #include "src/maglev/maglev-ir.h"
+#ifdef V8_ENABLE_MAGLEV
 #include "src/maglev/maglev-regalloc-data.h"
+#endif
 #include "src/maglev/maglev-register-frame-array.h"
 #include "src/zone/zone.h"
 
@@ -83,12 +87,12 @@ class NodeInfo {
   }
 
   NodeType type() const { return type_; }
-  NodeType CombineType(NodeType other) {
-    return type_ = maglev::CombineType(type_, other);
-  }
   NodeType IntersectType(NodeType other) {
     return type_ = maglev::IntersectType(type_, other);
   }
+  NodeType UnionType(NodeType other) {
+    return type_ = maglev::UnionType(type_, other);
+  }
 
   // Optional alternative nodes with the equivalent value but a different
   // representation.
@@ -101,7 +105,7 @@ class NodeInfo {
   V(int32, Int32)                                      \
   V(truncated_int32_to_number, TruncatedInt32ToNumber) \
   V(float64, Float64)                                  \
-  V(constant, Constant)
+  V(checked_value, CheckedValue)
 
     enum Kind {
 #define KIND(name, Name) k##Name,
@@ -166,7 +170,7 @@ class NodeInfo {
   // being a node info that is the subset of information valid in both inputs.
   void MergeWith(const NodeInfo& other, Zone* zone,
                  bool& any_merged_map_is_unstable) {
-    IntersectType(other.type_);
+    UnionType(other.type_);
     alternative_.MergeWith(other.alternative_);
     if (possible_maps_are_known_) {
       if (other.possible_maps_are_known_) {
@@ -192,9 +196,21 @@ class NodeInfo {
     if (!any_map_is_unstable_) return;
     possible_maps_.clear();
     possible_maps_are_known_ = false;
+    type_ = MakeTypeStable(type_);
     any_map_is_unstable_ = false;
   }
 
+  template <typename Function>
+  void ClearUnstableMapsIfAny(const Function& condition) {
+    if (!any_map_is_unstable_) return;
+    for (auto map : possible_maps_) {
+      if (condition(map)) {
+        ClearUnstableMaps();
+        return;
+      }
+    }
+  }
+
   bool possible_maps_are_known() const { return possible_maps_are_known_; }
 
   const PossibleMaps& possible_maps() const {
@@ -205,15 +221,16 @@ class NodeInfo {
   }
 
   void SetPossibleMaps(const PossibleMaps& possible_maps,
-                       bool any_map_is_unstable, NodeType possible_type) {
+                       bool any_map_is_unstable, NodeType possible_type,
+                       compiler::JSHeapBroker* broker) {
     possible_maps_ = possible_maps;
     possible_maps_are_known_ = true;
     any_map_is_unstable_ = any_map_is_unstable;
 #ifdef DEBUG
     if (possible_maps.size()) {
-      NodeType expected = StaticTypeForMap(*possible_maps.begin());
+      NodeType expected = StaticTypeForMap(*possible_maps.begin(), broker);
       for (auto map : possible_maps) {
-        expected = maglev::IntersectType(StaticTypeForMap(map), expected);
+        expected = maglev::UnionType(StaticTypeForMap(map, broker), expected);
       }
       // Ensure the claimed type is not narrower than what can be learned from
       // the map checks.
@@ -222,11 +239,16 @@ class NodeInfo {
       DCHECK_EQ(possible_type, NodeType::kUnknown);
     }
 #endif
-    CombineType(possible_type);
+    IntersectType(possible_type);
   }
 
   bool any_map_is_unstable() const { return any_map_is_unstable_; }
 
+  void set_node_type_is_unstable() {
+    // Re-use any_map_is_unstable to signal that the node type is unstable.
+    any_map_is_unstable_ = true;
+  }
+
  private:
   NodeType type_ = NodeType::kUnknown;
 
@@ -242,18 +264,12 @@ class NodeInfo {
   AlternativeNodes alternative_;
 };
 
+struct LoopEffects;
+
 struct KnownNodeAspects {
   // Permanently valid if checked in a dominator.
   using NodeInfos = ZoneMap<ValueNode*, NodeInfo>;
 
-  explicit KnownNodeAspects(Zone* zone)
-      : any_map_for_any_node_is_unstable(false),
-        loaded_constant_properties(zone),
-        loaded_properties(zone),
-        loaded_context_constants(zone),
-        loaded_context_slots(zone),
-        node_infos(zone) {}
-
   // Copy constructor is defaulted but private so that we explicitly call the
   // Clone method.
   KnownNodeAspects& operator=(const KnownNodeAspects& other) = delete;
@@ -268,25 +284,11 @@ struct KnownNodeAspects {
   // invalidated in the loop body, and similarly stable maps will have
   // dependencies installed. Unstable maps however might be invalidated by
   // calls, and we don't know about these until it's too late.
-  KnownNodeAspects* CloneForLoopHeader(Zone* zone) const {
-    KnownNodeAspects* clone = zone->New<KnownNodeAspects>(zone);
-    if (!any_map_for_any_node_is_unstable) {
-      clone->node_infos = node_infos;
-#ifdef DEBUG
-      for (const auto& it : node_infos) {
-        DCHECK(!it.second.any_map_is_unstable());
-      }
-#endif
-    } else {
-      for (const auto& it : node_infos) {
-        clone->node_infos.emplace(it.first,
-                                  NodeInfo::ClearUnstableMapsOnCopy{it.second});
-      }
-    }
-    clone->loaded_constant_properties = loaded_constant_properties;
-    clone->loaded_context_constants = loaded_context_constants;
-    return clone;
-  }
+  KnownNodeAspects* CloneForLoopHeader(bool optimistic_initial_state,
+                                       LoopEffects* loop_effects,
+                                       Zone* zone) const;
+
+  void ClearUnstableNodeAspects(bool is_tracing_enabled);
 
   void ClearUnstableMaps() {
     // A side effect could change existing objects' maps. For stable maps we
@@ -302,6 +304,16 @@ struct KnownNodeAspects {
     any_map_for_any_node_is_unstable = false;
   }
 
+  template <typename Function>
+  void ClearUnstableMapsIfAny(const Function& condition) {
+    if (!any_map_for_any_node_is_unstable) return;
+    for (auto& it : node_infos) {
+      it.second.ClearUnstableMapsIfAny(condition);
+    }
+  }
+
+  void ClearAvailableExpressions() { available_expressions.clear(); }
+
   NodeInfos::iterator FindInfo(ValueNode* node) {
     return node_infos.find(node);
   }
@@ -316,26 +328,157 @@ struct KnownNodeAspects {
   const NodeInfo* TryGetInfoFor(ValueNode* node) const {
     return const_cast<KnownNodeAspects*>(this)->TryGetInfoFor(node);
   }
+
   NodeInfo* TryGetInfoFor(ValueNode* node) {
     auto info_it = FindInfo(node);
     if (!IsValid(info_it)) return nullptr;
     return &info_it->second;
   }
-  NodeInfo* GetOrCreateInfoFor(ValueNode* node) {
+
+  NodeInfo* GetOrCreateInfoFor(compiler::JSHeapBroker* broker,
+                               ValueNode* node) {
     auto info_it = FindInfo(node);
     if (IsValid(info_it)) return &info_it->second;
-    return &node_infos.emplace(node, NodeInfo()).first->second;
+    auto res = &node_infos.emplace(node, NodeInfo()).first->second;
+    res->IntersectType(node->GetStaticType(broker));
+    if (auto alloc = node->TryCast<InlinedAllocation>()) {
+      if (alloc->object()->has_static_map()) {
+        compiler::MapRef map = alloc->object()->map();
+        res->SetPossibleMaps(PossibleMaps{map}, !map.is_stable(),
+                             StaticTypeForMap(map, broker), broker);
+      }
+    }
+    return res;
   }
 
-  NodeType NodeTypeFor(ValueNode* node) const {
-    if (auto info = TryGetInfoFor(node)) {
-      return info->type();
+  std::optional<PossibleMaps> TryGetPossibleMaps(ValueNode* node) {
+    DCHECK_NOT_NULL(node);
+    if (NodeInfo* info = TryGetInfoFor(node)) {
+      if (info->possible_maps_are_known()) {
+        return info->possible_maps();
+      }
+      return {};
+    }
+    if (auto alloc = node->TryCast<InlinedAllocation>()) {
+      if (alloc->object()->has_static_map()) {
+        return PossibleMaps{alloc->object()->map()};
+      }
+    }
+    return {};
+  }
+
+  NodeType GetType(compiler::JSHeapBroker* broker, ValueNode* node) const {
+    // We first check the KnownNodeAspects in order to return the most precise
+    // type possible.
+    auto info = TryGetInfoFor(node);
+    if (info == nullptr) {
+      // If this node has no NodeInfo (or not known type in its NodeInfo), we
+      // fall back to its static type.
+      return node->GetStaticType(broker);
+    }
+    NodeType actual_type = info->type();
+    if (auto phi = node->TryCast<Phi>()) {
+      actual_type = IntersectType(actual_type, phi->type());
+    }
+#ifdef DEBUG
+    NodeType static_type = node->GetStaticType(broker);
+    if (!NodeTypeIs(actual_type, static_type)) {
+      // In case we needed a numerical alternative of a smi value, the type
+      // must generalize. In all other cases the node info type should reflect
+      // the actual type.
+      DCHECK(static_type == NodeType::kSmi &&
+             actual_type == NodeType::kNumber &&
+             !TryGetInfoFor(node)->alternative().has_none());
+    }
+#endif  // DEBUG
+    return actual_type;
+  }
+
+  bool CheckType(compiler::JSHeapBroker* broker, ValueNode* node, NodeType type,
+                 NodeType* current_type = nullptr) {
+    NodeType static_type = node->GetStaticType(broker);
+    if (current_type) *current_type = static_type;
+    if (NodeTypeIs(static_type, type)) return true;
+    if (IsEmptyNodeType(IntersectType(static_type, type))) return false;
+    auto it = FindInfo(node);
+    if (!IsValid(it)) return false;
+    if (current_type) *current_type = it->second.type();
+    return NodeTypeIs(it->second.type(), type);
+  }
+
+  NodeType CheckTypes(compiler::JSHeapBroker* broker, ValueNode* node,
+                      std::initializer_list<NodeType> types) {
+    auto it = FindInfo(node);
+    bool has_kna = IsValid(it);
+    for (NodeType type : types) {
+      if (node->StaticTypeIs(broker, type)) return type;
+      if (has_kna) {
+        if (NodeTypeIs(it->second.type(), type)) return type;
+      }
     }
     return NodeType::kUnknown;
   }
 
+  bool MayBeNullOrUndefined(compiler::JSHeapBroker* broker, ValueNode* node) {
+    NodeType static_type = node->GetStaticType(broker);
+    if (!NodeTypeMayBeNullOrUndefined(static_type)) return false;
+    auto it = FindInfo(node);
+    if (!IsValid(it)) return true;
+    return NodeTypeMayBeNullOrUndefined(it->second.type());
+  }
+
+  bool EnsureType(compiler::JSHeapBroker* broker, ValueNode* node,
+                  NodeType type, NodeType* old_type = nullptr) {
+    NodeType static_type = node->GetStaticType(broker);
+    if (old_type) *old_type = static_type;
+    if (NodeTypeIs(static_type, type)) return true;
+    NodeInfo* known_info = GetOrCreateInfoFor(broker, node);
+    if (old_type) *old_type = known_info->type();
+    if (NodeTypeIs(known_info->type(), type)) return true;
+    known_info->IntersectType(type);
+    if (auto phi = node->TryCast<Phi>()) {
+      known_info->IntersectType(phi->type());
+    }
+    if (NodeTypeIsUnstable(type)) {
+      known_info->set_node_type_is_unstable();
+      any_map_for_any_node_is_unstable = true;
+    }
+    return false;
+  }
+
+  template <typename Function>
+  bool EnsureType(compiler::JSHeapBroker* broker, ValueNode* node,
+                  NodeType type, Function ensure_new_type) {
+    if (node->StaticTypeIs(broker, type)) return true;
+    NodeInfo* known_info = GetOrCreateInfoFor(broker, node);
+    if (NodeTypeIs(known_info->type(), type)) return true;
+    ensure_new_type(known_info->type());
+    known_info->IntersectType(type);
+    if (NodeTypeIsUnstable(type)) {
+      known_info->set_node_type_is_unstable();
+      any_map_for_any_node_is_unstable = true;
+    }
+    return false;
+  }
+
+  // Returns true if we statically know that {lhs} and {rhs} have disjoint
+  // types.
+  bool HaveDisjointTypes(compiler::JSHeapBroker* broker, ValueNode* lhs,
+                         ValueNode* rhs) {
+    return HasDisjointType(broker, lhs, GetType(broker, rhs));
+  }
+
+  bool HasDisjointType(compiler::JSHeapBroker* broker, ValueNode* lhs,
+                       NodeType rhs_type) {
+    return IsEmptyNodeType(IntersectType(GetType(broker, lhs), rhs_type));
+  }
+
   void Merge(const KnownNodeAspects& other, Zone* zone);
 
+  // If IsCompatibleWithLoopHeader(other) returns true, it means that
+  // Merge(other) would not remove any information from `this`.
+  bool IsCompatibleWithLoopHeader(const KnownNodeAspects& other) const;
+
   // TODO(leszeks): Store these more efficiently than with std::map -- in
   // particular, clear out entries that are no longer reachable, perhaps also
   // allow lookup by interpreter register rather than by node pointer.
@@ -344,10 +487,71 @@ struct KnownNodeAspects {
 
   // Cached property loads.
 
-  // Maps name->object->value, so that stores to a name can invalidate all loads
-  // of that name (in case the objects are aliasing).
+  // Represents a key into the cache. This is either a NameRef, or an enum
+  // value.
+  class LoadedPropertyMapKey {
+   public:
+    enum Type {
+      // kName must be zero so that pointers are unaffected.
+      kName = 0,
+      kElements,
+      kTypedArrayLength,
+      // TODO(leszeks): We could probably share kStringLength with
+      // kTypedArrayLength if needed.
+      kStringLength
+    };
+    static constexpr int kTypeMask = 0x3;
+    static_assert((kName & ~kTypeMask) == 0);
+    static_assert((kElements & ~kTypeMask) == 0);
+    static_assert((kTypedArrayLength & ~kTypeMask) == 0);
+    static_assert((kStringLength & ~kTypeMask) == 0);
+
+    static LoadedPropertyMapKey Elements() {
+      return LoadedPropertyMapKey(kElements);
+    }
+
+    static LoadedPropertyMapKey TypedArrayLength() {
+      return LoadedPropertyMapKey(kTypedArrayLength);
+    }
+
+    static LoadedPropertyMapKey StringLength() {
+      return LoadedPropertyMapKey(kStringLength);
+    }
+
+    // Allow implicit conversion from NameRef to key, so that callers in the
+    // common path can use a NameRef directly.
+    // NOLINTNEXTLINE
+    LoadedPropertyMapKey(compiler::NameRef ref)
+        : data_(reinterpret_cast<Address>(ref.data())) {
+      DCHECK_EQ(data_ & kTypeMask, kName);
+    }
+
+    bool operator==(const LoadedPropertyMapKey& other) const {
+      return data_ == other.data_;
+    }
+    bool operator<(const LoadedPropertyMapKey& other) const {
+      return data_ < other.data_;
+    }
+
+    compiler::NameRef name() {
+      DCHECK_EQ(type(), kName);
+      return compiler::NameRef(reinterpret_cast<compiler::ObjectData*>(data_),
+                               false);
+    }
+
+    Type type() { return static_cast<Type>(data_ & kTypeMask); }
+
+   private:
+    explicit LoadedPropertyMapKey(Type type) : data_(type) {
+      DCHECK_NE(type, kName);
+    }
+
+    Address data_;
+  };
+  // Maps key->object->value, so that stores to a key can invalidate all loads
+  // of that key (in case the objects are aliasing).
   using LoadedPropertyMap =
-      ZoneMap<compiler::NameRef, ZoneMap<ValueNode*, ValueNode*>>;
+      ZoneMap<LoadedPropertyMapKey, ZoneMap<ValueNode*, ValueNode*>>;
 
   // Valid across side-effecting calls, as long as we install a dependency.
   LoadedPropertyMap loaded_constant_properties;
@@ -356,29 +560,95 @@ struct KnownNodeAspects {
 
   // Unconditionally valid across side-effecting calls.
   ZoneMap<std::tuple<ValueNode*, int>, ValueNode*> loaded_context_constants;
+  enum class ContextSlotLoadsAlias : uint8_t {
+    kNone,
+    kOnlyLoadsRelativeToCurrentContext,
+    kOnlyLoadsRelativeToConstant,
+    kYes,
+  };
+  ContextSlotLoadsAlias may_have_aliasing_contexts() const {
+    return may_have_aliasing_contexts_;
+  }
+  static ContextSlotLoadsAlias ContextSlotLoadsAliasMerge(
+      ContextSlotLoadsAlias m1, ContextSlotLoadsAlias m2) {
+    if (m1 == m2) return m1;
+    if (m1 == ContextSlotLoadsAlias::kNone) return m2;
+    if (m2 == ContextSlotLoadsAlias::kNone) return m1;
+    return ContextSlotLoadsAlias::kYes;
+  }
+  void UpdateMayHaveAliasingContexts(compiler::JSHeapBroker* broker,
+                                     LocalIsolate* local_isolate,
+                                     ValueNode* context);
+
   // Flushed after side-effecting calls.
-  ZoneMap<std::tuple<ValueNode*, int>, ValueNode*> loaded_context_slots;
+  using LoadedContextSlotsKey = std::tuple<ValueNode*, int>;
+  using LoadedContextSlots = ZoneMap<LoadedContextSlotsKey, ValueNode*>;
+  LoadedContextSlots loaded_context_slots;
+
+  struct AvailableExpression {
+    NodeBase* node;
+    uint32_t effect_epoch;
+  };
+  ZoneMap<uint32_t, AvailableExpression> available_expressions;
+  uint32_t effect_epoch() const { return effect_epoch_; }
+  static constexpr uint32_t kEffectEpochForPureInstructions =
+      std::numeric_limits<uint32_t>::max();
+  static constexpr uint32_t kEffectEpochOverflow =
+      kEffectEpochForPureInstructions - 1;
+  void increment_effect_epoch() {
+    if (effect_epoch_ < kEffectEpochOverflow) effect_epoch_++;
+  }
+
+  explicit KnownNodeAspects(Zone* zone)
+      : any_map_for_any_node_is_unstable(false),
+        loaded_constant_properties(zone),
+        loaded_properties(zone),
+        loaded_context_constants(zone),
+        loaded_context_slots(zone),
+        available_expressions(zone),
+        may_have_aliasing_contexts_(ContextSlotLoadsAlias::kNone),
+        effect_epoch_(0),
+        node_infos(zone) {}
 
  private:
+  // This field indicates if the current state of loaded_context_slots might
+  // contain contexts aliases. If that is the case, then we need to be more
+  // conservative about updating the state on stores.
+  ContextSlotLoadsAlias may_have_aliasing_contexts_;
+  uint32_t effect_epoch_;
+
   NodeInfos node_infos;
 
   friend KnownNodeAspects* Zone::New<KnownNodeAspects, const KnownNodeAspects&>(
       const KnownNodeAspects&);
   KnownNodeAspects(const KnownNodeAspects& other) V8_NOEXCEPT = default;
+  // Copy constructor for CloneForLoopHeader
+  friend KnownNodeAspects* Zone::New<KnownNodeAspects, const KnownNodeAspects&,
+                                     bool&, LoopEffects*&, Zone*&>(
+      const KnownNodeAspects&, bool&, maglev::LoopEffects*&, Zone*&);
+  KnownNodeAspects(const KnownNodeAspects& other, bool optimistic_initial_state,
+                   LoopEffects* loop_effects, Zone* zone);
 };
 
 class InterpreterFrameState {
  public:
   InterpreterFrameState(const MaglevCompilationUnit& info,
-                        KnownNodeAspects* known_node_aspects)
-      : frame_(info), known_node_aspects_(known_node_aspects) {}
+                        KnownNodeAspects* known_node_aspects,
+                        VirtualObjectList virtual_objects)
+      : frame_(info),
+        known_node_aspects_(known_node_aspects),
+        virtual_objects_(virtual_objects) {
+    frame_[interpreter::Register::virtual_accumulator()] = nullptr;
+  }
 
   explicit InterpreterFrameState(const MaglevCompilationUnit& info)
-      : InterpreterFrameState(
-            info, info.zone()->New<KnownNodeAspects>(info.zone())) {}
+      : InterpreterFrameState(info,
+                              info.zone()->New<KnownNodeAspects>(info.zone()),
+                              VirtualObjectList()) {}
 
   inline void CopyFrom(const MaglevCompilationUnit& info,
-                       MergePointInterpreterFrameState& state);
+                       MergePointInterpreterFrameState& state,
+                       bool preserve_known_node_aspects, Zone* zone);
 
   void set_accumulator(ValueNode* value) {
     // Conversions should be stored in known_node_aspects/NodeInfo.
@@ -422,9 +692,16 @@ class InterpreterFrameState {
 
   void clear_known_node_aspects() { known_node_aspects_ = nullptr; }
 
+  void add_object(VirtualObject* vobject) { virtual_objects_.Add(vobject); }
+  const VirtualObjectList& virtual_objects() const { return virtual_objects_; }
+  void set_virtual_objects(const VirtualObjectList& virtual_objects) {
+    virtual_objects_ = virtual_objects;
+  }
+
  private:
   RegisterFrameArray<ValueNode*> frame_;
   KnownNodeAspects* known_node_aspects_;
+  VirtualObjectList virtual_objects_;
 };
 
 class CompactInterpreterFrameState {
@@ -433,12 +710,14 @@ class CompactInterpreterFrameState {
                                const compiler::BytecodeLivenessState* liveness)
       : live_registers_and_accumulator_(
             info.zone()->AllocateArray<ValueNode*>(SizeFor(info, liveness))),
-        liveness_(liveness) {}
+        liveness_(liveness),
+        virtual_objects_() {}
 
   CompactInterpreterFrameState(const MaglevCompilationUnit& info,
                                const compiler::BytecodeLivenessState* liveness,
                                const InterpreterFrameState& state)
       : CompactInterpreterFrameState(info, liveness) {
+    virtual_objects_ = state.virtual_objects();
     ForEachValue(info, [&](ValueNode*& entry, interpreter::Register reg) {
       entry = state.get(reg);
     });
@@ -568,6 +847,12 @@ class CompactInterpreterFrameState {
     return SizeFor(info, liveness_);
   }
 
+  const VirtualObjectList& virtual_objects() const { return virtual_objects_; }
+  VirtualObjectList& virtual_objects() { return virtual_objects_; }
+  void set_virtual_objects(const VirtualObjectList& vos) {
+    virtual_objects_ = vos;
+  }
+
  private:
   static size_t SizeFor(const MaglevCompilationUnit& info,
                         const compiler::BytecodeLivenessState* liveness) {
@@ -580,9 +865,12 @@ class CompactInterpreterFrameState {
   static const int context_register_count_ = 1;
   ValueNode** const live_registers_and_accumulator_;
   const compiler::BytecodeLivenessState* const liveness_;
+  VirtualObjectList virtual_objects_;
 };
 
 class MergePointRegisterState {
+#ifdef V8_ENABLE_MAGLEV
+
  public:
   bool is_initialized() const { return values_[0].GetPayload().is_initialized; }
 
@@ -608,6 +896,7 @@ class MergePointRegisterState {
  private:
   RegisterState values_[kAllocatableGeneralRegisterCount] = {{}};
   RegisterState double_values_[kAllocatableDoubleRegisterCount] = {{}};
+#endif  // V8_ENABLE_MAGLEV
 };
 
 class MergePointInterpreterFrameState {
@@ -616,6 +905,7 @@ class MergePointInterpreterFrameState {
     kDefault,
     kLoopHeader,
     kExceptionHandlerStart,
+    kUnusedExceptionHandlerStart,
   };
 
   static MergePointInterpreterFrameState* New(
@@ -624,7 +914,7 @@ class MergePointInterpreterFrameState {
       const compiler::BytecodeLivenessState* liveness);
 
   static MergePointInterpreterFrameState* NewForLoop(
-      const InterpreterFrameState& start_state,
+      const InterpreterFrameState& start_state, Graph* graph,
       const MaglevCompilationUnit& info, int merge_offset,
       int predecessor_count, const compiler::BytecodeLivenessState* liveness,
       const compiler::LoopInfo* loop_info, bool has_been_peeled = false);
@@ -632,7 +922,7 @@ class MergePointInterpreterFrameState {
   static MergePointInterpreterFrameState* NewForCatchBlock(
       const MaglevCompilationUnit& unit,
       const compiler::BytecodeLivenessState* liveness, int handler_offset,
-      interpreter::Register context_register, Graph* graph);
+      bool was_used, interpreter::Register context_register, Graph* graph);
 
   // Merges an unmerged framestate with a possibly merged framestate into |this|
   // framestate.
@@ -641,6 +931,12 @@ class MergePointInterpreterFrameState {
   void Merge(MaglevGraphBuilder* graph_builder,
              MaglevCompilationUnit& compilation_unit,
              InterpreterFrameState& unmerged, BasicBlock* predecessor);
+  void InitializeLoop(MaglevGraphBuilder* graph_builder,
+                      MaglevCompilationUnit& compilation_unit,
+                      InterpreterFrameState& unmerged, BasicBlock* predecessor,
+                      bool optimistic_initial_state = false,
+                      LoopEffects* loop_effects = nullptr);
+  void InitializeWithBasicBlock(BasicBlock* current_block);
 
   // Merges an unmerged framestate with a possibly merged framestate into |this|
   // framestate.
@@ -651,25 +947,36 @@ class MergePointInterpreterFrameState {
                  MaglevCompilationUnit& compilation_unit,
                  InterpreterFrameState& loop_end_state,
                  BasicBlock* loop_end_block);
+  void set_loop_effects(LoopEffects* loop_effects);
+  const LoopEffects* loop_effects();
+  // Merges a frame-state that might not be mergable, in which case we need to
+  // re-compile the loop again. Calls FinishBlock only if the merge succeeded.
+  bool TryMergeLoop(MaglevGraphBuilder* graph_builder,
+                    InterpreterFrameState& loop_end_state,
+                    const std::function<BasicBlock*()>& FinishBlock);
 
-  // Merges an unmerged framestate with a possibly merged framestate into |this|
-  // framestate.
-  void MergeThrow(MaglevGraphBuilder* builder,
+  // Merges an unmerged framestate into a possibly merged framestate at the
+  // start of the target catchblock.
+  void MergeThrow(MaglevGraphBuilder* handler_builder,
                   const MaglevCompilationUnit* handler_unit,
-                  InterpreterFrameState& unmerged);
+                  const KnownNodeAspects& known_node_aspects,
+                  const VirtualObjectList virtual_objects);
 
   // Merges a dead framestate (e.g. one which has been early terminated with a
   // deopt).
-  void MergeDead(const MaglevCompilationUnit& compilation_unit) {
-    DCHECK_GE(predecessor_count_, 1);
+  void MergeDead(const MaglevCompilationUnit& compilation_unit,
+                 unsigned num = 1) {
+    DCHECK_GE(predecessor_count_, num);
     DCHECK_LT(predecessors_so_far_, predecessor_count_);
-    predecessor_count_--;
+    ReducePhiPredecessorCount(num);
+    predecessor_count_ -= num;
     DCHECK_LE(predecessors_so_far_, predecessor_count_);
+  }
 
-    frame_state_.ForEachValue(compilation_unit,
-                              [&](ValueNode* value, interpreter::Register reg) {
-                                ReducePhiPredecessorCount(reg, value);
-                              });
+  void clear_is_loop() {
+    bitfield_ =
+        kBasicBlockTypeBits::update(bitfield_, BasicBlockType::kDefault);
+    bitfield_ = kIsResumableLoopBit::update(bitfield_, false);
   }
 
   // Merges a dead loop framestate (e.g. one where the block containing the
@@ -680,10 +987,20 @@ class MergePointInterpreterFrameState {
     DCHECK(is_unmerged_loop());
     MergeDead(compilation_unit);
     // This means that this is no longer a loop.
-    bitfield_ =
-        kBasicBlockTypeBits::update(bitfield_, BasicBlockType::kDefault);
+    clear_is_loop();
   }
 
+  // Clears dead loop state, after all merges have already be done.
+  void TurnLoopIntoRegularBlock() {
+    DCHECK(is_loop());
+    predecessor_count_--;
+    predecessors_so_far_--;
+    ReducePhiPredecessorCount(1);
+    clear_is_loop();
+  }
+
+  void RemovePredecessorAt(int predecessor_id);
+
   // Returns and clears the known node aspects on this state. Expects to only
   // ever be called once, when starting a basic block with this state.
   KnownNodeAspects* TakeKnownNodeAspects() {
@@ -691,6 +1008,10 @@ class MergePointInterpreterFrameState {
     return std::exchange(known_node_aspects_, nullptr);
   }
 
+  KnownNodeAspects* CloneKnownNodeAspects(Zone* zone) {
+    return known_node_aspects_->Clone(zone);
+  }
+
   const CompactInterpreterFrameState& frame_state() const {
     return frame_state_;
   }
@@ -699,28 +1020,54 @@ class MergePointInterpreterFrameState {
   bool has_phi() const { return !phis_.is_empty(); }
   Phi::List* phis() { return &phis_; }
 
-  void SetPhis(Phi::List&& phis) {
-    // Move the collected phis to the live interpreter frame.
-    DCHECK(phis_.is_empty());
-    phis_.MoveTail(&phis, phis.begin());
-  }
+  uint32_t predecessor_count() const { return predecessor_count_; }
 
-  int predecessor_count() const { return predecessor_count_; }
+  uint32_t predecessors_so_far() const { return predecessors_so_far_; }
 
   BasicBlock* predecessor_at(int i) const {
-    // DCHECK_EQ(predecessors_so_far_, predecessor_count_);
-    DCHECK_LT(i, predecessor_count_);
+    DCHECK_LE(predecessors_so_far_, predecessor_count_);
+    DCHECK_LT(i, predecessors_so_far_);
     return predecessors_[i];
   }
+  void set_predecessor_at(int i, BasicBlock* val) {
+    DCHECK_LE(predecessors_so_far_, predecessor_count_);
+    DCHECK_LT(i, predecessors_so_far_);
+    predecessors_[i] = val;
+  }
+
+  void set_virtual_objects(const VirtualObjectList& vos) {
+    frame_state_.set_virtual_objects(vos);
+  }
+
+  void PrintVirtualObjects(const MaglevCompilationUnit& unit,
+                           VirtualObjectList from_ifs,
+                           const char* prelude = nullptr) {
+    if (V8_LIKELY(!v8_flags.trace_maglev_graph_building ||
+                  !unit.is_tracing_enabled())) {
+      return;
+    }
+    if (prelude) {
+      std::cout << prelude << std::endl;
+    }
+    from_ifs.Print(std::cout, "* VOs (Interpreter Frame State): ");
+    frame_state_.virtual_objects().Print(std::cout,
+                                         "* VOs (Merge Frame State): ");
+  }
 
   bool is_loop() const {
     return basic_block_type() == BasicBlockType::kLoopHeader;
   }
 
-  bool is_exception_handler() const {
+  bool exception_handler_was_used() const {
+    DCHECK(is_exception_handler());
     return basic_block_type() == BasicBlockType::kExceptionHandlerStart;
   }
 
+  bool is_exception_handler() const {
+    return basic_block_type() == BasicBlockType::kExceptionHandlerStart ||
+           basic_block_type() == BasicBlockType::kUnusedExceptionHandlerStart;
+  }
+
   bool is_unmerged_loop() const {
     // If this is a loop and not all predecessors are set, then the loop isn't
     // merged yet.
@@ -728,19 +1075,25 @@ class MergePointInterpreterFrameState {
     return is_loop() && predecessors_so_far_ < predecessor_count_;
   }
 
-  bool is_unreachable_loop() const {
+  bool is_unmerged_unreachable_loop() const {
     // If there is only one predecessor, and it's not set, then this is a loop
     // merge with no forward control flow entering it.
-    return is_loop() && !is_resumable_loop() && predecessor_count_ == 1 &&
-           predecessors_so_far_ == 0;
+    return is_unmerged_loop() && !is_resumable_loop() &&
+           predecessor_count_ == 1 && predecessors_so_far_ == 0;
   }
 
+  bool IsUnreachableByForwardEdge() const;
+  bool IsUnreachable() const;
+
   BasicBlockType basic_block_type() const {
     return kBasicBlockTypeBits::decode(bitfield_);
   }
   bool is_resumable_loop() const {
-    return kIsResumableLoopBit::decode(bitfield_);
+    bool res = kIsResumableLoopBit::decode(bitfield_);
+    DCHECK_IMPLIES(res, is_loop());
+    return res;
   }
+  void set_is_resumable_loop(Graph* graph);
   bool is_loop_with_peeled_iteration() const {
     return kIsLoopWithPeeledIterationBit::decode(bitfield_);
   }
@@ -750,8 +1103,13 @@ class MergePointInterpreterFrameState {
   DeoptFrame* backedge_deopt_frame() const { return backedge_deopt_frame_; }
 
   const compiler::LoopInfo* loop_info() const {
-    DCHECK(loop_info_.has_value());
-    return loop_info_.value();
+    DCHECK(loop_metadata_.has_value());
+    DCHECK_NOT_NULL(loop_metadata_->loop_info);
+    return loop_metadata_->loop_info;
+  }
+  void ClearLoopInfo() { loop_metadata_->loop_info = nullptr; }
+  bool HasLoopInfo() const {
+    return loop_metadata_.has_value() && loop_metadata_->loop_info;
   }
 
   interpreter::Register catch_block_context_register() const {
@@ -797,19 +1155,43 @@ class MergePointInterpreterFrameState {
       int predecessor_count, int predecessors_so_far, BasicBlock** predecessors,
       BasicBlockType type, const compiler::BytecodeLivenessState* liveness);
 
-  ValueNode* MergeValue(MaglevGraphBuilder* graph_builder,
+  void MergePhis(MaglevGraphBuilder* builder,
+                 MaglevCompilationUnit& compilation_unit,
+                 InterpreterFrameState& unmerged, BasicBlock* predecessor,
+                 bool optimistic_loop_phis);
+  void MergeVirtualObjects(MaglevGraphBuilder* builder,
+                           MaglevCompilationUnit& compilation_unit,
+                           InterpreterFrameState& unmerged,
+                           BasicBlock* predecessor);
+
+  ValueNode* MergeValue(const MaglevGraphBuilder* graph_builder,
                         interpreter::Register owner,
                         const KnownNodeAspects& unmerged_aspects,
                         ValueNode* merged, ValueNode* unmerged,
-                        Alternatives::List* per_predecessor_alternatives);
+                        Alternatives::List* per_predecessor_alternatives,
+                        bool optimistic_loop_phis = false);
 
-  void ReducePhiPredecessorCount(interpreter::Register owner,
-                                 ValueNode* merged);
+  void ReducePhiPredecessorCount(unsigned num);
+
+  void MergeVirtualObjects(MaglevGraphBuilder* builder,
+                           MaglevCompilationUnit& compilation_unit,
+                           const VirtualObjectList unmerged_vos,
+                           const KnownNodeAspects& unmerged_aspects);
+
+  void MergeVirtualObject(MaglevGraphBuilder* builder,
+                          const VirtualObjectList unmerged_vos,
+                          const KnownNodeAspects& unmerged_aspects,
+                          VirtualObject* merged, VirtualObject* unmerged);
+
+  std::optional<ValueNode*> MergeVirtualObjectValue(
+      const MaglevGraphBuilder* graph_builder,
+      const KnownNodeAspects& unmerged_aspects, ValueNode* merged,
+      ValueNode* unmerged);
 
   void MergeLoopValue(MaglevGraphBuilder* graph_builder,
                       interpreter::Register owner,
-                      KnownNodeAspects& unmerged_aspects, ValueNode* merged,
-                      ValueNode* unmerged);
+                      const KnownNodeAspects& unmerged_aspects,
+                      ValueNode* merged, ValueNode* unmerged);
 
   ValueNode* NewLoopPhi(Zone* zone, interpreter::Register reg);
 
@@ -823,8 +1205,8 @@ class MergePointInterpreterFrameState {
 
   int merge_offset_;
 
-  int predecessor_count_;
-  int predecessors_so_far_;
+  uint32_t predecessor_count_;
+  uint32_t predecessors_so_far_;
 
   uint32_t bitfield_;
 
@@ -851,18 +1233,80 @@ class MergePointInterpreterFrameState {
     interpreter::Register catch_block_context_register_;
   };
 
-  base::Optional<const compiler::LoopInfo*> loop_info_ = base::nullopt;
+  struct LoopMetadata {
+    const compiler::LoopInfo* loop_info;
+    const LoopEffects* loop_effects;
+  };
+  std::optional<LoopMetadata> loop_metadata_ = std::nullopt;
 };
 
-void InterpreterFrameState::CopyFrom(const MaglevCompilationUnit& info,
-                                     MergePointInterpreterFrameState& state) {
+struct LoopEffects {
+  explicit LoopEffects(int loop_header, Zone* zone)
+      :
+#ifdef DEBUG
+        loop_header(loop_header),
+#endif
+        context_slot_written(zone),
+        objects_written(zone),
+        keys_cleared(zone),
+        allocations(zone) {
+  }
+#ifdef DEBUG
+  int loop_header;
+#endif
+  ZoneSet<KnownNodeAspects::LoadedContextSlotsKey> context_slot_written;
+  ZoneSet<ValueNode*> objects_written;
+  ZoneSet<KnownNodeAspects::LoadedPropertyMapKey> keys_cleared;
+  ZoneSet<InlinedAllocation*> allocations;
+  bool unstable_aspects_cleared = false;
+  bool may_have_aliasing_contexts = false;
+  void Merge(const LoopEffects* other) {
+    if (!unstable_aspects_cleared) {
+      unstable_aspects_cleared = other->unstable_aspects_cleared;
+    }
+    if (!may_have_aliasing_contexts) {
+      may_have_aliasing_contexts = other->may_have_aliasing_contexts;
+    }
+    context_slot_written.insert(other->context_slot_written.begin(),
+                                other->context_slot_written.end());
+    objects_written.insert(other->objects_written.begin(),
+                           other->objects_written.end());
+    keys_cleared.insert(other->keys_cleared.begin(), other->keys_cleared.end());
+    allocations.insert(other->allocations.begin(), other->allocations.end());
+  }
+};
+
+void InterpreterFrameState::CopyFrom(const MaglevCompilationUnit& unit,
+                                     MergePointInterpreterFrameState& state,
+                                     bool preserve_known_node_aspects = false,
+                                     Zone* zone = nullptr) {
+  DCHECK_IMPLIES(preserve_known_node_aspects, zone);
+  if (V8_UNLIKELY(v8_flags.trace_maglev_graph_building &&
+                  unit.is_tracing_enabled())) {
+    std::cout << "- Copying frame state from merge @" << &state << std::endl;
+    state.PrintVirtualObjects(unit, virtual_objects());
+  }
+  virtual_objects_.Snapshot();
   state.frame_state().ForEachValue(
-      info, [&](ValueNode* value, interpreter::Register reg) {
+      unit, [&](ValueNode* value, interpreter::Register reg) {
         frame_[reg] = value;
       });
-  // Move "what we know" across without copying -- we can safely mutate it
-  // now, as we won't be entering this merge point again.
-  known_node_aspects_ = state.TakeKnownNodeAspects();
+  if (preserve_known_node_aspects) {
+    known_node_aspects_ = state.CloneKnownNodeAspects(zone);
+  } else {
+    // Move "what we know" across without copying -- we can safely mutate it
+    // now, as we won't be entering this merge point again.
+    known_node_aspects_ = state.TakeKnownNodeAspects();
+  }
+  virtual_objects_ = state.frame_state().virtual_objects();
+}
+
+inline VirtualObjectList DeoptFrame::GetVirtualObjects() const {
+  if (type() == DeoptFrame::FrameType::kInterpretedFrame) {
+    return as_interpreted().frame_state()->virtual_objects();
+  }
+  DCHECK_NOT_NULL(parent());
+  return parent()->GetVirtualObjects();
 }
 
 }  // namespace maglev
