@@ -5,13 +5,16 @@
 #ifndef V8_OBJECTS_FEEDBACK_VECTOR_H_
 #define V8_OBJECTS_FEEDBACK_VECTOR_H_
 
+#include <optional>
 #include <vector>
 
 #include "src/base/bit-field.h"
 #include "src/base/logging.h"
 #include "src/base/macros.h"
+#include "src/base/small-vector.h"
 #include "src/common/globals.h"
 #include "src/objects/elements-kind.h"
+#include "src/objects/feedback-cell.h"
 #include "src/objects/map.h"
 #include "src/objects/maybe-object.h"
 #include "src/objects/name.h"
@@ -21,8 +24,7 @@
 // Has to be the last include (doesn't have include guards):
 #include "src/objects/object-macros.h"
 
-namespace v8 {
-namespace internal {
+namespace v8::internal {
 
 class IsCompiledScope;
 class FeedbackVectorSpec;
@@ -70,7 +72,9 @@ enum class FeedbackSlotKind : uint8_t {
   kLiteral,
   kForIn,
   kInstanceOf,
+  kTypeOf,
   kCloneObject,
+  kStringAddAndInternalize,
   kJumpLoop,
 
   kLast = kJumpLoop  // Always update this if the list above changes.
@@ -79,8 +83,102 @@ enum class FeedbackSlotKind : uint8_t {
 static constexpr int kFeedbackSlotKindCount =
     static_cast<int>(FeedbackSlotKind::kLast) + 1;
 
-using MapAndHandler = std::pair<Handle<Map>, MaybeObjectHandle>;
-using MapAndFeedback = std::pair<Handle<Map>, MaybeObjectHandle>;
+using MapAndHandler = std::pair<DirectHandle<Map>, MaybeObjectDirectHandle>;
+
+class MapsAndHandlers {
+ public:
+  explicit MapsAndHandlers(Isolate* isolate)
+      : maps_(isolate), handlers_(isolate) {}
+
+  bool empty() const { return maps_.empty(); }
+  size_t size() const { return maps_.size(); }
+  void reserve(size_t capacity) {
+    maps_.reserve(capacity);
+    handlers_.reserve(capacity);
+  }
+
+  MapAndHandler operator[](size_t i) const {
+    DCHECK_LT(i, size());
+    MaybeObjectDirectHandle handler;
+    switch (handlers_reference_types_[i]) {
+      case HeapObjectReferenceType::STRONG:
+        handler = MaybeObjectDirectHandle(handlers_[i]);
+        break;
+      case HeapObjectReferenceType::WEAK:
+        handler = MaybeObjectDirectHandle::Weak(handlers_[i]);
+        break;
+    }
+    return MapAndHandler(maps_[i], handler);
+  }
+
+  void set_map(size_t i, DirectHandle<Map> map) {
+    DCHECK_LT(i, size());
+    maps_[i] = map;
+  }
+
+  void set_handler(size_t i, MaybeObjectDirectHandle handler) {
+    DCHECK_LT(i, size());
+    handlers_[i] =
+        handler.is_null() ? DirectHandle<Object>() : handler.object();
+    handlers_reference_types_[i] = handler.reference_type();
+  }
+
+  class Iterator final
+      : public base::iterator<std::input_iterator_tag, MapAndHandler> {
+   public:
+    constexpr Iterator() = default;
+
+    constexpr bool operator==(const Iterator& other) {
+      return index_ == other.index_;
+    }
+    constexpr bool operator!=(const Iterator& other) {
+      return index_ != other.index_;
+    }
+
+    constexpr Iterator& operator++() {
+      ++index_;
+      return *this;
+    }
+
+    constexpr Iterator operator++(int) {
+      Iterator temp = *this;
+      ++*this;
+      return temp;
+    }
+
+    value_type operator*() const {
+      DCHECK_NOT_NULL(container_);
+      return (*container_)[index_];
+    }
+
+   private:
+    friend class MapsAndHandlers;
+
+    constexpr Iterator(const MapsAndHandlers* container, size_t i)
+        : container_(container), index_(i) {}
+
+    const MapsAndHandlers* container_ = nullptr;
+    size_t index_ = 0;
+  };
+
+  void emplace_back(DirectHandle<Map> map, MaybeObjectDirectHandle handler) {
+    maps_.push_back(map);
+    handlers_.push_back(handler.is_null() ? DirectHandle<Object>()
+                                          : handler.object());
+    handlers_reference_types_.push_back(handler.reference_type());
+  }
+
+  Iterator begin() const { return Iterator(this, 0); }
+  Iterator end() const { return Iterator(this, size()); }
+
+  base::Vector<DirectHandle<Map>> maps() { return base::VectorOf(maps_); }
+
+ private:
+  DirectHandleSmallVector<Map, DEFAULT_MAX_POLYMORPHIC_MAP_COUNT> maps_;
+  DirectHandleSmallVector<Object, DEFAULT_MAX_POLYMORPHIC_MAP_COUNT> handlers_;
+  base::SmallVector<HeapObjectReferenceType, DEFAULT_MAX_POLYMORPHIC_MAP_COUNT>
+      handlers_reference_types_;
+};
 
 inline bool IsCallICKind(FeedbackSlotKind kind) {
   return kind == FeedbackSlotKind::kCall;
@@ -172,27 +270,36 @@ class FeedbackMetadata;
 
 #include "torque-generated/src/objects/feedback-vector-tq.inc"
 
-// ClosureFeedbackCellArray is a FixedArray that contains feedback cells used
-// when creating closures from a function. This is created once the function is
-// compiled and is either held by the feedback vector (if allocated) or by the
-// FeedbackCell of the closure.
-class ClosureFeedbackCellArray : public FixedArray {
+class ClosureFeedbackCellArrayShape final : public AllStatic {
  public:
-  NEVER_READ_ONLY_SPACE
+  using ElementT = FeedbackCell;
+  using CompressionScheme = V8HeapCompressionScheme;
+  static constexpr RootIndex kMapRootIndex =
+      RootIndex::kClosureFeedbackCellArrayMap;
+  static constexpr bool kLengthEqualsCapacity = true;
+};
 
-  DECL_CAST(ClosureFeedbackCellArray)
+// ClosureFeedbackCellArray contains feedback cells used when creating closures
+// from a function. This is created once the function is compiled and is either
+// held by the feedback vector (if allocated) or by the FeedbackCell of the
+// closure.
+class ClosureFeedbackCellArray
+    : public TaggedArrayBase<ClosureFeedbackCellArray,
+                             ClosureFeedbackCellArrayShape> {
+  using Super =
+      TaggedArrayBase<ClosureFeedbackCellArray, ClosureFeedbackCellArrayShape>;
 
-  V8_EXPORT_PRIVATE static Handle<ClosureFeedbackCellArray> New(
-      Isolate* isolate, Handle<SharedFunctionInfo> shared);
+ public:
+  using Shape = ClosureFeedbackCellArrayShape;
 
-  inline Handle<FeedbackCell> GetFeedbackCell(int index);
-  inline Tagged<FeedbackCell> cell(int index);
+  V8_EXPORT_PRIVATE static DirectHandle<ClosureFeedbackCellArray> New(
+      Isolate* isolate, DirectHandle<SharedFunctionInfo> shared,
+      AllocationType allocation = AllocationType::kYoung);
 
   DECL_VERIFIER(ClosureFeedbackCellArray)
   DECL_PRINTER(ClosureFeedbackCellArray)
 
- private:
-  OBJECT_CONSTRUCTORS(ClosureFeedbackCellArray, FixedArray);
+  class BodyDescriptor;
 };
 
 class NexusConfig;
@@ -202,10 +309,11 @@ class NexusConfig;
 class FeedbackVector
     : public TorqueGeneratedFeedbackVector<FeedbackVector, HeapObject> {
  public:
-  NEVER_READ_ONLY_SPACE
   DEFINE_TORQUE_GENERATED_OSR_STATE()
   DEFINE_TORQUE_GENERATED_FEEDBACK_VECTOR_FLAGS()
-  static_assert(TieringState::kLastTieringState <= TieringStateBits::kMax);
+
+#ifndef V8_ENABLE_LEAPTIERING
+  static_assert(TieringStateBits::is_valid(TieringState::kLastTieringState));
 
   static constexpr uint32_t kFlagsMaybeHasTurbofanCode =
       FeedbackVector::MaybeHasTurbofanCodeBit::kMask;
@@ -219,6 +327,10 @@ class FeedbackVector
   static constexpr uint32_t kFlagsLogNextExecution =
       FeedbackVector::LogNextExecutionBit::kMask;
 
+  static constexpr inline uint32_t FlagMaskForNeedsProcessingCheckFrom(
+      CodeKind code_kind);
+#endif  // !V8_ENABLE_LEAPTIERING
+
   inline bool is_empty() const;
 
   DECL_GETTER(metadata, Tagged<FeedbackMetadata>)
@@ -229,12 +341,19 @@ class FeedbackVector
   using TorqueGeneratedFeedbackVector::set_invocation_count;
   DECL_RELAXED_INT32_ACCESSORS(invocation_count)
   inline void clear_invocation_count(RelaxedStoreTag tag);
+  using TorqueGeneratedFeedbackVector::invocation_count_before_stable;
+  using TorqueGeneratedFeedbackVector::set_invocation_count_before_stable;
+  DECL_RELAXED_UINT8_ACCESSORS(invocation_count_before_stable)
+
+  // In case a function deoptimizes we set invocation_count_before_stable to
+  // this sentinel.
+  static constexpr uint8_t kInvocationCountBeforeStableDeoptSentinel = 0xff;
 
   // The [osr_urgency] controls when OSR is attempted, and is incremented as
   // the function becomes hotter. When the current loop depth is less than the
   // osr_urgency, JumpLoop calls into runtime to attempt OSR optimization.
   static constexpr int kMaxOsrUrgency = 6;
-  static_assert(kMaxOsrUrgency <= OsrUrgencyBits::kMax);
+  static_assert(OsrUrgencyBits::is_valid(kMaxOsrUrgency));
   inline int osr_urgency() const;
   inline void set_osr_urgency(int urgency);
   inline void reset_osr_urgency();
@@ -252,12 +371,14 @@ class FeedbackVector
   // The `osr_state` contains the osr_urgency and maybe_has_optimized_osr_code.
   inline void reset_osr_state();
 
-  inline Tagged<Code> optimized_code() const;
+#ifndef V8_ENABLE_LEAPTIERING
+  inline bool log_next_execution() const;
+  inline void set_log_next_execution(bool value = true);
+
+  inline Tagged<Code> optimized_code(IsolateForSandbox isolate) const;
   // Whether maybe_optimized_code contains a cached Code object.
   inline bool has_optimized_code() const;
 
-  inline bool log_next_execution() const;
-  inline void set_log_next_execution(bool value = true);
   // Similar to above, but represented internally as a bit that can be
   // efficiently checked by generated code. May lag behind the actual state of
   // the world, thus 'maybe'.
@@ -266,24 +387,43 @@ class FeedbackVector
   inline bool maybe_has_turbofan_code() const;
   inline void set_maybe_has_turbofan_code(bool value);
 
-  void SetOptimizedCode(Tagged<Code> code);
+  void SetOptimizedCode(IsolateForSandbox isolate, Tagged<Code> code);
   void EvictOptimizedCodeMarkedForDeoptimization(
       Isolate* isolate, Tagged<SharedFunctionInfo> shared, const char* reason);
   void ClearOptimizedCode();
+#endif  // !V8_ENABLE_LEAPTIERING
 
   // Optimized OSR'd code is cached in JumpLoop feedback vector slots. The
   // slots either contain a Code object or the ClearedValue.
-  inline base::Optional<Tagged<Code>> GetOptimizedOsrCode(Isolate* isolate,
-                                                          FeedbackSlot slot);
+  inline std::optional<Tagged<Code>> GetOptimizedOsrCode(
+      Isolate* isolate, Handle<BytecodeArray> bytecode_array,
+      FeedbackSlot slot);
   void SetOptimizedOsrCode(Isolate* isolate, FeedbackSlot slot,
                            Tagged<Code> code);
+  inline void RecomputeOptimizedOsrCodeFlags(
+      Isolate* isolate, Handle<BytecodeArray> bytecode_array);
 
+#ifdef V8_ENABLE_LEAPTIERING
+  inline bool tiering_in_progress() const;
+  void set_tiering_in_progress(bool);
+#else
   inline TieringState tiering_state() const;
-  void set_tiering_state(TieringState state);
-  void reset_tiering_state();
+  V8_EXPORT_PRIVATE void set_tiering_state(TieringState state);
+  inline void reset_tiering_state();
+#endif  // !V8_ENABLE_LEAPTIERING
 
-  TieringState osr_tiering_state();
-  void set_osr_tiering_state(TieringState marker);
+  bool osr_tiering_in_progress();
+  void set_osr_tiering_in_progress(bool osr_in_progress);
+
+  inline bool interrupt_budget_reset_by_ic_change() const;
+  inline void set_interrupt_budget_reset_by_ic_change(bool value);
+
+  // Check if this function was ever deoptimized. This flag can be used as a
+  // blanked bailout for optimizations which are not guaranteed to be deopt-loop
+  // free (such as hoisting checks out of loops).
+  // TODO(olivf): Have a more granular (e.g., per loop) mechanism.
+  inline bool was_once_deoptimized() const;
+  inline void set_was_once_deoptimized();
 
   void reset_flags();
 
@@ -293,18 +433,18 @@ class FeedbackVector
   // Conversion from an integer index to the underlying array to a slot.
   static inline FeedbackSlot ToSlot(intptr_t index);
 
-  inline MaybeObject SynchronizedGet(FeedbackSlot slot) const;
-  inline void SynchronizedSet(FeedbackSlot slot, MaybeObject value,
-                              WriteBarrierMode mode = UPDATE_WRITE_BARRIER);
-  inline void SynchronizedSet(FeedbackSlot slot, Tagged<Object> value,
+  inline Tagged<MaybeObject> SynchronizedGet(FeedbackSlot slot) const;
+  inline void SynchronizedSet(FeedbackSlot slot, Tagged<MaybeObject> value,
                               WriteBarrierMode mode = UPDATE_WRITE_BARRIER);
 
-  inline MaybeObject Get(FeedbackSlot slot) const;
-  inline MaybeObject Get(PtrComprCageBase cage_base, FeedbackSlot slot) const;
+  inline Tagged<MaybeObject> Get(FeedbackSlot slot) const;
+  inline Tagged<MaybeObject> Get(PtrComprCageBase cage_base,
+                                 FeedbackSlot slot) const;
 
   // Returns the feedback cell at |index| that is used to create the
   // closure.
-  inline Handle<FeedbackCell> GetClosureFeedbackCell(int index) const;
+  inline DirectHandle<FeedbackCell> GetClosureFeedbackCell(Isolate* isolate,
+                                                           int index) const;
   inline Tagged<FeedbackCell> closure_feedback_cell(int index) const;
 
   // Gives access to raw memory which stores the array's data.
@@ -316,9 +456,9 @@ class FeedbackVector
                                              AcquireLoadTag tag) const;
 
   V8_EXPORT_PRIVATE static Handle<FeedbackVector> New(
-      Isolate* isolate, Handle<SharedFunctionInfo> shared,
-      Handle<ClosureFeedbackCellArray> closure_feedback_cell_array,
-      Handle<FeedbackCell> parent_feedback_cell,
+      Isolate* isolate, DirectHandle<SharedFunctionInfo> shared,
+      DirectHandle<ClosureFeedbackCellArray> closure_feedback_cell_array,
+      DirectHandle<FeedbackCell> parent_feedback_cell,
       IsCompiledScope* is_compiled_scope);
 
   V8_EXPORT_PRIVATE static Handle<FeedbackVector> NewForTesting(
@@ -356,6 +496,12 @@ class FeedbackVector
 
   void FeedbackSlotPrint(std::ostream& os, FeedbackSlot slot);
 
+#ifdef V8_TRACE_FEEDBACK_UPDATES
+  static void TraceFeedbackChange(Isolate* isolate,
+                                  Tagged<FeedbackVector> vector,
+                                  FeedbackSlot slot, const char* reason);
+#endif
+
   // Clears the vector slots. Return true if feedback has changed.
   bool ClearSlots(Isolate* isolate) {
     return ClearSlots(isolate, ClearBehavior::kDefault);
@@ -367,13 +513,13 @@ class FeedbackVector
   }
 
   // The object that indicates an uninitialized cache.
-  static inline Handle<Symbol> UninitializedSentinel(Isolate* isolate);
+  static inline DirectHandle<Symbol> UninitializedSentinel(Isolate* isolate);
 
   // The object that indicates a megamorphic state.
   static inline Handle<Symbol> MegamorphicSentinel(Isolate* isolate);
 
   // The object that indicates a MegaDOM state.
-  static inline Handle<Symbol> MegaDOMSentinel(Isolate* isolate);
+  static inline DirectHandle<Symbol> MegaDOMSentinel(Isolate* isolate);
 
   // A raw version of the uninitialized sentinel that's safe to read during
   // garbage collection (e.g., for patching the cache).
@@ -393,19 +539,17 @@ class FeedbackVector
  private:
   bool ClearSlots(Isolate* isolate, ClearBehavior behavior);
 
-  static void AddToVectorsForProfilingTools(Isolate* isolate,
-                                            Handle<FeedbackVector> vector);
+  static void AddToVectorsForProfilingTools(
+      Isolate* isolate, DirectHandle<FeedbackVector> vector);
 
   // Private for initializing stores in FeedbackVector::New().
-  inline void Set(FeedbackSlot slot, MaybeObject value,
-                  WriteBarrierMode mode = UPDATE_WRITE_BARRIER);
-  inline void Set(FeedbackSlot slot, Tagged<Object> value,
+  inline void Set(FeedbackSlot slot, Tagged<MaybeObject> value,
                   WriteBarrierMode mode = UPDATE_WRITE_BARRIER);
 
 #ifdef DEBUG
   // Returns true if value is a non-HashTable FixedArray. We want to
   // make sure not to store such objects in the vector.
-  inline static bool IsOfLegacyType(MaybeObject value);
+  inline static bool IsOfLegacyType(Tagged<MaybeObject> value);
 #endif  // DEBUG
 
   // NexusConfig controls setting slots in the vector.
@@ -417,14 +561,24 @@ class FeedbackVector
 
 class V8_EXPORT_PRIVATE FeedbackVectorSpec {
  public:
-  explicit FeedbackVectorSpec(Zone* zone) : slot_kinds_(zone) {
+  explicit FeedbackVectorSpec(Zone* zone)
+      : slot_kinds_(zone), create_closure_parameter_counts_(zone) {
     slot_kinds_.reserve(16);
   }
 
   int slot_count() const { return static_cast<int>(slot_kinds_.size()); }
-  int create_closure_slot_count() const { return create_closure_slot_count_; }
+  int create_closure_slot_count() const {
+    return static_cast<int>(create_closure_parameter_counts_.size());
+  }
 
-  int AddCreateClosureSlot() { return create_closure_slot_count_++; }
+  int AddCreateClosureParameterCount(uint16_t parameter_count) {
+    create_closure_parameter_counts_.push_back(parameter_count);
+    return create_closure_slot_count() - 1;
+  }
+
+  uint16_t GetCreateClosureParameterCount(int index) const {
+    return create_closure_parameter_counts_.at(index);
+  }
 
   FeedbackSlotKind GetKind(FeedbackSlot slot) const {
     return slot_kinds_.at(slot.ToInt());
@@ -505,6 +659,8 @@ class V8_EXPORT_PRIVATE FeedbackVectorSpec {
     return AddSlot(FeedbackSlotKind::kInstanceOf);
   }
 
+  FeedbackSlot AddTypeOfSlot() { return AddSlot(FeedbackSlotKind::kTypeOf); }
+
   FeedbackSlot AddLiteralSlot() { return AddSlot(FeedbackSlotKind::kLiteral); }
 
   FeedbackSlot AddDefineKeyedOwnPropertyInLiteralICSlot() {
@@ -519,6 +675,10 @@ class V8_EXPORT_PRIVATE FeedbackVectorSpec {
     return AddSlot(FeedbackSlotKind::kJumpLoop);
   }
 
+  FeedbackSlot AddStringAddAndInternalizeICSlot() {
+    return AddSlot(FeedbackSlotKind::kStringAddAndInternalize);
+  }
+
 #ifdef OBJECT_PRINT
   // For gdb debugging.
   void Print();
@@ -533,7 +693,8 @@ class V8_EXPORT_PRIVATE FeedbackVectorSpec {
 
   static_assert(sizeof(FeedbackSlotKind) == sizeof(uint8_t));
   ZoneVector<FeedbackSlotKind> slot_kinds_;
-  int create_closure_slot_count_ = 0;
+  // A vector containing the parameter count for every create closure slot.
+  ZoneVector<uint16_t> create_closure_parameter_counts_;
 
   friend class SharedFeedbackSlot;
 };
@@ -561,10 +722,15 @@ class SharedFeedbackSlot {
 // of int32 data. The length is never stored - it is always calculated from
 // slot_count. All instances are created through the static New function, and
 // the number of slots is static once an instance is created.
+//
+// Besides the feedback slots, the FeedbackMetadata also stores the parameter
+// count for every CreateClosure slot as that is required for allocating the
+// FeedbackCells for the closres. This data doesn't necessarily need to live in
+// this object (it could, for example, also be stored on the Bytecode), but
+// keeping it here is somewhat efficient as the uint16s can just be stored
+// after the int32s of the slots.
 class FeedbackMetadata : public HeapObject {
  public:
-  DECL_CAST(FeedbackMetadata)
-
   // The number of slots that this metadata contains. Stored as an int32.
   DECL_INT32_ACCESSORS(slot_count)
 
@@ -577,6 +743,9 @@ class FeedbackMetadata : public HeapObject {
   // Get slot_count using an acquire load.
   inline int32_t slot_count(AcquireLoadTag) const;
 
+  // Get create_closure_slot_count using an acquire load.
+  inline int32_t create_closure_slot_count(AcquireLoadTag) const;
+
   // Returns number of feedback vector elements used by given slot kind.
   static inline int GetSlotSize(FeedbackSlotKind kind);
 
@@ -587,10 +756,14 @@ class FeedbackMetadata : public HeapObject {
   // Returns slot kind for given slot.
   V8_EXPORT_PRIVATE FeedbackSlotKind GetKind(FeedbackSlot slot) const;
 
+  // Returns the parameter count for the create closure slot with the given
+  // index.
+  V8_EXPORT_PRIVATE uint16_t GetCreateClosureParameterCount(int index) const;
+
   // If {spec} is null, then it is considered empty.
   template <typename IsolateT>
   V8_EXPORT_PRIVATE static Handle<FeedbackMetadata> New(
-      IsolateT* isolate, const FeedbackVectorSpec* spec = nullptr);
+      IsolateT* isolate, const FeedbackVectorSpec* spec);
 
   DECL_PRINTER(FeedbackMetadata)
   DECL_VERIFIER(FeedbackMetadata)
@@ -600,8 +773,12 @@ class FeedbackMetadata : public HeapObject {
   // Garbage collection support.
   // This includes any necessary padding at the end of the object for pointer
   // size alignment.
-  static int SizeFor(int slot_count) {
-    return OBJECT_POINTER_ALIGN(kHeaderSize + length(slot_count) * kInt32Size);
+  inline int AllocatedSize();
+
+  static int SizeFor(int slot_count, int create_closure_slot_count) {
+    return OBJECT_POINTER_ALIGN(kHeaderSize +
+                                word_count(slot_count) * kInt32Size +
+                                create_closure_slot_count * kUInt16Size);
   }
 
 #define FIELDS(V)                              \
@@ -623,16 +800,18 @@ class FeedbackMetadata : public HeapObject {
 
   // The number of int32 data fields needed to store {slot_count} slots.
   // Does not include any extra padding for pointer size alignment.
-  static int length(int slot_count) {
+  static int word_count(int slot_count) {
     return VectorICComputer::word_count(slot_count);
   }
-  inline int length() const;
+  inline int word_count() const;
 
   static const int kFeedbackSlotKindBits = 5;
   static_assert(kFeedbackSlotKindCount <= (1 << kFeedbackSlotKindBits));
 
   void SetKind(FeedbackSlot slot, FeedbackSlotKind kind);
 
+  void SetCreateClosureParameterCount(int index, uint16_t parameter_count);
+
   using VectorICComputer =
       base::BitSetComputer<FeedbackSlotKind, kFeedbackSlotKindBits,
                            kInt32Size * kBitsPerByte, uint32_t>;
@@ -654,7 +833,8 @@ class FeedbackMetadataIterator {
         next_slot_(FeedbackSlot(0)),
         slot_kind_(FeedbackSlotKind::kInvalid) {}
 
-  explicit FeedbackMetadataIterator(Tagged<FeedbackMetadata> metadata)
+  FeedbackMetadataIterator(Tagged<FeedbackMetadata> metadata,
+                           const DisallowGarbageCollection& no_gc)
       : metadata_(metadata),
         next_slot_(FeedbackSlot(0)),
         slot_kind_(FeedbackSlotKind::kInvalid) {}
@@ -702,11 +882,13 @@ class FeedbackMetadataIterator {
 class V8_EXPORT_PRIVATE NexusConfig {
  public:
   static NexusConfig FromMainThread(Isolate* isolate) {
+    DCHECK_NOT_NULL(isolate);
     return NexusConfig(isolate);
   }
 
   static NexusConfig FromBackgroundThread(Isolate* isolate,
                                           LocalHeap* local_heap) {
+    DCHECK_NOT_NULL(isolate);
     return NexusConfig(isolate, local_heap);
   }
 
@@ -718,23 +900,23 @@ class V8_EXPORT_PRIVATE NexusConfig {
 
   Isolate* isolate() const { return isolate_; }
 
-  MaybeObjectHandle NewHandle(MaybeObject object) const;
+  MaybeObjectHandle NewHandle(Tagged<MaybeObject> object) const;
   template <typename T>
   Handle<T> NewHandle(Tagged<T> object) const;
 
   bool can_write() const { return mode() == MainThread; }
 
-  inline MaybeObject GetFeedback(Tagged<FeedbackVector> vector,
-                                 FeedbackSlot slot) const;
+  inline Tagged<MaybeObject> GetFeedback(Tagged<FeedbackVector> vector,
+                                         FeedbackSlot slot) const;
   inline void SetFeedback(Tagged<FeedbackVector> vector, FeedbackSlot slot,
-                          MaybeObject object,
+                          Tagged<MaybeObject> object,
                           WriteBarrierMode mode = UPDATE_WRITE_BARRIER) const;
 
-  std::pair<MaybeObject, MaybeObject> GetFeedbackPair(
+  std::pair<Tagged<MaybeObject>, Tagged<MaybeObject>> GetFeedbackPair(
       Tagged<FeedbackVector> vector, FeedbackSlot slot) const;
   void SetFeedbackPair(Tagged<FeedbackVector> vector, FeedbackSlot start_slot,
-                       MaybeObject feedback, WriteBarrierMode mode,
-                       MaybeObject feedback_extra,
+                       Tagged<MaybeObject> feedback, WriteBarrierMode mode,
+                       Tagged<MaybeObject> feedback_extra,
                        WriteBarrierMode mode_extra) const;
 
  private:
@@ -751,8 +933,9 @@ class V8_EXPORT_PRIVATE NexusConfig {
 class V8_EXPORT_PRIVATE FeedbackNexus final {
  public:
   // For use on the main thread. A null {vector} is accepted as well.
-  FeedbackNexus(Handle<FeedbackVector> vector, FeedbackSlot slot);
-  FeedbackNexus(Tagged<FeedbackVector> vector, FeedbackSlot slot);
+  FeedbackNexus(Isolate* isolate, Handle<FeedbackVector> vector,
+                FeedbackSlot slot);
+  FeedbackNexus(Isolate*, Tagged<FeedbackVector> vector, FeedbackSlot slot);
 
   // For use on the main or background thread as configured by {config}.
   // {vector} must be valid.
@@ -760,7 +943,7 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
                 const NexusConfig& config);
 
   const NexusConfig* config() const { return &config_; }
-  Handle<FeedbackVector> vector_handle() const {
+  DirectHandle<FeedbackVector> vector_handle() const {
     DCHECK(vector_.is_null());
     return vector_handle_;
   }
@@ -794,15 +977,13 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
   // the extra feedback. This is used by ICs when updating the handlers.
   using TryUpdateHandler = std::function<MaybeHandle<Map>(Handle<Map>)>;
   int ExtractMapsAndHandlers(
-      std::vector<MapAndHandler>* maps_and_handlers,
+      MapsAndHandlers* maps_and_handlers,
       TryUpdateHandler map_handler = TryUpdateHandler()) const;
-  MaybeObjectHandle FindHandlerForMap(Handle<Map> map) const;
-  // Used to obtain maps and the associated feedback stored in the feedback
-  // vector. The returned feedback need not be always a handler. It could be a
-  // name in the case of StoreDataInPropertyLiteral. This is used by TurboFan to
-  // get all the feedback stored in the vector.
-  int ExtractMapsAndFeedback(
-      std::vector<MapAndFeedback>* maps_and_feedback) const;
+  MaybeObjectDirectHandle FindHandlerForMap(DirectHandle<Map> map) const;
+  // Used to obtain maps. This is used by compilers to get all the feedback
+  // stored in the vector.
+  template <typename F>
+  void IterateMapsWithUnclearedHandler(F) const;
 
   bool IsCleared() const {
     InlineCacheState state = ic_state();
@@ -817,23 +998,24 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
   bool ConfigureMegamorphic();
   bool ConfigureMegamorphic(IcCheckType property_type);
 
-  inline MaybeObject GetFeedback() const;
-  inline MaybeObject GetFeedbackExtra() const;
-  inline std::pair<MaybeObject, MaybeObject> GetFeedbackPair() const;
+  inline Tagged<MaybeObject> GetFeedback() const;
+  inline Tagged<MaybeObject> GetFeedbackExtra() const;
+  inline std::pair<Tagged<MaybeObject>, Tagged<MaybeObject>> GetFeedbackPair()
+      const;
 
-  inline Isolate* GetIsolate() const;
+  void ConfigureMonomorphic(DirectHandle<Name> name,
+                            DirectHandle<Map> receiver_map,
+                            const MaybeObjectDirectHandle& handler);
 
-  void ConfigureMonomorphic(Handle<Name> name, Handle<Map> receiver_map,
-                            const MaybeObjectHandle& handler);
+  void ConfigurePolymorphic(DirectHandle<Name> name,
+                            MapsAndHandlers const& maps_and_handlers);
 
-  void ConfigurePolymorphic(
-      Handle<Name> name, std::vector<MapAndHandler> const& maps_and_handlers);
-
-  void ConfigureMegaDOM(const MaybeObjectHandle& handler);
+  void ConfigureMegaDOM(const MaybeObjectDirectHandle& handler);
   MaybeObjectHandle ExtractMegaDOMHandler();
 
   BinaryOperationHint GetBinaryOperationFeedback() const;
   CompareOperationHint GetCompareOperationFeedback() const;
+  TypeOfFeedback::Result GetTypeOfFeedback() const;
   ForInHint GetForInFeedback() const;
 
   // For KeyedLoad ICs.
@@ -845,10 +1027,12 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
   // For KeyedLoad and KeyedStore ICs.
   IcCheckType GetKeyType() const;
   Tagged<Name> GetName() const;
+  bool IsOneMapManyNames() const;
 
   // For Call ICs.
   int GetCallCount();
   void SetSpeculationMode(SpeculationMode mode);
+  void NextSpeculationMode(SpeculationMode mode);
   SpeculationMode GetSpeculationMode();
   CallFeedbackContent GetCallFeedbackContent();
 
@@ -856,23 +1040,23 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
   // count (taken from the type feedback vector).
   float ComputeCallFrequency();
 
-  using SpeculationModeField = base::BitField<SpeculationMode, 0, 1>;
-  using CallFeedbackContentField = base::BitField<CallFeedbackContent, 1, 1>;
-  using CallCountField = base::BitField<uint32_t, 2, 30>;
+  using SpeculationModeField = base::BitField<SpeculationMode, 0, 2>;
+  using CallFeedbackContentField = base::BitField<CallFeedbackContent, 2, 1>;
+  using CallCountField = base::BitField<uint32_t, 3, 29>;
 
   // For InstanceOf ICs.
-  MaybeHandle<JSObject> GetConstructorFeedback() const;
+  MaybeDirectHandle<JSObject> GetConstructorFeedback() const;
 
   // For Global Load and Store ICs.
-  void ConfigurePropertyCellMode(Handle<PropertyCell> cell);
+  void ConfigurePropertyCellMode(DirectHandle<PropertyCell> cell);
   // Returns false if given combination of indices is not allowed.
   bool ConfigureLexicalVarMode(int script_context_index, int context_slot_index,
                                bool immutable);
-  void ConfigureHandlerMode(const MaybeObjectHandle& handler);
+  void ConfigureHandlerMode(const MaybeObjectDirectHandle& handler);
 
   // For CloneObject ICs
   static constexpr int kCloneObjectPolymorphicEntrySize = 2;
-  void ConfigureCloneObject(Handle<Map> source_map,
+  void ConfigureCloneObject(DirectHandle<Map> source_map,
                             const MaybeObjectHandle& handler);
 
 // Bit positions in a smi that encodes lexical environment variable access.
@@ -889,23 +1073,23 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
 
  private:
   template <typename FeedbackType>
-  inline void SetFeedback(FeedbackType feedback,
+  inline void SetFeedback(Tagged<FeedbackType> feedback,
                           WriteBarrierMode mode = UPDATE_WRITE_BARRIER);
   template <typename FeedbackType, typename FeedbackExtraType>
-  inline void SetFeedback(FeedbackType feedback, WriteBarrierMode mode,
-                          FeedbackExtraType feedback_extra,
+  inline void SetFeedback(Tagged<FeedbackType> feedback, WriteBarrierMode mode,
+                          Tagged<FeedbackExtraType> feedback_extra,
                           WriteBarrierMode mode_extra = UPDATE_WRITE_BARRIER);
 
-  inline MaybeObject UninitializedSentinel() const;
-  inline MaybeObject MegamorphicSentinel() const;
-  inline MaybeObject MegaDOMSentinel() const;
+  inline Tagged<MaybeObject> UninitializedSentinel() const;
+  inline Tagged<MaybeObject> MegamorphicSentinel() const;
+  inline Tagged<MaybeObject> MegaDOMSentinel() const;
 
   // Create an array. The caller must install it in a feedback vector slot.
-  Handle<WeakFixedArray> CreateArrayOfSize(int length);
+  DirectHandle<WeakFixedArray> CreateArrayOfSize(int length);
 
   // Helpers to maintain feedback_cache_.
-  inline MaybeObject FromHandle(MaybeObjectHandle slot) const;
-  inline MaybeObjectHandle ToHandle(MaybeObject value) const;
+  inline Tagged<MaybeObject> FromHandle(MaybeObjectDirectHandle slot) const;
+  inline MaybeObjectHandle ToHandle(Tagged<MaybeObject> value) const;
 
   // The reason for having a vector handle and a raw pointer is that we can and
   // should use handles during IC miss, but not during GC when we clear ICs. If
@@ -917,9 +1101,10 @@ class V8_EXPORT_PRIVATE FeedbackNexus final {
   FeedbackSlotKind kind_;
   // When using the background-thread configuration, a cache is used to
   // guarantee a consistent view of the feedback to FeedbackNexus methods.
-  mutable base::Optional<std::pair<MaybeObjectHandle, MaybeObjectHandle>>
+  mutable std::optional<std::pair<MaybeObjectHandle, MaybeObjectHandle>>
       feedback_cache_;
   NexusConfig config_;
+  Isolate* isolate_;
 };
 
 class V8_EXPORT_PRIVATE FeedbackIterator final {
@@ -928,7 +1113,7 @@ class V8_EXPORT_PRIVATE FeedbackIterator final {
   void Advance();
   bool done() { return done_; }
   Tagged<Map> map() { return map_; }
-  MaybeObject handler() { return handler_; }
+  Tagged<MaybeObject> handler() { return handler_; }
 
   static int SizeFor(int number_of_entries) {
     CHECK_GT(number_of_entries, 0);
@@ -952,9 +1137,9 @@ class V8_EXPORT_PRIVATE FeedbackIterator final {
   void AdvancePolymorphic();
   enum State { kMonomorphic, kPolymorphic, kOther };
 
-  Handle<WeakFixedArray> polymorphic_feedback_;
+  DirectHandle<WeakFixedArray> polymorphic_feedback_;
   Tagged<Map> map_;
-  MaybeObject handler_;
+  Tagged<MaybeObject> handler_;
   bool done_;
   int index_;
   State state_;
@@ -964,8 +1149,7 @@ inline BinaryOperationHint BinaryOperationHintFromFeedback(int type_feedback);
 inline CompareOperationHint CompareOperationHintFromFeedback(int type_feedback);
 inline ForInHint ForInHintFromFeedback(ForInFeedback type_feedback);
 
-}  // namespace internal
-}  // namespace v8
+}  // namespace v8::internal
 
 #include "src/objects/object-macros-undef.h"
 
