@@ -10,9 +10,15 @@
 
 #include "src/base/bits.h"
 #include "src/base/ieee754.h"
-#include "src/base/safe_conversions.h"
+#include "src/base/numerics/safe_conversions.h"
 #include "src/common/assert-scope.h"
+#include "src/execution/pointer-authentication.h"
+#include "src/numbers/conversions.h"
+#include "src/numbers/ieee754.h"
+#include "src/roots/roots-inl.h"
 #include "src/utils/memcopy.h"
+#include "src/wasm/float16.h"
+#include "src/wasm/wasm-engine.h"
 #include "src/wasm/wasm-objects-inl.h"
 
 #if defined(ADDRESS_SANITIZER) || defined(MEMORY_SANITIZER) || \
@@ -32,6 +38,7 @@
 #endif
 
 #include "src/base/memory.h"
+#include "src/base/overflowing-math.h"
 #include "src/utils/utils.h"
 #include "src/wasm/wasm-external-refs.h"
 
@@ -281,6 +288,14 @@ void float64_to_uint64_sat_wrapper(Address data) {
   WriteUnalignedValue<uint64_t>(data, 0);
 }
 
+void float16_to_float32_wrapper(Address data) {
+  WriteUnalignedValue<float>(data, Float16::Read(data).ToFloat32());
+}
+
+void float32_to_float16_wrapper(Address data) {
+  Float16::FromFloat32(ReadUnalignedValue<float>(data)).Write(data);
+}
+
 int32_t int64_div_wrapper(Address data) {
   int64_t dividend = ReadUnalignedValue<int64_t>(data);
   int64_t divisor = ReadUnalignedValue<int64_t>(data + sizeof(dividend));
@@ -328,52 +343,26 @@ int32_t uint64_mod_wrapper(Address data) {
   return 1;
 }
 
-uint32_t word32_ctz_wrapper(Address data) {
-  return base::bits::CountTrailingZeros(ReadUnalignedValue<uint32_t>(data));
+uint32_t word32_rol_wrapper(uint32_t input, uint32_t shift) {
+  return (input << (shift & 31)) | (input >> ((32 - shift) & 31));
 }
 
-uint32_t word64_ctz_wrapper(Address data) {
-  return base::bits::CountTrailingZeros(ReadUnalignedValue<uint64_t>(data));
+uint32_t word32_ror_wrapper(uint32_t input, uint32_t shift) {
+  return (input >> (shift & 31)) | (input << ((32 - shift) & 31));
 }
 
-uint32_t word32_popcnt_wrapper(Address data) {
-  return base::bits::CountPopulation(ReadUnalignedValue<uint32_t>(data));
+uint64_t word64_rol_wrapper(uint64_t input, uint32_t shift) {
+  return (input << (shift & 63)) | (input >> ((64 - shift) & 63));
 }
 
-uint32_t word64_popcnt_wrapper(Address data) {
-  return base::bits::CountPopulation(ReadUnalignedValue<uint64_t>(data));
-}
-
-uint32_t word32_rol_wrapper(Address data) {
-  uint32_t input = ReadUnalignedValue<uint32_t>(data);
-  uint32_t shift = ReadUnalignedValue<uint32_t>(data + sizeof(input)) & 31;
-  return (input << shift) | (input >> ((32 - shift) & 31));
-}
-
-uint32_t word32_ror_wrapper(Address data) {
-  uint32_t input = ReadUnalignedValue<uint32_t>(data);
-  uint32_t shift = ReadUnalignedValue<uint32_t>(data + sizeof(input)) & 31;
-  return (input >> shift) | (input << ((32 - shift) & 31));
-}
-
-void word64_rol_wrapper(Address data) {
-  uint64_t input = ReadUnalignedValue<uint64_t>(data);
-  uint64_t shift = ReadUnalignedValue<uint64_t>(data + sizeof(input)) & 63;
-  uint64_t result = (input << shift) | (input >> ((64 - shift) & 63));
-  WriteUnalignedValue<uint64_t>(data, result);
-}
-
-void word64_ror_wrapper(Address data) {
-  uint64_t input = ReadUnalignedValue<uint64_t>(data);
-  uint64_t shift = ReadUnalignedValue<uint64_t>(data + sizeof(input)) & 63;
-  uint64_t result = (input >> shift) | (input << ((64 - shift) & 63));
-  WriteUnalignedValue<uint64_t>(data, result);
+uint64_t word64_ror_wrapper(uint64_t input, uint32_t shift) {
+  return (input >> (shift & 63)) | (input << ((64 - shift) & 63));
 }
 
 void float64_pow_wrapper(Address data) {
   double x = ReadUnalignedValue<double>(data);
   double y = ReadUnalignedValue<double>(data + sizeof(x));
-  WriteUnalignedValue<double>(data, base::ieee754::pow(x, y));
+  WriteUnalignedValue<double>(data, math::pow(x, y));
 }
 
 template <typename T, T (*float_round_op)(T)>
@@ -421,45 +410,302 @@ void f32x4_nearest_int_wrapper(Address data) {
   simd_float_round_wrapper<float, &nearbyintf>(data);
 }
 
-namespace {
-class V8_NODISCARD ThreadNotInWasmScope {
-// Asan on Windows triggers exceptions to allocate shadow memory lazily. When
-// this function is called from WebAssembly, these exceptions would be handled
-// by the trap handler before they get handled by Asan, and thereby confuse the
-// thread-in-wasm flag. Therefore we disable ASAN for this function.
-// Alternatively we could reset the thread-in-wasm flag before calling this
-// function. However, as this is only a problem with Asan on Windows, we did not
-// consider it worth the overhead.
-#if defined(RESET_THREAD_IN_WASM_FLAG_FOR_ASAN_ON_WINDOWS)
+Float16 f16_abs(Float16 a) {
+  return Float16::FromFloat32(std::abs(a.ToFloat32()));
+}
 
- public:
-  ThreadNotInWasmScope() : thread_was_in_wasm_(trap_handler::IsThreadInWasm()) {
-    if (thread_was_in_wasm_) {
-      trap_handler::ClearThreadInWasm();
-    }
+void f16x8_abs_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_abs>(data);
+}
+
+Float16 f16_neg(Float16 a) { return Float16::FromFloat32(-(a.ToFloat32())); }
+
+void f16x8_neg_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_neg>(data);
+}
+
+Float16 f16_sqrt(Float16 a) {
+  return Float16::FromFloat32(std::sqrt(a.ToFloat32()));
+}
+
+void f16x8_sqrt_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_sqrt>(data);
+}
+
+Float16 f16_ceil(Float16 a) {
+  return Float16::FromFloat32(ceilf(a.ToFloat32()));
+}
+
+void f16x8_ceil_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_ceil>(data);
+}
+
+Float16 f16_floor(Float16 a) {
+  return Float16::FromFloat32(floorf(a.ToFloat32()));
+}
+
+void f16x8_floor_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_floor>(data);
+}
+
+Float16 f16_trunc(Float16 a) {
+  return Float16::FromFloat32(truncf(a.ToFloat32()));
+}
+
+void f16x8_trunc_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_trunc>(data);
+}
+
+Float16 f16_nearest_int(Float16 a) {
+  return Float16::FromFloat32(nearbyintf(a.ToFloat32()));
+}
+
+void f16x8_nearest_int_wrapper(Address data) {
+  simd_float_round_wrapper<Float16, &f16_nearest_int>(data);
+}
+
+template <typename R, R (*float_bin_op)(Float16, Float16)>
+void simd_float16_bin_wrapper(Address data) {
+  constexpr int n = kSimd128Size / sizeof(Float16);
+  for (int i = 0; i < n; i++) {
+    Float16 lhs = Float16::Read(data + (i * sizeof(Float16)));
+    Float16 rhs = Float16::Read(data + kSimd128Size + (i * sizeof(Float16)));
+    R value = float_bin_op(lhs, rhs);
+    WriteUnalignedValue<R>(data + (i * sizeof(R)), value);
   }
+}
 
-  ~ThreadNotInWasmScope() {
-    if (thread_was_in_wasm_) {
-      trap_handler::SetThreadInWasm();
-    }
+int16_t f16_eq(Float16 a, Float16 b) {
+  return a.ToFloat32() == b.ToFloat32() ? -1 : 0;
+}
+
+void f16x8_eq_wrapper(Address data) {
+  simd_float16_bin_wrapper<int16_t, &f16_eq>(data);
+}
+
+int16_t f16_ne(Float16 a, Float16 b) {
+  return a.ToFloat32() != b.ToFloat32() ? -1 : 0;
+}
+
+void f16x8_ne_wrapper(Address data) {
+  simd_float16_bin_wrapper<int16_t, &f16_ne>(data);
+}
+
+int16_t f16_lt(Float16 a, Float16 b) {
+  return a.ToFloat32() < b.ToFloat32() ? -1 : 0;
+}
+
+void f16x8_lt_wrapper(Address data) {
+  simd_float16_bin_wrapper<int16_t, &f16_lt>(data);
+}
+
+int16_t f16_le(Float16 a, Float16 b) {
+  return a.ToFloat32() <= b.ToFloat32() ? -1 : 0;
+}
+
+void f16x8_le_wrapper(Address data) {
+  simd_float16_bin_wrapper<int16_t, &f16_le>(data);
+}
+
+Float16 f16_add(Float16 a, Float16 b) {
+  return Float16::FromFloat32(a.ToFloat32() + b.ToFloat32());
+}
+
+void f16x8_add_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_add>(data);
+}
+
+Float16 f16_sub(Float16 a, Float16 b) {
+  return Float16::FromFloat32(a.ToFloat32() - b.ToFloat32());
+}
+
+void f16x8_sub_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_sub>(data);
+}
+
+Float16 f16_mul(Float16 a, Float16 b) {
+  return Float16::FromFloat32(a.ToFloat32() * b.ToFloat32());
+}
+
+void f16x8_mul_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_mul>(data);
+}
+
+Float16 f16_div(Float16 a, Float16 b) {
+  return Float16::FromFloat32(base::Divide(a.ToFloat32(), b.ToFloat32()));
+}
+
+void f16x8_div_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_div>(data);
+}
+
+Float16 f16_min(Float16 a, Float16 b) {
+  return Float16::FromFloat32(JSMin(a.ToFloat32(), b.ToFloat32()));
+}
+
+void f16x8_min_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_min>(data);
+}
+
+Float16 f16_max(Float16 a, Float16 b) {
+  return Float16::FromFloat32(JSMax(a.ToFloat32(), b.ToFloat32()));
+}
+
+void f16x8_max_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_max>(data);
+}
+
+Float16 f16_pmin(Float16 a, Float16 b) {
+  return Float16::FromFloat32(std::min(a.ToFloat32(), b.ToFloat32()));
+}
+
+void f16x8_pmin_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_pmin>(data);
+}
+
+Float16 f16_pmax(Float16 a, Float16 b) {
+  return Float16::FromFloat32(std::max(a.ToFloat32(), b.ToFloat32()));
+}
+
+void f16x8_pmax_wrapper(Address data) {
+  simd_float16_bin_wrapper<Float16, &f16_pmax>(data);
+}
+
+template <typename T, typename R, R (*float_un_op)(T)>
+void simd_float_un_wrapper(Address data) {
+  constexpr int n = kSimd128Size / sizeof(T);
+  for (int i = 0; i < n; i++) {
+    T input = ReadUnalignedValue<T>(data + (i * sizeof(T)));
+    R value = float_un_op(input);
+    WriteUnalignedValue<R>(data + (i * sizeof(T)), value);
   }
+}
 
- private:
-  bool thread_was_in_wasm_;
+int16_t ConvertToIntS(Float16 val) {
+  float f32 = val.ToFloat32();
+  if (std::isnan(f32)) return 0;
+  if (f32 > float{kMaxInt16}) return kMaxInt16;
+  if (f32 < float{kMinInt16}) return kMinInt16;
+  return static_cast<int16_t>(f32);
+}
+
+uint16_t ConvertToIntU(Float16 val) {
+  float f32 = val.ToFloat32();
+  if (std::isnan(f32)) return 0;
+  if (f32 > float{kMaxUInt16}) return kMaxUInt16;
+  if (f32 < float{0}) return 0;
+  return static_cast<uint16_t>(f32);
+}
+
+void i16x8_sconvert_f16x8_wrapper(Address data) {
+  simd_float_un_wrapper<Float16, int16_t, &ConvertToIntS>(data);
+}
+
+void i16x8_uconvert_f16x8_wrapper(Address data) {
+  simd_float_un_wrapper<Float16, uint16_t, &ConvertToIntU>(data);
+}
+
+Float16 ConvertToF16S(int16_t val) { return Float16::FromFloat32(val); }
+
+void f16x8_sconvert_i16x8_wrapper(Address data) {
+  simd_float_un_wrapper<int16_t, Float16, &ConvertToF16S>(data);
+}
+
+Float16 ConvertToF16U(uint16_t val) { return Float16::FromFloat32(val); }
+
+void f16x8_uconvert_i16x8_wrapper(Address data) {
+  simd_float_un_wrapper<uint16_t, Float16, &ConvertToF16U>(data);
+}
+
+void f32x4_promote_low_f16x8_wrapper(Address data) {
+  // Result is stored in the same buffer, so read all values to local
+  // stack variables first.
+  Float16 a = Float16::Read(data);
+  Float16 b = Float16::Read(data + sizeof(Float16));
+  Float16 c = Float16::Read(data + 2 * sizeof(Float16));
+  Float16 d = Float16::Read(data + 3 * sizeof(Float16));
+
+  WriteUnalignedValue<float>(data, a.ToFloat32());
+  WriteUnalignedValue<float>(data + sizeof(float), b.ToFloat32());
+  WriteUnalignedValue<float>(data + (2 * sizeof(float)), c.ToFloat32());
+  WriteUnalignedValue<float>(data + (3 * sizeof(float)), d.ToFloat32());
+}
+
+void f16x8_demote_f32x4_zero_wrapper(Address data) {
+#if V8_TARGET_BIG_ENDIAN
+  for (int i = 3, j = 7; i >= 0; i--, j--) {
+    float input = ReadUnalignedValue<float>(data + (i * sizeof(float)));
+    Float16::FromFloat32(input).Write(data + (j * sizeof(Float16)));
+  }
+  for (int i = 0; i < 4; i++) {
+    WriteUnalignedValue<Float16>(data + (i * sizeof(Float16)),
+                                 Float16::FromFloat32(0));
+  }
 #else
-
- public:
-  ThreadNotInWasmScope() {
-    // This is needed to avoid compilation errors (unused variable).
-    USE(this);
+  for (int i = 0; i < 4; i++) {
+    float input = ReadUnalignedValue<float>(data + (i * sizeof(float)));
+    Float16::FromFloat32(input).Write(data + (i * sizeof(Float16)));
+  }
+  for (int i = 4; i < 8; i++) {
+    WriteUnalignedValue<Float16>(data + (i * sizeof(Float16)),
+                                 Float16::FromFloat32(0));
   }
 #endif
-};
+}
 
-inline uint8_t* EffectiveAddress(Tagged<WasmInstanceObject> instance,
+void f16x8_demote_f64x2_zero_wrapper(Address data) {
+#if V8_TARGET_BIG_ENDIAN
+  for (int i = 1, j = 7; i >= 0; i--, j--) {
+    double input = ReadUnalignedValue<double>(data + (i * sizeof(double)));
+    WriteUnalignedValue<uint16_t>(data + (j * sizeof(uint16_t)),
+                                  DoubleToFloat16(input));
+  }
+  for (int i = 0; i < 6; i++) {
+    WriteUnalignedValue<Float16>(data + (i * sizeof(Float16)),
+                                 Float16::FromFloat32(0));
+  }
+#else
+  for (int i = 0; i < 2; i++) {
+    double input = ReadUnalignedValue<double>(data + (i * sizeof(double)));
+    WriteUnalignedValue<uint16_t>(data + (i * sizeof(uint16_t)),
+                                  DoubleToFloat16(input));
+  }
+  for (int i = 2; i < 8; i++) {
+    WriteUnalignedValue<Float16>(data + (i * sizeof(Float16)),
+                                 Float16::FromFloat32(0));
+  }
+#endif
+}
+
+template <float (*float_fma_op)(float, float, float)>
+void simd_float16_fma_wrapper(Address data) {
+  constexpr int n = kSimd128Size / sizeof(Float16);
+  for (int i = 0; i < n; i++) {
+    Address offset = data + i * sizeof(Float16);
+    Float16 a = Float16::Read(offset);
+    Float16 b = Float16::Read(offset + kSimd128Size);
+    Float16 c = Float16::Read(offset + 2 * kSimd128Size);
+    float value = float_fma_op(a.ToFloat32(), b.ToFloat32(), c.ToFloat32());
+    Float16::FromFloat32(value).Write(offset);
+  }
+}
+
+float Qfma(float a, float b, float c) { return a * b + c; }
+
+void f16x8_qfma_wrapper(Address data) {
+  return simd_float16_fma_wrapper<&Qfma>(data);
+}
+
+float Qfms(float a, float b, float c) { return -(a * b) + c; }
+
+void f16x8_qfms_wrapper(Address data) {
+  return simd_float16_fma_wrapper<&Qfms>(data);
+}
+
+namespace {
+inline uint8_t* EffectiveAddress(Tagged<WasmTrustedInstanceData> trusted_data,
                                  uint32_t mem_index, uintptr_t index) {
-  return instance->memory_base(mem_index) + index;
+  return trusted_data->memory_base(mem_index) + index;
 }
 
 template <typename V>
@@ -473,71 +719,56 @@ constexpr int32_t kSuccess = 1;
 constexpr int32_t kOutOfBounds = 0;
 }  // namespace
 
-int32_t memory_init_wrapper(Address data) {
-  ThreadNotInWasmScope thread_not_in_wasm_scope;
+int32_t memory_init_wrapper(Address trusted_data_addr, uint32_t mem_index,
+                            uintptr_t dst, uint32_t src, uint32_t seg_index,
+                            uint32_t size) {
   DisallowGarbageCollection no_gc;
-  size_t offset = 0;
-  Tagged<WasmInstanceObject> instance = WasmInstanceObject::cast(
-      ReadAndIncrementOffset<Tagged<Object>>(data, &offset));
-  uint32_t mem_index = ReadAndIncrementOffset<uint32_t>(data, &offset);
-  uintptr_t dst = ReadAndIncrementOffset<uintptr_t>(data, &offset);
-  uint32_t src = ReadAndIncrementOffset<uint32_t>(data, &offset);
-  uint32_t seg_index = ReadAndIncrementOffset<uint32_t>(data, &offset);
-  uint32_t size = ReadAndIncrementOffset<uint32_t>(data, &offset);
+  Tagged<WasmTrustedInstanceData> trusted_data =
+      Cast<WasmTrustedInstanceData>(Tagged<Object>{trusted_data_addr});
 
-  uint64_t mem_size = instance->memory_size(mem_index);
+  uint64_t mem_size = trusted_data->memory_size(mem_index);
   if (!base::IsInBounds<uint64_t>(dst, size, mem_size)) return kOutOfBounds;
 
-  uint32_t seg_size = instance->data_segment_sizes()->get(seg_index);
+  uint32_t seg_size = trusted_data->data_segment_sizes()->get(seg_index);
   if (!base::IsInBounds<uint32_t>(src, size, seg_size)) return kOutOfBounds;
 
   uint8_t* seg_start = reinterpret_cast<uint8_t*>(
-      instance->data_segment_starts()->get(seg_index));
-  std::memcpy(EffectiveAddress(instance, mem_index, dst), seg_start + src,
+      trusted_data->data_segment_starts()->get(seg_index));
+  std::memcpy(EffectiveAddress(trusted_data, mem_index, dst), seg_start + src,
               size);
   return kSuccess;
 }
 
-int32_t memory_copy_wrapper(Address data) {
-  ThreadNotInWasmScope thread_not_in_wasm_scope;
+int32_t memory_copy_wrapper(Address trusted_data_addr, uint32_t dst_mem_index,
+                            uint32_t src_mem_index, uintptr_t dst,
+                            uintptr_t src, uintptr_t size) {
   DisallowGarbageCollection no_gc;
-  size_t offset = 0;
-  Tagged<WasmInstanceObject> instance = WasmInstanceObject::cast(
-      ReadAndIncrementOffset<Tagged<Object>>(data, &offset));
-  uint32_t dst_mem_index = ReadAndIncrementOffset<uint32_t>(data, &offset);
-  uint32_t src_mem_index = ReadAndIncrementOffset<uint32_t>(data, &offset);
-  uintptr_t dst = ReadAndIncrementOffset<uintptr_t>(data, &offset);
-  uintptr_t src = ReadAndIncrementOffset<uintptr_t>(data, &offset);
-  uintptr_t size = ReadAndIncrementOffset<uintptr_t>(data, &offset);
+  Tagged<WasmTrustedInstanceData> trusted_data =
+      Cast<WasmTrustedInstanceData>(Tagged<Object>{trusted_data_addr});
 
-  uint64_t dst_mem_size = instance->memory_size(dst_mem_index);
-  uint64_t src_mem_size = instance->memory_size(src_mem_index);
-  if (!base::IsInBounds<uint64_t>(dst, size, dst_mem_size)) return kOutOfBounds;
-  if (!base::IsInBounds<uint64_t>(src, size, src_mem_size)) return kOutOfBounds;
+  size_t dst_mem_size = trusted_data->memory_size(dst_mem_index);
+  size_t src_mem_size = trusted_data->memory_size(src_mem_index);
+  static_assert(std::is_same_v<size_t, uintptr_t>);
+  if (!base::IsInBounds<size_t>(dst, size, dst_mem_size)) return kOutOfBounds;
+  if (!base::IsInBounds<size_t>(src, size, src_mem_size)) return kOutOfBounds;
 
   // Use std::memmove, because the ranges can overlap.
-  std::memmove(EffectiveAddress(instance, dst_mem_index, dst),
-               EffectiveAddress(instance, src_mem_index, src), size);
+  std::memmove(EffectiveAddress(trusted_data, dst_mem_index, dst),
+               EffectiveAddress(trusted_data, src_mem_index, src), size);
   return kSuccess;
 }
 
-int32_t memory_fill_wrapper(Address data) {
-  ThreadNotInWasmScope thread_not_in_wasm_scope;
+int32_t memory_fill_wrapper(Address trusted_data_addr, uint32_t mem_index,
+                            uintptr_t dst, uint8_t value, uintptr_t size) {
   DisallowGarbageCollection no_gc;
 
-  size_t offset = 0;
-  Tagged<WasmInstanceObject> instance = WasmInstanceObject::cast(
-      ReadAndIncrementOffset<Tagged<Object>>(data, &offset));
-  uint32_t mem_index = ReadAndIncrementOffset<uint32_t>(data, &offset);
-  uintptr_t dst = ReadAndIncrementOffset<uintptr_t>(data, &offset);
-  uint8_t value =
-      static_cast<uint8_t>(ReadAndIncrementOffset<uint32_t>(data, &offset));
-  uintptr_t size = ReadAndIncrementOffset<uintptr_t>(data, &offset);
+  Tagged<WasmTrustedInstanceData> trusted_data =
+      Cast<WasmTrustedInstanceData>(Tagged<Object>{trusted_data_addr});
 
-  uint64_t mem_size = instance->memory_size(mem_index);
+  uint64_t mem_size = trusted_data->memory_size(mem_index);
   if (!base::IsInBounds<uint64_t>(dst, size, mem_size)) return kOutOfBounds;
 
-  std::memset(EffectiveAddress(instance, mem_index, dst), value, size);
+  std::memset(EffectiveAddress(trusted_data, mem_index, dst), value, size);
   return kSuccess;
 }
 
@@ -553,32 +784,30 @@ inline void* ArrayElementAddress(Tagged<WasmArray> array, uint32_t index,
 }
 }  // namespace
 
-void array_copy_wrapper(Address raw_instance, Address raw_dst_array,
-                        uint32_t dst_index, Address raw_src_array,
-                        uint32_t src_index, uint32_t length) {
+void array_copy_wrapper(Address raw_dst_array, uint32_t dst_index,
+                        Address raw_src_array, uint32_t src_index,
+                        uint32_t length) {
   DCHECK_GT(length, 0);
-  ThreadNotInWasmScope thread_not_in_wasm_scope;
   DisallowGarbageCollection no_gc;
-  Tagged<WasmArray> dst_array = WasmArray::cast(Tagged<Object>(raw_dst_array));
-  Tagged<WasmArray> src_array = WasmArray::cast(Tagged<Object>(raw_src_array));
+  Tagged<WasmArray> dst_array = Cast<WasmArray>(Tagged<Object>(raw_dst_array));
+  Tagged<WasmArray> src_array = Cast<WasmArray>(Tagged<Object>(raw_src_array));
 
   bool overlapping_ranges =
       dst_array.ptr() == src_array.ptr() &&
       (dst_index < src_index ? dst_index + length > src_index
                              : src_index + length > dst_index);
-  wasm::ValueType element_type = src_array->type()->element_type();
+  wasm::CanonicalValueType element_type =
+      src_array->map()->wasm_type_info()->element_type();
   if (element_type.is_reference()) {
-    Tagged<WasmInstanceObject> instance =
-        WasmInstanceObject::cast(Tagged<Object>(raw_instance));
-    Isolate* isolate = instance->GetIsolate();
     ObjectSlot dst_slot = dst_array->ElementSlot(dst_index);
     ObjectSlot src_slot = src_array->ElementSlot(src_index);
+    Heap* heap = Isolate::Current()->heap();
     if (overlapping_ranges) {
-      isolate->heap()->MoveRange(dst_array, dst_slot, src_slot, length,
-                                 UPDATE_WRITE_BARRIER);
+      heap->MoveRange(dst_array, dst_slot, src_slot, length,
+                      UPDATE_WRITE_BARRIER);
     } else {
-      isolate->heap()->CopyRange(dst_array, dst_slot, src_slot, length,
-                                 UPDATE_WRITE_BARRIER);
+      heap->CopyRange(dst_array, dst_slot, src_slot, length,
+                      UPDATE_WRITE_BARRIER);
     }
   } else {
     int element_size_bytes = element_type.value_kind_size();
@@ -596,62 +825,90 @@ void array_copy_wrapper(Address raw_instance, Address raw_dst_array,
 void array_fill_wrapper(Address raw_array, uint32_t index, uint32_t length,
                         uint32_t emit_write_barrier, uint32_t raw_type,
                         Address initial_value_addr) {
-  ThreadNotInWasmScope thread_not_in_wasm_scope;
   DisallowGarbageCollection no_gc;
   ValueType type = ValueType::FromRawBitField(raw_type);
   int8_t* initial_element_address = reinterpret_cast<int8_t*>(
       ArrayElementAddress(raw_array, index, type.value_kind_size()));
-  int64_t initial_value = *reinterpret_cast<int64_t*>(initial_value_addr);
   const int bytes_to_set = length * type.value_kind_size();
 
-  // If the initial value is zero, we memset the array.
-  if (type.is_numeric() && initial_value == 0) {
-    std::memset(initial_element_address, 0, bytes_to_set);
-    return;
-  }
-
   // We implement the general case by setting the first 8 bytes manually, then
   // filling the rest by exponentially growing {memcpy}s.
 
-  DCHECK_GE(static_cast<size_t>(bytes_to_set), sizeof(int64_t));
+  CHECK_GE(static_cast<size_t>(bytes_to_set), sizeof(int64_t));
 
   switch (type.kind()) {
     case kI64:
     case kF64: {
-      *reinterpret_cast<int64_t*>(initial_element_address) = initial_value;
+      // Stack pointers are only aligned to 4 bytes.
+      int64_t initial_value =
+          base::ReadUnalignedValue<int64_t>(initial_value_addr);
+      if (initial_value == 0) {
+        std::memset(initial_element_address, 0, bytes_to_set);
+        return;
+      }
+      // Array elements are only aligned to 4 bytes, therefore
+      // `initial_element_address` may be misaligned as a 64-bit pointer.
+      base::WriteUnalignedValue<int64_t>(
+          reinterpret_cast<Address>(initial_element_address), initial_value);
       break;
     }
     case kI32:
     case kF32: {
+      int32_t initial_value = *reinterpret_cast<int32_t*>(initial_value_addr);
+      if (initial_value == 0) {
+        std::memset(initial_element_address, 0, bytes_to_set);
+        return;
+      }
       int32_t* base = reinterpret_cast<int32_t*>(initial_element_address);
-      base[0] = base[1] = static_cast<int32_t>(initial_value);
+      base[0] = base[1] = initial_value;
       break;
     }
+    case kF16:
     case kI16: {
+      // The array.fill input is an i32!
+      int16_t initial_value = *reinterpret_cast<int32_t*>(initial_value_addr);
+      if (initial_value == 0) {
+        std::memset(initial_element_address, 0, bytes_to_set);
+        return;
+      }
       int16_t* base = reinterpret_cast<int16_t*>(initial_element_address);
-      base[0] = base[1] = base[2] = base[3] =
-          static_cast<int16_t>(initial_value);
+      base[0] = base[1] = base[2] = base[3] = initial_value;
       break;
     }
     case kI8: {
+      // The array.fill input is an i32!
+      int8_t initial_value = *reinterpret_cast<int32_t*>(initial_value_addr);
+      if (initial_value == 0) {
+        std::memset(initial_element_address, 0, bytes_to_set);
+        return;
+      }
       int8_t* base = reinterpret_cast<int8_t*>(initial_element_address);
       for (size_t i = 0; i < sizeof(int64_t); i++) {
-        base[i] = static_cast<int8_t>(initial_value);
+        base[i] = initial_value;
       }
       break;
     }
     case kRefNull:
-    case kRef:
+    case kRef: {
+      intptr_t uncompressed_pointer =
+          base::ReadUnalignedValue<intptr_t>(initial_value_addr);
       if constexpr (kTaggedSize == 4) {
         int32_t* base = reinterpret_cast<int32_t*>(initial_element_address);
-        base[0] = base[1] = static_cast<int32_t>(initial_value);
+        base[0] = base[1] = static_cast<int32_t>(uncompressed_pointer);
       } else {
-        *reinterpret_cast<int64_t*>(initial_element_address) = initial_value;
+        base::WriteUnalignedValue(
+            reinterpret_cast<Address>(initial_element_address),
+            uncompressed_pointer);
       }
       break;
+    }
     case kS128:
-    case kRtt:
+      // S128 can only be filled with zeros.
+      DCHECK_EQ(base::ReadUnalignedValue<int64_t>(initial_value_addr), 0);
+      std::memset(initial_element_address, 0, bytes_to_set);
+      return;
     case kVoid:
+    case kTop:
     case kBottom:
       UNREACHABLE();
   }
@@ -671,37 +928,48 @@ void array_fill_wrapper(Address raw_array, uint32_t index, uint32_t length,
 
   if (emit_write_barrier) {
     DCHECK(type.is_reference());
-    Tagged<WasmArray> array = WasmArray::cast(Tagged<Object>(raw_array));
-    Isolate* isolate = array->GetIsolate();
+    Tagged<WasmArray> array = Cast<WasmArray>(Tagged<Object>(raw_array));
+    Isolate* isolate = Isolate::Current();
     ObjectSlot start(reinterpret_cast<Address>(initial_element_address));
     ObjectSlot end(
         reinterpret_cast<Address>(initial_element_address + bytes_to_set));
-    isolate->heap()->WriteBarrierForRange(array, start, end);
+    WriteBarrier::ForRange(isolate->heap(), array, start, end);
   }
 }
 
 double flat_string_to_f64(Address string_address) {
-  Tagged<String> s = String::cast(Tagged<Object>(string_address));
+  Tagged<String> s = Cast<String>(Tagged<Object>(string_address));
   return FlatStringToDouble(s, ALLOW_TRAILING_JUNK,
                             std::numeric_limits<double>::quiet_NaN());
 }
 
-void sync_stack_limit(Isolate* isolate) {
-  CHECK(v8_flags.experimental_wasm_stack_switching);
-  DisallowGarbageCollection no_gc;
+void start_or_suspend_stack(Isolate* isolate, wasm::StackMemory* from) {
+  isolate->SwitchStacks(from, isolate->isolate_data()->active_stack());
+}
 
-  isolate->SyncStackLimit();
+void resume_stack(Isolate* isolate, wasm::StackMemory* from,
+                  Address suspender_raw) {
+  Tagged<Object> suspender_obj(suspender_raw);
+  auto suspender = Tagged<WasmSuspenderObject>::cast(suspender_obj);
+  Tagged<Object> active_suspender = isolate->isolate_data()->active_suspender();
+  if (active_suspender == Smi::zero()) {
+    suspender->clear_parent();
+  } else {
+    suspender->set_parent(Tagged<WasmSuspenderObject>::cast(active_suspender));
+  }
+  isolate->isolate_data()->set_active_suspender(suspender);
+  isolate->SwitchStacks(from, isolate->isolate_data()->active_stack());
+}
+
+void return_stack(Isolate* isolate, wasm::StackMemory* from) {
+  isolate->SwitchStacks(from, isolate->isolate_data()->active_stack());
+  isolate->RetireWasmStack(from);
 }
 
 intptr_t switch_to_the_central_stack(Isolate* isolate, uintptr_t current_sp) {
-  CHECK(v8_flags.experimental_wasm_stack_switching);
-
   ThreadLocalTop* thread_local_top = isolate->thread_local_top();
   StackGuard* stack_guard = isolate->stack_guard();
 
-  CHECK_EQ(thread_local_top->secondary_stack_sp_, 0);
-  CHECK_EQ(thread_local_top->secondary_stack_limit_, 0);
-
   auto secondary_stack_limit = stack_guard->real_jslimit();
 
   stack_guard->SetStackLimitForStackSwitching(
@@ -718,8 +986,6 @@ intptr_t switch_to_the_central_stack(Isolate* isolate, uintptr_t current_sp) {
 }
 
 void switch_from_the_central_stack(Isolate* isolate) {
-  CHECK(v8_flags.experimental_wasm_stack_switching);
-
   ThreadLocalTop* thread_local_top = isolate->thread_local_top();
   CHECK_NE(thread_local_top->secondary_stack_sp_, 0);
   CHECK_NE(thread_local_top->secondary_stack_limit_, 0);
@@ -733,6 +999,99 @@ void switch_from_the_central_stack(Isolate* isolate) {
   stack_guard->SetStackLimitForStackSwitching(secondary_stack_limit);
 }
 
+intptr_t switch_to_the_central_stack_for_js(Isolate* isolate, Address fp) {
+  ThreadLocalTop* thread_local_top = isolate->thread_local_top();
+  StackGuard* stack_guard = isolate->stack_guard();
+  wasm::StackMemory* stack = isolate->isolate_data()->active_stack();
+  Address central_stack_sp = thread_local_top->central_stack_sp_;
+  stack->set_stack_switch_info(fp, central_stack_sp);
+  stack_guard->SetStackLimitForStackSwitching(
+      thread_local_top->central_stack_limit_);
+  thread_local_top->is_on_central_stack_flag_ = true;
+  return central_stack_sp;
+}
+
+void switch_from_the_central_stack_for_js(Isolate* isolate) {
+  // The stack only contains wasm frames after this JS call.
+  wasm::StackMemory* stack = isolate->isolate_data()->active_stack();
+  stack->clear_stack_switch_info();
+  ThreadLocalTop* thread_local_top = isolate->thread_local_top();
+  thread_local_top->is_on_central_stack_flag_ = false;
+  StackGuard* stack_guard = isolate->stack_guard();
+  stack_guard->SetStackLimitForStackSwitching(
+      reinterpret_cast<uintptr_t>(stack->jslimit()));
+}
+
+// frame_size includes param slots area and extra frame slots above FP.
+Address grow_stack(Isolate* isolate, void* current_sp, size_t frame_size,
+                   size_t gap, Address current_fp) {
+  // Check if this is a real stack overflow.
+  StackLimitCheck check(isolate);
+  if (check.WasmHasOverflowed(gap)) {
+    wasm::StackMemory* active_stack = isolate->isolate_data()->active_stack();
+    if (isolate->IsOnCentralStack()) {
+      // Should not grow the central stack.
+      return 0;
+    }
+    DCHECK(active_stack->IsActive());
+    // Grow by at least the new frame size plus the stack limit margin.
+    size_t min =
+        gap + frame_size + StackMemory::JSGrowableStackLimitMarginKB() * KB;
+    if (!active_stack->Grow(current_fp, min)) {
+      return 0;
+    }
+
+    Address new_sp = active_stack->base() - frame_size;
+    // Here we assume stack values don't refer other moved stack slots.
+    // A stack grow event happens right in the beginning of the function
+    // call so moved slots contain only incoming params and frame header.
+    // So, it is reasonable to assume no self references.
+    std::memcpy(reinterpret_cast<void*>(new_sp), current_sp, frame_size);
+
+#if V8_TARGET_ARCH_ARM64
+    Address new_fp =
+        new_sp + (current_fp - reinterpret_cast<Address>(current_sp));
+    Address old_pc_address = current_fp + CommonFrameConstants::kCallerPCOffset;
+    Address new_pc_address = new_fp + CommonFrameConstants::kCallerPCOffset;
+    Address old_signed_pc = base::Memory<Address>(old_pc_address);
+    Address new_signed_pc = PointerAuthentication::MoveSignedPC(
+        isolate, old_signed_pc, new_pc_address + kSystemPointerSize,
+        old_pc_address + kSystemPointerSize);
+    WriteUnalignedValue<Address>(new_pc_address, new_signed_pc);
+#endif
+
+    isolate->stack_guard()->SetStackLimitForStackSwitching(
+        reinterpret_cast<uintptr_t>(active_stack->jslimit()));
+    return new_sp;
+  }
+
+  return 0;
+}
+
+Address shrink_stack(Isolate* isolate) {
+  // If there is no parent, then the current stack is the main isolate stack.
+  wasm::StackMemory* active_stack = isolate->isolate_data()->active_stack();
+  if (active_stack->jmpbuf()->parent == nullptr) {
+    return 0;
+  }
+  DCHECK(active_stack->IsActive());
+  Address old_fp = active_stack->Shrink();
+
+  isolate->stack_guard()->SetStackLimitForStackSwitching(
+      reinterpret_cast<uintptr_t>(active_stack->jslimit()));
+  return old_fp;
+}
+
+Address load_old_fp(Isolate* isolate) {
+  // If there is no parent, then the current stack is the main isolate stack.
+  wasm::StackMemory* active_stack = isolate->isolate_data()->active_stack();
+  if (active_stack->jmpbuf()->parent == nullptr) {
+    return 0;
+  }
+  DCHECK_EQ(active_stack->jmpbuf()->state, wasm::JumpBuffer::Active);
+  return active_stack->old_fp();
+}
+
 }  // namespace v8::internal::wasm
 
 #undef V8_WITH_SANITIZER
