@@ -7,6 +7,7 @@
 #include "src/codegen/signature.h"
 #include "src/wasm/function-body-decoder.h"
 #include "src/wasm/leb-helper.h"
+#include "src/wasm/struct-types.h"
 #include "src/wasm/wasm-constants.h"
 #include "src/wasm/wasm-module.h"
 #include "src/zone/zone-containers.h"
@@ -33,12 +34,162 @@ void FixupSection(ZoneBuffer* buffer, size_t start) {
                                                   kPaddedVarInt32Size));
 }
 
+WasmOpcode FromInitExprOperator(WasmInitExpr::Operator op) {
+  switch (op) {
+    case WasmInitExpr::kGlobalGet:
+      return kExprGlobalGet;
+    case WasmInitExpr::kI32Const:
+      return kExprI32Const;
+    case WasmInitExpr::kI64Const:
+      return kExprI64Const;
+    case WasmInitExpr::kF32Const:
+      return kExprF32Const;
+    case WasmInitExpr::kF64Const:
+      return kExprF64Const;
+    case WasmInitExpr::kS128Const:
+      return kExprS128Const;
+    case WasmInitExpr::kI32Add:
+      return kExprI32Add;
+    case WasmInitExpr::kI32Sub:
+      return kExprI32Sub;
+    case WasmInitExpr::kI32Mul:
+      return kExprI32Mul;
+    case WasmInitExpr::kI64Add:
+      return kExprI64Add;
+    case WasmInitExpr::kI64Sub:
+      return kExprI64Sub;
+    case WasmInitExpr::kI64Mul:
+      return kExprI64Mul;
+    case WasmInitExpr::kRefNullConst:
+      return kExprRefNull;
+    case WasmInitExpr::kRefFuncConst:
+      return kExprRefFunc;
+    case WasmInitExpr::kStructNew:
+      return kExprStructNew;
+    case WasmInitExpr::kStructNewDefault:
+      return kExprStructNewDefault;
+    case WasmInitExpr::kArrayNew:
+      return kExprArrayNew;
+    case WasmInitExpr::kArrayNewDefault:
+      return kExprArrayNewDefault;
+    case WasmInitExpr::kArrayNewFixed:
+      return kExprArrayNewFixed;
+    case WasmInitExpr::kRefI31:
+      return kExprRefI31;
+    case WasmInitExpr::kStringConst:
+      return kExprStringConst;
+    case WasmInitExpr::kAnyConvertExtern:
+      return kExprAnyConvertExtern;
+    case WasmInitExpr::kExternConvertAny:
+      return kExprExternConvertAny;
+  }
+}
+
+void WriteInitializerExpressionWithoutEnd(ZoneBuffer* buffer,
+                                          const WasmInitExpr& init) {
+  switch (init.kind()) {
+    case WasmInitExpr::kI32Const:
+      buffer->write_u8(kExprI32Const);
+      buffer->write_i32v(init.immediate().i32_const);
+      break;
+    case WasmInitExpr::kI64Const:
+      buffer->write_u8(kExprI64Const);
+      buffer->write_i64v(init.immediate().i64_const);
+      break;
+    case WasmInitExpr::kF32Const:
+      buffer->write_u8(kExprF32Const);
+      buffer->write_f32(init.immediate().f32_const);
+      break;
+    case WasmInitExpr::kF64Const:
+      buffer->write_u8(kExprF64Const);
+      buffer->write_f64(init.immediate().f64_const);
+      break;
+    case WasmInitExpr::kS128Const:
+      buffer->write_u8(kSimdPrefix);
+      buffer->write_u8(kExprS128Const & 0xFF);
+      buffer->write(init.immediate().s128_const.data(), kSimd128Size);
+      break;
+    case WasmInitExpr::kI32Add:
+    case WasmInitExpr::kI32Sub:
+    case WasmInitExpr::kI32Mul:
+    case WasmInitExpr::kI64Add:
+    case WasmInitExpr::kI64Sub:
+    case WasmInitExpr::kI64Mul:
+      WriteInitializerExpressionWithoutEnd(buffer, (*init.operands())[0]);
+      WriteInitializerExpressionWithoutEnd(buffer, (*init.operands())[1]);
+      buffer->write_u8(FromInitExprOperator(init.kind()));
+      break;
+    case WasmInitExpr::kGlobalGet:
+      buffer->write_u8(kExprGlobalGet);
+      buffer->write_u32v(init.immediate().index);
+      break;
+    case WasmInitExpr::kRefNullConst:
+      buffer->write_u8(kExprRefNull);
+      if (init.heap_type().encoding_needs_exact()) buffer->write_u8(kExactCode);
+      buffer->write_i32v(init.heap_type().code());
+      break;
+    case WasmInitExpr::kRefFuncConst:
+      buffer->write_u8(kExprRefFunc);
+      buffer->write_u32v(init.immediate().index);
+      break;
+    case WasmInitExpr::kStructNew:
+    case WasmInitExpr::kStructNewDefault:
+    case WasmInitExpr::kArrayNew:
+    case WasmInitExpr::kArrayNewDefault: {
+      if (init.operands() != nullptr) {
+        for (const WasmInitExpr& operand : *init.operands()) {
+          WriteInitializerExpressionWithoutEnd(buffer, operand);
+        }
+      }
+      WasmOpcode opcode = FromInitExprOperator(init.kind());
+      DCHECK_EQ(opcode >> 8, kGCPrefix);
+      DCHECK_EQ(opcode & 0x80, 0);
+      buffer->write_u8(kGCPrefix);
+      buffer->write_u8(static_cast<uint8_t>(opcode));
+      buffer->write_u32v(init.immediate().index);
+      break;
+    }
+    case WasmInitExpr::kArrayNewFixed: {
+      static_assert((kExprArrayNewFixed >> 8) == kGCPrefix);
+      static_assert((kExprArrayNewFixed & 0x80) == 0);
+      for (const WasmInitExpr& operand : *init.operands()) {
+        WriteInitializerExpressionWithoutEnd(buffer, operand);
+      }
+      buffer->write_u8(kGCPrefix);
+      buffer->write_u8(static_cast<uint8_t>(kExprArrayNewFixed));
+      buffer->write_u32v(init.immediate().index);
+      buffer->write_u32v(static_cast<uint32_t>(init.operands()->size()));
+      break;
+    }
+    case WasmInitExpr::kRefI31:
+    case WasmInitExpr::kAnyConvertExtern:
+    case WasmInitExpr::kExternConvertAny: {
+      WriteInitializerExpressionWithoutEnd(buffer, (*init.operands())[0]);
+      WasmOpcode opcode = FromInitExprOperator(init.kind());
+      DCHECK_EQ(opcode >> 8, kGCPrefix);
+      DCHECK_EQ(opcode & 0x80, 0);
+      buffer->write_u8(kGCPrefix);
+      buffer->write_u8(opcode);
+      break;
+    }
+    case WasmInitExpr::kStringConst:
+      buffer->write_u8(kGCPrefix);
+      buffer->write_u32v(kExprStringConst & 0xFF);
+      buffer->write_u32v(init.immediate().index);
+      break;
+  }
+}
+
+void WriteInitializerExpression(ZoneBuffer* buffer, const WasmInitExpr& init) {
+  WriteInitializerExpressionWithoutEnd(buffer, init);
+  buffer->write_u8(kExprEnd);
+}
 }  // namespace
 
 WasmFunctionBuilder::WasmFunctionBuilder(WasmModuleBuilder* builder)
     : builder_(builder),
       locals_(builder->zone()),
-      signature_index_(0),
+      signature_index_{0},
       func_index_(static_cast<uint32_t>(builder->functions_.size())),
       body_(builder->zone(), 256),
       i32_temps_(builder->zone()),
@@ -54,17 +205,19 @@ void WasmFunctionBuilder::EmitI32V(int32_t val) { body_.write_i32v(val); }
 
 void WasmFunctionBuilder::EmitU32V(uint32_t val) { body_.write_u32v(val); }
 
+void WasmFunctionBuilder::EmitU64V(uint64_t val) { body_.write_u64v(val); }
+
 void WasmFunctionBuilder::SetSignature(const FunctionSig* sig) {
   DCHECK(!locals_.has_sig());
   locals_.set_sig(sig);
   signature_index_ = builder_->AddSignature(sig, true);
 }
 
-void WasmFunctionBuilder::SetSignature(uint32_t sig_index) {
+void WasmFunctionBuilder::SetSignature(ModuleTypeIndex sig_index) {
   DCHECK(!locals_.has_sig());
-  DCHECK_EQ(builder_->types_[sig_index].kind, TypeDefinition::kFunction);
+  DCHECK_EQ(builder_->types_[sig_index.index].kind, TypeDefinition::kFunction);
   signature_index_ = sig_index;
-  locals_.set_sig(builder_->types_[sig_index].function_sig);
+  locals_.set_sig(builder_->types_[sig_index.index].function_sig);
 }
 
 uint32_t WasmFunctionBuilder::AddLocal(ValueType type) {
@@ -88,6 +241,10 @@ void WasmFunctionBuilder::EmitCode(const uint8_t* code, uint32_t code_size) {
   body_.write(code, code_size);
 }
 
+void WasmFunctionBuilder::EmitCode(std::initializer_list<const uint8_t> code) {
+  body_.write(code.begin(), code.size());
+}
+
 void WasmFunctionBuilder::Emit(WasmOpcode opcode) {
   DCHECK_LE(opcode, 0xFF);
   body_.write_u8(opcode);
@@ -129,17 +286,25 @@ void WasmFunctionBuilder::EmitWithU32V(WasmOpcode opcode, uint32_t immediate) {
 }
 
 namespace {
+void WriteHeapType(ZoneBuffer* buffer, HeapType type) {
+  if (type.encoding_needs_exact()) buffer->write_u8(kExactCode);
+  buffer->write_i32v(type.code());
+}
 void WriteValueType(ZoneBuffer* buffer, const ValueType& type) {
   buffer->write_u8(type.value_type_code());
-  if (type.encoding_needs_heap_type()) {
-    buffer->write_i32v(type.heap_type().code());
+  if (type.encoding_needs_shared()) {
+    buffer->write_u8(kSharedFlagCode);
   }
-  if (type.is_rtt()) {
-    buffer->write_u32v(type.ref_index());
+  if (type.encoding_needs_heap_type()) {
+    WriteHeapType(buffer, type.heap_type());
   }
 }
 }  // namespace
 
+void WasmFunctionBuilder::EmitHeapType(HeapType type) {
+  WriteHeapType(&body_, type);
+}
+
 void WasmFunctionBuilder::EmitValueType(ValueType type) {
   WriteValueType(&body_, type);
 }
@@ -172,6 +337,11 @@ void WasmFunctionBuilder::EmitDirectCallIndex(uint32_t index) {
   EmitCode(placeholder_bytes, arraysize(placeholder_bytes));
 }
 
+void WasmFunctionBuilder::EmitFromInitializerExpression(
+    const WasmInitExpr& init_expr) {
+  WriteInitializerExpression(&body_, init_expr);
+}
+
 void WasmFunctionBuilder::SetName(base::Vector<const char> name) {
   name_ = name;
 }
@@ -207,16 +377,6 @@ void WasmFunctionBuilder::SetAsmFunctionStartPosition(
   last_asm_source_position_ = function_position_u32;
 }
 
-void WasmFunctionBuilder::SetCompilationHint(
-    WasmCompilationHintStrategy strategy, WasmCompilationHintTier baseline,
-    WasmCompilationHintTier top_tier) {
-  uint8_t hint_byte = static_cast<uint8_t>(strategy) |
-                      static_cast<uint8_t>(baseline) << 2 |
-                      static_cast<uint8_t>(top_tier) << 4;
-  DCHECK_NE(hint_byte, kNoCompilationHint);
-  hint_ = hint_byte;
-}
-
 void WasmFunctionBuilder::DeleteCodeAfter(size_t position) {
   DCHECK_LE(position, body_.size());
   body_.Truncate(position);
@@ -270,18 +430,15 @@ WasmModuleBuilder::WasmModuleBuilder(Zone* zone)
       exports_(zone),
       functions_(zone),
       tables_(zone),
+      memories_(zone),
       data_segments_(zone),
       element_segments_(zone),
       globals_(zone),
-      exceptions_(zone),
+      tags_(zone),
       signature_map_(zone),
       current_recursive_group_start_(-1),
       recursive_groups_(zone),
-      start_function_index_(-1),
-      min_memory_size_(16),
-      max_memory_size_(0),
-      has_max_memory_size_(false),
-      has_shared_memory_(false) {}
+      start_function_index_(-1) {}
 
 WasmFunctionBuilder* WasmModuleBuilder::AddFunction(const FunctionSig* sig) {
   functions_.push_back(zone_->New<WasmFunctionBuilder>(this));
@@ -290,7 +447,7 @@ WasmFunctionBuilder* WasmModuleBuilder::AddFunction(const FunctionSig* sig) {
   return functions_.back();
 }
 
-WasmFunctionBuilder* WasmModuleBuilder::AddFunction(uint32_t sig_index) {
+WasmFunctionBuilder* WasmModuleBuilder::AddFunction(ModuleTypeIndex sig_index) {
   functions_.push_back(zone_->New<WasmFunctionBuilder>(this));
   functions_.back()->SetSignature(sig_index);
   return functions_.back();
@@ -298,7 +455,7 @@ WasmFunctionBuilder* WasmModuleBuilder::AddFunction(uint32_t sig_index) {
 
 void WasmModuleBuilder::AddDataSegment(const uint8_t* data, uint32_t size,
                                        uint32_t dest) {
-  data_segments_.push_back({ZoneVector<uint8_t>(zone()), dest});
+  data_segments_.push_back({.data = ZoneVector<uint8_t>(zone()), .dest = dest});
   ZoneVector<uint8_t>& vec = data_segments_.back().data;
   for (uint32_t i = 0; i < size; i++) {
     vec.push_back(data[i]);
@@ -307,56 +464,58 @@ void WasmModuleBuilder::AddDataSegment(const uint8_t* data, uint32_t size,
 
 void WasmModuleBuilder::AddPassiveDataSegment(const uint8_t* data,
                                               uint32_t size) {
-  data_segments_.push_back({ZoneVector<uint8_t>(zone()), 0, false});
+  data_segments_.push_back(
+      {.data = ZoneVector<uint8_t>(zone()), .dest = 0, .is_active = false});
   ZoneVector<uint8_t>& vec = data_segments_.back().data;
   for (uint32_t i = 0; i < size; i++) {
     vec.push_back(data[i]);
   }
 }
 
-uint32_t WasmModuleBuilder::ForceAddSignature(const FunctionSig* sig,
-                                              bool is_final,
-                                              uint32_t supertype) {
-  uint32_t index = static_cast<uint32_t>(types_.size());
+ModuleTypeIndex WasmModuleBuilder::ForceAddSignature(
+    const FunctionSig* sig, bool is_final, ModuleTypeIndex supertype) {
+  ModuleTypeIndex index{static_cast<uint32_t>(types_.size())};
   signature_map_.emplace(*sig, index);
-  types_.emplace_back(sig, supertype, is_final);
+  types_.emplace_back(sig, supertype, is_final, false);
   return index;
 }
 
-uint32_t WasmModuleBuilder::AddSignature(const FunctionSig* sig, bool is_final,
-                                         uint32_t supertype) {
+ModuleTypeIndex WasmModuleBuilder::AddSignature(const FunctionSig* sig,
+                                                bool is_final,
+                                                ModuleTypeIndex supertype) {
   auto sig_entry = signature_map_.find(*sig);
   if (sig_entry != signature_map_.end()) return sig_entry->second;
   return ForceAddSignature(sig, is_final, supertype);
 }
 
-uint32_t WasmModuleBuilder::AddException(const FunctionSig* type) {
+uint32_t WasmModuleBuilder::AddTag(const FunctionSig* type) {
   DCHECK_EQ(0, type->return_count());
-  int type_index = AddSignature(type, true);
-  uint32_t except_index = static_cast<uint32_t>(exceptions_.size());
-  exceptions_.push_back(type_index);
+  ModuleTypeIndex type_index = AddSignature(type, true);
+  uint32_t except_index = static_cast<uint32_t>(tags_.size());
+  tags_.push_back(type_index);
   return except_index;
 }
 
-uint32_t WasmModuleBuilder::AddStructType(StructType* type, bool is_final,
-                                          uint32_t supertype) {
+ModuleTypeIndex WasmModuleBuilder::AddStructType(StructType* type,
+                                                 bool is_final,
+                                                 ModuleTypeIndex supertype) {
   uint32_t index = static_cast<uint32_t>(types_.size());
-  types_.emplace_back(type, supertype, is_final);
-  return index;
+  types_.emplace_back(type, supertype, is_final, false);
+  return ModuleTypeIndex{index};
 }
 
-uint32_t WasmModuleBuilder::AddArrayType(ArrayType* type, bool is_final,
-                                         uint32_t supertype) {
+ModuleTypeIndex WasmModuleBuilder::AddArrayType(ArrayType* type, bool is_final,
+                                                ModuleTypeIndex supertype) {
   uint32_t index = static_cast<uint32_t>(types_.size());
-  types_.emplace_back(type, supertype, is_final);
-  return index;
+  types_.emplace_back(type, supertype, is_final, false);
+  return ModuleTypeIndex{index};
 }
 
 uint32_t WasmModuleBuilder::IncreaseTableMinSize(uint32_t table_index,
                                                  uint32_t count) {
   DCHECK_LT(table_index, tables_.size());
   uint32_t old_min_size = tables_[table_index].min_size;
-  if (count > v8_flags.wasm_max_table_size - old_min_size) {
+  if (count > wasm::max_table_size() - old_min_size) {
     return std::numeric_limits<uint32_t>::max();
   }
   tables_[table_index].min_size = old_min_size + count;
@@ -366,22 +525,59 @@ uint32_t WasmModuleBuilder::IncreaseTableMinSize(uint32_t table_index,
 }
 
 uint32_t WasmModuleBuilder::AddTable(ValueType type, uint32_t min_size) {
-  tables_.push_back({type, min_size, 0, false, {}});
+  tables_.push_back({.type = type, .min_size = min_size});
   return static_cast<uint32_t>(tables_.size() - 1);
 }
 
 uint32_t WasmModuleBuilder::AddTable(ValueType type, uint32_t min_size,
-                                     uint32_t max_size) {
-  tables_.push_back({type, min_size, max_size, true, {}});
+                                     uint32_t max_size,
+                                     AddressType address_type) {
+  tables_.push_back({.type = type,
+                     .min_size = min_size,
+                     .max_size = max_size,
+                     .has_maximum = true,
+                     .address_type = address_type});
   return static_cast<uint32_t>(tables_.size() - 1);
 }
 
 uint32_t WasmModuleBuilder::AddTable(ValueType type, uint32_t min_size,
-                                     uint32_t max_size, WasmInitExpr init) {
-  tables_.push_back({type, min_size, max_size, true, {init}});
+                                     uint32_t max_size, WasmInitExpr init,
+                                     AddressType address_type) {
+  tables_.push_back({.type = type,
+                     .min_size = min_size,
+                     .max_size = max_size,
+                     .has_maximum = true,
+                     .address_type = address_type,
+                     .init = {init}});
   return static_cast<uint32_t>(tables_.size() - 1);
 }
 
+uint32_t WasmModuleBuilder::AddMemory(uint32_t min_pages) {
+  memories_.push_back({.min_pages = min_pages});
+  return static_cast<uint32_t>(memories_.size() - 1);
+}
+
+uint32_t WasmModuleBuilder::AddMemory(uint32_t min_pages, uint32_t max_pages) {
+  memories_.push_back(
+      {.min_pages = min_pages, .max_pages = max_pages, .has_max_pages = true});
+  return static_cast<uint32_t>(memories_.size() - 1);
+}
+
+uint32_t WasmModuleBuilder::AddMemory64(uint32_t min_pages) {
+  memories_.push_back(
+      {.min_pages = min_pages, .address_type = AddressType::kI64});
+  return static_cast<uint32_t>(memories_.size() - 1);
+}
+
+uint32_t WasmModuleBuilder::AddMemory64(uint32_t min_pages,
+                                        uint32_t max_pages) {
+  memories_.push_back({.min_pages = min_pages,
+                       .max_pages = max_pages,
+                       .has_max_pages = true,
+                       .address_type = AddressType::kI64});
+  return static_cast<uint32_t>(memories_.size() - 1);
+}
+
 uint32_t WasmModuleBuilder::AddElementSegment(WasmElemSegment segment) {
   element_segments_.push_back(std::move(segment));
   return static_cast<uint32_t>(element_segments_.size() - 1);
@@ -400,17 +596,21 @@ void WasmModuleBuilder::SetIndirectFunction(
 }
 
 uint32_t WasmModuleBuilder::AddImport(base::Vector<const char> name,
-                                      FunctionSig* sig,
+                                      const FunctionSig* sig,
                                       base::Vector<const char> module) {
   DCHECK(adding_imports_allowed_);
-  function_imports_.push_back({module, name, AddSignature(sig, true)});
+  function_imports_.push_back(
+      {.module = module, .name = name, .sig_index = AddSignature(sig, true)});
   return static_cast<uint32_t>(function_imports_.size() - 1);
 }
 
 uint32_t WasmModuleBuilder::AddGlobalImport(base::Vector<const char> name,
                                             ValueType type, bool mutability,
                                             base::Vector<const char> module) {
-  global_imports_.push_back({module, name, type.value_type_code(), mutability});
+  global_imports_.push_back({.module = module,
+                             .name = name,
+                             .type_code = type.value_type_code(),
+                             .mutability = mutability});
   return static_cast<uint32_t>(global_imports_.size() - 1);
 }
 
@@ -421,7 +621,8 @@ void WasmModuleBuilder::MarkStartFunction(WasmFunctionBuilder* function) {
 void WasmModuleBuilder::AddExport(base::Vector<const char> name,
                                   ImportExportKindCode kind, uint32_t index) {
   DCHECK_LE(index, std::numeric_limits<int>::max());
-  exports_.push_back({name, kind, static_cast<int>(index)});
+  exports_.push_back(
+      {.name = name, .kind = kind, .index = static_cast<int>(index)});
 }
 
 uint32_t WasmModuleBuilder::AddExportedGlobal(ValueType type, bool mutability,
@@ -439,207 +640,53 @@ void WasmModuleBuilder::ExportImportedFunction(base::Vector<const char> name,
   adding_imports_allowed_ = false;
 #endif
   exports_.push_back(
-      {name, kExternalFunction,
-       import_index - static_cast<int>(function_imports_.size())});
+      {.name = name,
+       .kind = kExternalFunction,
+       .index = import_index - static_cast<int>(function_imports_.size())});
 }
 
 uint32_t WasmModuleBuilder::AddGlobal(ValueType type, bool mutability,
                                       WasmInitExpr init) {
-  globals_.push_back({type, mutability, init});
+  globals_.push_back({.type = type, .mutability = mutability, .init = init});
   return static_cast<uint32_t>(globals_.size() - 1);
 }
 
-void WasmModuleBuilder::SetMinMemorySize(uint32_t value) {
-  min_memory_size_ = value;
-}
-
-void WasmModuleBuilder::SetMaxMemorySize(uint32_t value) {
-  has_max_memory_size_ = true;
-  max_memory_size_ = value;
-}
-
-void WasmModuleBuilder::SetHasSharedMemory() { has_shared_memory_ = true; }
-
-namespace {
-
-WasmOpcode FromInitExprOperator(WasmInitExpr::Operator op) {
-  switch (op) {
-    case WasmInitExpr::kGlobalGet:
-      return kExprGlobalGet;
-    case WasmInitExpr::kI32Const:
-      return kExprI32Const;
-    case WasmInitExpr::kI64Const:
-      return kExprI64Const;
-    case WasmInitExpr::kF32Const:
-      return kExprF32Const;
-    case WasmInitExpr::kF64Const:
-      return kExprF64Const;
-    case WasmInitExpr::kS128Const:
-      return kExprS128Const;
-    case WasmInitExpr::kI32Add:
-      return kExprI32Add;
-    case WasmInitExpr::kI32Sub:
-      return kExprI32Sub;
-    case WasmInitExpr::kI32Mul:
-      return kExprI32Mul;
-    case WasmInitExpr::kI64Add:
-      return kExprI64Add;
-    case WasmInitExpr::kI64Sub:
-      return kExprI64Sub;
-    case WasmInitExpr::kI64Mul:
-      return kExprI64Mul;
-    case WasmInitExpr::kRefNullConst:
-      return kExprRefNull;
-    case WasmInitExpr::kRefFuncConst:
-      return kExprRefFunc;
-    case WasmInitExpr::kStructNew:
-      return kExprStructNew;
-    case WasmInitExpr::kStructNewDefault:
-      return kExprStructNewDefault;
-    case WasmInitExpr::kArrayNew:
-      return kExprArrayNew;
-    case WasmInitExpr::kArrayNewDefault:
-      return kExprArrayNewDefault;
-    case WasmInitExpr::kArrayNewFixed:
-      return kExprArrayNewFixed;
-    case WasmInitExpr::kRefI31:
-      return kExprRefI31;
-    case WasmInitExpr::kStringConst:
-      return kExprStringConst;
-    case WasmInitExpr::kExternInternalize:
-      return kExprExternInternalize;
-    case WasmInitExpr::kExternExternalize:
-      return kExprExternExternalize;
-  }
-}
-
-void WriteInitializerExpressionWithoutEnd(ZoneBuffer* buffer,
-                                          const WasmInitExpr& init) {
-  switch (init.kind()) {
-    case WasmInitExpr::kI32Const:
-      buffer->write_u8(kExprI32Const);
-      buffer->write_i32v(init.immediate().i32_const);
-      break;
-    case WasmInitExpr::kI64Const:
-      buffer->write_u8(kExprI64Const);
-      buffer->write_i64v(init.immediate().i64_const);
-      break;
-    case WasmInitExpr::kF32Const:
-      buffer->write_u8(kExprF32Const);
-      buffer->write_f32(init.immediate().f32_const);
-      break;
-    case WasmInitExpr::kF64Const:
-      buffer->write_u8(kExprF64Const);
-      buffer->write_f64(init.immediate().f64_const);
-      break;
-    case WasmInitExpr::kS128Const:
-      buffer->write_u8(kSimdPrefix);
-      buffer->write_u8(kExprS128Const & 0xFF);
-      buffer->write(init.immediate().s128_const.data(), kSimd128Size);
-      break;
-    case WasmInitExpr::kI32Add:
-    case WasmInitExpr::kI32Sub:
-    case WasmInitExpr::kI32Mul:
-    case WasmInitExpr::kI64Add:
-    case WasmInitExpr::kI64Sub:
-    case WasmInitExpr::kI64Mul:
-      WriteInitializerExpressionWithoutEnd(buffer, (*init.operands())[0]);
-      WriteInitializerExpressionWithoutEnd(buffer, (*init.operands())[1]);
-      buffer->write_u8(FromInitExprOperator(init.kind()));
-      break;
-    case WasmInitExpr::kGlobalGet:
-      buffer->write_u8(kExprGlobalGet);
-      buffer->write_u32v(init.immediate().index);
-      break;
-    case WasmInitExpr::kRefNullConst:
-      buffer->write_u8(kExprRefNull);
-      buffer->write_i32v(HeapType(init.immediate().heap_type).code());
-      break;
-    case WasmInitExpr::kRefFuncConst:
-      buffer->write_u8(kExprRefFunc);
-      buffer->write_u32v(init.immediate().index);
-      break;
-    case WasmInitExpr::kStructNew:
-    case WasmInitExpr::kStructNewDefault:
-    case WasmInitExpr::kArrayNew:
-    case WasmInitExpr::kArrayNewDefault: {
-      if (init.operands() != nullptr) {
-        for (const WasmInitExpr& operand : *init.operands()) {
-          WriteInitializerExpressionWithoutEnd(buffer, operand);
-        }
-      }
-      WasmOpcode opcode = FromInitExprOperator(init.kind());
-      DCHECK_EQ(opcode >> 8, kGCPrefix);
-      DCHECK_EQ(opcode & 0x80, 0);
-      buffer->write_u8(kGCPrefix);
-      buffer->write_u8(static_cast<uint8_t>(opcode));
-      buffer->write_u32v(init.immediate().index);
-      break;
-    }
-    case WasmInitExpr::kArrayNewFixed: {
-      static_assert((kExprArrayNewFixed >> 8) == kGCPrefix);
-      static_assert((kExprArrayNewFixed & 0x80) == 0);
-      for (const WasmInitExpr& operand : *init.operands()) {
-        WriteInitializerExpressionWithoutEnd(buffer, operand);
-      }
-      buffer->write_u8(kGCPrefix);
-      buffer->write_u8(static_cast<uint8_t>(kExprArrayNewFixed));
-      buffer->write_u32v(init.immediate().index);
-      buffer->write_u32v(static_cast<uint32_t>(init.operands()->size()));
-      break;
-    }
-    case WasmInitExpr::kRefI31:
-    case WasmInitExpr::kExternInternalize:
-    case WasmInitExpr::kExternExternalize: {
-      WriteInitializerExpressionWithoutEnd(buffer, (*init.operands())[0]);
-      WasmOpcode opcode = FromInitExprOperator(init.kind());
-      DCHECK_EQ(opcode >> 8, kGCPrefix);
-      DCHECK_EQ(opcode & 0x80, 0);
-      buffer->write_u8(kGCPrefix);
-      buffer->write_u8(opcode);
-      break;
-    }
-    case WasmInitExpr::kStringConst:
-      buffer->write_u8(kGCPrefix);
-      buffer->write_u32v(kExprStringConst & 0xFF);
-      buffer->write_u32v(init.immediate().index);
-      break;
-  }
-}
-
-void WriteInitializerExpression(ZoneBuffer* buffer, const WasmInitExpr& init) {
-  WriteInitializerExpressionWithoutEnd(buffer, init);
-  buffer->write_u8(kExprEnd);
-}
-}  // namespace
-
 void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
   // == Emit magic =============================================================
   buffer->write_u32(kWasmMagic);
   buffer->write_u32(kWasmVersion);
 
   // == Emit types =============================================================
-  if (types_.size() > 0) {
+  // Check that the last `StartRecursiveTypeGroup()` was followed by
+  // `EndRecursiveTypeGroup()`.
+  DCHECK_EQ(-1, current_recursive_group_start_);
+  if (!types_.empty()) {
     size_t start = EmitSection(kTypeSectionCode, buffer);
-    size_t type_count = types_.size();
-    for (auto pair : recursive_groups_) {
-      // Every rec. group counts as one type entry.
-      type_count -= pair.second - 1;
+    // Every recursion group occupies one type entry.
+    size_t type_count = types_.size() + recursive_groups_.size();
+    // Types inside recursion groups occupy no additional type entry.
+    for (auto [first_index, size] : recursive_groups_) {
+      type_count -= size;
     }
 
     buffer->write_size(type_count);
 
-    for (uint32_t i = 0; i < types_.size(); i++) {
-      auto recursive_group = recursive_groups_.find(i);
+    const RecGroup* next_rec_group =
+        recursive_groups_.empty() ? nullptr : recursive_groups_.data();
 
-      if (recursive_group != recursive_groups_.end()) {
+    for (uint32_t i = 0; i < types_.size(); i++) {
+      // Note: while loop, because recgroups can be empty.
+      while (next_rec_group && i == next_rec_group->start_index) {
         buffer->write_u8(kWasmRecursiveTypeGroupCode);
-        buffer->write_u32v(recursive_group->second);
+        buffer->write_u32v(next_rec_group->size);
+        next_rec_group = next_rec_group == &recursive_groups_.back()
+                             ? nullptr
+                             : next_rec_group + 1;
       }
 
       const TypeDefinition& type = types_[i];
 
-      if (type.supertype != kNoSuperType) {
+      if (type.supertype.valid()) {
         buffer->write_u8(type.is_final ? kWasmSubtypeFinalCode
                                        : kWasmSubtypeCode);
         buffer->write_u8(1);
@@ -666,9 +713,9 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
           const StructType* struct_type = type.struct_type;
           buffer->write_u8(kWasmStructTypeCode);
           buffer->write_size(struct_type->field_count());
-          for (uint32_t i = 0; i < struct_type->field_count(); i++) {
-            WriteValueType(buffer, struct_type->field(i));
-            buffer->write_u8(struct_type->mutability(i) ? 1 : 0);
+          for (uint32_t j = 0; j < struct_type->field_count(); j++) {
+            WriteValueType(buffer, struct_type->field(j));
+            buffer->write_u8(struct_type->mutability(j) ? 1 : 0);
           }
           break;
         }
@@ -679,8 +726,25 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
           buffer->write_u8(array_type->mutability() ? 1 : 0);
           break;
         }
+        case TypeDefinition::kCont: {
+          const ContType* cont_type = type.cont_type;
+          buffer->write_u8(kWasmContTypeCode);
+          buffer->write_u32v(cont_type->contfun_typeindex());
+          break;
+        }
       }
     }
+
+    // Handle empty recursion groups defined after all types.
+    while (next_rec_group) {
+      DCHECK_EQ(types_.size(), next_rec_group->start_index);
+      DCHECK_EQ(0, next_rec_group->size);
+      buffer->write_u8(kWasmRecursiveTypeGroupCode);
+      buffer->write_u32v(0);
+      if (next_rec_group == &recursive_groups_.back()) break;
+      ++next_rec_group;
+    }
+
     FixupSection(buffer, start);
   }
 
@@ -706,7 +770,7 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
 
   // == Emit function signatures ===============================================
   uint32_t num_function_names = 0;
-  if (functions_.size() > 0) {
+  if (!functions_.empty()) {
     size_t start = EmitSection(kFunctionSectionCode, buffer);
     buffer->write_size(functions_.size());
     for (auto* function : functions_) {
@@ -717,7 +781,7 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
   }
 
   // == Emit tables ============================================================
-  if (tables_.size() > 0) {
+  if (!tables_.empty()) {
     size_t start = EmitSection(kTableSectionCode, buffer);
     buffer->write_size(tables_.size());
     for (const WasmTable& table : tables_) {
@@ -726,9 +790,17 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
         buffer->write_u8(0x00);  // reserved byte
       }
       WriteValueType(buffer, table.type);
-      buffer->write_u8(table.has_maximum ? kWithMaximum : kNoMaximum);
-      buffer->write_size(table.min_size);
-      if (table.has_maximum) buffer->write_size(table.max_size);
+      uint8_t limits_byte = (table.is_table64() ? 4 : 0) |
+                            (table.is_shared ? 2 : 0) |
+                            (table.has_maximum ? 1 : 0);
+      buffer->write_u8(limits_byte);
+      auto WriteValToBuffer = [&](uint32_t val) {
+        table.is_table64() ? buffer->write_u64v(val) : buffer->write_u32v(val);
+      };
+      WriteValToBuffer(table.min_size);
+      if (table.has_maximum) {
+        WriteValToBuffer(table.max_size);
+      }
       if (table.init) {
         WriteInitializerExpression(buffer, *table.init);
       }
@@ -737,27 +809,31 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
   }
 
   // == Emit memory declaration ================================================
-  {
+  if (!memories_.empty()) {
     size_t start = EmitSection(kMemorySectionCode, buffer);
-    buffer->write_u8(1);  // memory count
-    if (has_shared_memory_) {
-      buffer->write_u8(has_max_memory_size_ ? kSharedWithMaximum
-                                            : kSharedNoMaximum);
-    } else {
-      buffer->write_u8(has_max_memory_size_ ? kWithMaximum : kNoMaximum);
-    }
-    buffer->write_u32v(min_memory_size_);
-    if (has_max_memory_size_) {
-      buffer->write_u32v(max_memory_size_);
+    buffer->write_size(memories_.size());
+    for (const WasmMemory& memory : memories_) {
+      uint8_t limits_byte = (memory.is_memory64() ? 4 : 0) |
+                            (memory.is_shared ? 2 : 0) |
+                            (memory.has_max_pages ? 1 : 0);
+      buffer->write_u8(limits_byte);
+      auto WriteValToBuffer = [&](uint32_t val) {
+        memory.is_memory64() ? buffer->write_u64v(val)
+                             : buffer->write_u32v(val);
+      };
+      WriteValToBuffer(memory.min_pages);
+      if (memory.has_max_pages) {
+        WriteValToBuffer(memory.max_pages);
+      }
     }
     FixupSection(buffer, start);
   }
 
-  // Emit event section.
-  if (exceptions_.size() > 0) {
+  // == Emit event section =====================================================
+  if (!tags_.empty()) {
     size_t start = EmitSection(kTagSectionCode, buffer);
-    buffer->write_size(exceptions_.size());
-    for (int type : exceptions_) {
+    buffer->write_size(tags_.size());
+    for (ModuleTypeIndex type : tags_) {
       buffer->write_u32v(kExceptionAttribute);
       buffer->write_u32v(type);
     }
@@ -765,7 +841,7 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
   }
 
   // == Emit globals ===========================================================
-  if (globals_.size() > 0) {
+  if (!globals_.empty()) {
     size_t start = EmitSection(kGlobalSectionCode, buffer);
     buffer->write_size(globals_.size());
 
@@ -777,8 +853,8 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
     FixupSection(buffer, start);
   }
 
-  // == emit exports ===========================================================
-  if (exports_.size() > 0) {
+  // == Emit exports ===========================================================
+  if (!exports_.empty()) {
     size_t start = EmitSection(kExportSectionCode, buffer);
     buffer->write_size(exports_.size());
     for (auto ex : exports_) {
@@ -804,15 +880,15 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
     FixupSection(buffer, start);
   }
 
-  // == emit start function index ==============================================
+  // == Emit start function index ==============================================
   if (start_function_index_ >= 0) {
     size_t start = EmitSection(kStartSectionCode, buffer);
     buffer->write_size(start_function_index_ + function_imports_.size());
     FixupSection(buffer, start);
   }
 
-  // == emit element segments ==================================================
-  if (element_segments_.size() > 0) {
+  // == Emit element segments ==================================================
+  if (!element_segments_.empty()) {
     size_t start = EmitSection(kElementSectionCode, buffer);
     buffer->write_size(element_segments_.size());
     for (const WasmElemSegment& segment : element_segments_) {
@@ -856,7 +932,7 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
     FixupSection(buffer, start);
   }
 
-  // == emit data segment count section ========================================
+  // == Emit data segment count section ========================================
   if (std::any_of(
           data_segments_.begin(), data_segments_.end(),
           [](const WasmDataSegment& segment) { return !segment.is_active; })) {
@@ -865,34 +941,8 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
     buffer->write_u32v(static_cast<uint32_t>(data_segments_.size()));
   }
 
-  // == emit compilation hints section =========================================
-  bool emit_compilation_hints = false;
-  for (auto* fn : functions_) {
-    if (fn->hint_ != kNoCompilationHint) {
-      emit_compilation_hints = true;
-      break;
-    }
-  }
-  if (emit_compilation_hints) {
-    // Emit the section code.
-    buffer->write_u8(kUnknownSectionCode);
-    // Emit a placeholder for section length.
-    size_t start = buffer->reserve_u32v();
-    // Emit custom section name.
-    buffer->write_string(base::CStrVector("compilationHints"));
-    // Emit hint count.
-    buffer->write_size(functions_.size());
-    // Emit hint bytes.
-    for (auto* fn : functions_) {
-      uint8_t hint_byte =
-          fn->hint_ != kNoCompilationHint ? fn->hint_ : kDefaultCompilationHint;
-      buffer->write_u8(hint_byte);
-    }
-    FixupSection(buffer, start);
-  }
-
-  // == emit code ==============================================================
-  if (functions_.size() > 0) {
+  // == Emit code ==============================================================
+  if (!functions_.empty()) {
     size_t start = EmitSection(kCodeSectionCode, buffer);
     buffer->write_size(functions_.size());
     for (auto* function : functions_) {
@@ -901,8 +951,8 @@ void WasmModuleBuilder::WriteTo(ZoneBuffer* buffer) const {
     FixupSection(buffer, start);
   }
 
-  // == emit data segments =====================================================
-  if (data_segments_.size() > 0) {
+  // == Emit data segments =====================================================
+  if (!data_segments_.empty()) {
     size_t start = EmitSection(kDataSectionCode, buffer);
     buffer->write_size(data_segments_.size());
 
