@@ -39,8 +39,10 @@
 #include "src/heap/factory.h"
 #include "src/heap/heap.h"
 #include "src/heap/large-spaces.h"
+#include "src/heap/main-allocator.h"
 #include "src/heap/memory-allocator.h"
-#include "src/heap/memory-chunk.h"
+#include "src/heap/memory-pool.h"
+#include "src/heap/mutable-page-metadata.h"
 #include "src/heap/spaces-inl.h"
 #include "src/heap/spaces.h"
 #include "src/objects/free-space.h"
@@ -67,7 +69,8 @@ class V8_NODISCARD TestMemoryAllocatorScope {
     isolate->heap()->memory_allocator_.reset(new MemoryAllocator(
         isolate,
         page_allocator != nullptr ? page_allocator : isolate->page_allocator(),
-        max_capacity));
+        page_allocator != nullptr ? page_allocator : isolate->page_allocator(),
+        isolate->isolate_group()->memory_pool(), max_capacity));
     if (page_allocator != nullptr) {
       isolate->heap()->memory_allocator_->data_page_allocator_ = page_allocator;
     }
@@ -76,6 +79,7 @@ class V8_NODISCARD TestMemoryAllocatorScope {
   MemoryAllocator* allocator() { return isolate_->heap()->memory_allocator(); }
 
   ~TestMemoryAllocatorScope() {
+    isolate_->heap()->memory_allocator()->ReleasePooledChunksImmediately();
     isolate_->heap()->memory_allocator()->TearDown();
     isolate_->heap()->memory_allocator_.swap(old_allocator_);
     isolate_->GetCodePages()->swap(code_pages_);
@@ -129,24 +133,21 @@ static void VerifyMemoryChunk(Isolate* isolate, Heap* heap,
 
   size_t allocatable_memory_area_offset =
       MemoryChunkLayout::ObjectStartOffsetInMemoryChunk(space->identity());
-  size_t guard_size =
-      (executable == EXECUTABLE) ? MemoryChunkLayout::CodePageGuardSize() : 0;
 
-  MemoryChunk* memory_chunk =
-      memory_allocator->AllocateLargePage(space, area_size, executable);
+  MutablePageMetadata* memory_chunk = memory_allocator->AllocateLargePage(
+      space, area_size, executable, AllocationHint());
   size_t reserved_size =
       ((executable == EXECUTABLE))
           ? RoundUp(allocatable_memory_area_offset +
-                        RoundUp(area_size, page_allocator->CommitPageSize()) +
-                        guard_size,
+                        RoundUp(area_size, page_allocator->CommitPageSize()),
                     page_allocator->CommitPageSize())
           : RoundUp(allocatable_memory_area_offset + area_size,
                     page_allocator->CommitPageSize());
   CHECK(memory_chunk->size() == reserved_size);
   CHECK(memory_chunk->area_start() <
-        memory_chunk->address() + memory_chunk->size());
+        memory_chunk->ChunkAddress() + memory_chunk->size());
   CHECK(memory_chunk->area_end() <=
-        memory_chunk->address() + memory_chunk->size());
+        memory_chunk->ChunkAddress() + memory_chunk->size());
   CHECK(static_cast<size_t>(memory_chunk->area_size()) == area_size);
 
   memory_allocator->Free(MemoryAllocator::FreeMode::kImmediately, memory_chunk);
@@ -158,8 +159,7 @@ static unsigned int PseudorandomAreaSize() {
   return lo & 0xFFFFF;
 }
 
-
-TEST(MemoryChunk) {
+TEST(MutablePageMetadata) {
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
   IsolateSafepointScope safepoint(heap);
@@ -167,23 +167,35 @@ TEST(MemoryChunk) {
   v8::PageAllocator* page_allocator = GetPlatformPageAllocator();
   size_t area_size;
 
-  bool jitless = isolate->jitless();
-
   for (int i = 0; i < 100; i++) {
     area_size =
         RoundUp(PseudorandomAreaSize(), page_allocator->CommitPageSize());
 
-    // With CodeRange.
     const size_t code_range_size = 32 * MB;
+#ifdef V8_ENABLE_SANDBOX
+    // When the sandbox is enabled, the code assumes that there's only a single
+    // code range for easy metadata lookup, so use the process wide code range
+    // in this case.
+    CodeRange* code_range =
+        IsolateGroup::current()->EnsureCodeRange(code_range_size);
+    base::BoundedPageAllocator* bounded_page_allocator =
+        code_range->page_allocator();
+#else
+    // With CodeRange.
+    bool jitless = isolate->jitless();
     VirtualMemory code_range_reservation(
-        page_allocator, code_range_size, nullptr, MemoryChunk::kAlignment,
-        jitless ? JitPermission::kNoJit : JitPermission::kMapAsJittable);
+        page_allocator, code_range_size, PageAllocator::AllocationHint(),
+        MemoryChunk::GetAlignmentForAllocation(),
+        jitless ? PageAllocator::Permission::kNoAccess
+                : PageAllocator::Permission::kNoAccessWillJitLater);
 
+    base::PageInitializationMode page_initialization_mode =
+        base::PageInitializationMode::kAllocatedPagesCanBeUninitialized;
     base::PageFreeingMode page_freeing_mode =
         base::PageFreeingMode::kMakeInaccessible;
 
-    // On MacOS on ARM64 the code range reservation must be committed as RWX.
-    if (V8_HEAP_USE_PTHREAD_JIT_WRITE_PROTECT && !jitless) {
+    if (!jitless) {
+      page_initialization_mode = base::PageInitializationMode::kRecommitOnly;
       page_freeing_mode = base::PageFreeingMode::kDiscard;
       void* base = reinterpret_cast<void*>(code_range_reservation.address());
       CHECK(page_allocator->SetPermissions(base, code_range_size,
@@ -195,35 +207,31 @@ TEST(MemoryChunk) {
 
     base::BoundedPageAllocator code_page_allocator(
         page_allocator, code_range_reservation.address(),
-        code_range_reservation.size(), MemoryChunk::kAlignment,
-        base::PageInitializationMode::kAllocatedPagesCanBeUninitialized,
-        page_freeing_mode);
+        code_range_reservation.size(), MemoryChunk::GetAlignmentForAllocation(),
+        page_initialization_mode, page_freeing_mode);
+    base::BoundedPageAllocator* bounded_page_allocator = &code_page_allocator;
+#endif
 
-    // Modification of pages in code_range_reservation requires write access.
-    RwxMemoryWriteScopeForTesting rwx_write_scope;
-
-    VerifyMemoryChunk(isolate, heap, &code_page_allocator, area_size,
+    VerifyMemoryChunk(isolate, heap, bounded_page_allocator, area_size,
                       EXECUTABLE, PageSize::kLarge, heap->code_lo_space());
 
-    VerifyMemoryChunk(isolate, heap, &code_page_allocator, area_size,
+    VerifyMemoryChunk(isolate, heap, bounded_page_allocator, area_size,
                       NOT_EXECUTABLE, PageSize::kLarge, heap->lo_space());
   }
 }
 
-
 TEST(MemoryAllocator) {
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
 
   TestMemoryAllocatorScope test_allocator_scope(isolate, heap->MaxReserved());
   MemoryAllocator* memory_allocator = test_allocator_scope.allocator();
-  LinearAllocationArea allocation_info;
 
   int total_pages = 0;
-  OldSpace faked_space(heap, allocation_info);
+  OldSpace faked_space(heap);
   CHECK(!faked_space.first_page());
   CHECK(!faked_space.last_page());
-  Page* first_page = memory_allocator->AllocatePage(
+  PageMetadata* first_page = memory_allocator->AllocatePage(
       MemoryAllocator::AllocationMode::kRegular,
       static_cast<PagedSpace*>(&faked_space), NOT_EXECUTABLE);
 
@@ -231,66 +239,69 @@ TEST(MemoryAllocator) {
   CHECK(first_page->next_page() == nullptr);
   total_pages++;
 
-  for (Page* p = first_page; p != nullptr; p = p->next_page()) {
+  for (PageMetadata* p = first_page; p != nullptr; p = p->next_page()) {
     CHECK(p->owner() == &faked_space);
   }
 
   // Again, we should get n or n - 1 pages.
-  Page* other = memory_allocator->AllocatePage(
+  PageMetadata* other = memory_allocator->AllocatePage(
       MemoryAllocator::AllocationMode::kRegular,
       static_cast<PagedSpace*>(&faked_space), NOT_EXECUTABLE);
   total_pages++;
   faked_space.memory_chunk_list().PushBack(other);
   int page_count = 0;
-  for (Page* p = first_page; p != nullptr; p = p->next_page()) {
+  for (PageMetadata* p = first_page; p != nullptr; p = p->next_page()) {
     CHECK(p->owner() == &faked_space);
     page_count++;
   }
   CHECK(total_pages == page_count);
 
-  Page* second_page = first_page->next_page();
+  PageMetadata* second_page = first_page->next_page();
   CHECK_NOT_NULL(second_page);
 
   // OldSpace's destructor will tear down the space and free up all pages.
 }
 
 TEST(ComputeDiscardMemoryAreas) {
-  base::AddressRegion memory_area;
+  std::optional<base::AddressRegion> discard_area;
   size_t page_size = MemoryAllocator::GetCommitPageSize();
-  size_t free_header_size = FreeSpace::kSize;
 
-  memory_area = MemoryAllocator::ComputeDiscardMemoryArea(0, 0);
-  CHECK_EQ(memory_area.begin(), 0);
-  CHECK_EQ(memory_area.size(), 0);
+  discard_area = Sweeper::ComputeDiscardMemoryArea(0, 0);
+  CHECK(!discard_area);
 
-  memory_area = MemoryAllocator::ComputeDiscardMemoryArea(
-      0, page_size + free_header_size);
-  CHECK_EQ(memory_area.begin(), 0);
-  CHECK_EQ(memory_area.size(), 0);
+  discard_area = Sweeper::ComputeDiscardMemoryArea(0, page_size);
+  CHECK_EQ(discard_area->begin(), 0);
+  CHECK_EQ(discard_area->size(), page_size);
 
-  memory_area = MemoryAllocator::ComputeDiscardMemoryArea(
-      page_size - free_header_size, page_size + free_header_size);
-  CHECK_EQ(memory_area.begin(), page_size);
-  CHECK_EQ(memory_area.size(), page_size);
+  discard_area = Sweeper::ComputeDiscardMemoryArea(page_size, 2 * page_size);
+  CHECK_EQ(discard_area->begin(), page_size);
+  CHECK_EQ(discard_area->size(), page_size);
 
-  memory_area = MemoryAllocator::ComputeDiscardMemoryArea(page_size, page_size);
-  CHECK_EQ(memory_area.begin(), 0);
-  CHECK_EQ(memory_area.size(), 0);
+  discard_area =
+      Sweeper::ComputeDiscardMemoryArea(page_size - kTaggedSize, 2 * page_size);
+  CHECK_EQ(discard_area->begin(), page_size);
+  CHECK_EQ(discard_area->size(), page_size);
 
-  memory_area = MemoryAllocator::ComputeDiscardMemoryArea(
-      page_size / 2, page_size + page_size / 2);
-  CHECK_EQ(memory_area.begin(), page_size);
-  CHECK_EQ(memory_area.size(), page_size);
+  discard_area =
+      Sweeper::ComputeDiscardMemoryArea(page_size, 2 * page_size + kTaggedSize);
+  CHECK_EQ(discard_area->begin(), page_size);
+  CHECK_EQ(discard_area->size(), page_size);
 
-  memory_area = MemoryAllocator::ComputeDiscardMemoryArea(
-      page_size / 2, page_size + page_size / 4);
-  CHECK_EQ(memory_area.begin(), 0);
-  CHECK_EQ(memory_area.size(), 0);
+  discard_area = Sweeper::ComputeDiscardMemoryArea(page_size, page_size);
+  CHECK(!discard_area);
 
-  memory_area =
-      MemoryAllocator::ComputeDiscardMemoryArea(page_size / 2, page_size * 3);
-  CHECK_EQ(memory_area.begin(), page_size);
-  CHECK_EQ(memory_area.size(), page_size * 2);
+  discard_area = Sweeper::ComputeDiscardMemoryArea(page_size / 2,
+                                                   page_size + page_size / 2);
+  CHECK(!discard_area);
+
+  discard_area = Sweeper::ComputeDiscardMemoryArea(page_size / 2,
+                                                   page_size + page_size / 4);
+  CHECK(!discard_area);
+
+  discard_area =
+      Sweeper::ComputeDiscardMemoryArea(page_size / 2, page_size * 3);
+  CHECK_EQ(discard_area->begin(), page_size);
+  CHECK_EQ(discard_area->size(), page_size * 2);
 }
 
 TEST(SemiSpaceNewSpace) {
@@ -303,23 +314,27 @@ TEST(SemiSpaceNewSpace) {
 
   auto new_space = std::make_unique<SemiSpaceNewSpace>(
       heap, heap->InitialSemiSpaceSize(), heap->InitialSemiSpaceSize(),
-      allocation_info);
+      heap->InitialSemiSpaceSize());
+  MainAllocator allocator(heap->main_thread_local_heap(), new_space.get(),
+                          MainAllocator::IsNewGeneration::kYes,
+                          &allocation_info);
   CHECK(new_space->MaximumCapacity());
 
   size_t successful_allocations = 0;
   while (new_space->Available() >= kMaxRegularHeapObjectSize) {
     AllocationResult allocation =
-        new_space->AllocateRaw(kMaxRegularHeapObjectSize, kTaggedAligned);
+        allocator.AllocateRaw(kMaxRegularHeapObjectSize, kTaggedAligned,
+                              AllocationOrigin::kRuntime, AllocationHint());
     if (allocation.IsFailure()) break;
     successful_allocations++;
     Tagged<Object> obj = allocation.ToObjectChecked();
-    Tagged<HeapObject> ho = HeapObject::cast(obj);
+    Tagged<HeapObject> ho = Cast<HeapObject>(obj);
     CHECK(new_space->Contains(ho));
   }
   CHECK_LT(0, successful_allocations);
 
   new_space.reset();
-  memory_allocator->unmapper()->EnsureUnmappingCompleted();
+  memory_allocator->ReleasePooledChunksImmediately();
 }
 
 TEST(PagedNewSpace) {
@@ -333,25 +348,27 @@ TEST(PagedNewSpace) {
 
   auto new_space = std::make_unique<PagedNewSpace>(
       heap, heap->InitialSemiSpaceSize(), heap->InitialSemiSpaceSize(),
-      allocation_info);
-  CHECK(new_space->MaximumCapacity());
-  CHECK(new_space->EnsureCurrentCapacity());
-  CHECK_LT(0, new_space->TotalCapacity());
+      heap->InitialSemiSpaceSize());
+  MainAllocator allocator(heap->main_thread_local_heap(), new_space.get(),
+                          MainAllocator::IsNewGeneration::kYes,
+                          &allocation_info);
+  GrowNewSpaceToMaximumCapacity(heap);
 
   size_t successful_allocations = 0;
   while (true) {
     AllocationResult allocation =
-        new_space->AllocateRaw(kMaxRegularHeapObjectSize, kTaggedAligned);
+        allocator.AllocateRaw(kMaxRegularHeapObjectSize, kTaggedAligned,
+                              AllocationOrigin::kRuntime, AllocationHint());
     if (allocation.IsFailure()) break;
     successful_allocations++;
     Tagged<Object> obj = allocation.ToObjectChecked();
-    Tagged<HeapObject> ho = HeapObject::cast(obj);
+    Tagged<HeapObject> ho = Cast<HeapObject>(obj);
     CHECK(new_space->Contains(ho));
   }
   CHECK_LT(0, successful_allocations);
 
   new_space.reset();
-  memory_allocator->unmapper()->EnsureUnmappingCompleted();
+  memory_allocator->ReleasePooledChunksImmediately();
 }
 
 TEST(OldSpace) {
@@ -367,18 +384,21 @@ TEST(OldSpace) {
   TestMemoryAllocatorScope test_allocator_scope(isolate, heap->MaxReserved());
   LinearAllocationArea allocation_info;
 
-  auto old_space = std::make_unique<OldSpace>(heap, allocation_info);
+  auto old_space = std::make_unique<OldSpace>(heap);
+  MainAllocator allocator(heap->main_thread_local_heap(), old_space.get(),
+                          MainAllocator::IsNewGeneration::kNo,
+                          &allocation_info);
   const int obj_size = kMaxRegularHeapObjectSize;
 
   size_t successful_allocations = 0;
 
   while (true) {
-    AllocationResult allocation =
-        old_space->AllocateRaw(obj_size, kTaggedAligned);
+    AllocationResult allocation = allocator.AllocateRaw(
+        obj_size, kTaggedAligned, AllocationOrigin::kRuntime, AllocationHint());
     if (allocation.IsFailure()) break;
     successful_allocations++;
     Tagged<Object> obj = allocation.ToObjectChecked();
-    Tagged<HeapObject> ho = HeapObject::cast(obj);
+    Tagged<HeapObject> ho = Cast<HeapObject>(obj);
     CHECK(old_space->Contains(ho));
   }
   CHECK_LT(0, successful_allocations);
@@ -397,32 +417,36 @@ TEST(OldLargeObjectSpace) {
   Heap* heap = isolate->heap();
 
   auto lo = std::make_unique<OldLargeObjectSpace>(heap);
-  const int lo_size = Page::kPageSize;
+  const int lo_size = PageMetadata::kPageSize;
 
   HandleScope handle_scope(isolate);
   Tagged<Map> map = ReadOnlyRoots(isolate).fixed_double_array_map();
   size_t successful_allocations = 0;
 
   while (true) {
-    AllocationResult allocation = lo->AllocateRaw(lo_size);
+    AllocationResult allocation = lo->AllocateRaw(
+        heap->main_thread_local_heap(), lo_size, AllocationHint());
     if (allocation.IsFailure()) break;
     successful_allocations++;
     Tagged<Object> obj = allocation.ToObjectChecked();
     CHECK(IsHeapObject(obj));
-    Tagged<HeapObject> ho = HeapObject::cast(obj);
+    Tagged<HeapObject> ho = Cast<HeapObject>(obj);
     CHECK(lo->Contains(ho));
     CHECK_EQ(0, Heap::GetFillToAlign(ho.address(), kTaggedAligned));
     // All large objects have the same alignment because they start at the
     // same offset within a page. Fixed double arrays have the most strict
     // alignment requirements.
-    CHECK_EQ(0, Heap::GetFillToAlign(ho.address(),
-                                     HeapObject::RequiredAlignment(map)));
-    Handle<HeapObject> keep_alive(ho, isolate);
+    CHECK_EQ(0,
+             Heap::GetFillToAlign(ho.address(), HeapObject::RequiredAlignment(
+                                                    lo->identity(), map)));
+    DirectHandle<HeapObject> keep_alive(ho, isolate);
   }
   CHECK_LT(0, successful_allocations);
 
   CHECK(!lo->IsEmpty());
-  CHECK(lo->AllocateRaw(lo_size).IsFailure());
+  CHECK(
+      lo->AllocateRaw(heap->main_thread_local_heap(), lo_size, AllocationHint())
+          .IsFailure());
 }
 
 #ifndef DEBUG
@@ -454,7 +478,7 @@ TEST(SizeOfInitialHeap) {
 // snapshot.
 // In PPC the page size is 64K, causing more internal fragmentation
 // hence requiring a larger limit.
-#if V8_OS_LINUX && (V8_HOST_ARCH_PPC || V8_HOST_ARCH_PPC64)
+#if V8_OS_LINUX && V8_HOST_ARCH_PPC64
   const size_t kMaxInitialSizePerSpace = 3 * MB;
 #else
   const size_t kMaxInitialSizePerSpace = 2 * MB;
@@ -504,6 +528,7 @@ HEAP_TEST(Regress777177) {
   OldSpace* old_space = heap->old_space();
   MainAllocator* old_space_allocator = heap->allocator()->old_space_allocator();
   Observer observer(128);
+  old_space_allocator->FreeLinearAllocationArea();
   old_space_allocator->AddAllocationObserver(&observer);
 
   int area_size = old_space->AreaSize();
@@ -514,8 +539,9 @@ HEAP_TEST(Regress777177) {
     // Ensure a new linear allocation area on a fresh page.
     AlwaysAllocateScopeForTesting always_allocate(heap);
     heap::SimulateFullSpace(old_space);
-    AllocationResult result =
-        old_space->AllocateRaw(filler_size, kTaggedAligned);
+    AllocationResult result = old_space_allocator->AllocateRaw(
+        filler_size, kTaggedAligned, AllocationOrigin::kRuntime,
+        AllocationHint());
     Tagged<HeapObject> obj = result.ToObjectChecked();
     heap->CreateFillerObjectAt(obj.address(), filler_size);
   }
@@ -523,19 +549,21 @@ HEAP_TEST(Regress777177) {
   {
     // Allocate all bytes of the linear allocation area. This moves top_ and
     // top_on_previous_step_ to the next page.
-    AllocationResult result =
-        old_space->AllocateRaw(max_object_size, kTaggedAligned);
+    AllocationResult result = old_space_allocator->AllocateRaw(
+        max_object_size, kTaggedAligned, AllocationOrigin::kRuntime,
+        AllocationHint());
     Tagged<HeapObject> obj = result.ToObjectChecked();
     // Simulate allocation folding moving the top pointer back.
-    old_space->SetTopAndLimit(
+    old_space_allocator->ResetLab(
         obj.address(), heap->allocator()->old_space_allocator()->limit(),
         heap->allocator()->old_space_allocator()->limit());
   }
 
   {
     // This triggers assert in crbug.com/777177.
-    AllocationResult result =
-        old_space->AllocateRaw(filler_size, kTaggedAligned);
+    AllocationResult result = old_space_allocator->AllocateRaw(
+        filler_size, kTaggedAligned, AllocationOrigin::kRuntime,
+        AllocationHint());
     Tagged<HeapObject> obj = result.ToObjectChecked();
     heap->CreateFillerObjectAt(obj.address(), filler_size);
   }
@@ -548,9 +576,8 @@ HEAP_TEST(Regress791582) {
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
   HandleScope scope(isolate);
-  NewSpace* new_space = heap->new_space();
   MainAllocator* new_space_allocator = heap->allocator()->new_space_allocator();
-  GrowNewSpace(heap);
+  GrowNewSpaceToMaximumCapacity(heap);
 
   int until_page_end =
       static_cast<int>(heap->NewSpaceLimit() - heap->NewSpaceTop());
@@ -563,11 +590,13 @@ HEAP_TEST(Regress791582) {
   }
 
   Observer observer(128);
+  new_space_allocator->FreeLinearAllocationArea();
   new_space_allocator->AddAllocationObserver(&observer);
 
   {
-    AllocationResult result =
-        new_space->AllocateRaw(until_page_end, kTaggedAligned);
+    AllocationResult result = new_space_allocator->AllocateRaw(
+        until_page_end, kTaggedAligned, AllocationOrigin::kRuntime,
+        AllocationHint());
     Tagged<HeapObject> obj = result.ToObjectChecked();
     heap->CreateFillerObjectAt(obj.address(), until_page_end);
     // Simulate allocation folding moving the top pointer back.
@@ -576,125 +605,14 @@ HEAP_TEST(Regress791582) {
 
   {
     // This triggers assert in crbug.com/791582
-    AllocationResult result = new_space->AllocateRaw(256, kTaggedAligned);
+    AllocationResult result = new_space_allocator->AllocateRaw(
+        256, kTaggedAligned, AllocationOrigin::kRuntime, AllocationHint());
     Tagged<HeapObject> obj = result.ToObjectChecked();
     heap->CreateFillerObjectAt(obj.address(), 256);
   }
   new_space_allocator->RemoveAllocationObserver(&observer);
 }
 
-TEST(ShrinkPageToHighWaterMarkFreeSpaceEnd) {
-  v8_flags.stress_incremental_marking = false;
-  v8_flags.stress_concurrent_allocation = false;  // For SealCurrentObjects.
-  CcTest::InitializeVM();
-  Isolate* isolate = CcTest::i_isolate();
-  HandleScope scope(isolate);
-
-  heap::SealCurrentObjects(CcTest::heap());
-
-  // Prepare page that only contains a single object and a trailing FreeSpace
-  // filler.
-  Handle<FixedArray> array =
-      isolate->factory()->NewFixedArray(128, AllocationType::kOld);
-  Page* page = Page::FromHeapObject(*array);
-
-  // Reset space so high water mark is consistent.
-  PagedSpace* old_space = CcTest::heap()->old_space();
-  old_space->FreeLinearAllocationArea();
-  old_space->ResetFreeList();
-
-  Tagged<HeapObject> filler =
-      HeapObject::FromAddress(array->address() + array->Size());
-  CHECK(IsFreeSpace(filler));
-  size_t shrunk = old_space->ShrinkPageToHighWaterMark(page);
-  size_t should_have_shrunk = RoundDown(
-      static_cast<size_t>(MemoryChunkLayout::AllocatableMemoryInDataPage() -
-                          array->Size()),
-      CommitPageSize());
-  CHECK_EQ(should_have_shrunk, shrunk);
-}
-
-TEST(ShrinkPageToHighWaterMarkNoFiller) {
-  v8_flags.stress_concurrent_allocation = false;  // For SealCurrentObjects.
-  CcTest::InitializeVM();
-  Isolate* isolate = CcTest::i_isolate();
-  HandleScope scope(isolate);
-  heap::SealCurrentObjects(CcTest::heap());
-
-  const int kFillerSize = 0;
-  std::vector<Handle<FixedArray>> arrays =
-      heap::FillOldSpacePageWithFixedArrays(CcTest::heap(), kFillerSize);
-  Handle<FixedArray> array = arrays.back();
-  Page* page = Page::FromHeapObject(*array);
-  CHECK_EQ(page->area_end(), array->address() + array->Size() + kFillerSize);
-
-  // Reset space so high water mark and fillers are consistent.
-  PagedSpace* old_space = CcTest::heap()->old_space();
-  old_space->ResetFreeList();
-  old_space->FreeLinearAllocationArea();
-
-  size_t shrunk = old_space->ShrinkPageToHighWaterMark(page);
-  CHECK_EQ(0u, shrunk);
-}
-
-TEST(ShrinkPageToHighWaterMarkOneWordFiller) {
-  v8_flags.stress_concurrent_allocation = false;  // For SealCurrentObjects.
-  CcTest::InitializeVM();
-  Isolate* isolate = CcTest::i_isolate();
-  HandleScope scope(isolate);
-
-  heap::SealCurrentObjects(CcTest::heap());
-
-  const int kFillerSize = kTaggedSize;
-  std::vector<Handle<FixedArray>> arrays =
-      heap::FillOldSpacePageWithFixedArrays(CcTest::heap(), kFillerSize);
-  Handle<FixedArray> array = arrays.back();
-  Page* page = Page::FromHeapObject(*array);
-  CHECK_EQ(page->area_end(), array->address() + array->Size() + kFillerSize);
-
-  // Reset space so high water mark and fillers are consistent.
-  PagedSpace* old_space = CcTest::heap()->old_space();
-  old_space->FreeLinearAllocationArea();
-  old_space->ResetFreeList();
-
-  Tagged<HeapObject> filler =
-      HeapObject::FromAddress(array->address() + array->Size());
-  CHECK_EQ(filler->map(),
-           ReadOnlyRoots(CcTest::heap()).one_pointer_filler_map());
-
-  size_t shrunk = old_space->ShrinkPageToHighWaterMark(page);
-  CHECK_EQ(0u, shrunk);
-}
-
-TEST(ShrinkPageToHighWaterMarkTwoWordFiller) {
-  v8_flags.stress_concurrent_allocation = false;  // For SealCurrentObjects.
-  CcTest::InitializeVM();
-  Isolate* isolate = CcTest::i_isolate();
-  HandleScope scope(isolate);
-
-  heap::SealCurrentObjects(CcTest::heap());
-
-  const int kFillerSize = 2 * kTaggedSize;
-  std::vector<Handle<FixedArray>> arrays =
-      heap::FillOldSpacePageWithFixedArrays(CcTest::heap(), kFillerSize);
-  Handle<FixedArray> array = arrays.back();
-  Page* page = Page::FromHeapObject(*array);
-  CHECK_EQ(page->area_end(), array->address() + array->Size() + kFillerSize);
-
-  // Reset space so high water mark and fillers are consistent.
-  PagedSpace* old_space = CcTest::heap()->old_space();
-  old_space->FreeLinearAllocationArea();
-  old_space->ResetFreeList();
-
-  Tagged<HeapObject> filler =
-      HeapObject::FromAddress(array->address() + array->Size());
-  CHECK_EQ(filler->map(),
-           ReadOnlyRoots(CcTest::heap()).two_pointer_filler_map());
-
-  size_t shrunk = old_space->ShrinkPageToHighWaterMark(page);
-  CHECK_EQ(0u, shrunk);
-}
-
 namespace {
 // PageAllocator that always fails.
 class FailingPageAllocator : public v8::PageAllocator {
@@ -720,6 +638,7 @@ class FailingPageAllocator : public v8::PageAllocator {
     return false;
   }
   bool DecommitPages(void* address, size_t length) override { return false; }
+  bool SealPages(void* address, size_t length) override { return false; }
 };
 }  // namespace
 
@@ -731,9 +650,8 @@ TEST(NoMemoryForNewPage) {
   FailingPageAllocator failing_allocator;
   TestMemoryAllocatorScope test_allocator_scope(isolate, 0, &failing_allocator);
   MemoryAllocator* memory_allocator = test_allocator_scope.allocator();
-  LinearAllocationArea allocation_info;
-  OldSpace faked_space(heap, allocation_info);
-  Page* page = memory_allocator->AllocatePage(
+  OldSpace faked_space(heap);
+  PageMetadata* page = memory_allocator->AllocatePage(
       MemoryAllocator::AllocationMode::kRegular,
       static_cast<PagedSpace*>(&faked_space), NOT_EXECUTABLE);
 
@@ -783,19 +701,17 @@ TEST(ReadOnlySpaceMetrics_OnePage) {
   // Allocated objects size.
   CHECK_EQ(faked_space->Size(), 16);
 
-  size_t committed_memory =
-      RoundUp(MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage() +
-                  faked_space->Size(),
-              MemoryAllocator::GetCommitPageSize());
+  size_t committed_memory = RoundUp(
+      MemoryChunkLayout::ObjectStartOffsetInDataPage() + faked_space->Size(),
+      MemoryAllocator::GetCommitPageSize());
 
   // Amount of OS allocated memory.
   CHECK_EQ(faked_space->CommittedMemory(), committed_memory);
   CHECK_EQ(faked_space->CommittedPhysicalMemory(), committed_memory);
 
   // Capacity will be one OS page minus the page header.
-  CHECK_EQ(
-      faked_space->Capacity(),
-      committed_memory - MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage());
+  CHECK_EQ(faked_space->Capacity(),
+           committed_memory - MemoryChunkLayout::ObjectStartOffsetInDataPage());
 }
 
 TEST(ReadOnlySpaceMetrics_AlignedAllocations) {
@@ -818,19 +734,19 @@ TEST(ReadOnlySpaceMetrics_AlignedAllocations) {
   int object_size =
       static_cast<int>(MemoryAllocator::GetCommitPageSize() - kApiTaggedSize);
 
-  int alignment = USE_ALLOCATION_ALIGNMENT_BOOL ? kDoubleSize : kTaggedSize;
+  const int kExpectedAlignment = kDoubleSize;
 
   Tagged<HeapObject> object =
       faked_space->AllocateRaw(object_size, kDoubleAligned).ToObjectChecked();
-  CHECK_EQ(object.address() % alignment, 0);
+  CHECK_EQ(object.address() % kExpectedAlignment, 0);
   object =
       faked_space->AllocateRaw(object_size, kDoubleAligned).ToObjectChecked();
-  CHECK_EQ(object.address() % alignment, 0);
+  CHECK_EQ(object.address() % kExpectedAlignment, 0);
 
   // Calculate size of allocations based on area_start.
   Address area_start = faked_space->pages().back()->GetAreaStart();
-  Address top = RoundUp(area_start, alignment) + object_size;
-  top = RoundUp(top, alignment) + object_size;
+  Address top = RoundUp(area_start, kExpectedAlignment) + object_size;
+  top = RoundUp(top, kExpectedAlignment) + object_size;
   size_t expected_size = top - area_start;
 
   faked_space->ShrinkPages();
@@ -840,18 +756,16 @@ TEST(ReadOnlySpaceMetrics_AlignedAllocations) {
   // with pointer compression.
   CHECK_EQ(faked_space->Size(), expected_size);
 
-  size_t committed_memory =
-      RoundUp(MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage() +
-                  faked_space->Size(),
-              MemoryAllocator::GetCommitPageSize());
+  size_t committed_memory = RoundUp(
+      MemoryChunkLayout::ObjectStartOffsetInDataPage() + faked_space->Size(),
+      MemoryAllocator::GetCommitPageSize());
 
   CHECK_EQ(faked_space->CommittedMemory(), committed_memory);
   CHECK_EQ(faked_space->CommittedPhysicalMemory(), committed_memory);
 
   // Capacity will be 3 OS pages minus the page header.
-  CHECK_EQ(
-      faked_space->Capacity(),
-      committed_memory - MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage());
+  CHECK_EQ(faked_space->Capacity(),
+           committed_memory - MemoryChunkLayout::ObjectStartOffsetInDataPage());
 }
 
 TEST(ReadOnlySpaceMetrics_TwoPages) {
@@ -890,9 +804,9 @@ TEST(ReadOnlySpaceMetrics_TwoPages) {
   CHECK_EQ(faked_space->Size(), object_size * 2);
 
   // Amount of OS allocated memory.
-  size_t committed_memory_per_page = RoundUp(
-      MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage() + object_size,
-      MemoryAllocator::GetCommitPageSize());
+  size_t committed_memory_per_page =
+      RoundUp(MemoryChunkLayout::ObjectStartOffsetInDataPage() + object_size,
+              MemoryAllocator::GetCommitPageSize());
   CHECK_EQ(faked_space->CommittedMemory(), 2 * committed_memory_per_page);
   CHECK_EQ(faked_space->CommittedPhysicalMemory(),
            2 * committed_memory_per_page);
@@ -900,10 +814,9 @@ TEST(ReadOnlySpaceMetrics_TwoPages) {
   // Capacity will be the space up to the amount of committed memory minus the
   // page headers.
   size_t capacity_per_page =
-      RoundUp(
-          MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage() + object_size,
-          MemoryAllocator::GetCommitPageSize()) -
-      MemoryChunkLayout::ObjectStartOffsetInReadOnlyPage();
+      RoundUp(MemoryChunkLayout::ObjectStartOffsetInDataPage() + object_size,
+              MemoryAllocator::GetCommitPageSize()) -
+      MemoryChunkLayout::ObjectStartOffsetInDataPage();
   CHECK_EQ(faked_space->Capacity(), 2 * capacity_per_page);
 }
 
