@@ -121,6 +121,17 @@ class StringResource : public v8::String::ExternalStringResource {
 
 }  // namespace
 
+class TestDebugHelper {
+ public:
+  static Address MetadataTableAddress() {
+#ifdef V8_ENABLE_SANDBOX
+    return reinterpret_cast<Address>(MemoryChunk::MetadataTableAddress());
+#else
+    return 0;
+#endif
+  }
+};
+
 TEST(GetObjectProperties) {
   CcTest::InitializeVM();
   v8::Isolate* isolate = CcTest::isolate();
@@ -129,10 +140,10 @@ TEST(GetObjectProperties) {
   v8::HandleScope scope(isolate);
   LocalContext context;
   // Claim we don't know anything about the heap layout.
-  d::HeapAddresses heap_addresses{0, 0, 0, 0};
+  d::HeapAddresses heap_addresses{0, 0, 0, 0, 0};
 
   v8::Local<v8::Value> v = CompileRun("42");
-  Handle<Object> o = v8::Utils::OpenHandle(*v);
+  DirectHandle<Object> o = v8::Utils::OpenDirectHandle(*v);
   d::ObjectPropertiesResultPtr props =
       d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   CHECK(props->type_check_result == d::TypeCheckResult::kSmi);
@@ -141,7 +152,7 @@ TEST(GetObjectProperties) {
   CHECK_EQ(props->num_properties, 0);
 
   v = CompileRun("[\"a\", \"bc\"]");
-  o = v8::Utils::OpenHandle(*v);
+  o = v8::Utils::OpenDirectHandle(*v);
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   CHECK(props->type_check_result == d::TypeCheckResult::kUsedMap);
   CHECK(props->type == std::string("v8::internal::JSArray"));
@@ -305,7 +316,7 @@ TEST(GetObjectProperties) {
   v = CompileRun(R"(
     const alphabet = "abcdefghijklmnopqrstuvwxyz";
     alphabet.substr(3,20) + alphabet.toUpperCase().substr(5,15) + "7")");
-  o = v8::Utils::OpenHandle(*v);
+  o = v8::Utils::OpenDirectHandle(*v);
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   CHECK(Contains(props->brief, "\"defghijklmnopqrstuvwFGHIJKLMNOPQRST7\""));
 
@@ -322,31 +333,34 @@ TEST(GetObjectProperties) {
 
   // Build a very long string.
   v = CompileRun("'a'.repeat(1000)");
-  o = v8::Utils::OpenHandle(*v);
+  o = v8::Utils::OpenDirectHandle(*v);
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   CHECK(Contains(props->brief, "\"" + std::string(80, 'a') + "...\""));
 
   // GetObjectProperties can read cacheable external strings.
+  heap_addresses.metadata_pointer_table =
+      TestDebugHelper::MetadataTableAddress();
   StringResource* string_resource = new StringResource(true);
-  auto external_string =
+  auto cachable_external_string =
       v8::String::NewExternalTwoByte(isolate, string_resource);
-  o = v8::Utils::OpenHandle(*external_string.ToLocalChecked());
+  o = v8::Utils::OpenDirectHandle(*cachable_external_string.ToLocalChecked());
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   CHECK(Contains(props->brief, "\"abcde\""));
   CheckProp(*props->properties[5], "char16_t", "raw_characters",
             d::PropertyKind::kArrayOfKnownSize, string_resource->length());
   CHECK_EQ(props->properties[5]->address,
            reinterpret_cast<uintptr_t>(string_resource->data()));
+
   // GetObjectProperties cannot read uncacheable external strings.
-  external_string =
+  auto external_string =
       v8::String::NewExternalTwoByte(isolate, new StringResource(false));
-  o = v8::Utils::OpenHandle(*external_string.ToLocalChecked());
+  o = v8::Utils::OpenDirectHandle(*external_string.ToLocalChecked());
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   CHECK_EQ(std::string(props->brief).find("\""), std::string::npos);
 
   // Build a basic JS object and get its properties.
   v = CompileRun("({a: 1, b: 2})");
-  o = v8::Utils::OpenHandle(*v);
+  o = v8::Utils::OpenDirectHandle(*v);
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
 
   // Objects constructed from literals get their properties placed inline, so
@@ -371,7 +385,8 @@ TEST(GetObjectProperties) {
   props = d::GetObjectProperties(
       ReadProp<i::Tagged_t>(*props, "instance_descriptors"), &ReadMemory,
       heap_addresses);
-  CHECK_EQ(props->num_properties, 6);
+  int padding = TAGGED_SIZE_8_BYTES ? 1 : 0;
+  CHECK_EQ(props->num_properties, 7 + padding);
   // It should have at least two descriptors (possibly plus slack).
   CheckProp(*props->properties[1], "uint16_t", "number_of_all_descriptors");
   uint16_t number_of_all_descriptors =
@@ -379,7 +394,7 @@ TEST(GetObjectProperties) {
   CHECK_GE(number_of_all_descriptors, 2);
   // The "descriptors" property should describe the struct layout for each
   // element in the array.
-  const d::ObjectProperty& descriptors = *props->properties[5];
+  const d::ObjectProperty& descriptors = *props->properties[6 + padding];
   // No C++ type is reported directly because there may not be an actual C++
   // struct with this layout, hence the empty string in this check.
   CheckProp(descriptors, /*type=*/"", "descriptors",
@@ -400,7 +415,7 @@ TEST(GetObjectProperties) {
   // Build a basic JS function and get its properties. This will allow us to
   // exercise bitfield functionality.
   v = CompileRun("(function () {})");
-  o = v8::Utils::OpenHandle(*v);
+  o = v8::Utils::OpenDirectHandle(*v);
   props = d::GetObjectProperties((*o).ptr(), &ReadMemory, heap_addresses);
   props = d::GetObjectProperties(
       ReadProp<i::Tagged_t>(*props, "shared_function_info"), &ReadMemory,
@@ -413,9 +428,9 @@ TEST(GetObjectProperties) {
   CheckStructProp(*flags.struct_fields[2], "bool", "is_strict", 0, 1, 6);
 
   // Get data about a different bitfield struct which is contained within a smi.
-  Handle<i::JSFunction> function = Handle<i::JSFunction>::cast(o);
-  Handle<i::SharedFunctionInfo> shared(function->shared(), i_isolate);
-  Handle<i::DebugInfo> debug_info =
+  DirectHandle<i::JSFunction> function = Cast<i::JSFunction>(o);
+  DirectHandle<i::SharedFunctionInfo> shared(function->shared(), i_isolate);
+  DirectHandle<i::DebugInfo> debug_info =
       i_isolate->debug()->GetOrCreateDebugInfo(shared);
   props =
       d::GetObjectProperties(debug_info->ptr(), &ReadMemory, heap_addresses);
@@ -427,32 +442,14 @@ TEST(GetObjectProperties) {
                   0, 1, i::kSmiTagSize + i::kSmiShiftSize + 4);
 }
 
-TEST(ListObjectClasses) {
-  CcTest::InitializeVM();
-
-  // The ListObjectClasses result will change as classes are added, removed, or
-  // renamed. Just check that a few expected classes are included in the list,
-  // and that there are no duplicates.
-  const d::ClassList* class_list = d::ListObjectClasses();
-  std::unordered_set<std::string> class_set;
-  for (size_t i = 0; i < class_list->num_class_names; ++i) {
-    CHECK_WITH_MSG(class_set.insert(class_list->class_names[i]).second,
-                   "there should be no duplicate entries");
-  }
-  CHECK_NE(class_set.find("v8::internal::HeapObject"), class_set.end());
-  CHECK_NE(class_set.find("v8::internal::String"), class_set.end());
-  CHECK_NE(class_set.find("v8::internal::JSRegExp"), class_set.end());
-}
-
 static void FrameIterationCheck(
-    v8::Local<v8::String> name,
-    const v8::PropertyCallbackInfo<v8::Value>& info) {
+    v8::Local<v8::Name> name, const v8::PropertyCallbackInfo<v8::Value>& info) {
   i::StackFrameIterator iter(reinterpret_cast<i::Isolate*>(info.GetIsolate()));
   for (int i = 0; !iter.done(); i++) {
     i::StackFrame* frame = iter.frame();
     CHECK(i != 0 || (frame->type() == i::StackFrame::EXIT));
     d::StackFrameResultPtr props = d::GetStackFrame(frame->fp(), &ReadMemory);
-    if (frame->is_java_script()) {
+    if (frame->is_javascript()) {
       JavaScriptFrame* js_frame = JavaScriptFrame::cast(frame);
       CHECK_EQ(props->num_properties, 5);
       auto js_function = js_frame->function();
@@ -461,7 +458,7 @@ static void FrameIterationCheck(
                 "v8::internal::Tagged<v8::internal::JSFunction>",
                 "currently_executing_jsfunction", js_function.ptr());
       auto shared_function_info = js_function->shared();
-      auto script = i::Script::cast(shared_function_info->script());
+      auto script = i::Cast<i::Script>(shared_function_info->script());
       CheckProp(*props->properties[1],
                 "v8::internal::TaggedMember<v8::internal::Object>",
                 "script_name", static_cast<i::Tagged_t>(script->name().ptr()));
@@ -495,12 +492,12 @@ static void FrameIterationCheck(
 
 THREADED_TEST(GetFrameStack) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   i::Isolate* i_isolate = reinterpret_cast<Isolate*>(isolate);
   PtrComprCageAccessScope ptr_compr_cage_access_scope(i_isolate);
   v8::HandleScope scope(isolate);
   v8::Local<v8::ObjectTemplate> obj = v8::ObjectTemplate::New(isolate);
-  obj->SetAccessor(v8_str("xxx"), FrameIterationCheck);
+  obj->SetNativeDataProperty(v8_str("xxx"), FrameIterationCheck);
   CHECK(env->Global()
             ->Set(env.local(), v8_str("obj"),
                   obj->NewInstance(env.local()).ToLocalChecked())
@@ -516,19 +513,19 @@ THREADED_TEST(GetFrameStack) {
 
 TEST(SmallOrderedHashSetGetObjectProperties) {
   LocalContext context;
-  Isolate* isolate = reinterpret_cast<Isolate*>((*context)->GetIsolate());
+  Isolate* isolate = context.i_isolate();
   Factory* factory = isolate->factory();
   PtrComprCageAccessScope ptr_compr_cage_access_scope(isolate);
   HandleScope scope(isolate);
 
-  Handle<SmallOrderedHashSet> set = factory->NewSmallOrderedHashSet();
+  DirectHandle<SmallOrderedHashSet> set = factory->NewSmallOrderedHashSet();
   const size_t number_of_buckets = 2;
   CHECK_EQ(number_of_buckets, set->NumberOfBuckets());
   CHECK_EQ(0, set->NumberOfElements());
 
   // Verify with the definition of SmallOrderedHashSet in
   // src\objects\ordered-hash-table.tq.
-  d::HeapAddresses heap_addresses{0, 0, 0, 0};
+  d::HeapAddresses heap_addresses{0, 0, 0, 0, 0};
   d::ObjectPropertiesResultPtr props =
       d::GetObjectProperties(set->ptr(), &ReadMemory, heap_addresses);
   CHECK_EQ(props->type_check_result, d::TypeCheckResult::kUsedMap);
