@@ -35,6 +35,7 @@
 
 using ::v8::base::CStrVector;
 using ::v8::base::Vector;
+using ::v8::internal::DirectHandle;
 using ::v8::internal::Factory;
 using ::v8::internal::Handle;
 using ::v8::internal::Heap;
@@ -49,19 +50,16 @@ static void CheckFunctionName(v8::Local<v8::Script> script,
   i::Isolate* isolate = CcTest::i_isolate();
 
   // Get script source.
-  Handle<i::Object> obj = v8::Utils::OpenHandle(*script);
-  Handle<SharedFunctionInfo> shared_function;
-  if (IsSharedFunctionInfo(*obj)) {
-    shared_function =
-        Handle<SharedFunctionInfo>(SharedFunctionInfo::cast(*obj), isolate);
-  } else {
-    shared_function =
-        Handle<SharedFunctionInfo>(JSFunction::cast(*obj)->shared(), isolate);
-  }
-  Handle<i::Script> i_script(i::Script::cast(shared_function->script()),
+  DirectHandle<i::Object> obj = v8::Utils::OpenDirectHandle(*script);
+  DirectHandle<SharedFunctionInfo> shared_function(
+      IsSharedFunctionInfo(*obj) ? Cast<SharedFunctionInfo>(*obj)
+                                 : Cast<JSFunction>(*obj)->shared(),
+      isolate);
+  Handle<i::Script> i_script(i::Cast<i::Script>(shared_function->script()),
                              isolate);
   CHECK(IsString(i_script->source()));
-  Handle<i::String> script_src(i::String::cast(i_script->source()), isolate);
+  DirectHandle<i::String> script_src(i::Cast<i::String>(i_script->source()),
+                                     isolate);
 
   // Find the position of a given func source substring in the source.
   int func_pos;
@@ -76,10 +74,9 @@ static void CheckFunctionName(v8::Local<v8::Script> script,
   CHECK_NE(0, func_pos);
 
   // Obtain SharedFunctionInfo for the function.
-  Handle<SharedFunctionInfo> shared_func_info =
-      Handle<SharedFunctionInfo>::cast(
-          isolate->debug()->FindInnermostContainingFunctionInfo(i_script,
-                                                                func_pos));
+  DirectHandle<SharedFunctionInfo> shared_func_info = Cast<SharedFunctionInfo>(
+      isolate->debug()->FindInnermostContainingFunctionInfo(i_script,
+                                                            func_pos));
 
   // Verify inferred function name.
   std::unique_ptr<char[]> inferred_name =
