@@ -32,6 +32,7 @@
 #include "src/base/utils/random-number-generator.h"
 #include "src/codegen/assembler-inl.h"
 #include "src/codegen/macro-assembler.h"
+#include "src/compiler/access-builder.h"
 #include "src/deoptimizer/deoptimizer.h"
 #include "src/execution/simulator.h"
 #include "src/objects/objects-inl.h"
@@ -740,7 +741,7 @@ TEST(cvt_d_l_Trunc_l_ud) {
   CcTest::InitializeVM();
   FOR_INT64_INPUTS(i, cvt_trunc_int64_test_values) {
     int64_t input = *i;
-    uint64_t abs_input = (input < 0) ? -input : input;
+    uint64_t abs_input = (input >= 0 || input == INT64_MIN) ? input : -input;
     auto fn = [](MacroAssembler* masm) {
       __ dmtc1(a0, f4);
       __ cvt_d_l(f0, f4);
@@ -934,7 +935,8 @@ TEST(min_max_nan) {
   auto handle_dnan = [masm](FPURegister dst, Label* nan, Label* back) {
     __ bind(nan);
     __ LoadRoot(t8, RootIndex::kNanValue);
-    __ Ldc1(dst, FieldMemOperand(t8, HeapNumber::kValueOffset));
+    __ Ldc1(dst, FieldMemOperand(
+                     t8, compiler::AccessBuilder::ForHeapNumberValue().offset));
     __ Branch(back);
   };
 
