@@ -134,8 +134,8 @@ WASM_EXEC_TEST(MemoryInitOutOfBounds) {
 WASM_EXEC_TEST(MemoryCopy) {
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   uint8_t* mem = r.builder().AddMemory(kWasmPageSize);
-  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
-                            WASM_LOCAL_GET(2)),
+  r.Build({WASM_MEMORY0_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                             WASM_LOCAL_GET(2)),
            kExprI32Const, 0});
 
   const uint8_t initial[] = {0, 11, 22, 33, 44, 55, 66, 77};
@@ -161,8 +161,8 @@ WASM_EXEC_TEST(MemoryCopy) {
 WASM_EXEC_TEST(MemoryCopyOverlapping) {
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   uint8_t* mem = r.builder().AddMemory(kWasmPageSize);
-  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
-                            WASM_LOCAL_GET(2)),
+  r.Build({WASM_MEMORY0_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                             WASM_LOCAL_GET(2)),
            kExprI32Const, 0});
 
   const uint8_t initial[] = {10, 20, 30};
@@ -182,8 +182,8 @@ WASM_EXEC_TEST(MemoryCopyOverlapping) {
 WASM_EXEC_TEST(MemoryCopyOutOfBoundsData) {
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   uint8_t* mem = r.builder().AddMemory(kWasmPageSize);
-  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
-                            WASM_LOCAL_GET(2)),
+  r.Build({WASM_MEMORY0_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                             WASM_LOCAL_GET(2)),
            kExprI32Const, 0});
 
   const uint8_t data[] = {11, 22, 33, 44, 55, 66, 77, 88};
@@ -209,8 +209,8 @@ WASM_EXEC_TEST(MemoryCopyOutOfBoundsData) {
 WASM_EXEC_TEST(MemoryCopyOutOfBounds) {
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
-                            WASM_LOCAL_GET(2)),
+  r.Build({WASM_MEMORY0_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                             WASM_LOCAL_GET(2)),
            kExprI32Const, 0});
 
   // Copy full range is OK.
@@ -360,24 +360,25 @@ WASM_COMPILED_EXEC_TEST(TableCopyInboundsFrom6To6) {
 
 namespace {
 template <typename... Args>
-void CheckTable(Isolate* isolate, Handle<WasmTableObject> table, Args... args) {
+void CheckTable(Isolate* isolate, DirectHandle<WasmTableObject> table,
+                Args... args) {
   uint32_t args_length = static_cast<uint32_t>(sizeof...(args));
   CHECK_EQ(table->current_length(), args_length);
-  Handle<Object> handles[] = {args...};
+  DirectHandle<Object> handles[] = {args...};
   for (uint32_t i = 0; i < args_length; ++i) {
     CHECK(WasmTableObject::Get(isolate, table, i).is_identical_to(handles[i]));
   }
 }
 
 template <typename WasmRunner, typename... Args>
-void CheckTableCall(Isolate* isolate, Handle<WasmTableObject> table,
+void CheckTableCall(Isolate* isolate, DirectHandle<WasmTableObject> table,
                     WasmRunner* r, uint32_t function_index, Args... args) {
   uint32_t args_length = static_cast<uint32_t>(sizeof...(args));
   CHECK_EQ(table->current_length(), args_length);
   double expected[] = {args...};
   for (uint32_t i = 0; i < args_length; ++i) {
-    Handle<Object> buffer[] = {isolate->factory()->NewNumber(i)};
-    r->CheckCallApplyViaJS(expected[i], function_index, buffer, 1);
+    DirectHandle<Object> buffer[] = {isolate->factory()->NewNumber(i)};
+    r->CheckCallApplyViaJS(expected[i], function_index, base::VectorOf(buffer));
   }
 }
 }  // namespace
@@ -390,7 +391,7 @@ void TestTableCopyElems(TestExecutionTier execution_tier, int table_dst,
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   const uint32_t kTableSize = 5;
   uint16_t function_indexes[kTableSize];
-  const uint32_t sig_index = r.builder().AddSignature(sigs.i_v());
+  const ModuleTypeIndex sig_index = r.builder().AddSignature(sigs.i_v());
 
   for (uint32_t i = 0; i < kTableSize; ++i) {
     WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
@@ -410,8 +411,8 @@ void TestTableCopyElems(TestExecutionTier execution_tier, int table_dst,
   r.builder().InitializeWrapperCache();
 
   auto table =
-      handle(WasmTableObject::cast(
-                 r.builder().instance_object()->tables()->get(table_dst)),
+      handle(Cast<WasmTableObject>(
+                 r.builder().trusted_instance_data()->tables()->get(table_dst)),
              isolate);
   r.CheckCallViaJS(0, 0, 0, kTableSize);
   auto f0 = WasmTableObject::Get(isolate, table, 0);
@@ -467,7 +468,7 @@ void TestTableCopyCalls(TestExecutionTier execution_tier, int table_dst,
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   const uint32_t kTableSize = 5;
   uint16_t function_indexes[kTableSize];
-  const uint32_t sig_index = r.builder().AddSignature(sigs.i_v());
+  const ModuleTypeIndex sig_index = r.builder().AddSignature(sigs.i_v());
 
   for (uint32_t i = 0; i < kTableSize; ++i) {
     WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
@@ -490,8 +491,8 @@ void TestTableCopyCalls(TestExecutionTier execution_tier, int table_dst,
            kExprI32Const, 0});
 
   auto table =
-      handle(WasmTableObject::cast(
-                 r.builder().instance_object()->tables()->get(table_dst)),
+      handle(Cast<WasmTableObject>(
+                 r.builder().trusted_instance_data()->tables()->get(table_dst)),
              isolate);
 
   if (table_dst == table_src) {
@@ -537,7 +538,7 @@ void TestTableCopyOobWrites(TestExecutionTier execution_tier, int table_dst,
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   const uint32_t kTableSize = 5;
   uint16_t function_indexes[kTableSize];
-  const uint32_t sig_index = r.builder().AddSignature(sigs.i_v());
+  const ModuleTypeIndex sig_index = r.builder().AddSignature(sigs.i_v());
 
   for (uint32_t i = 0; i < kTableSize; ++i) {
     WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
@@ -557,8 +558,8 @@ void TestTableCopyOobWrites(TestExecutionTier execution_tier, int table_dst,
   r.builder().InitializeWrapperCache();
 
   auto table =
-      handle(WasmTableObject::cast(
-                 r.builder().instance_object()->tables()->get(table_dst)),
+      handle(Cast<WasmTableObject>(
+                 r.builder().trusted_instance_data()->tables()->get(table_dst)),
              isolate);
   // Fill the dst table with values from the src table, to make checks easier.
   r.CheckCallViaJS(0, 0, 0, kTableSize);
