@@ -6,16 +6,16 @@ Debug = debug.Debug;
 
 Debug.setListener(function() {});
 
-function factory() {
-  return function* generator() {
+function main() {
+  function* boo() {
     debugger;
     yield 42;
   }
+
+  var gen = boo();
+  gen.next();
+  Debug.setBreakPoint(main, 0, 0);
+  gen.next();
 }
 
-Debug.setBreakPoint(function(){}, 0, 0);
-var generator = factory();
-var obj = generator();
-obj.next();
-Debug.setBreakPoint(factory, 0, 0);
-obj.next();
+main();
