@@ -114,7 +114,6 @@ class DebugWrapper {
 
   // Returns the resulting breakpoint id.
   setBreakPoint(func, opt_line, opt_column, opt_condition) {
-    assertTrue(%IsFunction(func));
     assertFalse(%FunctionIsAPIFunction(func));
 
     const scriptid = %FunctionGetScriptId(func);
@@ -144,8 +143,6 @@ class DebugWrapper {
   }
 
   showBreakPoints(f) {
-    if (!%IsFunction(f)) throw new Error("Not passed a Function");
-
     const source = %FunctionGetSourceCode(f);
     const offset = %FunctionGetScriptSourcePosition(f);
     const locations = %GetBreakLocations(f);
