@@ -2,7 +2,7 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-InspectorTest.start('Don\'t crash when pausing in a untrusted session.');
+InspectorTest.start('Don\'t crash when pausing in an untrusted session.');
 
 (async () => {
   const contextGroup = new InspectorTest.ContextGroup();
