@@ -18,7 +18,10 @@
 // Test JSCallReducer::ReduceJSCallWithArrayLike.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
@@ -58,7 +61,10 @@
 // Test with holey array.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend
+  // on ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
@@ -80,7 +86,10 @@
 // Test with holey-double array.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend
+  // on ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + (c ? c : .0) + d;
@@ -95,27 +104,19 @@
   %OptimizeFunctionOnNextCall(foo);
   assertTrue(sum_js_got_interpreted);
 
-  // This is expected to deoptimize
+  // We optimize in the next call and  sum_js stays inlined.
   assertEquals(45.31, foo(16.11, 26.06));
-  assertTrue(sum_js_got_interpreted);
-  assertUnoptimized(foo);
-
-  // Optimize again
-  %PrepareFunctionForOptimization(foo);
-  assertEquals(45.31, foo(16.11, 26.06));
-  %OptimizeFunctionOnNextCall(foo);
-  assertTrue(sum_js_got_interpreted);
-
-  // This should stay optimized, but with the call not inlined.
-  assertEquals(45.31, foo(16.11, 26.06));
-  assertTrue(sum_js_got_interpreted);
+  assertFalse(sum_js_got_interpreted);
   assertOptimized(foo);
 })();
 
 // Test deopt when array size changes.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
@@ -147,7 +148,10 @@
 // Test with FixedDoubleArray.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
@@ -169,7 +173,10 @@
 // Test with empty array.
 (function () {
   "use strict";
-  var got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var got_interpreted = null;
+  got_interpreted = true;
   function fortytwo() {
     got_interpreted = %IsBeingInterpreted();
     return 42;
@@ -191,13 +198,19 @@
 // Test with empty array that changes size.
 (function () {
   "use strict";
-  var got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var got_interpreted = null;
+  got_interpreted = true;
   function fortytwo() {
     got_interpreted = %IsBeingInterpreted();
     return 42 + arguments.length;
   }
 
-  var len = 2;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var len = null;
+  len = 2;
   function foo() {
     let args = []
     for (var i = 0; i < len; i++) { args.push(1); }
@@ -279,7 +292,10 @@
 // Test Reflect.apply().
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
@@ -301,7 +317,10 @@
 // Test Reflect.apply() with empty array.
 (function () {
   "use strict";
-  var got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var got_interpreted = null;
+  got_interpreted = true;
   function fortytwo() {
     got_interpreted = %IsBeingInterpreted();
     return 42;
@@ -323,13 +342,17 @@
 // Test Reflect.apply() with empty array that changes size.
 (function () {
   "use strict";
-  var got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var got_interpreted = null;
+  got_interpreted = true;
   function fortytwo() {
     got_interpreted = %IsBeingInterpreted();
     return 42 + arguments.length;
   }
 
-  var len = 2;
+  var len = null;
+  len = 2;
   function foo() {
     let args = []
     for (var i = 0; i < len; i++) { args.push(1); }
@@ -355,7 +378,10 @@
 // Test JSCallReducer::ReduceJSCallWithSpread.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
@@ -378,7 +404,10 @@
 // Test spread call with empty array.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c;
@@ -401,13 +430,17 @@
 // Test spread call with empty array that changes size.
 (function () {
   "use strict";
-  var max_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var max_got_interpreted = null;
+  max_got_interpreted = true;
   function max() {
     max_got_interpreted = %IsBeingInterpreted();
     return Math.max(...arguments);
   }
 
-  var len = 2;
+  var len = null;
+  len = 2;
   function foo(x, y, z) {
     let args = [];
     for (var i = 0; i < len; i++) { args.push(4 + i); }
@@ -432,7 +465,10 @@
 // Test spread call with more args.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true;
   function sum_js(a, b, c, d, e, f, g) {
     assertEquals(7, arguments.length);
     sum_js_got_interpreted = %IsBeingInterpreted();
@@ -532,7 +568,10 @@
 // Test apply on JSCreateClosure.
 (function () {
   "use strict";
-  var sum_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_got_interpreted = null;
+  sum_got_interpreted = true;
   function foo_closure() {
     return function(a, b, c, d) {
       sum_got_interpreted = %IsBeingInterpreted();
@@ -560,7 +599,10 @@
 // Test apply with JSBoundFunction
 (function () {
   "use strict";
-  var sum_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_got_interpreted = null;
+  sum_got_interpreted = true;
   function sum_js(a, b, c, d, e) {
     sum_got_interpreted = %IsBeingInterpreted();
     return this.x + a + b + c + d + e;
@@ -584,7 +626,10 @@
 // Test apply with nested bindings
 (function () {
   "use strict";
-  var sum_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_got_interpreted = null;
+  sum_got_interpreted = true;
   function sum_js(a, b, c, d, e) {
     sum_got_interpreted = %IsBeingInterpreted();
     return this.x + a + b + c + d + e;
@@ -608,7 +653,10 @@
 // Test apply on bound function (JSCreateBoundFunction).
 (function () {
   "use strict";
-  var sum_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_got_interpreted = null;
+  sum_got_interpreted = true;
   function sum_js(a, b, c, d) {
     sum_got_interpreted = %IsBeingInterpreted();
     return this.x + a + b + c + d;
@@ -631,7 +679,10 @@
 // Test apply on bound function (JSCreateBoundFunction) with args.
 (function () {
   "use strict";
-  var sum_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_got_interpreted = null;
+  sum_got_interpreted = true
   function sum_js(a, b, c, d, e, f) {
     sum_got_interpreted = %IsBeingInterpreted();
     return this.x + a + b + c + d + e + f;
@@ -654,7 +705,10 @@
 // Test call with array-like under-application.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d + arguments.length;
@@ -676,7 +730,10 @@
 // Test call with array-like over-application.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d + arguments.length;
@@ -698,7 +755,10 @@
 // Test call with spread under-application.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d + arguments.length;
@@ -721,7 +781,10 @@
 // Test call with spread over-application.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d + arguments.length;
@@ -744,7 +807,10 @@
 // Test calling function that has rest parameters.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true
   function sum_js(a, b, ...moreArgs) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + moreArgs[0] + moreArgs[1] + moreArgs[2] + moreArgs[3];
@@ -790,7 +856,10 @@
 // Test with inlined calls.
 (function () {
   "use strict";
-  var sum_js_got_interpreted = true;
+  // Introduce an indirection, so that we don't depend on
+  // ContextCells constness.
+  var sum_js_got_interpreted = null;
+  sum_js_got_interpreted = true
   function sum_js(a, b, c, d) {
     sum_js_got_interpreted = %IsBeingInterpreted();
     return a + b + c + d;
