@@ -2,20 +2,18 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 //
-// Flags: --allow-natives-syntax --maglev --maglev-escape-analysis
-
-function foo(o) {
-  o.g.h;
-  return o;
-}
+// Flags: --allow-natives-syntax --maglev --maglev-escape-analysis --verify-heap
 
 function bar() {
-  foo({ f:42, g:{h:3} });
+  return arguments;
+}
+function foo() {
+  let v = bar();
+  bar(v);
 }
-
-%PrepareFunctionForOptimization(foo);
 %PrepareFunctionForOptimization(bar);
-bar();
-bar();
-%OptimizeMaglevOnNextCall(bar);
-bar();
+%PrepareFunctionForOptimization(foo);
+foo();
+foo();
+%OptimizeMaglevOnNextCall(foo);
+foo();
