@@ -3,6 +3,6 @@
 // found in the LICENSE file.
 
 try {
-  load("test/mjsunit/regress/regress-444805.js-script");
+  d8.file.execute("test/mjsunit/regress/regress-444805.js-script");
 } catch (e) {
 }
