@@ -2,16 +2,18 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-// Flags: --experimental-wasm-gc --experimental-wasm-stringref
+// Flags: --wasm-staging
 
 d8.file.execute('test/mjsunit/wasm/wasm-module-builder.js');
 
 const builder = new WasmModuleBuilder();
 
-builder.addFunction("main", kSig_i_v).exportFunc().addBodyWithEnd([
-  kExprRefNull, kStringViewIterCode,
-  kGCPrefix, kExprRefTestNull, kAnyRefCode,
-  kExprEnd,
-]);
+builder.addFunction("main", makeSig([kWasmStringRef], [kWasmI32])).exportFunc()
+  .addBodyWithEnd([
+    kExprLocalGet, 0,
+    ...GCInstr(kExprStringAsIter),
+    kGCPrefix, kExprRefTestNull, kAnyRefCode,
+    kExprEnd,
+  ]);
 const instance = builder.instantiate();
-assertEquals(0, instance.exports.main());
+assertEquals(0, instance.exports.main("foo"));
