@@ -14,7 +14,7 @@ class C extends bar {
 }
 
 function foo() {
-  new C(-8);
+  new C(4.25);
 }
 
 %PrepareFunctionForOptimization(bar);
