@@ -20,6 +20,7 @@
 #include "src/objects/allocation-site-inl.h"
 #include "src/objects/objects-inl.h"
 #include "src/objects/shared-function-info.h"
+#include "test/unittests/heap/heap-utils.h"  // For ManualGCScope.
 #include "test/unittests/test-utils.h"
 #include "testing/gtest/include/gtest/gtest.h"
 
@@ -28,7 +29,7 @@ namespace v8 {
 using CompilerTest = TestWithContext;
 namespace internal {
 
-static Handle<Object> GetGlobalProperty(const char* name) {
+static DirectHandle<Object> GetGlobalProperty(const char* name) {
   Isolate* isolate = reinterpret_cast<i::Isolate*>(v8::Isolate::GetCurrent());
   return JSReceiver::GetProperty(isolate, isolate->global_object(), name)
       .ToHandleChecked();
@@ -36,25 +37,27 @@ static Handle<Object> GetGlobalProperty(const char* name) {
 
 static void SetGlobalProperty(const char* name, Tagged<Object> value) {
   Isolate* isolate = reinterpret_cast<i::Isolate*>(v8::Isolate::GetCurrent());
-  Handle<Object> object(value, isolate);
-  Handle<String> internalized_name =
+  DirectHandle<Object> object(value, isolate);
+  DirectHandle<String> internalized_name =
       isolate->factory()->InternalizeUtf8String(name);
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   Runtime::SetObjectProperty(isolate, global, internalized_name, object,
                              StoreOrigin::kMaybeKeyed, Just(kDontThrow))
       .Check();
 }
 
-static Handle<JSFunction> Compile(const char* source) {
+static DirectHandle<JSFunction> Compile(const char* source) {
   Isolate* isolate = reinterpret_cast<i::Isolate*>(v8::Isolate::GetCurrent());
   Handle<String> source_code = isolate->factory()
                                    ->NewStringFromUtf8(base::CStrVector(source))
                                    .ToHandleChecked();
-  Handle<SharedFunctionInfo> shared =
+  ScriptCompiler::CompilationDetails compilation_details;
+  DirectHandle<SharedFunctionInfo> shared =
       Compiler::GetSharedFunctionInfoForScript(
           isolate, source_code, ScriptDetails(),
           v8::ScriptCompiler::kNoCompileOptions,
-          ScriptCompiler::kNoCacheNoReason, NOT_NATIVES_CODE)
+          ScriptCompiler::kNoCacheNoReason, NOT_NATIVES_CODE,
+          &compilation_details)
           .ToHandleChecked();
   return Factory::JSFunctionBuilder{isolate, shared, isolate->native_context()}
       .Build();
@@ -65,14 +68,14 @@ static double Inc(Isolate* isolate, int x) {
   base::EmbeddedVector<char, 512> buffer;
   SNPrintF(buffer, source, x);
 
-  Handle<JSFunction> fun = Compile(buffer.begin());
+  DirectHandle<JSFunction> fun = Compile(buffer.begin());
   if (fun.is_null()) return -1;
 
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   Execution::CallScript(isolate, fun, global,
                         isolate->factory()->empty_fixed_array())
       .Check();
-  return Object::Number(*GetGlobalProperty("result"));
+  return Object::NumberValue(*GetGlobalProperty("result"));
 }
 
 TEST_F(CompilerTest, Inc) {
@@ -81,16 +84,16 @@ TEST_F(CompilerTest, Inc) {
 }
 
 static double Add(Isolate* isolate, int x, int y) {
-  Handle<JSFunction> fun = Compile("result = x + y;");
+  DirectHandle<JSFunction> fun = Compile("result = x + y;");
   if (fun.is_null()) return -1;
 
   SetGlobalProperty("x", Smi::FromInt(x));
   SetGlobalProperty("y", Smi::FromInt(y));
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   Execution::CallScript(isolate, fun, global,
                         isolate->factory()->empty_fixed_array())
       .Check();
-  return Object::Number(*GetGlobalProperty("result"));
+  return Object::NumberValue(*GetGlobalProperty("result"));
 }
 
 TEST_F(CompilerTest, Add) {
@@ -99,15 +102,16 @@ TEST_F(CompilerTest, Add) {
 }
 
 static double Abs(Isolate* isolate, int x) {
-  Handle<JSFunction> fun = Compile("if (x < 0) result = -x; else result = x;");
+  DirectHandle<JSFunction> fun =
+      Compile("if (x < 0) result = -x; else result = x;");
   if (fun.is_null()) return -1;
 
   SetGlobalProperty("x", Smi::FromInt(x));
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   Execution::CallScript(isolate, fun, global,
                         isolate->factory()->empty_fixed_array())
       .Check();
-  return Object::Number(*GetGlobalProperty("result"));
+  return Object::NumberValue(*GetGlobalProperty("result"));
 }
 
 TEST_F(CompilerTest, Abs) {
@@ -116,16 +120,16 @@ TEST_F(CompilerTest, Abs) {
 }
 
 static double Sum(Isolate* isolate, int n) {
-  Handle<JSFunction> fun =
+  DirectHandle<JSFunction> fun =
       Compile("s = 0; while (n > 0) { s += n; n -= 1; }; result = s;");
   if (fun.is_null()) return -1;
 
   SetGlobalProperty("n", Smi::FromInt(n));
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   Execution::CallScript(isolate, fun, global,
                         isolate->factory()->empty_fixed_array())
       .Check();
-  return Object::Number(*GetGlobalProperty("result"));
+  return Object::NumberValue(*GetGlobalProperty("result"));
 }
 
 TEST_F(CompilerTest, Sum) {
@@ -143,9 +147,10 @@ TEST_F(CompilerPrintTest, Print) {
   v8::Local<v8::Context> context = v8::Context::New(isolate(), &config);
   v8::Context::Scope context_scope(context);
   const char* source = "for (n = 0; n < 100; ++n) print(n, 1, 2);";
-  Handle<JSFunction> fun = Compile(source);
+  DirectHandle<JSFunction> fun = Compile(source);
   if (fun.is_null()) return;
-  Handle<JSObject> global(i_isolate()->context()->global_object(), i_isolate());
+  DirectHandle<JSObject> global(i_isolate()->context()->global_object(),
+                                i_isolate());
   Execution::CallScript(i_isolate(), fun, global,
                         i_isolate()->factory()->empty_fixed_array())
       .Check();
@@ -174,27 +179,28 @@ TEST_F(CompilerTest, Stuff) {
       "function Cons2(x, y) { this.sum = x + y; }\n"
       "if (new Cons2(3,4).sum == 7) r+=256;";  // 256
 
-  Handle<JSFunction> fun = Compile(source);
+  DirectHandle<JSFunction> fun = Compile(source);
   EXPECT_TRUE(!fun.is_null());
-  Handle<JSObject> global(i_isolate()->context()->global_object(), i_isolate());
+  DirectHandle<JSObject> global(i_isolate()->context()->global_object(),
+                                i_isolate());
   Execution::CallScript(i_isolate(), fun, global,
                         i_isolate()->factory()->empty_fixed_array())
       .Check();
-  EXPECT_EQ(511.0, Object::Number(*GetGlobalProperty("r")));
+  EXPECT_EQ(511.0, Object::NumberValue(*GetGlobalProperty("r")));
 }
 
 TEST_F(CompilerTest, UncaughtThrow) {
   v8::HandleScope scope(isolate());
 
   const char* source = "throw 42;";
-  Handle<JSFunction> fun = Compile(source);
+  DirectHandle<JSFunction> fun = Compile(source);
   EXPECT_TRUE(!fun.is_null());
-  Isolate* isolate = fun->GetIsolate();
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  Isolate* isolate = i_isolate();
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   EXPECT_TRUE(Execution::CallScript(isolate, fun, global,
                                     isolate->factory()->empty_fixed_array())
                   .is_null());
-  EXPECT_EQ(42.0, Object::Number(isolate->pending_exception()));
+  EXPECT_EQ(42.0, Object::NumberValue(isolate->exception()));
 }
 
 using CompilerC2JSFramesTest = WithPrintExtensionMixin<v8::TestWithIsolate>;
@@ -216,25 +222,24 @@ TEST_F(CompilerC2JSFramesTest, C2JSFrames) {
 
   const char* source = "function foo(a) { gc(), print(a); }";
 
-  Handle<JSFunction> fun0 = Compile(source);
+  DirectHandle<JSFunction> fun0 = Compile(source);
   EXPECT_TRUE(!fun0.is_null());
-  Isolate* isolate = fun0->GetIsolate();
+  Isolate* isolate = i_isolate();
 
   // Run the generated code to populate the global object with 'foo'.
-  Handle<JSObject> global(isolate->context()->global_object(), isolate);
+  DirectHandle<JSObject> global(isolate->context()->global_object(), isolate);
   Execution::CallScript(isolate, fun0, global,
                         isolate->factory()->empty_fixed_array())
       .Check();
 
-  Handle<Object> fun1 =
+  DirectHandle<Object> fun1 =
       JSReceiver::GetProperty(isolate, isolate->global_object(), "foo")
           .ToHandleChecked();
   EXPECT_TRUE(IsJSFunction(*fun1));
 
-  Handle<Object> argv[] = {
+  DirectHandle<Object> args[] = {
       isolate->factory()->InternalizeString(base::StaticCharVector("hello"))};
-  Execution::Call(isolate, Handle<JSFunction>::cast(fun1), global,
-                  arraysize(argv), argv)
+  Execution::Call(isolate, Cast<JSFunction>(fun1), global, base::VectorOf(args))
       .Check();
 }
 
@@ -244,7 +249,7 @@ TEST_F(CompilerTest, Regression236) {
   Factory* factory = i_isolate()->factory();
   v8::HandleScope scope(isolate());
 
-  Handle<Script> script = factory->NewScript(factory->undefined_value());
+  DirectHandle<Script> script = factory->NewScript(factory->undefined_value());
   EXPECT_EQ(-1, Script::GetLineNumber(script, 0));
   EXPECT_EQ(-1, Script::GetLineNumber(script, 100));
   EXPECT_EQ(-1, Script::GetLineNumber(script, -1));
@@ -252,7 +257,7 @@ TEST_F(CompilerTest, Regression236) {
 
 TEST_F(CompilerTest, GetScriptLineNumber) {
   v8::HandleScope scope(isolate());
-  v8::ScriptOrigin origin = v8::ScriptOrigin(isolate(), NewString("test"));
+  v8::ScriptOrigin origin = v8::ScriptOrigin(NewString("test"));
   const char function_f[] = "function f() {}";
   const int max_rows = 1000;
   const int buffer_size = max_rows + sizeof(function_f);
@@ -287,17 +292,16 @@ TEST_F(CompilerTest, FeedbackVectorPreservedAcrossRecompiles) {
       "%PrepareFunctionForOptimization(f);"
       "function f(a) { a(); } f(fun1);");
 
-  Handle<JSFunction> f = Handle<JSFunction>::cast(v8::Utils::OpenHandle(
-      *v8::Local<v8::Function>::Cast(context()
-                                         ->Global()
-                                         ->Get(context(), NewString("f"))
-                                         .ToLocalChecked())));
+  DirectHandle<JSFunction> f = Cast<
+      JSFunction>(v8::Utils::OpenDirectHandle(*v8::Local<v8::Function>::Cast(
+      context()->Global()->Get(context(), NewString("f")).ToLocalChecked())));
 
   // Verify that we gathered feedback.
-  Handle<FeedbackVector> feedback_vector(f->feedback_vector(), f->GetIsolate());
+  DirectHandle<FeedbackVector> feedback_vector(f->feedback_vector(),
+                                               i_isolate());
   EXPECT_TRUE(!feedback_vector->is_empty());
   FeedbackSlot slot_for_a(0);
-  MaybeObject object = feedback_vector->Get(slot_for_a);
+  Tagged<MaybeObject> object = feedback_vector->Get(slot_for_a);
   {
     Tagged<HeapObject> heap_object;
     EXPECT_TRUE(object.GetHeapObjectIfWeak(&heap_object));
@@ -308,7 +312,7 @@ TEST_F(CompilerTest, FeedbackVectorPreservedAcrossRecompiles) {
 
   // Verify that the feedback is still "gathered" despite a recompilation
   // of the full code.
-  EXPECT_TRUE(f->HasAttachedOptimizedCode());
+  EXPECT_TRUE(f->HasAttachedOptimizedCode(i_isolate()));
   object = f->feedback_vector()->Get(slot_for_a);
   {
     Tagged<HeapObject> heap_object;
@@ -337,8 +341,8 @@ TEST_F(CompilerTest, FeedbackVectorUnaffectedByScopeChanges) {
       "}"
       "morphing_call = builder();");
 
-  Handle<JSFunction> f = Handle<JSFunction>::cast(
-      v8::Utils::OpenHandle(*v8::Local<v8::Function>::Cast(
+  DirectHandle<JSFunction> f = Cast<JSFunction>(
+      v8::Utils::OpenDirectHandle(*v8::Local<v8::Function>::Cast(
           context()
               ->Global()
               ->Get(context(), NewString("morphing_call"))
@@ -378,23 +382,23 @@ TEST_F(CompilerTest, OptimizedCodeSharing1) {
         "%DebugPrint(closure0());"
         "closure1();"
         "var closure2 = MakeClosure(); closure2();");
-    Handle<JSFunction> fun1 = Handle<JSFunction>::cast(
-        v8::Utils::OpenHandle(*v8::Local<v8::Function>::Cast(
+    DirectHandle<JSFunction> fun1 = Cast<JSFunction>(
+        v8::Utils::OpenDirectHandle(*v8::Local<v8::Function>::Cast(
             context()
                 ->Global()
                 ->Get(context(), NewString("closure1"))
                 .ToLocalChecked())));
-    Handle<JSFunction> fun2 = Handle<JSFunction>::cast(
-        v8::Utils::OpenHandle(*v8::Local<v8::Function>::Cast(
+    DirectHandle<JSFunction> fun2 = Cast<JSFunction>(
+        v8::Utils::OpenDirectHandle(*v8::Local<v8::Function>::Cast(
             context()
                 ->Global()
                 ->Get(context(), NewString("closure2"))
                 .ToLocalChecked())));
-    EXPECT_TRUE(fun1->HasAttachedOptimizedCode() ||
+    EXPECT_TRUE(fun1->HasAttachedOptimizedCode(i_isolate()) ||
                 !i_isolate()->use_optimizer());
-    EXPECT_TRUE(fun2->HasAttachedOptimizedCode() ||
+    EXPECT_TRUE(fun2->HasAttachedOptimizedCode(i_isolate()) ||
                 !i_isolate()->use_optimizer());
-    EXPECT_EQ(fun1->code(), fun2->code());
+    EXPECT_EQ(fun1->code(i_isolate()), fun2->code(i_isolate()));
   }
 }
 
@@ -605,19 +609,20 @@ TEST_F(CompilerTest, CompileFunctionQuirks) {
 
 TEST_F(CompilerTest, CompileFunctionScriptOrigin) {
   v8::HandleScope scope(isolate());
-  v8::ScriptOrigin origin(isolate(), NewString("test"), 22, 41);
+  v8::ScriptOrigin origin(NewString("test"), 22, 41);
   v8::ScriptCompiler::Source script_source(NewString("throw new Error()"),
                                            origin);
   v8::Local<v8::Function> fun =
       v8::ScriptCompiler::CompileFunction(context(), &script_source)
           .ToLocalChecked();
   EXPECT_TRUE(!fun.IsEmpty());
-  auto fun_i = i::Handle<i::JSFunction>::cast(Utils::OpenHandle(*fun));
+  auto fun_i = i::Cast<i::JSFunction>(Utils::OpenHandle(*fun));
   EXPECT_TRUE(IsSharedFunctionInfo(fun_i->shared()));
-  EXPECT_TRUE(Utils::ToLocal(
-                  i::handle(i::Script::cast(fun_i->shared()->script())->name(),
-                            i_isolate()))
-                  ->StrictEquals(NewString("test")));
+  EXPECT_TRUE(
+      Utils::ToLocal(i::direct_handle(
+                         i::Cast<i::Script>(fun_i->shared()->script())->name(),
+                         i_isolate()))
+          ->StrictEquals(NewString("test")));
   v8::TryCatch try_catch(isolate());
   isolate()->SetCaptureStackTraceForUncaughtExceptions(true);
   EXPECT_TRUE(fun->Call(context(), context()->Global(), 0, nullptr).IsEmpty());
@@ -649,7 +654,7 @@ TEST_F(CompilerTest, CompileFunctionFunctionToString) {
 
     // Regression test for v8:6190
     {
-      v8::ScriptOrigin origin(isolate(), NewString("test"), 22, 41);
+      v8::ScriptOrigin origin(NewString("test"), 22, 41);
       v8::ScriptCompiler::Source script_source(NewString("return event"),
                                                origin);
 
@@ -675,7 +680,7 @@ TEST_F(CompilerTest, CompileFunctionFunctionToString) {
 
     // With no parameters:
     {
-      v8::ScriptOrigin origin(isolate(), NewString("test"), 17, 31);
+      v8::ScriptOrigin origin(NewString("test"), 17, 31);
       v8::ScriptCompiler::Source script_source(NewString("return 0"), origin);
 
       v8::TryCatch try_catch(isolate());
@@ -698,7 +703,7 @@ TEST_F(CompilerTest, CompileFunctionFunctionToString) {
 
     // With a name:
     {
-      v8::ScriptOrigin origin(isolate(), NewString("test"), 17, 31);
+      v8::ScriptOrigin origin(NewString("test"), 17, 31);
       v8::ScriptCompiler::Source script_source(NewString("return 0"), origin);
 
       v8::TryCatch try_catch(isolate());
@@ -737,7 +742,7 @@ TEST_F(CompilerTest, InvocationCount) {
       "function foo() { return bar(); };"
       "%EnsureFeedbackVectorForFunction(foo);"
       "foo();");
-  Handle<JSFunction> foo = Handle<JSFunction>::cast(GetGlobalProperty("foo"));
+  DirectHandle<JSFunction> foo = Cast<JSFunction>(GetGlobalProperty("foo"));
   EXPECT_EQ(1, foo->feedback_vector()->invocation_count());
   RunJS("foo()");
   EXPECT_EQ(2, foo->feedback_vector()->invocation_count());
@@ -824,39 +829,46 @@ TEST_F(CompilerTest, DeepEagerCompilationPeakMemory) {
 
   v8::HeapStatistics heap_statistics;
   isolate()->GetHeapStatistics(&heap_statistics);
-  size_t peak_mem_1 = heap_statistics.peak_malloced_memory();
-  printf("peak memory after init:          %8zu\n", peak_mem_1);
+  size_t peak_mem_after_init = heap_statistics.peak_malloced_memory();
+  printf("peak memory after init:          %8zu\n", peak_mem_after_init);
 
-  v8::ScriptCompiler::Compile(context(), &script_source,
-                              v8::ScriptCompiler::kNoCompileOptions)
-      .ToLocalChecked();
+  // Peak memory during lazy compilation should converge to the same value
+  // (usually after 1-2 iterations).
+  std::vector<size_t> peak_mem_after_lazy_compile;
+  const int kNumLazyCompiles = 5;
+  for (int i = 0; i < kNumLazyCompiles; i++) {
+    v8::ScriptCompiler::Compile(context(), &script_source,
+                                v8::ScriptCompiler::kNoCompileOptions)
+        .ToLocalChecked();
 
-  isolate()->GetHeapStatistics(&heap_statistics);
-  size_t peak_mem_2 = heap_statistics.peak_malloced_memory();
-  printf("peak memory after lazy compile:  %8zu\n", peak_mem_2);
-
-  v8::ScriptCompiler::Compile(context(), &script_source,
-                              v8::ScriptCompiler::kNoCompileOptions)
-      .ToLocalChecked();
-
-  isolate()->GetHeapStatistics(&heap_statistics);
-  size_t peak_mem_3 = heap_statistics.peak_malloced_memory();
-  printf("peak memory after lazy compile:  %8zu\n", peak_mem_3);
+    isolate()->GetHeapStatistics(&heap_statistics);
+    size_t peak_mem = heap_statistics.peak_malloced_memory();
+    printf("peak memory after lazy compile:  %8zu\n", peak_mem);
+    peak_mem_after_lazy_compile.push_back(peak_mem);
+  }
+  size_t peak_mem_after_first_lazy_compile = peak_mem_after_lazy_compile[0];
+  size_t peak_mem_after_second_to_last_lazy_compile =
+      peak_mem_after_lazy_compile[kNumLazyCompiles - 2];
+  size_t peak_mem_after_last_lazy_compile =
+      peak_mem_after_lazy_compile[kNumLazyCompiles - 1];
 
   v8::ScriptCompiler::Compile(context(), &script_source,
                               v8::ScriptCompiler::kEagerCompile)
       .ToLocalChecked();
 
   isolate()->GetHeapStatistics(&heap_statistics);
-  size_t peak_mem_4 = heap_statistics.peak_malloced_memory();
-  printf("peak memory after eager compile: %8zu\n", peak_mem_4);
+  size_t peak_mem_after_eager_compile = heap_statistics.peak_malloced_memory();
+  printf("peak memory after eager compile: %8zu\n",
+         peak_mem_after_eager_compile);
 
-  EXPECT_LE(peak_mem_1, peak_mem_2);
-  EXPECT_EQ(peak_mem_2, peak_mem_3);
-  EXPECT_LE(peak_mem_3, peak_mem_4);
+  EXPECT_LE(peak_mem_after_init, peak_mem_after_first_lazy_compile);
+  EXPECT_EQ(peak_mem_after_second_to_last_lazy_compile,
+            peak_mem_after_last_lazy_compile);
+  EXPECT_LE(peak_mem_after_last_lazy_compile, peak_mem_after_eager_compile);
   // Check that eager compilation does not cause significantly higher (+100%)
   // peak memory than lazy compilation.
-  EXPECT_LE(peak_mem_4 - peak_mem_3, peak_mem_3);
+  EXPECT_LE(peak_mem_after_eager_compile - peak_mem_after_last_lazy_compile,
+            peak_mem_after_last_lazy_compile);
 }
 
 namespace {
@@ -915,15 +927,232 @@ TEST_F(CompilerTest, ProfilerEnabledDuringBackgroundCompile) {
   v8::Local<v8::Script> script =
       v8::ScriptCompiler::Compile(isolate()->GetCurrentContext(),
                                   &streamed_source, NewString(source),
-                                  v8::ScriptOrigin(isolate(), NewString("foo")))
+                                  v8::ScriptOrigin(NewString("foo")))
           .ToLocalChecked();
 
-  i::Handle<i::Object> obj = Utils::OpenHandle(*script);
-  EXPECT_TRUE(i::JSFunction::cast(*obj)->shared()->AreSourcePositionsAvailable(
-      i_isolate()));
+  i::DirectHandle<i::Object> obj = Utils::OpenDirectHandle(*script);
+  EXPECT_TRUE(
+      i::Cast<i::JSFunction>(*obj)->shared()->AreSourcePositionsAvailable(
+          i_isolate()));
 
   cpu_profiler->StopProfiling(profile);
 }
 
+using BackgroundMergeTest = TestWithNativeContext;
+
+// Tests that a GC during merge doesn't break the merge.
+TEST_F(BackgroundMergeTest, GCDuringMerge) {
+  v8_flags.verify_code_merge = true;
+
+  HandleScope handle_scope(isolate());
+  const char* source =
+      // f is compiled eagerly thanks to the IIFE hack.
+      "f = (function f(x) {"
+      "  let b = x;"
+      // f is compiled eagerly, so g's SFI exists. But, it is not compiled.
+      "  return function g() {"
+      // g isn't compiled, so h's SFI does not exist.
+      "    return function h() {"
+      "      return b;"
+      "    }"
+      "  }"
+      "})";
+  IndirectHandle<String> source_string =
+      isolate()
+          ->factory()
+          ->NewStringFromUtf8(base::CStrVector(source))
+          .ToHandleChecked();
+
+  const int kTopLevelId = 0;
+  const int kFId = 1;
+  const int kGId = 2;
+  const int kHId = 3;
+
+  // Compile the script once to warm up the compilation cache.
+  IndirectHandle<JSFunction> old_g;
+  ([&]() V8_NOINLINE {
+    // Compile in a new handle scope inside a non-inlined function, so that the
+    // script can die while select inner functions stay alive.
+    HandleScope scope(isolate());
+    ScriptCompiler::CompilationDetails compilation_details;
+    DirectHandle<SharedFunctionInfo> top_level_sfi =
+        Compiler::GetSharedFunctionInfoForScript(
+            isolate(), source_string, ScriptDetails(),
+            v8::ScriptCompiler::kNoCompileOptions,
+            ScriptCompiler::kNoCacheNoReason, NOT_NATIVES_CODE,
+            &compilation_details)
+            .ToHandleChecked();
+
+    {
+      Tagged<Script> script = Cast<Script>(top_level_sfi->script());
+      CHECK(!script->infos()->get(kTopLevelId).IsCleared());
+      CHECK(!script->infos()->get(kFId).IsCleared());
+      CHECK(!script->infos()->get(kGId).IsCleared());
+      // h in the script infos list was never initialized by the compilation, so
+      // it's the default value for a WeakFixedArray, which is `undefined`.
+      CHECK(Is<Undefined>(script->infos()->get(kHId)));
+    }
+
+    DirectHandle<JSFunction> top_level =
+        Factory::JSFunctionBuilder{isolate(), top_level_sfi,
+                                   isolate()->native_context()}
+            .Build();
+
+    DirectHandle<JSObject> global(isolate()->context()->global_object(),
+                                  isolate());
+    Execution::CallScript(isolate(), top_level, global,
+                          isolate()->factory()->empty_fixed_array())
+        .Check();
+
+    DirectHandle<JSFunction> f = Cast<JSFunction>(
+        JSObject::GetProperty(isolate(), global, "f").ToHandleChecked());
+
+    CHECK(f->is_compiled(isolate()));
+
+    // Execute f to get g's SFI (no g bytecode yet)
+    IndirectHandle<JSFunction> g = Cast<JSFunction>(
+        Execution::Call(isolate(), f, global, {}).ToHandleChecked());
+    CHECK(!g->is_compiled(isolate()));
+
+    // Execute g's SFI to initialize g's bytecode, and to get h.
+    DirectHandle<JSFunction> h = Cast<JSFunction>(
+        Execution::Call(isolate(), g, global, {}).ToHandleChecked());
+    CHECK(g->is_compiled(isolate()));
+    CHECK(!h->is_compiled(isolate()));
+
+    CHECK_EQ(top_level->shared()->function_literal_id(kRelaxedLoad),
+             kTopLevelId);
+    CHECK_EQ(f->shared()->function_literal_id(kRelaxedLoad), kFId);
+    CHECK_EQ(g->shared()->function_literal_id(kRelaxedLoad), kGId);
+    CHECK_EQ(h->shared()->function_literal_id(kRelaxedLoad), kHId);
+
+    // Age everything so that subsequent GCs can pick it up if possible.
+    SharedFunctionInfo::EnsureOldForTesting(top_level->shared());
+    SharedFunctionInfo::EnsureOldForTesting(f->shared());
+    SharedFunctionInfo::EnsureOldForTesting(g->shared());
+    SharedFunctionInfo::EnsureOldForTesting(h->shared());
+
+    old_g = scope.CloseAndEscape(g);
+  })();
+
+  IndirectHandle<Script> old_script(Cast<Script>(old_g->shared()->script()),
+                                    isolate());
+
+  // Make sure bytecode is cleared...
+  for (int i = 0; i < 3; ++i) {
+    // We need to invoke GC without stack, otherwise some objects may not be
+    // reclaimed because of conservative stack scanning.
+    DisableConservativeStackScanningScopeForTesting no_stack_scanning(
+        isolate()->heap());
+    InvokeMajorGC();
+  }
+  CHECK(!old_g->is_compiled(isolate()));
+
+  // The top-level script should now be dead.
+  CHECK(old_script->infos()->get(kTopLevelId).IsCleared());
+  // f should still be alive by global reference.
+  CHECK(!old_script->infos()->get(kFId).IsCleared());
+  // g should be kept alive by our old_g handle.
+  CHECK(!old_script->infos()->get(kGId).IsCleared());
+  // h should be dead since g's bytecode was flushed.
+  CHECK(old_script->infos()->get(kHId).IsCleared());
+
+  // Copy the old_script_infos WeakFixedArray, so that we can inspect it after
+  // the merge mutated the original.
+  DirectHandle<WeakFixedArray> unmutated_old_script_list =
+      isolate()->factory()->CopyWeakFixedArray(
+          direct_handle(old_script->infos(), isolate()));
+
+  {
+    HandleScope scope(isolate());
+    ScriptStreamingData streamed_source(
+        std::make_unique<DummySourceStream>(source),
+        v8::ScriptCompiler::StreamedSource::UTF8);
+    ScriptCompiler::CompilationDetails details;
+    streamed_source.task = std::make_unique<i::BackgroundCompileTask>(
+        &streamed_source, isolate(), ScriptType::kClassic,
+        ScriptCompiler::CompileOptions::kNoCompileOptions, &details);
+
+    streamed_source.task->RunOnMainThread(isolate());
+
+    DirectHandle<SharedFunctionInfo> top_level_sfi;
+    {
+      // Use a manual GC scope, because we want to test a GC in a very precise
+      // spot in the merge.
+      ManualGCScope manual_gc(isolate());
+      // There's one more reference to the old_g -- clear it so that nothing is
+      // keeping it alive
+      CHECK(!old_script->infos()->get(kGId).IsCleared());
+      CHECK(!unmutated_old_script_list->get(kGId).IsCleared());
+      old_g.PatchValue({});
+      CHECK(!old_script->infos()->get(kFId).IsCleared());
+
+      BackgroundMergeTask::ForceGCDuringNextMergeForTesting();
+
+      top_level_sfi = streamed_source.task
+                          ->FinalizeScript(isolate(), source_string,
+                                           ScriptDetails(), old_script)
+                          .ToHandleChecked();
+      CHECK(!old_script->infos()->get(kFId).IsCleared());
+    }
+
+    CHECK_EQ(top_level_sfi->script(), *old_script);
+
+    DirectHandle<JSFunction> top_level =
+        Factory::JSFunctionBuilder{isolate(), top_level_sfi,
+                                   isolate()->native_context()}
+            .Build();
+
+    DirectHandle<JSObject> global(isolate()->context()->global_object(),
+                                  isolate());
+
+    DirectHandle<JSFunction> f = Cast<JSFunction>(
+        JSObject::GetProperty(isolate(), global, "f").ToHandleChecked());
+
+    // f should normally be compiled (with the old shared function info but the
+    // new bytecode). However, the extra GCs in finalization might cause it to
+    // be flushed, so we can't guarantee this check.
+    // CHECK(f->is_compiled(isolate()));
+
+    // Execute f to get g's SFI (no g bytecode yet)
+    DirectHandle<JSFunction> g = Cast<JSFunction>(
+        Execution::Call(isolate(), f, global, {}).ToHandleChecked());
+    CHECK(!g->is_compiled(isolate()));
+
+    // Execute g's SFI to initialize g's bytecode, and to get h.
+    DirectHandle<JSFunction> h = Cast<JSFunction>(
+        Execution::Call(isolate(), g, global, {}).ToHandleChecked());
+    CHECK(g->is_compiled(isolate()));
+    CHECK(!h->is_compiled(isolate()));
+
+    CHECK_EQ(top_level->shared()->function_literal_id(kRelaxedLoad),
+             kTopLevelId);
+    CHECK_EQ(f->shared()->function_literal_id(kRelaxedLoad), kFId);
+    CHECK_EQ(g->shared()->function_literal_id(kRelaxedLoad), kGId);
+    CHECK_EQ(h->shared()->function_literal_id(kRelaxedLoad), kHId);
+
+    CHECK_EQ(top_level->shared()->script(), *old_script);
+    CHECK_EQ(f->shared()->script(), *old_script);
+    CHECK_EQ(g->shared()->script(), *old_script);
+    CHECK_EQ(h->shared()->script(), *old_script);
+
+    CHECK_EQ(MakeWeak(top_level->shared()),
+             old_script->infos()->get(kTopLevelId));
+    CHECK_EQ(MakeWeak(f->shared()), old_script->infos()->get(kFId));
+    CHECK_EQ(MakeWeak(g->shared()), old_script->infos()->get(kGId));
+    CHECK_EQ(MakeWeak(h->shared()), old_script->infos()->get(kHId));
+
+    // The old top-level died, so we have a new one.
+    CHECK_NE(MakeWeak(top_level->shared()),
+             unmutated_old_script_list->get(kTopLevelId));
+    // The old f was still alive, so it's the same.
+    CHECK_EQ(MakeWeak(f->shared()), unmutated_old_script_list->get(kFId));
+    // The old g was still alive, so it's the same.
+    CHECK_EQ(MakeWeak(g->shared()), unmutated_old_script_list->get(kGId));
+    // The old h died, so it's different.
+    CHECK_NE(MakeWeak(h->shared()), unmutated_old_script_list->get(kHId));
+  }
+}
+
 }  // namespace internal
 }  // namespace v8
