@@ -3,12 +3,11 @@
 // found in the LICENSE file.
 
 #include "src/compiler/scheduler.h"
+
 #include "src/codegen/tick-counter.h"
 #include "src/compiler/access-builder.h"
 #include "src/compiler/common-operator.h"
 #include "src/compiler/compiler-source-position-table.h"
-#include "src/compiler/graph-visualizer.h"
-#include "src/compiler/graph.h"
 #include "src/compiler/js-operator.h"
 #include "src/compiler/node-origin-table.h"
 #include "src/compiler/node.h"
@@ -16,6 +15,8 @@
 #include "src/compiler/operator.h"
 #include "src/compiler/schedule.h"
 #include "src/compiler/simplified-operator.h"
+#include "src/compiler/turbofan-graph-visualizer.h"
+#include "src/compiler/turbofan-graph.h"
 #include "src/compiler/verifier.h"
 #include "test/unittests/compiler/compiler-test-utils.h"
 #include "test/unittests/test-utils.h"
@@ -30,11 +31,7 @@ namespace compiler {
 class SchedulerTest : public TestWithIsolateAndZone {
  public:
   SchedulerTest()
-      : TestWithIsolateAndZone(kCompressGraphZone),
-        graph_(zone()),
-        common_(zone()),
-        simplified_(zone()),
-        js_(zone()) {}
+      : graph_(zone()), common_(zone()), simplified_(zone()), js_(zone()) {}
 
   Schedule* ComputeAndVerifySchedule(size_t expected) {
     if (v8_flags.trace_turbo) {
@@ -63,7 +60,7 @@ class SchedulerTest : public TestWithIsolateAndZone {
     return node_count;
   }
 
-  Graph* graph() { return &graph_; }
+  TFGraph* graph() { return &graph_; }
   CommonOperatorBuilder* common() { return &common_; }
   SimplifiedOperatorBuilder* simplified() { return &simplified_; }
   JSOperatorBuilder* js() { return &js_; }
@@ -71,7 +68,7 @@ class SchedulerTest : public TestWithIsolateAndZone {
 
  private:
   TickCounter tick_counter_;
-  Graph graph_;
+  TFGraph graph_;
   CommonOperatorBuilder common_;
   SimplifiedOperatorBuilder simplified_;
   JSOperatorBuilder js_;
@@ -117,7 +114,7 @@ TEST_F(SchedulerTest, BuildScheduleOneParameter) {
 
 namespace {
 
-Node* CreateDiamond(Graph* graph, CommonOperatorBuilder* common, Node* cond) {
+Node* CreateDiamond(TFGraph* graph, CommonOperatorBuilder* common, Node* cond) {
   Node* tv = graph->NewNode(common->Int32Constant(6));
   Node* fv = graph->NewNode(common->Int32Constant(7));
   Node* br = graph->NewNode(common->Branch(), cond, graph->start());
@@ -164,7 +161,7 @@ TARGET_TEST_F(SchedulerTest, FloatingDeadDiamond1) {
 }
 
 TARGET_TEST_F(SchedulerTest, FloatingDeadDiamond2) {
-  Graph* g = graph();
+  TFGraph* g = graph();
   Node* start = g->NewNode(common()->Start(1));
   g->SetStart(start);
 
