@@ -183,11 +183,7 @@ class TyperTest : public TypedGraphTest {
           for (int x1 = lmin; x1 < lmin + width; x1++) {
             for (int x2 = rmin; x2 < rmin + width; x2++) {
               double result_value = opfun(x1, x2);
-              Type result_type = Type::Constant(
-                  broker(),
-                  CanonicalHandle(
-                      isolate()->factory()->NewNumber(result_value)),
-                  zone());
+              Type result_type = Type::Constant(result_value, zone());
               EXPECT_TRUE(result_type.Is(expected_type));
             }
           }
@@ -207,25 +203,18 @@ class TyperTest : public TypedGraphTest {
         double x1 = RandomInt(r1.AsRange());
         double x2 = RandomInt(r2.AsRange());
         double result_value = opfun(x1, x2);
-        Type result_type = Type::Constant(
-            broker(),
-            CanonicalHandle(isolate()->factory()->NewNumber(result_value)),
-            zone());
+        Type result_type = Type::Constant(result_value, zone());
         EXPECT_TRUE(result_type.Is(expected_type));
       }
     }
     // Test extreme cases.
     double x1 = +1e-308;
     double x2 = -1e-308;
-    Type r1 = Type::Constant(
-        broker(), CanonicalHandle(isolate()->factory()->NewNumber(x1)), zone());
-    Type r2 = Type::Constant(
-        broker(), CanonicalHandle(isolate()->factory()->NewNumber(x2)), zone());
+    Type r1 = Type::Constant(x1, zone());
+    Type r2 = Type::Constant(x2, zone());
     Type expected_type = TypeBinaryOp(op, r1, r2);
     double result_value = opfun(x1, x2);
-    Type result_type = Type::Constant(
-        broker(),
-        CanonicalHandle(isolate()->factory()->NewNumber(result_value)), zone());
+    Type result_type = Type::Constant(result_value, zone());
     EXPECT_TRUE(result_type.Is(expected_type));
   }
 
@@ -258,10 +247,7 @@ class TyperTest : public TypedGraphTest {
         int32_t x1 = static_cast<int32_t>(RandomInt(r1.AsRange()));
         int32_t x2 = static_cast<int32_t>(RandomInt(r2.AsRange()));
         double result_value = opfun(x1, x2);
-        Type result_type = Type::Constant(
-            broker(),
-            CanonicalHandle(isolate()->factory()->NewNumber(result_value)),
-            zone());
+        Type result_type = Type::Constant(result_value, zone());
         EXPECT_TRUE(result_type.Is(expected_type));
       }
     }
