@@ -6,10 +6,12 @@
 
 #include <cmath>
 #include <limits>
+#include <optional>
 
 #include "src/base/platform/platform.h"
 #include "src/common/globals.h"
 #include "src/execution/isolate.h"
+#include "src/flags/flags.h"
 #include "src/heap/gc-tracer-inl.h"
 #include "test/unittests/test-utils.h"
 #include "testing/gtest/include/gtest/gtest.h"
@@ -36,7 +38,7 @@ enum class StartTracingMode {
 
 void StartTracing(GCTracer* tracer, GarbageCollector collector,
                   StartTracingMode mode,
-                  base::Optional<base::TimeTicks> time = {}) {
+                  std::optional<base::TimeTicks> time = {}) {
   DCHECK_IMPLIES(mode != StartTracingMode::kAtomic,
                  !Heap::IsYoungGenerationCollector(collector));
   // Start the cycle for incremental marking.
@@ -64,60 +66,35 @@ void StartTracing(GCTracer* tracer, GarbageCollector collector,
   }
 }
 
-void StopTracing(GCTracer* tracer, GarbageCollector collector,
-                 base::Optional<base::TimeTicks> time = {}) {
+void StopTracing(Heap* heap, GarbageCollector collector,
+                 std::optional<base::TimeTicks> time = {}) {
+  GCTracer* tracer = heap->tracer();
   tracer->StopAtomicPause();
   tracer->StopObservablePause(collector, time.value_or(base::TimeTicks::Now()));
   switch (collector) {
     case GarbageCollector::SCAVENGER:
-      tracer->StopYoungCycleIfNeeded();
-      break;
     case GarbageCollector::MINOR_MARK_SWEEPER:
-      tracer->NotifyYoungSweepingCompleted();
+      tracer->NotifyYoungSweepingCompletedAndStopCycleIfFinished();
       break;
     case GarbageCollector::MARK_COMPACTOR:
-      tracer->NotifyFullSweepingCompleted();
+      if (heap->cpp_heap()) {
+        using namespace cppgc::internal;
+        StatsCollector* stats_collector =
+            CppHeap::From(heap->cpp_heap())->stats_collector();
+        stats_collector->NotifyMarkingStarted(
+            CollectionType::kMajor, cppgc::Heap::MarkingType::kAtomic,
+            MarkingConfig::IsForcedGC::kNotForced);
+        stats_collector->NotifyMarkingCompleted(0);
+        stats_collector->NotifySweepingCompleted(
+            cppgc::Heap::SweepingType::kAtomic);
+      }
+      tracer->NotifyFullSweepingCompletedAndStopCycleIfFinished();
       break;
   }
 }
 
 }  // namespace
 
-TEST_F(GCTracerTest, AllocationThroughput) {
-  if (v8_flags.stress_incremental_marking) return;
-  // GCTracer::AllocationThroughputInBytesPerMillisecond ignores global memory.
-  GCTracer* tracer = i_isolate()->heap()->tracer();
-  tracer->ResetForTesting();
-
-  const int time1 = 100;
-  const size_t counter1 = 1000;
-  SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time1),
-                   counter1);
-  const int time2 = 200;
-  const size_t counter2 = 2000;
-  SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time2),
-                   counter2);
-  // Will only consider the current sample.
-  EXPECT_EQ(
-      2 * (counter2 - counter1) / (time2 - time1),
-      static_cast<size_t>(tracer->AllocationThroughputInBytesPerMillisecond(
-          base::TimeDelta::FromMilliseconds(100))));
-  const int time3 = 1000;
-  const size_t counter3 = 30000;
-  SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time3),
-                   counter3);
-  // Only consider last sample.
-  EXPECT_EQ(
-      2 * (counter3 - counter2) / (time3 - time2),
-      static_cast<size_t>(tracer->AllocationThroughputInBytesPerMillisecond(
-          base::TimeDelta::FromMilliseconds(800))));
-  // Considers last 2 samples.
-  EXPECT_EQ(
-      2 * (counter3 - counter1) / (time3 - time1),
-      static_cast<size_t>(tracer->AllocationThroughputInBytesPerMillisecond(
-          base::TimeDelta::FromMilliseconds(801))));
-}
-
 TEST_F(GCTracerTest, PerGenerationAllocationThroughput) {
   if (v8_flags.stress_incremental_marking) return;
   GCTracer* tracer = i_isolate()->heap()->tracer();
@@ -132,7 +109,7 @@ TEST_F(GCTracerTest, PerGenerationAllocationThroughput) {
   const size_t counter2 = 2000;
   SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time2),
                    counter2);
-  const size_t expected_throughput1 = (counter2 - counter1) / (200 - 100);
+  const size_t expected_throughput1 = counter2 / time2 * (1.0 - exp2(-2.0));
   EXPECT_EQ(expected_throughput1,
             static_cast<size_t>(
                 tracer->NewSpaceAllocationThroughputInBytesPerMillisecond()));
@@ -147,7 +124,11 @@ TEST_F(GCTracerTest, PerGenerationAllocationThroughput) {
   const size_t counter3 = 30000;
   SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time3),
                    counter3);
-  const size_t expected_throughput2 = counter3 / time3;
+  const size_t expected_throughput2 =
+      (counter3 - counter2) / (time3 - time2) * (1.0 - exp2(-8.0)) +
+      exp2(-8.0) * expected_throughput1;
+  EXPECT_GE(expected_throughput2, expected_throughput1);
+  EXPECT_LE(expected_throughput2, (counter3 - counter2) / (time3 - time2));
   EXPECT_EQ(expected_throughput2,
             static_cast<size_t>(
                 tracer->NewSpaceAllocationThroughputInBytesPerMillisecond()));
@@ -160,54 +141,6 @@ TEST_F(GCTracerTest, PerGenerationAllocationThroughput) {
                 tracer->EmbedderAllocationThroughputInBytesPerMillisecond()));
 }
 
-TEST_F(GCTracerTest, PerGenerationAllocationThroughputWithProvidedTime) {
-  if (v8_flags.stress_incremental_marking) return;
-  GCTracer* tracer = i_isolate()->heap()->tracer();
-  tracer->ResetForTesting();
-
-  const int time1 = 100;
-  const size_t counter1 = 1000;
-  SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time1),
-                   counter1);
-  const int time2 = 200;
-  const size_t counter2 = 2000;
-  SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time2),
-                   counter2);
-  const size_t expected_throughput1 = (counter2 - counter1) / (time2 - time1);
-  EXPECT_EQ(expected_throughput1,
-            static_cast<size_t>(
-                tracer->NewSpaceAllocationThroughputInBytesPerMillisecond(
-                    base::TimeDelta::FromMilliseconds(100))));
-  EXPECT_EQ(expected_throughput1,
-            static_cast<size_t>(
-                tracer->OldGenerationAllocationThroughputInBytesPerMillisecond(
-                    base::TimeDelta::FromMilliseconds(100))));
-  const int time3 = 1000;
-  const size_t counter3 = 30000;
-  SampleAllocation(tracer, base::TimeTicks::FromMsTicksForTesting(time3),
-                   counter3);
-  const size_t expected_throughput2 = (counter3 - counter2) / (time3 - time2);
-  // Only consider last sample.
-  EXPECT_EQ(expected_throughput2,
-            static_cast<size_t>(
-                tracer->NewSpaceAllocationThroughputInBytesPerMillisecond(
-                    base::TimeDelta::FromMilliseconds(800))));
-  EXPECT_EQ(expected_throughput2,
-            static_cast<size_t>(
-                tracer->OldGenerationAllocationThroughputInBytesPerMillisecond(
-                    base::TimeDelta::FromMilliseconds(800))));
-  const size_t expected_throughput3 = (counter3 - counter1) / (time3 - time1);
-  // Consider last two samples.
-  EXPECT_EQ(expected_throughput3,
-            static_cast<size_t>(
-                tracer->NewSpaceAllocationThroughputInBytesPerMillisecond(
-                    base::TimeDelta::FromMilliseconds(801))));
-  EXPECT_EQ(expected_throughput3,
-            static_cast<size_t>(
-                tracer->OldGenerationAllocationThroughputInBytesPerMillisecond(
-                    base::TimeDelta::FromMilliseconds(801))));
-}
-
 TEST_F(GCTracerTest, RegularScope) {
   if (v8_flags.stress_incremental_marking) return;
   GCTracer* tracer = i_isolate()->heap()->tracer();
@@ -222,7 +155,7 @@ TEST_F(GCTracerTest, RegularScope) {
                StartTracingMode::kAtomic);
   tracer->AddScopeSample(GCTracer::Scope::MC_MARK,
                          base::TimeDelta::FromMilliseconds(100));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR);
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(100),
             tracer->current_.scopes[GCTracer::Scope::MC_MARK]);
 }
@@ -233,17 +166,17 @@ TEST_F(GCTracerTest, IncrementalScope) {
   tracer->ResetForTesting();
 
   EXPECT_EQ(base::TimeDelta(),
-            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]);
+            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL]);
   // Sample is added because its ScopeId is listed as incremental sample.
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(100));
   StartTracing(tracer, GarbageCollector::MARK_COMPACTOR,
                StartTracingMode::kIncremental);
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(100));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR);
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(200),
-            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]);
+            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL]);
 }
 
 TEST_F(GCTracerTest, IncrementalMarkingDetails) {
@@ -252,53 +185,49 @@ TEST_F(GCTracerTest, IncrementalMarkingDetails) {
   tracer->ResetForTesting();
 
   // Round 1.
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(50));
   // Scavenger has no impact on incremental marking details.
   StartTracing(tracer, GarbageCollector::SCAVENGER, StartTracingMode::kAtomic);
-  StopTracing(tracer, GarbageCollector::SCAVENGER);
+  StopTracing(i_isolate()->heap(), GarbageCollector::SCAVENGER);
   StartTracing(tracer, GarbageCollector::MARK_COMPACTOR,
                StartTracingMode::kIncremental);
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(100));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR);
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(100),
-            tracer->current_
-                .incremental_scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]
+            tracer->current_.incremental_scopes[GCTracer::Scope::MC_INCREMENTAL]
                 .longest_step);
-  EXPECT_EQ(2, tracer->current_
-                   .incremental_scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]
-                   .steps);
+  EXPECT_EQ(2,
+            tracer->current_.incremental_scopes[GCTracer::Scope::MC_INCREMENTAL]
+                .steps);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(150),
-            tracer->current_
-                .incremental_scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]
+            tracer->current_.incremental_scopes[GCTracer::Scope::MC_INCREMENTAL]
                 .duration);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(150),
-            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]);
+            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL]);
 
   // Round 2. Numbers should be reset.
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(13));
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(15));
   StartTracing(tracer, GarbageCollector::MARK_COMPACTOR,
                StartTracingMode::kIncremental);
-  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL_FINALIZE,
+  tracer->AddScopeSample(GCTracer::Scope::MC_INCREMENTAL,
                          base::TimeDelta::FromMilliseconds(122));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR);
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(122),
-            tracer->current_
-                .incremental_scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]
+            tracer->current_.incremental_scopes[GCTracer::Scope::MC_INCREMENTAL]
                 .longest_step);
-  EXPECT_EQ(3, tracer->current_
-                   .incremental_scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]
-                   .steps);
+  EXPECT_EQ(3,
+            tracer->current_.incremental_scopes[GCTracer::Scope::MC_INCREMENTAL]
+                .steps);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(150),
-            tracer->current_
-                .incremental_scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]
+            tracer->current_.incremental_scopes[GCTracer::Scope::MC_INCREMENTAL]
                 .duration);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(150),
-            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL_FINALIZE]);
+            tracer->current_.scopes[GCTracer::Scope::MC_INCREMENTAL]);
 }
 
 TEST_F(GCTracerTest, IncrementalMarkingSpeed) {
@@ -319,31 +248,26 @@ TEST_F(GCTracerTest, IncrementalMarkingSpeed) {
   tracer->AddIncrementalMarkingStep(100, 1000000);
   EXPECT_EQ(1000000 / 100,
             tracer->IncrementalMarkingSpeedInBytesPerMillisecond());
-  // Scavenger has no impact on incremental marking details.
-  StartTracing(tracer, GarbageCollector::SCAVENGER, StartTracingMode::kAtomic);
-  StopTracing(tracer, GarbageCollector::SCAVENGER);
   // 1000000 bytes in 100ms.
   tracer->AddIncrementalMarkingStep(100, 1000000);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(300),
-            tracer->incremental_marking_duration_);
-  EXPECT_EQ(3000000u, tracer->incremental_marking_bytes_);
+            tracer->current_.incremental_marking_duration);
+  EXPECT_EQ(3000000u, tracer->current_.incremental_marking_bytes);
   EXPECT_EQ(1000000 / 100,
             tracer->IncrementalMarkingSpeedInBytesPerMillisecond());
   // 1000000 bytes in 100ms.
   tracer->AddIncrementalMarkingStep(100, 1000000);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(400),
-            tracer->incremental_marking_duration_);
-  EXPECT_EQ(4000000u, tracer->incremental_marking_bytes_);
+            tracer->current_.incremental_marking_duration);
+  EXPECT_EQ(4000000u, tracer->current_.incremental_marking_bytes);
   StartTracing(tracer, GarbageCollector::MARK_COMPACTOR,
                StartTracingMode::kIncrementalEnterPause,
                base::TimeTicks::FromMsTicksForTesting(500));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR,
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR,
               base::TimeTicks::FromMsTicksForTesting(600));
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(400),
             tracer->current_.incremental_marking_duration);
   EXPECT_EQ(4000000u, tracer->current_.incremental_marking_bytes);
-  EXPECT_TRUE(tracer->incremental_marking_duration_.IsZero());
-  EXPECT_EQ(0u, tracer->incremental_marking_bytes_);
   EXPECT_EQ(1000000 / 100,
             tracer->IncrementalMarkingSpeedInBytesPerMillisecond());
 
@@ -355,7 +279,7 @@ TEST_F(GCTracerTest, IncrementalMarkingSpeed) {
   StartTracing(tracer, GarbageCollector::MARK_COMPACTOR,
                StartTracingMode::kIncrementalEnterPause,
                base::TimeTicks::FromMsTicksForTesting(3000));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR,
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR,
               base::TimeTicks::FromMsTicksForTesting(3100));
   EXPECT_DOUBLE_EQ((4000000.0 / 400 + 1000.0 / 2000) / 2,
                    static_cast<double>(
@@ -414,7 +338,7 @@ TEST_F(GCTracerTest, BackgroundScavengerScope) {
   tracer->AddScopeSample(
       GCTracer::Scope::SCAVENGER_BACKGROUND_SCAVENGE_PARALLEL,
       base::TimeDelta::FromMilliseconds(1));
-  StopTracing(tracer, GarbageCollector::SCAVENGER);
+  StopTracing(i_isolate()->heap(), GarbageCollector::SCAVENGER);
   EXPECT_EQ(
       base::TimeDelta::FromMilliseconds(11),
       tracer->current_
@@ -431,7 +355,7 @@ TEST_F(GCTracerTest, BackgroundMinorMSScope) {
                          base::TimeDelta::FromMilliseconds(10));
   tracer->AddScopeSample(GCTracer::Scope::MINOR_MS_BACKGROUND_MARKING,
                          base::TimeDelta::FromMilliseconds(1));
-  StopTracing(tracer, GarbageCollector::MINOR_MARK_SWEEPER);
+  StopTracing(i_isolate()->heap(), GarbageCollector::MINOR_MARK_SWEEPER);
   EXPECT_EQ(
       base::TimeDelta::FromMilliseconds(11),
       tracer->current_.scopes[GCTracer::Scope::MINOR_MS_BACKGROUND_MARKING]);
@@ -449,9 +373,6 @@ TEST_F(GCTracerTest, BackgroundMajorMCScope) {
                          base::TimeDelta::FromMilliseconds(200));
   tracer->AddScopeSample(GCTracer::Scope::MC_BACKGROUND_MARKING,
                          base::TimeDelta::FromMilliseconds(10));
-  // Scavenger should not affect the major mark-compact scopes.
-  StartTracing(tracer, GarbageCollector::SCAVENGER, StartTracingMode::kAtomic);
-  StopTracing(tracer, GarbageCollector::SCAVENGER);
   tracer->AddScopeSample(GCTracer::Scope::MC_BACKGROUND_SWEEPING,
                          base::TimeDelta::FromMilliseconds(20));
   tracer->AddScopeSample(GCTracer::Scope::MC_BACKGROUND_MARKING,
@@ -470,7 +391,7 @@ TEST_F(GCTracerTest, BackgroundMajorMCScope) {
   tracer->AddScopeSample(
       GCTracer::Scope::MC_BACKGROUND_EVACUATE_UPDATE_POINTERS,
       base::TimeDelta::FromMilliseconds(4));
-  StopTracing(tracer, GarbageCollector::MARK_COMPACTOR);
+  StopTracing(i_isolate()->heap(), GarbageCollector::MARK_COMPACTOR);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(111),
             tracer->current_.scopes[GCTracer::Scope::MC_BACKGROUND_MARKING]);
   EXPECT_EQ(base::TimeDelta::FromMilliseconds(222),
@@ -599,4 +520,29 @@ TEST_F(GCTracerTest, RecordScavengerHistograms) {
   GcHistogram::CleanUp();
 }
 
+TEST_F(GCTracerTest, CyclePriorities) {
+  using Priority = v8::Isolate::Priority;
+  if (v8_flags.stress_incremental_marking) return;
+  GCTracer* tracer = i_isolate()->heap()->tracer();
+  CHECK_EQ(i_isolate()->priority(), Priority::kUserBlocking);
+  tracer->ResetForTesting();
+  EXPECT_TRUE(tracer->current_.priority.has_value());
+  EXPECT_EQ(tracer->current_.priority, Priority::kUserBlocking);
+  // Setting the same priority again doesn't change the cycle priority.
+  i_isolate()->SetPriority(Priority::kUserBlocking);
+  EXPECT_TRUE(tracer->current_.priority.has_value());
+  EXPECT_EQ(tracer->current_.priority, Priority::kUserBlocking);
+  // Setting a different priority resets the cycle priority.
+  i_isolate()->SetPriority(Priority::kUserVisible);
+  EXPECT_FALSE(tracer->current_.priority.has_value());
+  tracer->ResetForTesting();
+  // Initial cycle priority is the same as the isolate priority.
+  EXPECT_TRUE(tracer->current_.priority.has_value());
+  EXPECT_EQ(tracer->current_.priority, Priority::kUserVisible);
+  // Undoing a priority change doesn't restore a cycle priority.
+  i_isolate()->SetPriority(Priority::kUserBlocking);
+  i_isolate()->SetPriority(Priority::kUserVisible);
+  EXPECT_FALSE(tracer->current_.priority.has_value());
+}
+
 }  // namespace v8::internal
