@@ -2,7 +2,8 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-#include "src/heap/conservative-stack-visitor.h"
+#include "src/flags/flags.h"
+#include "src/heap/conservative-stack-visitor-inl.h"
 #include "src/heap/gc-tracer.h"
 #include "test/unittests/heap/heap-utils.h"
 #include "test/unittests/test-utils.h"
@@ -18,9 +19,10 @@ template <typename TMixin>
 class WithInnerPointerResolutionMixin : public TMixin {
  public:
   Address ResolveInnerPointer(Address maybe_inner_ptr) {
-    return ConservativeStackVisitor::ForTesting(
-               this->isolate(), GarbageCollector::MARK_COMPACTOR)
-        .FindBasePtr(maybe_inner_ptr);
+    // This can only resolve inner pointers in the regular cage.
+    PtrComprCageBase cage_base{this->isolate()};
+    return ConservativeStackVisitor(this->isolate(), nullptr)
+        .FindBasePtr(maybe_inner_ptr, cage_base);
   }
 };
 
@@ -56,8 +58,8 @@ class InnerPointerResolutionTest
   InnerPointerResolutionTest& operator=(const InnerPointerResolutionTest&) =
       delete;
 
-  Heap* heap() { return isolate()->heap(); }
-  MemoryAllocator* allocator() { return heap()->memory_allocator(); }
+  Heap* heap() const { return isolate()->heap(); }
+  MemoryAllocator* allocator() const { return heap()->memory_allocator(); }
 
   // Create, free and lookup pages, normal or large.
 
@@ -76,8 +78,8 @@ class InnerPointerResolutionTest
   int CreateLargePage(size_t size) {
     OldLargeObjectSpace* lo_space = heap()->lo_space();
     EXPECT_NE(nullptr, lo_space);
-    LargePage* page =
-        allocator()->AllocateLargePage(lo_space, size, NOT_EXECUTABLE);
+    LargePageMetadata* page = allocator()->AllocateLargePage(
+        lo_space, size, NOT_EXECUTABLE, AllocationHint());
     EXPECT_NE(nullptr, page);
     int page_id = next_page_id_++;
     DCHECK_EQ(pages_.end(), pages_.find(page_id));
@@ -93,7 +95,7 @@ class InnerPointerResolutionTest
     pages_.erase(it);
   }
 
-  MemoryChunk* LookupPage(int page_id) {
+  MutablePageMetadata* LookupPage(int page_id) {
     DCHECK_LE(0, page_id);
     auto it = pages_.find(page_id);
     DCHECK_NE(pages_.end(), it);
@@ -108,7 +110,7 @@ class InnerPointerResolutionTest
   // Creates a list of objects in a page and ensures that the page is iterable.
   int CreateObjectsInPage(const std::vector<ObjectRequest>& objects) {
     int page_id = CreateNormalPage();
-    MemoryChunk* page = LookupPage(page_id);
+    MutablePageMetadata* page = LookupPage(page_id);
     Address ptr = page->area_start();
     for (auto object : objects) {
       DCHECK_NE(ObjectRequest::LARGE, object.type);
@@ -179,7 +181,7 @@ class InnerPointerResolutionTest
     for (auto object : objects) {
       DCHECK_EQ(ObjectRequest::LARGE, object.type);
       int page_id = CreateLargePage(object.size);
-      MemoryChunk* page = LookupPage(page_id);
+      MutablePageMetadata* page = LookupPage(page_id);
       object.page_id = page_id;
       object.address = page->area_start();
       CHECK_EQ(object.address + object.size, page->area_end());
@@ -200,9 +202,10 @@ class InnerPointerResolutionTest
         DCHECK_LE(2 * kTaggedSize, object.size);
         ReadOnlyRoots roots(heap());
         Tagged<HeapObject> heap_object(HeapObject::FromAddress(object.address));
-        heap_object->set_map_after_allocation(roots.unchecked_fixed_array_map(),
+        heap_object->set_map_after_allocation(heap()->isolate(),
+                                              roots.unchecked_fixed_array_map(),
                                               SKIP_WRITE_BARRIER);
-        Tagged<FixedArray> arr(FixedArray::cast(heap_object));
+        Tagged<FixedArray> arr(Cast<FixedArray>(heap_object));
         arr->set_length((object.size - FixedArray::SizeFor(0)) / kTaggedSize);
         DCHECK_EQ(object.size, arr->AllocatedSize());
         break;
@@ -221,7 +224,7 @@ class InnerPointerResolutionTest
             HeapObject::FromAddress(object.address));
         break;
       case ObjectRequest::MARKED_AREA: {
-        MemoryChunk* page = LookupPage(object.page_id);
+        MutablePageMetadata* page = LookupPage(object.page_id);
         page->marking_bitmap()->SetRange<AccessMode::NON_ATOMIC>(
             MarkingBitmap::AddressToIndex(object.address),
             MarkingBitmap::LimitAddressToIndex(object.address + object.size));
@@ -257,17 +260,24 @@ class InnerPointerResolutionTest
       RunTestInside(object, object.size - 1);
     }
     for (auto [id, page] : pages_) {
-      const Address outside_ptr = page->area_start() - 42;
-      DCHECK_LE(page->address(), outside_ptr);
+      const Address outside_ptr = page->area_start() - 3;
+      DCHECK_LE(page->ChunkAddress(), outside_ptr);
       RunTestOutside(outside_ptr);
     }
     RunTestOutside(kNullAddress);
     RunTestOutside(static_cast<Address>(42));
-    RunTestOutside(static_cast<Address>(kZapValue));
+    if (!IsZapPageAllocated()) {
+      RunTestOutside(static_cast<Address>(kZapValue));
+    }
+  }
+
+  bool IsZapPageAllocated() const {
+    return allocator()->LookupChunkContainingAddress(
+               static_cast<Address>(kZapValue)) != nullptr;
   }
 
  private:
-  std::map<int, MemoryChunk*> pages_;
+  std::map<int, MutablePageMetadata*> pages_;
   int next_page_id_ = 0;
   std::vector<ObjectRequest> objects_;
 };
@@ -275,7 +285,6 @@ class InnerPointerResolutionTest
 }  // namespace
 
 TEST_F(InnerPointerResolutionTest, EmptyPage) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({});
   TestAll();
 }
@@ -283,7 +292,6 @@ TEST_F(InnerPointerResolutionTest, EmptyPage) {
 // Tests with some objects laid out randomly.
 
 TEST_F(InnerPointerResolutionTest, NothingMarked) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize},
       {12 * kTaggedSize},
@@ -299,7 +307,6 @@ TEST_F(InnerPointerResolutionTest, NothingMarked) {
 }
 
 TEST_F(InnerPointerResolutionTest, AllMarked) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED},
       {12 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED},
@@ -315,7 +322,6 @@ TEST_F(InnerPointerResolutionTest, AllMarked) {
 }
 
 TEST_F(InnerPointerResolutionTest, SomeMarked) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED},
       {12 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED},
@@ -331,7 +337,6 @@ TEST_F(InnerPointerResolutionTest, SomeMarked) {
 }
 
 TEST_F(InnerPointerResolutionTest, MarkedAreas) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED},
       {12 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA},
@@ -349,7 +354,6 @@ TEST_F(InnerPointerResolutionTest, MarkedAreas) {
 // Tests with specific object layout, to cover interesting and corner cases.
 
 TEST_F(InnerPointerResolutionTest, ThreeMarkedObjectsInSameCell) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       // Some initial large unmarked object, followed by a small marked object
       // towards the end of the cell.
@@ -369,7 +373,6 @@ TEST_F(InnerPointerResolutionTest, ThreeMarkedObjectsInSameCell) {
 }
 
 TEST_F(InnerPointerResolutionTest, ThreeMarkedAreasInSameCell) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       // Some initial large unmarked object, followed by a small marked area
       // towards the end of the cell.
@@ -390,7 +393,6 @@ TEST_F(InnerPointerResolutionTest, ThreeMarkedAreasInSameCell) {
 }
 
 TEST_F(InnerPointerResolutionTest, SmallMarkedAreaAtPageStart) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED, 30,
        ObjectRequest::PAD_MARKED},
@@ -400,7 +402,6 @@ TEST_F(InnerPointerResolutionTest, SmallMarkedAreaAtPageStart) {
 
 TEST_F(InnerPointerResolutionTest,
        SmallMarkedAreaAtPageStartUntilCellBoundary) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {2 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA},
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED, 0,
@@ -410,7 +411,6 @@ TEST_F(InnerPointerResolutionTest,
 }
 
 TEST_F(InnerPointerResolutionTest, LargeMarkedAreaAtPageStart) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {42 * FullCell, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA},
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED, 30,
@@ -421,7 +421,6 @@ TEST_F(InnerPointerResolutionTest, LargeMarkedAreaAtPageStart) {
 
 TEST_F(InnerPointerResolutionTest,
        LargeMarkedAreaAtPageStartUntilCellBoundary) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {42 * FullCell, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA},
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED, 0,
@@ -431,7 +430,6 @@ TEST_F(InnerPointerResolutionTest,
 }
 
 TEST_F(InnerPointerResolutionTest, SmallMarkedAreaStartingAtCellBoundary) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {5 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA, 0,
@@ -441,7 +439,6 @@ TEST_F(InnerPointerResolutionTest, SmallMarkedAreaStartingAtCellBoundary) {
 }
 
 TEST_F(InnerPointerResolutionTest, LargeMarkedAreaStartingAtCellBoundary) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {42 * FullCell + 16 * kTaggedSize, ObjectRequest::REGULAR,
@@ -451,7 +448,6 @@ TEST_F(InnerPointerResolutionTest, LargeMarkedAreaStartingAtCellBoundary) {
 }
 
 TEST_F(InnerPointerResolutionTest, SmallMarkedAreaEndingAtCellBoundary) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {2 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA, 13,
@@ -463,7 +459,6 @@ TEST_F(InnerPointerResolutionTest, SmallMarkedAreaEndingAtCellBoundary) {
 }
 
 TEST_F(InnerPointerResolutionTest, LargeMarkedAreaEndingAtCellBoundary) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {42 * FullCell + 16 * kTaggedSize, ObjectRequest::REGULAR,
@@ -475,7 +470,6 @@ TEST_F(InnerPointerResolutionTest, LargeMarkedAreaEndingAtCellBoundary) {
 }
 
 TEST_F(InnerPointerResolutionTest, TwoSmallMarkedAreasAtCellBoundaries) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {6 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA, 0,
@@ -489,7 +483,6 @@ TEST_F(InnerPointerResolutionTest, TwoSmallMarkedAreasAtCellBoundaries) {
 }
 
 TEST_F(InnerPointerResolutionTest, MarkedAreaOfOneCell) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {1 * FullCell, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA, 0,
@@ -499,7 +492,6 @@ TEST_F(InnerPointerResolutionTest, MarkedAreaOfOneCell) {
 }
 
 TEST_F(InnerPointerResolutionTest, MarkedAreaOfManyCells) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {128 * kTaggedSize},
       {17 * FullCell, ObjectRequest::REGULAR, ObjectRequest::MARKED_AREA, 0,
@@ -511,7 +503,6 @@ TEST_F(InnerPointerResolutionTest, MarkedAreaOfManyCells) {
 // Test with more pages, normal and large.
 
 TEST_F(InnerPointerResolutionTest, TwoPages) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED},
       {13 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED},
@@ -533,7 +524,6 @@ TEST_F(InnerPointerResolutionTest, TwoPages) {
 }
 
 TEST_F(InnerPointerResolutionTest, OneLargePage) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateLargeObjects({
       {1 * MB, ObjectRequest::LARGE, ObjectRequest::UNMARKED},
   });
@@ -541,7 +531,6 @@ TEST_F(InnerPointerResolutionTest, OneLargePage) {
 }
 
 TEST_F(InnerPointerResolutionTest, SeveralLargePages) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateLargeObjects({
       {1 * MB, ObjectRequest::LARGE, ObjectRequest::UNMARKED},
       {32 * MB, ObjectRequest::LARGE, ObjectRequest::MARKED},
@@ -550,7 +539,6 @@ TEST_F(InnerPointerResolutionTest, SeveralLargePages) {
 }
 
 TEST_F(InnerPointerResolutionTest, PagesOfBothKind) {
-  if (v8_flags.enable_third_party_heap) return;
   CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED},
       {13 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED},
@@ -576,7 +564,6 @@ TEST_F(InnerPointerResolutionTest, PagesOfBothKind) {
 }
 
 TEST_F(InnerPointerResolutionTest, FreePages) {
-  if (v8_flags.enable_third_party_heap) return;
   int some_normal_page = CreateObjectsInPage({
       {16 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::UNMARKED},
       {13 * kTaggedSize, ObjectRequest::REGULAR, ObjectRequest::MARKED},
@@ -609,18 +596,24 @@ using InnerPointerResolutionHeapTest =
     WithInnerPointerResolutionMixin<TestWithHeapInternalsAndContext>;
 
 TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
+  if (v8_flags.single_generation) return;
+
+  // Use predictable mode to prevent shrinking new space and releasing unused
+  // pages, which this test expects will remain allocated.
+  v8_flags.predictable = true;
+
   ManualGCScope manual_gc_scope(isolate());
   DisableConservativeStackScanningScopeForTesting no_stack_scanning(heap());
 
   Persistent<v8::FixedArray> weak1, weak2, strong;
   Address inner_ptr1, inner_ptr2, inner_ptr3, outside_ptr1, outside_ptr2;
-  Page *page1, *page2;
+  MemoryChunk *page1, *page2;
 
   auto allocator = heap()->memory_allocator();
 
   {
     PtrComprCageBase cage_base{isolate()};
-    HandleScope scope(isolate());
+    HandleScope handle_scope(isolate());
 
     // Allocate two objects, large enough that they fall in two different young
     // generation pages. Keep weak references to these objects.
@@ -636,10 +629,10 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
     weak2.SetWeak();
     auto obj1 = *h1;
     auto obj2 = *h2;
-    page1 = Page::FromHeapObject(obj1);
+    page1 = MemoryChunk::FromHeapObject(obj1);
     EXPECT_TRUE(!page1->IsLargePage());
     EXPECT_TRUE(v8_flags.minor_ms || page1->IsToPage());
-    page2 = Page::FromHeapObject(obj2);
+    page2 = MemoryChunk::FromHeapObject(obj2);
     EXPECT_TRUE(!page2->IsLargePage());
     EXPECT_TRUE(v8_flags.minor_ms || page2->IsToPage());
     EXPECT_NE(page1, page2);
@@ -649,12 +642,12 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
     auto h3 = factory()->NewFixedArray(16, AllocationType::kYoung);
     strong.Reset(v8_isolate(), Utils::FixedArrayToLocal(h3));
     auto obj3 = *h3;
-    auto page3 = Page::FromHeapObject(obj3);
+    auto page3 = MemoryChunk::FromHeapObject(obj3);
     EXPECT_TRUE(page3 == page1 || page3 == page2);
     if (page3 == page1) {
-      EXPECT_EQ(obj3.address(), obj1.address() + obj1->Size(cage_base));
+      EXPECT_EQ(obj3.address(), obj1.address() + obj1->Size());
     } else {
-      EXPECT_EQ(obj3.address(), obj2.address() + obj2->Size(cage_base));
+      EXPECT_EQ(obj3.address(), obj2.address() + obj2->Size());
     }
 
     // Keep inner pointers to all objects.
@@ -663,18 +656,21 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
     inner_ptr3 = obj3.address() + 7 * kTaggedSize;
 
     // Keep pointers to the end of the pages, after the objects.
-    outside_ptr1 = page1->area_end() - 3 * kTaggedSize;
-    outside_ptr2 = page2->area_end() - 2 * kTaggedSize;
-    EXPECT_LE(obj1.address() + obj1->Size(cage_base), outside_ptr1);
-    EXPECT_LE(obj2.address() + obj2->Size(cage_base), outside_ptr2);
+    outside_ptr1 = page1->Metadata()->area_end() - 3 * kTaggedSize;
+    outside_ptr2 = page2->Metadata()->area_end() - 2 * kTaggedSize;
+    EXPECT_LE(obj1.address() + obj1->Size(), outside_ptr1);
+    EXPECT_LE(obj2.address() + obj2->Size(), outside_ptr2);
     if (page3 == page1) {
-      EXPECT_LE(obj3.address() + obj3->Size(cage_base), outside_ptr1);
+      EXPECT_LE(obj3.address() + obj3->Size(), outside_ptr1);
     } else {
-      EXPECT_LE(obj3.address() + obj3->Size(cage_base), outside_ptr2);
+      EXPECT_LE(obj3.address() + obj3->Size(), outside_ptr2);
     }
 
     // Ensure the young generation space is iterable.
-    heap()->new_space()->main_allocator()->MakeLinearAllocationAreaIterable();
+    heap()
+        ->allocator()
+        ->new_space_allocator()
+        ->MakeLinearAllocationAreaIterable();
 
     // Inner pointer resolution should work now, finding the objects in the
     // case of the inner pointers.
@@ -707,8 +703,10 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
   // The two pages should still be around, in the new space.
   EXPECT_EQ(page1, allocator->LookupChunkContainingAddress(inner_ptr1));
   EXPECT_EQ(page2, allocator->LookupChunkContainingAddress(inner_ptr2));
-  EXPECT_EQ(AllocationSpace::NEW_SPACE, page1->owner_identity());
-  EXPECT_EQ(AllocationSpace::NEW_SPACE, page2->owner_identity());
+  EXPECT_EQ(AllocationSpace::NEW_SPACE,
+            MutablePageMetadata::cast(page1->Metadata())->owner_identity());
+  EXPECT_EQ(AllocationSpace::NEW_SPACE,
+            MutablePageMetadata::cast(page2->Metadata())->owner_identity());
   EXPECT_TRUE(v8_flags.minor_ms || page1->IsFromPage());
   EXPECT_TRUE(v8_flags.minor_ms || page2->IsFromPage());
 
@@ -723,8 +721,10 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
 
   // Garbage collection once more.
   InvokeAtomicMinorGC();
-  EXPECT_EQ(AllocationSpace::NEW_SPACE, page1->owner_identity());
-  EXPECT_EQ(AllocationSpace::NEW_SPACE, page2->owner_identity());
+  EXPECT_EQ(AllocationSpace::NEW_SPACE,
+            MutablePageMetadata::cast(page1->Metadata())->owner_identity());
+  EXPECT_EQ(AllocationSpace::NEW_SPACE,
+            MutablePageMetadata::cast(page2->Metadata())->owner_identity());
   // The two pages should still be around, in the new space.
   EXPECT_EQ(page1, allocator->LookupChunkContainingAddress(inner_ptr1));
   EXPECT_EQ(page2, allocator->LookupChunkContainingAddress(inner_ptr2));
@@ -742,6 +742,8 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedRegularYoungPages) {
 }
 
 TEST_F(InnerPointerResolutionHeapTest, UnusedLargeYoungPage) {
+  if (v8_flags.single_generation) return;
+
   ManualGCScope manual_gc_scope(isolate());
   DisableConservativeStackScanningScopeForTesting no_stack_scanning(heap());
 
@@ -761,9 +763,10 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedLargeYoungPage) {
     weak.Reset(v8_isolate(), Utils::FixedArrayToLocal(h));
     weak.SetWeak();
     auto obj = *h;
-    auto page = Page::FromHeapObject(obj);
+    auto page = MemoryChunk::FromHeapObject(obj);
     EXPECT_TRUE(page->IsLargePage());
-    EXPECT_EQ(AllocationSpace::NEW_LO_SPACE, page->owner_identity());
+    EXPECT_EQ(AllocationSpace::NEW_LO_SPACE,
+              MutablePageMetadata::cast(page->Metadata())->owner_identity());
     EXPECT_TRUE(v8_flags.minor_ms || page->IsToPage());
 
     // Keep inner pointer.
@@ -782,35 +785,6 @@ TEST_F(InnerPointerResolutionHeapTest, UnusedLargeYoungPage) {
   EXPECT_EQ(kNullAddress, ResolveInnerPointer(inner_ptr));
 }
 
-TEST_F(InnerPointerResolutionHeapTest, RegularPageAfterEnd) {
-  auto allocator = heap()->memory_allocator();
-
-  // Allocate a regular page.
-  OldSpace* old_space = heap()->old_space();
-  DCHECK_NE(nullptr, old_space);
-  auto* page = allocator->AllocatePage(
-      MemoryAllocator::AllocationMode::kRegular, old_space, NOT_EXECUTABLE);
-  EXPECT_NE(nullptr, page);
-
-  // The end of the page area is expected not to coincide with the beginning of
-  // the next page.
-  const int size = (1 << kPageSizeBits) / 2;
-  const Address mark = page->area_start() + size;
-  heap()->CreateFillerObjectAt(page->area_start(), size);
-  heap()->CreateFillerObjectAt(mark, static_cast<int>(page->area_end() - mark));
-  Page::UpdateHighWaterMark(mark);
-  page->ShrinkToHighWaterMark();
-  EXPECT_FALSE(Page::IsAlignedToPageSize(page->area_end()));
-
-  // Inner pointer resolution after the end of the page area should work.
-  Address inner_ptr = page->area_end() + kTaggedSize;
-  EXPECT_FALSE(Page::IsAlignedToPageSize(inner_ptr));
-  EXPECT_EQ(kNullAddress, ResolveInnerPointer(inner_ptr));
-
-  // Deallocate the page.
-  allocator->Free(MemoryAllocator::FreeMode::kImmediately, page);
-}
-
 TEST_F(InnerPointerResolutionHeapTest, LargePageAfterEnd) {
   auto allocator = heap()->memory_allocator();
 
@@ -818,17 +792,17 @@ TEST_F(InnerPointerResolutionHeapTest, LargePageAfterEnd) {
   OldLargeObjectSpace* lo_space = heap()->lo_space();
   EXPECT_NE(nullptr, lo_space);
   const int size = 3 * (1 << kPageSizeBits) / 2;
-  LargePage* page =
-      allocator->AllocateLargePage(lo_space, size, NOT_EXECUTABLE);
+  LargePageMetadata* page = allocator->AllocateLargePage(
+      lo_space, size, NOT_EXECUTABLE, AllocationHint());
   EXPECT_NE(nullptr, page);
 
   // The end of the page area is expected not to coincide with the beginning of
   // the next page.
-  EXPECT_FALSE(Page::IsAlignedToPageSize(page->area_end()));
+  EXPECT_FALSE(PageMetadata::IsAlignedToPageSize(page->area_end()));
 
   // Inner pointer resolution after the end of the pare area should work.
   Address inner_ptr = page->area_end() + kTaggedSize;
-  EXPECT_FALSE(Page::IsAlignedToPageSize(inner_ptr));
+  EXPECT_FALSE(PageMetadata::IsAlignedToPageSize(inner_ptr));
   EXPECT_EQ(kNullAddress, ResolveInnerPointer(inner_ptr));
 
   // Deallocate the page.
