@@ -2,16 +2,21 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-#include "src/base/optional.h"
+#include <optional>
+
 #include "src/base/platform/platform.h"
 #include "src/base/platform/semaphore.h"
 #include "src/heap/heap.h"
 #include "src/heap/parked-scope-inl.h"
+#include "src/objects/bytecode-array.h"
+#include "src/objects/fixed-array.h"
 #include "test/unittests/heap/heap-utils.h"
 #include "test/unittests/test-utils.h"
 #include "testing/gtest/include/gtest/gtest.h"
 
-#if V8_CAN_CREATE_SHARED_HEAP_BOOL
+// In multi-cage mode we create one cage per isolate
+// and we don't share objects between cages.
+#if V8_CAN_CREATE_SHARED_HEAP_BOOL && !COMPRESS_POINTERS_IN_MULTIPLE_CAGES_BOOL
 
 namespace v8 {
 namespace internal {
@@ -35,17 +40,18 @@ class SharedHeapNoClientsTest : public TestJSSharedMemoryWithPlatform {
 
  private:
   Isolate* shared_space_isolate_;
-  base::Optional<IsolateWrapper> shared_space_isolate_wrapper;
+  std::optional<IsolateWrapper> shared_space_isolate_wrapper;
 };
 
 namespace {
-const int kNumIterations = 2000;
+const int kDefaultNumIterations = 2000;
 
 template <typename Callback>
 void SetupClientIsolateAndRunCallback(Callback callback) {
-  IsolateWrapper isolate_wrapper(kNoCounters);
+  IsolateWrapper isolate_wrapper(kNoCounters, false);
   v8::Isolate* client_isolate = isolate_wrapper.isolate();
   Isolate* i_client_isolate = reinterpret_cast<Isolate*>(client_isolate);
+  v8::Isolate::Scope isolate_scope(client_isolate);
 
   callback(client_isolate, i_client_isolate);
 }
@@ -61,7 +67,7 @@ class SharedOldSpaceAllocationThread final : public ParkingThread {
         [](v8::Isolate* client_isolate, Isolate* i_client_isolate) {
           HandleScope scope(i_client_isolate);
 
-          for (int i = 0; i < kNumIterations; i++) {
+          for (int i = 0; i < kDefaultNumIterations; i++) {
             i_client_isolate->factory()->NewFixedArray(
                 10, AllocationType::kSharedOld);
           }
@@ -76,7 +82,7 @@ class SharedOldSpaceAllocationThread final : public ParkingThread {
 }  // namespace
 
 TEST_F(SharedHeapTest, ConcurrentAllocationInSharedOldSpace) {
-  i_isolate()->main_thread_local_isolate()->BlockMainThreadWhileParked(
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
       [](const ParkedScope& parked) {
         std::vector<std::unique_ptr<SharedOldSpaceAllocationThread>> threads;
         const int kThreads = 4;
@@ -91,6 +97,51 @@ TEST_F(SharedHeapTest, ConcurrentAllocationInSharedOldSpace) {
       });
 }
 
+namespace {
+class SharedTrustedSpaceAllocationThread final : public ParkingThread {
+ public:
+  SharedTrustedSpaceAllocationThread()
+      : ParkingThread(
+            base::Thread::Options("SharedTrustedSpaceAllocationThread")) {}
+
+  void Run() override {
+    constexpr int kNumIterations = 2000;
+
+    SetupClientIsolateAndRunCallback(
+        [](v8::Isolate* client_isolate, Isolate* i_client_isolate) {
+          HandleScope scope(i_client_isolate);
+
+          for (int i = 0; i < kNumIterations; i++) {
+            i_client_isolate->factory()->NewTrustedByteArray(
+                10, AllocationType::kSharedTrusted);
+          }
+
+          InvokeMajorGC(i_client_isolate);
+
+          v8::platform::PumpMessageLoop(i::V8::GetCurrentPlatform(),
+                                        client_isolate);
+        });
+  }
+};
+}  // namespace
+
+TEST_F(SharedHeapTest, ConcurrentAllocationInSharedTrustedSpace) {
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
+      [](const ParkedScope& parked) {
+        std::vector<std::unique_ptr<SharedTrustedSpaceAllocationThread>>
+            threads;
+        const int kThreads = 4;
+
+        for (int i = 0; i < kThreads; i++) {
+          auto thread = std::make_unique<SharedTrustedSpaceAllocationThread>();
+          CHECK(thread->Start());
+          threads.push_back(std::move(thread));
+        }
+
+        ParkingThread::ParkedJoinAll(parked, threads);
+      });
+}
+
 namespace {
 class SharedLargeOldSpaceAllocationThread final : public ParkingThread {
  public:
@@ -105,8 +156,8 @@ class SharedLargeOldSpaceAllocationThread final : public ParkingThread {
           const int kNumIterations = 50;
 
           for (int i = 0; i < kNumIterations; i++) {
-            HandleScope scope(i_client_isolate);
-            Handle<FixedArray> fixed_array =
+            HandleScope inner_scope(i_client_isolate);
+            DirectHandle<FixedArray> fixed_array =
                 i_client_isolate->factory()->NewFixedArray(
                     kMaxRegularHeapObjectSize / kTaggedSize,
                     AllocationType::kSharedOld);
@@ -123,7 +174,7 @@ class SharedLargeOldSpaceAllocationThread final : public ParkingThread {
 }  // namespace
 
 TEST_F(SharedHeapTest, ConcurrentAllocationInSharedLargeOldSpace) {
-  i_isolate()->main_thread_local_isolate()->BlockMainThreadWhileParked(
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
       [](const ParkedScope& parked) {
         std::vector<std::unique_ptr<SharedLargeOldSpaceAllocationThread>>
             threads;
@@ -139,6 +190,176 @@ TEST_F(SharedHeapTest, ConcurrentAllocationInSharedLargeOldSpace) {
       });
 }
 
+namespace {
+class SharedTrustedLargeObjectSpaceAllocationThread final
+    : public ParkingThread {
+ public:
+  SharedTrustedLargeObjectSpaceAllocationThread()
+      : ParkingThread(base::Thread::Options(
+            "SharedTrustedLargeObjectSpaceAllocationThread")) {}
+
+  void Run() override {
+    SetupClientIsolateAndRunCallback(
+        [](v8::Isolate* client_isolate, Isolate* i_client_isolate) {
+          HandleScope scope(i_client_isolate);
+          constexpr int kNumIterations = 50;
+
+          for (int i = 0; i < kNumIterations; i++) {
+            HandleScope inner_scope(i_client_isolate);
+            DirectHandle<TrustedByteArray> fixed_array =
+                i_client_isolate->factory()->NewTrustedByteArray(
+                    kMaxRegularHeapObjectSize, AllocationType::kSharedTrusted);
+            CHECK(MemoryChunk::FromHeapObject(*fixed_array)->IsLargePage());
+          }
+
+          InvokeMajorGC(i_client_isolate);
+
+          v8::platform::PumpMessageLoop(i::V8::GetCurrentPlatform(),
+                                        client_isolate);
+        });
+  }
+};
+}  // namespace
+
+TEST_F(SharedHeapTest, ConcurrentAllocationInSharedTrustedLargeObjectSpace) {
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
+      [](const ParkedScope& parked) {
+        std::vector<
+            std::unique_ptr<SharedTrustedLargeObjectSpaceAllocationThread>>
+            threads;
+        constexpr int kThreads = 4;
+
+        for (int i = 0; i < kThreads; i++) {
+          auto thread =
+              std::make_unique<SharedTrustedLargeObjectSpaceAllocationThread>();
+          CHECK(thread->Start());
+          threads.push_back(std::move(thread));
+        }
+
+        ParkingThread::ParkedJoinAll(parked, threads);
+      });
+}
+
+TEST_F(SharedHeapTest, TrustedToSharedTrustedPointer) {
+  Isolate* isolate = i_isolate();
+  Factory* factory = isolate->factory();
+
+  DirectHandle<TrustedFixedArray> constant_pool =
+      factory->NewTrustedFixedArray(0);
+  DirectHandle<TrustedByteArray> handler_table =
+      factory->NewTrustedByteArray(3, AllocationType::kSharedTrusted);
+  CHECK_EQ(MemoryChunk::FromHeapObject(*handler_table)
+               ->Metadata()
+               ->owner()
+               ->identity(),
+           SHARED_TRUSTED_SPACE);
+
+  // Use random bytes here since we don't ever run the bytecode.
+  constexpr uint8_t kRawBytes[] = {0x1, 0x2, 0x3, 0x4};
+  constexpr int kRawBytesSize = sizeof(kRawBytes);
+  constexpr int32_t kFrameSize = 32;
+  constexpr uint16_t kParameterCount = 2;
+  constexpr uint16_t kMaxArguments = 0;
+
+  Handle<BytecodeArray> bc = factory->NewBytecodeArray(
+      kRawBytesSize, kRawBytes, kFrameSize, kParameterCount, kMaxArguments,
+      constant_pool, handler_table);
+  CHECK_EQ(MemoryChunk::FromHeapObject(*bc)->Metadata()->owner()->identity(),
+           TRUSTED_SPACE);
+
+  InvokeMajorGC(isolate);
+
+  USE(bc);
+}
+
+namespace {
+class TrustedToSharedTrustedPointerOnClient final : public ParkingThread {
+ public:
+  explicit TrustedToSharedTrustedPointerOnClient(ParkingSemaphore* sem_ready,
+                                                 ParkingSemaphore* sema_done)
+      : ParkingThread(
+            base::Thread::Options("TrustedToSharedTrustedPointerOnClient")),
+        sema_ready_(sem_ready),
+        sema_done_(sema_done) {}
+
+  void Run() override {
+    SetupClientIsolateAndRunCallback([this](v8::Isolate* client_isolate,
+                                            Isolate* i_client_isolate) {
+      Factory* factory = i_client_isolate->factory();
+      HandleScope scope(i_client_isolate);
+      DirectHandle<BytecodeArray> keep_alive_bc;
+
+      {
+        HandleScope nested_scope(i_client_isolate);
+        DirectHandle<TrustedFixedArray> constant_pool =
+            factory->NewTrustedFixedArray(0);
+        DirectHandle<TrustedByteArray> handler_table =
+            factory->NewTrustedByteArray(3, AllocationType::kSharedTrusted);
+        CHECK_EQ(MemoryChunk::FromHeapObject(*handler_table)
+                     ->Metadata()
+                     ->owner()
+                     ->identity(),
+                 SHARED_TRUSTED_SPACE);
+
+        // Use random bytes here since we don't ever run the bytecode.
+        constexpr uint8_t kRawBytes[] = {0x1, 0x2, 0x3, 0x4};
+        constexpr int kRawBytesSize = sizeof(kRawBytes);
+        constexpr int32_t kFrameSize = 32;
+        constexpr uint16_t kParameterCount = 2;
+        constexpr uint16_t kMaxArguments = 0;
+
+        Handle<BytecodeArray> bc = factory->NewBytecodeArray(
+            kRawBytesSize, kRawBytes, kFrameSize, kParameterCount,
+            kMaxArguments, constant_pool, handler_table);
+        keep_alive_bc = nested_scope.CloseAndEscape(bc);
+      }
+
+      sema_ready_->Signal();
+      sema_done_->ParkedWait(i_client_isolate->main_thread_local_isolate());
+
+      Tagged<TrustedByteArray> handler_table = keep_alive_bc->handler_table();
+      CHECK(IsTrustedByteArray(handler_table));
+      CHECK_EQ(handler_table->length(), 3);
+
+      v8::platform::PumpMessageLoop(i::V8::GetCurrentPlatform(),
+                                    client_isolate);
+    });
+  }
+
+ private:
+  ParkingSemaphore* sema_ready_;
+  ParkingSemaphore* sema_done_;
+};
+}  // namespace
+
+TEST_F(SharedHeapTest, TrustedToSharedTrustedPointerOnClient) {
+  std::vector<std::unique_ptr<TrustedToSharedTrustedPointerOnClient>> threads;
+  const int kThreads = 4;
+
+  ParkingSemaphore sema_ready(0);
+  ParkingSemaphore sema_done(0);
+
+  for (int i = 0; i < kThreads; i++) {
+    auto thread = std::make_unique<TrustedToSharedTrustedPointerOnClient>(
+        &sema_ready, &sema_done);
+    CHECK(thread->Start());
+    threads.push_back(std::move(thread));
+  }
+
+  LocalIsolate* local_isolate = i_isolate()->main_thread_local_isolate();
+  for (int i = 0; i < kThreads; i++) {
+    sema_ready.ParkedWait(local_isolate);
+  }
+
+  InvokeMajorGC(i_isolate());
+
+  for (int i = 0; i < kThreads; i++) {
+    sema_done.Signal();
+  }
+
+  ParkingThread::ParkedJoinAll(local_isolate, threads);
+}
+
 namespace {
 class SharedMapSpaceAllocationThread final : public ParkingThread {
  public:
@@ -151,8 +372,8 @@ class SharedMapSpaceAllocationThread final : public ParkingThread {
         [](v8::Isolate* client_isolate, Isolate* i_client_isolate) {
           HandleScope scope(i_client_isolate);
 
-          for (int i = 0; i < kNumIterations; i++) {
-            i_client_isolate->factory()->NewMap(
+          for (int i = 0; i < kDefaultNumIterations; i++) {
+            i_client_isolate->factory()->NewContextlessMap(
                 NATIVE_CONTEXT_TYPE, kVariableSizeSentinel,
                 TERMINAL_FAST_ELEMENTS_KIND, 0, AllocationType::kSharedMap);
           }
@@ -167,7 +388,7 @@ class SharedMapSpaceAllocationThread final : public ParkingThread {
 }  // namespace
 
 TEST_F(SharedHeapTest, ConcurrentAllocationInSharedMapSpace) {
-  i_isolate()->main_thread_local_isolate()->BlockMainThreadWhileParked(
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
       [](const ParkedScope& parked) {
         std::vector<std::unique_ptr<SharedMapSpaceAllocationThread>> threads;
         const int kThreads = 4;
@@ -183,6 +404,8 @@ TEST_F(SharedHeapTest, ConcurrentAllocationInSharedMapSpace) {
 }
 
 TEST_F(SharedHeapNoClientsTest, SharedCollectionWithoutClients) {
+  // Set a "current isolate" so we can access pointer tables etc during GC.
+  ::i::SetCurrentIsolateScope isolate_scope{i_shared_space_isolate()};
   ::v8::internal::InvokeMajorGC(i_shared_space_isolate());
 }
 
@@ -193,11 +416,11 @@ void AllocateInSharedHeap(int iterations = 100) {
     std::vector<Handle<FixedArray>> arrays_in_handles;
     const int kKeptAliveInHandle = 1000;
     const int kKeptAliveInHeap = 100;
-    Handle<FixedArray> arrays_in_heap =
+    DirectHandle<FixedArray> arrays_in_heap =
         i_client_isolate->factory()->NewFixedArray(kKeptAliveInHeap,
                                                    AllocationType::kYoung);
 
-    for (int i = 0; i < kNumIterations * iterations; i++) {
+    for (int i = 0; i < kDefaultNumIterations * iterations; i++) {
       HandleScope scope(i_client_isolate);
       Handle<FixedArray> array = i_client_isolate->factory()->NewFixedArray(
           100, AllocationType::kSharedOld);
@@ -215,12 +438,12 @@ void AllocateInSharedHeap(int iterations = 100) {
       i_client_isolate->factory()->NewFixedArray(100, AllocationType::kYoung);
     }
 
-    for (Handle<FixedArray> array : arrays_in_handles) {
+    for (DirectHandle<FixedArray> array : arrays_in_handles) {
       CHECK_EQ(array->length(), 100);
     }
 
     for (int i = 0; i < kKeptAliveInHeap; i++) {
-      Tagged<FixedArray> array = FixedArray::cast(arrays_in_heap->get(i));
+      Tagged<FixedArray> array = Cast<FixedArray>(arrays_in_heap->get(i));
       CHECK_EQ(array->length(), 100);
     }
   });
@@ -228,7 +451,7 @@ void AllocateInSharedHeap(int iterations = 100) {
 
 TEST_F(SharedHeapTest, SharedCollectionWithOneClient) {
   v8_flags.max_old_space_size = 8;
-  i_isolate()->main_thread_local_isolate()->BlockMainThreadWhileParked(
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
       []() { AllocateInSharedHeap(); });
 }
 
@@ -246,7 +469,7 @@ class SharedFixedArrayAllocationThread final : public ParkingThread {
 TEST_F(SharedHeapTest, SharedCollectionWithMultipleClients) {
   v8_flags.max_old_space_size = 8;
 
-  i_isolate()->main_thread_local_isolate()->BlockMainThreadWhileParked(
+  i_isolate()->main_thread_local_isolate()->ExecuteMainThreadWhileParked(
       [](const ParkedScope& parked) {
         std::vector<std::unique_ptr<SharedFixedArrayAllocationThread>> threads;
         const int kThreads = 4;
@@ -316,6 +539,8 @@ class ConcurrentThread final : public ParkingThread {
     IsolateWrapper isolate_wrapper(kNoCounters);
     i_client_isolate_ = isolate_wrapper.i_isolate();
 
+    v8::Isolate::Scope isolate_scope(isolate_wrapper.isolate());
+
     // Allocate the state on the stack, so that handles, direct handles or raw
     // pointers are stack-allocated.
     State state;
@@ -328,7 +553,7 @@ class ConcurrentThread final : public ParkingThread {
       if (wait_while_parked_) {
         // Park and wait.
         i_client_isolate_->main_thread_local_isolate()
-            ->BlockMainThreadWhileParked(
+            ->ExecuteMainThreadWhileParked(
                 [this]() { sema_execute_start_->Wait(); });
       } else {
         // Do not park, but enter a safepoint every now and then.
@@ -462,7 +687,7 @@ namespace {
 // Testing the shared heap using ordinary (indirect) handles.
 
 struct StateWithHandle {
-  base::Optional<HandleScope> scope;
+  std::optional<HandleScope> scope;
   Handle<FixedArray> handle;
   Global<v8::FixedArray> weak;
 };
@@ -536,8 +761,6 @@ TEST_ALL_SCENARIA(SharedHeapTestStateWithHandleParked, ToEachTheirOwn,
 TEST_ALL_SCENARIA(SharedHeapTestStateWithHandleUnparked, ToEachTheirOwn,
                   ToEachTheirOwnWithHandle)
 
-#ifdef V8_ENABLE_CONSERVATIVE_STACK_SCANNING
-
 namespace {
 
 // Testing the shared heap using raw pointers.
@@ -552,9 +775,10 @@ template <AllocationType allocation, AllocationSpace space, int size>
 void AllocateWithRawPointer(Isolate* isolate, StateWithRawPointer* state) {
   // Allocate a fixed array, keep a raw pointer and a weak reference.
   HandleScope scope(isolate);
-  Handle<FixedArray> h = isolate->factory()->NewFixedArray(size, allocation);
+  DirectHandle<FixedArray> h =
+      isolate->factory()->NewFixedArray(size, allocation);
   state->ptr = (*h).ptr();
-  Local<v8::FixedArray> l = Utils::FixedArrayToLocal(h, isolate);
+  Local<v8::FixedArray> l = Utils::FixedArrayToLocal(h);
   state->weak.Reset(reinterpret_cast<v8::Isolate*>(isolate), l);
   state->weak.SetWeak();
 }
@@ -566,6 +790,8 @@ using SharedHeapTestStateWithRawPointerUnparked =
 
 template <typename TestType, AllocationType allocation, AllocationSpace space>
 void ToEachTheirOwnWithRawPointer(TestType* test) {
+  if (!v8_flags.conservative_stack_scanning) return;
+
   using ThreadType = typename TestType::ThreadType;
   ThreadType* thread = test->thread();
 
@@ -613,11 +839,58 @@ TEST_ALL_SCENARIA(SharedHeapTestStateWithRawPointerParked, ToEachTheirOwn,
 TEST_ALL_SCENARIA(SharedHeapTestStateWithRawPointerUnparked, ToEachTheirOwn,
                   ToEachTheirOwnWithRawPointer)
 
-#endif  // V8_ENABLE_CONSERVATIVE_STACK_SCANNING
-
 #undef TEST_SCENARIO
 #undef TEST_ALL_SCENARIA
 
+// TODO(358918874): Re-enable this test once allocation paths are using the
+// right tag for trusted pointers in shared objects.
+#if false
+TEST_F(SharedHeapTest, SharedUntrustedToSharedTrustedPointer) {
+  Isolate* isolate = i_isolate();
+  Factory* factory = isolate->factory();
+  ManualGCScope manual_gc_scope(isolate);
+
+  // Allocate an object in the shared trusted space.
+  // Use random bytes here since we don't ever run the bytecode.
+  constexpr uint8_t kRawBytes[] = {0x1, 0x2, 0x3, 0x4};
+  constexpr int kRawBytesSize = sizeof(kRawBytes);
+  constexpr int32_t kFrameSize = 32;
+  constexpr uint16_t kParameterCount = 2;
+  constexpr uint16_t kMaxArguments = 0;
+  Handle<TrustedFixedArray> constant_pool =
+      factory->NewTrustedFixedArray(0, AllocationType::kSharedTrusted);
+  Handle<TrustedByteArray> handler_table =
+      factory->NewTrustedByteArray(3, AllocationType::kSharedTrusted);
+  Handle<BytecodeArray> bytecode_array = factory->NewBytecodeArray(
+      kRawBytesSize, kRawBytes, kFrameSize, kParameterCount, kMaxArguments,
+      constant_pool, handler_table, AllocationType::kSharedTrusted);
+  CHECK_EQ(MemoryChunk::FromHeapObject(*bytecode_array)
+               ->Metadata()
+               ->owner()
+               ->identity(),
+           SHARED_TRUSTED_SPACE);
+
+  // Start incremental marking
+  isolate->heap()->StartIncrementalMarking(GCFlag::kNoFlags,
+                                           GarbageCollectionReason::kTesting);
+
+  // Allocate an object in the shared untrusted space.
+  Handle<BytecodeWrapper> bytecode_wrapper =
+      factory->NewBytecodeWrapper(AllocationType::kSharedOld);
+  CHECK_EQ(MemoryChunk::FromHeapObject(*bytecode_wrapper)
+               ->Metadata()
+               ->owner()
+               ->identity(),
+           SHARED_SPACE);
+
+  // Create a shared untrusted to shared trusted reference (with a write
+  // barrier)
+  bytecode_wrapper->set_bytecode(*bytecode_array);
+  bytecode_array->wrapper()->clear_bytecode();
+  bytecode_array->set_wrapper(*bytecode_wrapper);
+}
+#endif  // false
+
 }  // namespace internal
 }  // namespace v8
 
