@@ -53,10 +53,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, InvalidBeforeStart) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_1)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_1)
       .LoadNamedProperty(reg_1, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -105,10 +105,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, InvalidAfterEnd) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_1)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_1)
       .LoadNamedProperty(reg_1, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -157,10 +157,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, AccessesFirst) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_1)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_1)
       .LoadNamedProperty(reg_1, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -178,8 +178,9 @@ TEST_F(BytecodeArrayRandomIteratorTest, AccessesFirst) {
   EXPECT_EQ(iterator.current_index(), 0);
   EXPECT_EQ(iterator.current_offset(), 0);
   EXPECT_EQ(iterator.current_operand_scale(), OperandScale::kSingle);
-  EXPECT_EQ(Object::Number(*iterator.GetConstantForIndexOperand(0, isolate())),
-            heap_num_0);
+  EXPECT_EQ(
+      Object::NumberValue(*iterator.GetConstantForIndexOperand(0, isolate())),
+      heap_num_0);
   ASSERT_TRUE(iterator.IsValid());
 }
 
@@ -214,10 +215,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, AccessesLast) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_1)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_1)
       .LoadNamedProperty(reg_1, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -272,10 +273,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, RandomAccessValid) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_16)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_16)
       .LoadNamedProperty(reg_16, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -320,8 +321,9 @@ TEST_F(BytecodeArrayRandomIteratorTest, RandomAccessValid) {
   EXPECT_EQ(iterator.current_index(), 2);
   EXPECT_EQ(iterator.current_offset(), offset);
   EXPECT_EQ(iterator.current_operand_scale(), OperandScale::kSingle);
-  EXPECT_EQ(Object::Number(*iterator.GetConstantForIndexOperand(0, isolate())),
-            heap_num_1);
+  EXPECT_EQ(
+      Object::NumberValue(*iterator.GetConstantForIndexOperand(0, isolate())),
+      heap_num_1);
   ASSERT_TRUE(iterator.IsValid());
 
   iterator.GoToIndex(16);
@@ -450,10 +452,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, IteratesBytecodeArray) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_16)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_16)
       .LoadNamedProperty(reg_16, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -472,8 +474,9 @@ TEST_F(BytecodeArrayRandomIteratorTest, IteratesBytecodeArray) {
   EXPECT_EQ(iterator.current_index(), 0);
   EXPECT_EQ(iterator.current_offset(), offset);
   EXPECT_EQ(iterator.current_operand_scale(), OperandScale::kSingle);
-  EXPECT_EQ(Object::Number(*iterator.GetConstantForIndexOperand(0, isolate())),
-            heap_num_0);
+  EXPECT_EQ(
+      Object::NumberValue(*iterator.GetConstantForIndexOperand(0, isolate())),
+      heap_num_0);
   ASSERT_TRUE(iterator.IsValid());
   offset += Bytecodes::Size(Bytecode::kLdaConstant, OperandScale::kSingle);
   ++iterator;
@@ -490,8 +493,9 @@ TEST_F(BytecodeArrayRandomIteratorTest, IteratesBytecodeArray) {
   EXPECT_EQ(iterator.current_index(), 2);
   EXPECT_EQ(iterator.current_offset(), offset);
   EXPECT_EQ(iterator.current_operand_scale(), OperandScale::kSingle);
-  EXPECT_EQ(Object::Number(*iterator.GetConstantForIndexOperand(0, isolate())),
-            heap_num_1);
+  EXPECT_EQ(
+      Object::NumberValue(*iterator.GetConstantForIndexOperand(0, isolate())),
+      heap_num_1);
   ASSERT_TRUE(iterator.IsValid());
   offset += Bytecodes::Size(Bytecode::kLdaConstant, OperandScale::kSingle);
   ++iterator;
@@ -703,10 +707,10 @@ TEST_F(BytecodeArrayRandomIteratorTest, IteratesBytecodeArrayBackwards) {
       .LoadLiteral(smi_1)
       .StoreAccumulatorInRegister(reg_16)
       .LoadAccumulatorWithRegister(reg_0)
-      .BinaryOperation(Token::Value::ADD, reg_0, 2)
+      .BinaryOperation(Token::kAdd, reg_0, 2)
       .StoreAccumulatorInRegister(reg_16)
       .LoadNamedProperty(reg_16, name, feedback_slot)
-      .BinaryOperation(Token::Value::ADD, reg_0, 3)
+      .BinaryOperation(Token::kAdd, reg_0, 3)
       .StoreAccumulatorInRegister(param)
       .CallRuntimeForPair(Runtime::kLoadLookupSlotForCall, param, pair)
       .ForInPrepare(triple, feedback_slot)
@@ -902,8 +906,9 @@ TEST_F(BytecodeArrayRandomIteratorTest, IteratesBytecodeArrayBackwards) {
   EXPECT_EQ(iterator.current_index(), 2);
   EXPECT_EQ(iterator.current_offset(), offset);
   EXPECT_EQ(iterator.current_operand_scale(), OperandScale::kSingle);
-  EXPECT_EQ(Object::Number(*iterator.GetConstantForIndexOperand(0, isolate())),
-            heap_num_1);
+  EXPECT_EQ(
+      Object::NumberValue(*iterator.GetConstantForIndexOperand(0, isolate())),
+      heap_num_1);
   ASSERT_TRUE(iterator.IsValid());
   --iterator;
 
@@ -920,8 +925,9 @@ TEST_F(BytecodeArrayRandomIteratorTest, IteratesBytecodeArrayBackwards) {
   EXPECT_EQ(iterator.current_index(), 0);
   EXPECT_EQ(iterator.current_offset(), offset);
   EXPECT_EQ(iterator.current_operand_scale(), OperandScale::kSingle);
-  EXPECT_EQ(Object::Number(*iterator.GetConstantForIndexOperand(0, isolate())),
-            heap_num_0);
+  EXPECT_EQ(
+      Object::NumberValue(*iterator.GetConstantForIndexOperand(0, isolate())),
+      heap_num_0);
   ASSERT_TRUE(iterator.IsValid());
   --iterator;
   ASSERT_FALSE(iterator.IsValid());
