@@ -11,6 +11,7 @@
 #include "include/libplatform/libplatform.h"
 #include "include/v8-array-buffer.h"
 #include "include/v8-context.h"
+#include "include/v8-cppgc.h"
 #include "include/v8-extension.h"
 #include "include/v8-local-handle.h"
 #include "include/v8-object.h"
@@ -75,7 +76,10 @@ enum CountersMode { kNoCounters, kEnableCounters };
 // all client Isolates.
 class IsolateWrapper final {
  public:
-  explicit IsolateWrapper(CountersMode counters_mode);
+  // `use_statically_set_cpp_heap` exists to avoid TSAN issues, see the comment
+  // on `cpp_heap_`.
+  explicit IsolateWrapper(CountersMode counters_mode,
+                          bool use_statically_set_cpp_heap = true);
 
   ~IsolateWrapper();
   IsolateWrapper(const IsolateWrapper&) = delete;
@@ -86,16 +90,29 @@ class IsolateWrapper final {
     return reinterpret_cast<i::Isolate*>(isolate_);
   }
 
+  static void set_cpp_heap_for_next_isolate(std::unique_ptr<CppHeap> cpp_heap) {
+    cpp_heap_ = std::move(cpp_heap);
+  }
+
  private:
   std::unique_ptr<v8::ArrayBuffer::Allocator> array_buffer_allocator_;
   std::unique_ptr<CounterMap> counter_map_;
   v8::Isolate* isolate_;
+
+  // `cpp_heap_` is a side channel to pass a custom CppHeap to isolate creation.
+  // Ideally it could be passes as a parameter to the constructor, but with the
+  // MixIn design pattern, that's not possible. Using a static variable as a
+  // side channel causes problems with TSAN, however, when multiple
+  // IsolateWrapper are used at the same time in different threads. The
+  // parameter `use_statically_set_cpp_heap` of the constructor solves the TSAN
+  // issue by providing a way to avoid using `cpp_heap_`.
+  static std::unique_ptr<CppHeap> cpp_heap_;
 };
 
 class IsolateWithContextWrapper final {
  public:
   IsolateWithContextWrapper()
-      : isolate_wrapper_(kNoCounters),
+      : isolate_wrapper_(kNoCounters, false),
         isolate_scope_(isolate_wrapper_.isolate()),
         handle_scope_(isolate_wrapper_.isolate()),
         context_(v8::Context::New(isolate_wrapper_.isolate())),
@@ -154,7 +171,7 @@ class WithIsolateScopeMixin : public TMixin {
     return reinterpret_cast<v8::internal::Isolate*>(this->v8_isolate());
   }
 
-  i::Handle<i::String> MakeName(const char* str, int suffix) {
+  i::DirectHandle<i::String> MakeName(const char* str, int suffix) {
     v8::base::EmbeddedVector<char, 128> buffer;
     v8::base::SNPrintF(buffer, "%s%d", str, suffix);
     return MakeString(buffer.begin());
@@ -202,7 +219,7 @@ class WithIsolateScopeMixin : public TMixin {
                                   Local<String> origin_url,
                                   bool is_shared_cross_origin) {
     Isolate* isolate = Isolate::GetCurrent();
-    ScriptOrigin origin(isolate, origin_url, 0, 0, is_shared_cross_origin);
+    ScriptOrigin origin(origin_url, 0, 0, is_shared_cross_origin);
     ScriptCompiler::Source script_source(source, origin);
     return ScriptCompiler::Compile(isolate->GetCurrentContext(), &script_source)
         .ToLocalChecked();
@@ -257,13 +274,20 @@ class WithIsolateScopeMixin : public TMixin {
 template <typename TMixin>
 class WithContextMixin : public TMixin {
  public:
-  WithContextMixin()
-      : context_(Context::New(this->v8_isolate())), context_scope_(context_) {}
+  WithContextMixin() {
+    v8::Local<v8::Context> context = Context::New(this->v8_isolate());
+    context->Enter();
+    context_.Reset(this->v8_isolate(), context);
+  }
+  ~WithContextMixin() {
+    context_.Get(this->v8_isolate())->Exit();
+    context_.Reset();
+  }
   WithContextMixin(const WithContextMixin&) = delete;
   WithContextMixin& operator=(const WithContextMixin&) = delete;
 
-  const Local<Context>& context() const { return v8_context(); }
-  const Local<Context>& v8_context() const { return context_; }
+  Local<Context> context() const { return v8_context(); }
+  Local<Context> v8_context() const { return context_.Get(this->v8_isolate()); }
 
   void SetGlobalProperty(const char* name, v8::Local<v8::Value> value) {
     CHECK(v8_context()
@@ -273,8 +297,7 @@ class WithContextMixin : public TMixin {
   }
 
  private:
-  v8::Local<v8::Context> context_;
-  v8::Context::Scope context_scope_;
+  v8::Global<v8::Context> context_;
 };
 
 using TestWithPlatform =       //
@@ -382,13 +405,13 @@ class WithInternalIsolateMixin : public TMixin {
   Factory* factory() const { return isolate()->factory(); }
   Isolate* isolate() const { return TMixin::i_isolate(); }
 
-  Handle<NativeContext> native_context() const {
+  DirectHandle<NativeContext> native_context() const {
     return isolate()->native_context();
   }
 
   template <typename T = Object>
   Handle<T> RunJS(const char* source) {
-    return Handle<T>::cast(RunJSInternal(source));
+    return Cast<T>(RunJSInternal(source));
   }
 
   Handle<Object> RunJSInternal(const char* source) {
@@ -396,8 +419,8 @@ class WithInternalIsolateMixin : public TMixin {
   }
 
   template <typename T = Object>
-  Handle<T> RunJS(::v8::String::ExternalOneByteStringResource* source) {
-    return Handle<T>::cast(RunJSInternal(source));
+  DirectHandle<T> RunJS(::v8::String::ExternalOneByteStringResource* source) {
+    return Cast<T>(RunJSInternal(source));
   }
 
   Handle<Object> RunJSInternal(
@@ -413,8 +436,7 @@ class WithInternalIsolateMixin : public TMixin {
 template <typename TMixin>
 class WithZoneMixin : public TMixin {
  public:
-  explicit WithZoneMixin(bool support_zone_compression = false)
-      : zone_(&allocator_, ZONE_NAME, support_zone_compression) {}
+  WithZoneMixin() : zone_(&allocator_, ZONE_NAME) {}
   WithZoneMixin(const WithZoneMixin&) = delete;
   WithZoneMixin& operator=(const WithZoneMixin&) = delete;
 
@@ -529,7 +551,7 @@ class TestTransitionsAccessor : public TransitionsAccessor {
  public:
   TestTransitionsAccessor(Isolate* isolate, Tagged<Map> map)
       : TransitionsAccessor(isolate, map) {}
-  TestTransitionsAccessor(Isolate* isolate, Handle<Map> map)
+  TestTransitionsAccessor(Isolate* isolate, DirectHandle<Map> map)
       : TransitionsAccessor(isolate, *map) {}
 
   // Expose internals for tests.
@@ -555,14 +577,15 @@ class FeedbackVectorHelper {
       : vector_(vector) {
     int slot_count = vector->length();
     slots_.reserve(slot_count);
-    FeedbackMetadataIterator iter(vector->metadata());
+    DisallowGarbageCollection no_gc;
+    FeedbackMetadataIterator iter(vector->metadata(), no_gc);
     while (iter.HasNext()) {
       FeedbackSlot slot = iter.Next();
       slots_.push_back(slot);
     }
   }
 
-  Handle<FeedbackVector> vector() { return vector_; }
+  DirectHandle<FeedbackVector> vector() { return vector_; }
 
   // Returns slot identifier by numerical index.
   FeedbackSlot slot(int index) const { return slots_[index]; }
@@ -589,16 +612,18 @@ class FakeCodeEventLogger : public i::CodeEventLogger {
                      i::Tagged<i::InstructionStream> to) override {}
   void BytecodeMoveEvent(i::Tagged<i::BytecodeArray> from,
                          i::Tagged<i::BytecodeArray> to) override {}
-  void CodeDisableOptEvent(i::Handle<i::AbstractCode> code,
-                           i::Handle<i::SharedFunctionInfo> shared) override {}
+  void CodeDisableOptEvent(
+      i::DirectHandle<i::AbstractCode> code,
+      i::DirectHandle<i::SharedFunctionInfo> shared) override {}
 
  private:
-  void LogRecordedBuffer(i::Tagged<i::AbstractCode> code,
-                         i::MaybeHandle<i::SharedFunctionInfo> maybe_shared,
-                         const char* name, int length) override {}
+  void LogRecordedBuffer(
+      i::Tagged<i::AbstractCode> code,
+      i::MaybeDirectHandle<i::SharedFunctionInfo> maybe_shared,
+      const char* name, size_t length) override {}
 #if V8_ENABLE_WEBASSEMBLY
   void LogRecordedBuffer(const i::wasm::WasmCode* code, const char* name,
-                         int length) override {}
+                         size_t length) override {}
 #endif  // V8_ENABLE_WEBASSEMBLY
 };
 
@@ -622,18 +647,15 @@ class FakeCodeEventLogger : public i::CodeEventLogger {
 #elif V8_HOST_ARCH_MIPS64
 #define GET_STACK_POINTER_TO(sp_addr) \
   __asm__ __volatile__("sd $sp, %0" : "=g"(sp_addr))
+#elif V8_OS_ZOS
+#define GET_STACK_POINTER_TO(sp_addr) \
+  __asm__ __volatile__(" stg 15,%0" : "=m"(sp_addr))
 #elif defined(__s390x__) || defined(_ARCH_S390X)
 #define GET_STACK_POINTER_TO(sp_addr) \
   __asm__ __volatile__("stg %%r15, %0" : "=m"(sp_addr))
-#elif defined(__s390__) || defined(_ARCH_S390)
-#define GET_STACK_POINTER_TO(sp_addr) \
-  __asm__ __volatile__("st 15, %0" : "=m"(sp_addr))
 #elif defined(__PPC64__) || defined(_ARCH_PPC64)
 #define GET_STACK_POINTER_TO(sp_addr) \
   __asm__ __volatile__("std 1, %0" : "=m"(sp_addr))
-#elif defined(__PPC__) || defined(_ARCH_PPC)
-#define GET_STACK_POINTER_TO(sp_addr) \
-  __asm__ __volatile__("stw 1, %0" : "=m"(sp_addr))
 #elif V8_TARGET_ARCH_RISCV64
 #define GET_STACK_POINTER_TO(sp_addr) \
   __asm__ __volatile__("add %0, sp, x0" : "=r"(sp_addr))
