@@ -2,6 +2,8 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
+#include <optional>
+
 #include "src/torque/torque-compiler.h"
 #include "src/torque/utils.h"
 #include "test/unittests/test-utils.h"
@@ -46,7 +48,6 @@ type StrongTagged extends Tagged
 type Smi extends StrongTagged generates 'TNode<Smi>' constexpr 'Smi';
 type WeakHeapObject extends Tagged;
 type Weak<T : type extends HeapObject> extends WeakHeapObject;
-type Uninitialized extends Tagged;
 type TaggedIndex extends StrongTagged;
 type TaggedZeroPattern extends TaggedIndex;
 
@@ -84,8 +85,11 @@ type string constexpr 'const char*';
 type RawPtr generates 'TNode<RawPtrT>' constexpr 'void*';
 type ExternalPointer
     generates 'TNode<ExternalPointerT>' constexpr 'ExternalPointer_t';
-type IndirectPointer
-    generates 'TNode<IndirectPointerHandle>' constexpr 'IndirectPointerHandle';
+type CppHeapPointer
+    generates 'TNode<CppHeapPointerT>' constexpr 'CppHeapPointer_t';
+type TrustedPointer
+    generates 'TNode<TrustedPointerT>' constexpr 'TrustedPointer_t';
+type ProtectedPointer extends Tagged;
 type InstructionStream extends HeapObject generates 'TNode<InstructionStream>';
 type BuiltinPtr extends Smi generates 'TNode<BuiltinPtr>';
 type Context extends HeapObject generates 'TNode<Context>';
@@ -96,7 +100,8 @@ type HeapNumber extends HeapObject;
 type FixedArrayBase extends HeapObject;
 type Lazy<T: type>;
 
-struct float64_or_hole {
+struct float64_or_undefined_or_hole {
+  is_undefined: bool;
   is_hole: bool;
   value: float64;
 }
@@ -199,7 +204,7 @@ void ExpectFailingCompilation(std::string source,
   for (size_t i = 0; i < limit; ++i) {
     EXPECT_THAT(result.messages[i].message, message_patterns[i].first);
     if (message_patterns[i].second != LineAndColumn::Invalid()) {
-      base::Optional<SourcePosition> actual = result.messages[i].position;
+      std::optional<SourcePosition> actual = result.messages[i].position;
       EXPECT_TRUE(actual.has_value());
       EXPECT_EQ(actual->start, message_patterns[i].second);
     }
@@ -625,7 +630,7 @@ TEST(Torque, Enums) {
     extern enum MyEnum {
       kValue0,
       kValue1,
-      kValue2,
+      @sameEnumValueAs(kValue0) kValue2,
       kValue3
     }
   )");
