@@ -20,6 +20,7 @@ with different test suite extensions and build configurations.
 from collections import deque
 from pathlib import Path
 
+import re
 import sys
 import unittest
 from mock import patch
@@ -168,7 +169,7 @@ class StandardRunnerTest(TestRunnerTest):
     # With test processors we don't count reruns as separated failures.
     # TODO(majeski): fix it?
     result.stdout_includes('1 tests failed')
-    result.has_returncode(0)
+    result.has_returncode(1)
 
     # TODO(majeski): Previously we only reported the variant flags in the
     # flags field of the test result.
@@ -191,17 +192,17 @@ class StandardRunnerTest(TestRunnerTest):
           infra_staging=False,
       )
 
-      self.assertEquals(len(records), 3)
-      self.assertEquals(records[0]['testId'], 'sweet/bananaflakes//stress')
-      self.assertEquals(tag_dict(records[0]['tags'])['run'], '1')
+      self.assertEqual(len(records), 3)
+      self.assertEqual(records[0]['testId'], 'sweet/bananaflakes//stress')
+      self.assertEqual(tag_dict(records[0]['tags'])['run'], '1')
       self.assertFalse(records[0]['expected'])
 
-      self.assertEquals(records[1]['testId'], 'sweet/bananaflakes//stress')
-      self.assertEquals(tag_dict(records[1]['tags'])['run'], '2')
+      self.assertEqual(records[1]['testId'], 'sweet/bananaflakes//stress')
+      self.assertEqual(tag_dict(records[1]['tags'])['run'], '2')
       self.assertTrue(records[1]['expected'])
 
-      self.assertEquals(records[2]['testId'], 'sweet/bananaflakes//default')
-      self.assertEquals(tag_dict(records[2]['tags'])['run'], '1')
+      self.assertEqual(records[2]['testId'], 'sweet/bananaflakes//default')
+      self.assertEqual(tag_dict(records[2]['tags'])['run'], '1')
       self.assertTrue(records[2]['expected'])
 
   def testFlakeWithRerunAndJSON(self):
@@ -220,8 +221,14 @@ class StandardRunnerTest(TestRunnerTest):
     result.stdout_includes('=== sweet/bananaflakes (flaky) ===')
     result.stdout_includes('1 tests failed')
     result.stdout_includes('1 tests were flaky')
-    result.has_returncode(0)
+    result.has_returncode(1)
     result.json_content_equals('expected_test_results2.json')
+    self.assertTrue(re.search(
+        r'sweet/bananaflakes default: FAIL \(\d+\.\d+:\d+\.\d+\)',
+        result.test_schedule))
+    self.assertTrue(re.search(
+        r'sweet/bananaflakes default: PASS \(\d+\.\d+:\d+\.\d+\)',
+        result.test_schedule))
 
   def testAutoDetect(self):
     """Fake a build with several auto-detected options.
@@ -240,17 +247,32 @@ class StandardRunnerTest(TestRunnerTest):
             dcheck_always_on=True,
             has_webassembly=True,
             msan=True,
+            sandbox_hardware_support=True,
             target_cpu='x86',
             tsan=True,
             ubsan=True,
             use_sanitizer=True,
             v8_target_cpu='x86',
         ))
-    result.stdout_includes('>>> Autodetected:')
+    result.stdout_includes('>>> Statusfile variables:')
     result.stdout_includes(
-        'arch="ia32", asan, cfi, dcheck_always_on, has_webassembly, i18n, '
-        'msan, target_cpu="x86", tsan, ubsan, use_sanitizer, '
-        'v8_target_cpu="x86"')
+        "DEBUG_defined=False, all_arm64_features=False, "
+        "arch=ia32, asan=True, byteorder=little, "
+        "cfi=True, code_comments=False, component_build=False, "
+        "dcheck_always_on=True, debug_code=False, debugging_features=False, "
+        "deopt_fuzzer=False, device_type=None, "
+        "dict_property_const_tracking=False, disassembler=False, "
+        "endurance_fuzzer=False, full_debug=False, gc_fuzzer=False, "
+        "gc_stress=False, gdbjit=False, has_jitless=False, has_maglev=False, "
+        "has_turbofan=False, has_webassembly=True, i18n=True, "
+        "interrupt_fuzzer=False, is_android=False, is_ios=False, "
+        "isolates=False, lite_mode=False, mode=debug, msan=True, "
+        "no_harness=False, no_simd_hardware=False, novfp3=False, "
+        "optimize_for_size=False, sandbox_hardware_support=True, "
+        "simulator_run=False, slow_dchecks=False, "
+        "system=linux, target_cpu=x86, tsan=True, ubsan=True, "
+        "use_sanitizer=True, v8_target_cpu=x86, verify_heap=False, "
+        "verify_predictable=False")
     result.stdout_includes('>>> Running tests for ia32.release')
     result.has_returncode(0)
     # TODO(machenbach): Test some more implications of the auto-detected
@@ -275,8 +297,8 @@ class StandardRunnerTest(TestRunnerTest):
           '--progress=verbose',
           'sweet',
       )
-      result.stdout_includes('===>Starting stuff\n'
-                             '>>> Running tests for x64.release\n'
+      result.stdout_includes('===>Starting stuff')
+      result.stdout_includes('>>> Running tests for x64.release\n'
                              '>>> Running with test processors\n')
       result.stdout_includes('--- stdout ---\nfake stdout 1')
       result.stdout_includes('--- stderr ---\nfake stderr 1')
@@ -407,6 +429,32 @@ class StandardRunnerTest(TestRunnerTest):
     result.stdout_includes('--random-seed=123')
     result.has_returncode(1)
 
+  def testRandomSeedStressWithNumfuzz(self):
+    """Test using random-seed-stress feature with numfuzz flavor as used by
+    flake bisect for flakes on numfuzz.
+    """
+    result = self.run_tests(
+        '--progress=verbose',
+        '--framework=num_fuzzer',
+        '--variants=default',
+        '--random-seed-stress-count=2',
+        'sweet/bananas',
+        'sweet/apples',
+        infra_staging=False,
+        baseroot='testroot7'
+    )
+
+    # The bananas test is expected to pass when --fuzzing, one of the numfuzz
+    # default flags is present. The apples test is expected to fail with this
+    # flag.
+    result.stdout_includes('sweet/bananas default: PASS')
+    result.stdout_includes('sweet/apples default: FAIL')
+
+    # We get everything twice due to the stress count above set to 2.
+    result.stdout_includes('2 tests failed')
+    result.stdout_includes('4 tests ran')
+    result.has_returncode(1)
+
   def testSpecificVariants(self):
     """Test using NO_VARIANTS modifiers in status files skips the desire tests.
 
@@ -578,7 +626,7 @@ class NumFuzzerTest(TestRunnerTest):
   def testNumFuzzer(self):
     fuzz_flags = [
       f'{flag}=1' for flag in self.get_runner_options()
-      if flag.startswith('--stress-')
+      if flag.startswith('--stress-') or flag.startswith('--allocation')
     ]
     self.assertEqual(len(fuzz_flags), len(fuzzer.FUZZERS))
     for fuzz_flag in fuzz_flags:
@@ -601,9 +649,48 @@ class NumFuzzerTest(TestRunnerTest):
               'sweet/bananas',
             )
             result.has_returncode(0)
-            result.stdout_includes('>>> Autodetected')
+            result.stdout_includes('>>> Statusfile variables:')
             result.stdout_includes('11 tests ran')
 
+  def _run_test_with_random_skip(self, prob):
+    """Run a test root that marks sweet/apples as RARE and pass a probability
+    for random skipping rare tests.
+    """
+    with patch('testrunner.testproc.timeout.TimeoutProc.create',
+               lambda x: FakeTimeoutProc(10)):
+      return self.run_tests(
+          '--command-prefix',
+          sys.executable,
+          '--outdir',
+          'out/build',
+          '--variants=default',
+          '--fuzzer-random-seed=12345',
+          '--total-timeout-sec=60',
+          '--allocation-offset=1',
+          '--progress=verbose',
+          f'--skip-rare-tests-prob={prob}',
+          'sweet',
+          baseroot="testroot8")
+
+  def testRandomSkip_Includes(self):
+    """Ensure that a test case marked as FUZZ_RARE (here apples) is still
+    included if the probability for skipping is 0.
+    """
+    result = self._run_test_with_random_skip(0.0)
+    result.has_returncode(1)
+    result.stdout_includes('sweet/apples default: FAIL')
+    result.stdout_includes('sweet/bananas default: PASS')
+    result.stdout_includes('7 tests ran')
+
+  def testRandomSkip_Excludes(self):
+    """Ensure that a test case marked as FUZZ_RARE (here apples) is always
+    excluded if the probability for skipping is 1.
+    """
+    result = self._run_test_with_random_skip(1.0)
+    result.has_returncode(0)
+    result.stdout_excludes('sweet/apples')
+    result.stdout_includes('sweet/bananas default: PASS')
+    result.stdout_includes('6 tests ran')
 
 class OtherTest(TestRunnerTest):
   def testStatusFilePresubmit(self):
